@@ -2,7 +2,7 @@
 
   class.c -
 
-  $Author: usa $
+  $Author: nagachika $
   created at: Tue Aug 10 15:05:44 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -49,15 +49,16 @@ static ID id_attached;
 static VALUE
 class_alloc(VALUE flags, VALUE klass)
 {
-    rb_classext_t *ext = ALLOC(rb_classext_t);
-    NEWOBJ(obj, struct RClass);
-    OBJSETUP(obj, klass, flags);
-    obj->ptr = ext;
+    NEWOBJ_OF(obj, struct RClass, klass, flags);
+    obj->ptr = ALLOC(rb_classext_t);
     RCLASS_IV_TBL(obj) = 0;
     RCLASS_CONST_TBL(obj) = 0;
     RCLASS_M_TBL(obj) = 0;
     RCLASS_SUPER(obj) = 0;
+    RCLASS_ORIGIN(obj) = (VALUE)obj;
     RCLASS_IV_INDEX_TBL(obj) = 0;
+    RCLASS_REFINED_CLASS(obj) = Qnil;
+    RCLASS_EXT(obj)->allocator = 0;
     return (VALUE)obj;
 }
 
@@ -93,7 +94,7 @@ rb_class_boot(VALUE super)
 void
 rb_check_inheritable(VALUE super)
 {
-    if (TYPE(super) != T_CLASS) {
+    if (!RB_TYPE_P(super, T_CLASS)) {
 	rb_raise(rb_eTypeError, "superclass must be a Class (%s given)",
 		 rb_obj_classname(super));
     }
@@ -120,27 +121,44 @@ rb_class_new(VALUE super)
     return rb_class_boot(super);
 }
 
-struct clone_method_data {
-    st_table *tbl;
-    VALUE klass;
-};
+static NODE*
+rewrite_cref_stack(NODE *node, VALUE old_klass, VALUE new_klass)
+{
+    NODE *new_node;
+    if (!node) {
+	return NULL;
+    }
+    if (node->nd_clss == old_klass) {
+	new_node = NEW_CREF(new_klass);
+	new_node->nd_next = node->nd_next;
+    } else {
+	new_node = NEW_CREF(node->nd_clss);
+	new_node->nd_next = rewrite_cref_stack(node->nd_next, old_klass, new_klass);
+    }
+    return new_node;
+}
 
-VALUE rb_iseq_clone(VALUE iseqval, VALUE newcbase);
-
-static int
-clone_method(ID mid, const rb_method_entry_t *me, struct clone_method_data *data)
+static void
+clone_method(VALUE klass, ID mid, const rb_method_entry_t *me)
 {
     VALUE newiseqval;
     if (me->def && me->def->type == VM_METHOD_TYPE_ISEQ) {
 	rb_iseq_t *iseq;
-	newiseqval = rb_iseq_clone(me->def->body.iseq->self, data->klass);
+	newiseqval = rb_iseq_clone(me->def->body.iseq->self, klass);
 	GetISeqPtr(newiseqval, iseq);
-	rb_add_method(data->klass, mid, VM_METHOD_TYPE_ISEQ, iseq, me->flag);
+	iseq->cref_stack = rewrite_cref_stack(me->def->body.iseq->cref_stack, me->klass, klass);
+	rb_add_method(klass, mid, VM_METHOD_TYPE_ISEQ, iseq, me->flag);
 	RB_GC_GUARD(newiseqval);
     }
     else {
-	rb_method_entry_set(data->klass, mid, me, me->flag);
+	rb_method_entry_set(klass, mid, me, me->flag);
     }
+}
+
+static int
+clone_method_i(st_data_t key, st_data_t value, st_data_t data)
+{
+    clone_method((VALUE)data, (ID)key, (const rb_method_entry_t *)value);
     return ST_CONTINUE;
 }
 
@@ -186,6 +204,7 @@ rb_mod_init_copy(VALUE clone, VALUE orig)
 	rb_singleton_class_attached(RBASIC(clone)->klass, (VALUE)clone);
     }
     RCLASS_SUPER(clone) = RCLASS_SUPER(orig);
+    RCLASS_EXT(clone)->allocator = RCLASS_EXT(orig)->allocator;
     if (RCLASS_IV_TBL(orig)) {
 	st_data_t id;
 
@@ -193,6 +212,8 @@ rb_mod_init_copy(VALUE clone, VALUE orig)
 	    st_free_table(RCLASS_IV_TBL(clone));
 	}
 	RCLASS_IV_TBL(clone) = st_copy(RCLASS_IV_TBL(orig));
+	CONST_ID(id, "__tmp_classpath__");
+	st_delete(RCLASS_IV_TBL(clone), &id, 0);
 	CONST_ID(id, "__classpath__");
 	st_delete(RCLASS_IV_TBL(clone), &id, 0);
 	CONST_ID(id, "__classid__");
@@ -206,15 +227,11 @@ rb_mod_init_copy(VALUE clone, VALUE orig)
 	st_foreach(RCLASS_CONST_TBL(orig), clone_const_i, (st_data_t)RCLASS_CONST_TBL(clone));
     }
     if (RCLASS_M_TBL(orig)) {
-	struct clone_method_data data;
-
 	if (RCLASS_M_TBL(clone)) {
 	    rb_free_m_table(RCLASS_M_TBL(clone));
 	}
-	data.tbl = RCLASS_M_TBL(clone) = st_init_numtable();
-	data.klass = clone;
-	st_foreach(RCLASS_M_TBL(orig), clone_method,
-		   (st_data_t)&data);
+	RCLASS_M_TBL(clone) = st_init_numtable();
+	st_foreach(RCLASS_M_TBL(orig), clone_method_i, (st_data_t)clone);
     }
 
     return clone;
@@ -222,24 +239,30 @@ rb_mod_init_copy(VALUE clone, VALUE orig)
 
 VALUE
 rb_singleton_class_clone(VALUE obj)
+{
+    return rb_singleton_class_clone_and_attach(obj, Qundef);
+}
+
+VALUE
+rb_singleton_class_clone_and_attach(VALUE obj, VALUE attach)
 {
     VALUE klass = RBASIC(obj)->klass;
 
     if (!FL_TEST(klass, FL_SINGLETON))
 	return klass;
     else {
-	struct clone_method_data data;
 	/* copy singleton(unnamed) class */
-	VALUE clone = class_alloc((RBASIC(klass)->flags & ~(FL_MARK)), 0);
+	VALUE clone = class_alloc(RBASIC(klass)->flags, 0);
 
 	if (BUILTIN_TYPE(obj) == T_CLASS) {
-	    RBASIC(clone)->klass = (VALUE)clone;
+	    RBASIC(clone)->klass = clone;
 	}
 	else {
 	    RBASIC(clone)->klass = rb_singleton_class_clone(klass);
 	}
 
 	RCLASS_SUPER(clone) = RCLASS_SUPER(klass);
+	RCLASS_EXT(clone)->allocator = RCLASS_EXT(klass)->allocator;
 	if (RCLASS_IV_TBL(klass)) {
 	    RCLASS_IV_TBL(clone) = st_copy(RCLASS_IV_TBL(klass));
 	}
@@ -247,14 +270,14 @@ rb_singleton_class_clone(VALUE obj)
 	    RCLASS_CONST_TBL(clone) = st_init_numtable();
 	    st_foreach(RCLASS_CONST_TBL(klass), clone_const_i, (st_data_t)RCLASS_CONST_TBL(clone));
 	}
+	if (attach != Qundef) {
+	    rb_singleton_class_attached(clone, attach);
+	}
 	RCLASS_M_TBL(clone) = st_init_numtable();
-	data.tbl = RCLASS_M_TBL(clone);
-	data.klass = (VALUE)clone;
-	st_foreach(RCLASS_M_TBL(klass), clone_method,
-		   (st_data_t)&data);
-	rb_singleton_class_attached(RBASIC(clone)->klass, (VALUE)clone);
+	st_foreach(RCLASS_M_TBL(klass), clone_method_i, (st_data_t)clone);
+	rb_singleton_class_attached(RBASIC(clone)->klass, clone);
 	FL_SET(clone, FL_SINGLETON);
-	return (VALUE)clone;
+	return clone;
     }
 }
 
@@ -284,6 +307,14 @@ rb_singleton_class_attached(VALUE klass, VALUE obj)
  */
 #define META_CLASS_OF_CLASS_CLASS_P(k)  (METACLASS_OF(k) == (k))
 
+/*!
+ * whether k has a metaclass
+ * @retval 1 if \a k has a metaclass
+ * @retval 0 otherwise
+ */
+#define HAVE_METACLASS_P(k) \
+    (FL_TEST(METACLASS_OF(k), FL_SINGLETON) && \
+     rb_ivar_get(METACLASS_OF(k), id_attached) == (k))
 
 /*!
  * ensures \a klass belongs to its own eigenclass.
@@ -293,7 +324,7 @@ rb_singleton_class_attached(VALUE klass, VALUE obj)
  * @note this macro creates a new eigenclass if necessary.
  */
 #define ENSURE_EIGENCLASS(klass) \
- (rb_ivar_get(METACLASS_OF(klass), id_attached) == (klass) ? METACLASS_OF(klass) : make_metaclass(klass))
+    (HAVE_METACLASS_P(klass) ? METACLASS_OF(klass) : make_metaclass(klass))
 
 
 /*!
@@ -473,7 +504,7 @@ rb_define_class(const char *name, VALUE super)
     id = rb_intern(name);
     if (rb_const_defined(rb_cObject, id)) {
 	klass = rb_const_get(rb_cObject, id);
-	if (TYPE(klass) != T_CLASS) {
+	if (!RB_TYPE_P(klass, T_CLASS)) {
 	    rb_raise(rb_eTypeError, "%s is not a class", name);
 	}
 	if (rb_class_real(RCLASS_SUPER(klass)) != super) {
@@ -540,7 +571,7 @@ rb_define_class_id_under(VALUE outer, ID id, VALUE super)
 
     if (rb_const_defined_at(outer, id)) {
 	klass = rb_const_get_at(outer, id);
-	if (TYPE(klass) != T_CLASS) {
+	if (!RB_TYPE_P(klass, T_CLASS)) {
 	    rb_raise(rb_eTypeError, "%s is not a class", rb_id2name(id));
 	}
 	if (rb_class_real(RCLASS_SUPER(klass)) != super) {
@@ -591,7 +622,7 @@ rb_define_module(const char *name)
     id = rb_intern(name);
     if (rb_const_defined(rb_cObject, id)) {
 	module = rb_const_get(rb_cObject, id);
-	if (TYPE(module) == T_MODULE)
+	if (RB_TYPE_P(module, T_MODULE))
 	    return module;
 	rb_raise(rb_eTypeError, "%s is not a module", rb_obj_classname(module));
     }
@@ -615,7 +646,7 @@ rb_define_module_id_under(VALUE outer, ID id)
 
     if (rb_const_defined_at(outer, id)) {
 	module = rb_const_get_at(outer, id);
-	if (TYPE(module) == T_MODULE)
+	if (RB_TYPE_P(module, T_MODULE))
 	    return module;
 	rb_raise(rb_eTypeError, "%s::%s is not a module",
 		 rb_class2name(outer), rb_obj_classname(module));
@@ -628,8 +659,8 @@ rb_define_module_id_under(VALUE outer, ID id)
     return module;
 }
 
-static VALUE
-include_class_new(VALUE module, VALUE super)
+VALUE
+rb_include_class_new(VALUE module, VALUE super)
 {
     VALUE klass = class_alloc(T_ICLASS, rb_cClass);
 
@@ -644,9 +675,9 @@ include_class_new(VALUE module, VALUE super)
     }
     RCLASS_IV_TBL(klass) = RCLASS_IV_TBL(module);
     RCLASS_CONST_TBL(klass) = RCLASS_CONST_TBL(module);
-    RCLASS_M_TBL(klass) = RCLASS_M_TBL(module);
+    RCLASS_M_TBL(klass) = RCLASS_M_TBL(RCLASS_ORIGIN(module));
     RCLASS_SUPER(klass) = super;
-    if (TYPE(module) == T_ICLASS) {
+    if (RB_TYPE_P(module, T_ICLASS)) {
 	RBASIC(klass)->klass = RBASIC(module)->klass;
     }
     else {
@@ -658,10 +689,11 @@ include_class_new(VALUE module, VALUE super)
     return (VALUE)klass;
 }
 
+static int include_modules_at(const VALUE klass, VALUE c, VALUE module);
+
 void
 rb_include_module(VALUE klass, VALUE module)
 {
-    VALUE p, c;
     int changed = 0;
 
     rb_frozen_class_p(klass);
@@ -669,17 +701,39 @@ rb_include_module(VALUE klass, VALUE module)
 	rb_secure(4);
     }
 
-    if (TYPE(module) != T_MODULE) {
+    if (!RB_TYPE_P(module, T_MODULE)) {
 	Check_Type(module, T_MODULE);
     }
 
     OBJ_INFECT(klass, module);
-    c = klass;
+
+    changed = include_modules_at(klass, RCLASS_ORIGIN(klass), module);
+    if (changed < 0)
+	rb_raise(rb_eArgError, "cyclic include detected");
+    if (changed) rb_clear_cache();
+}
+
+static int
+add_refined_method_entry_i(st_data_t key, st_data_t value, st_data_t data)
+{
+    rb_add_refined_method_entry((VALUE) data, (ID) key);
+    return ST_CONTINUE;
+}
+
+static int
+include_modules_at(const VALUE klass, VALUE c, VALUE module)
+{
+    VALUE p;
+    int changed = 0;
+    const st_table *const klass_m_tbl = RCLASS_M_TBL(RCLASS_ORIGIN(klass));
+
     while (module) {
 	int superclass_seen = FALSE;
 
-	if (RCLASS_M_TBL(klass) == RCLASS_M_TBL(module))
-	    rb_raise(rb_eArgError, "cyclic include detected");
+	if (RCLASS_ORIGIN(module) != module)
+	    goto skip;
+	if (klass_m_tbl && klass_m_tbl == RCLASS_M_TBL(module))
+	    return -1;
 	/* ignore if the module included already in superclasses */
 	for (p = RCLASS_SUPER(klass); p; p = RCLASS_SUPER(p)) {
 	    switch (BUILTIN_TYPE(p)) {
@@ -696,13 +750,87 @@ rb_include_module(VALUE klass, VALUE module)
 		break;
 	    }
 	}
-	c = RCLASS_SUPER(c) = include_class_new(module, RCLASS_SUPER(c));
+	c = RCLASS_SUPER(c) = rb_include_class_new(module, RCLASS_SUPER(c));
+	if (FL_TEST(klass, RMODULE_IS_REFINEMENT)) {
+	    VALUE refined_class =
+		rb_refinement_module_get_refined_class(klass);
+
+	    st_foreach(RMODULE_M_TBL(module), add_refined_method_entry_i,
+		       (st_data_t) refined_class);
+	    FL_SET(c, RMODULE_INCLUDED_INTO_REFINEMENT);
+	}
 	if (RMODULE_M_TBL(module) && RMODULE_M_TBL(module)->num_entries)
 	    changed = 1;
+	if (RMODULE_CONST_TBL(module) && RMODULE_CONST_TBL(module)->num_entries)
+	    changed = 1;
       skip:
 	module = RCLASS_SUPER(module);
     }
-    if (changed) rb_clear_cache();
+
+    return changed;
+}
+
+static int
+move_refined_method(st_data_t key, st_data_t value, st_data_t data)
+{
+    rb_method_entry_t *me = (rb_method_entry_t *) value;
+    st_table *tbl = (st_table *) data;
+
+    if (me->def->type == VM_METHOD_TYPE_REFINED) {
+	if (me->def->body.orig_me) {
+	    rb_method_entry_t *orig_me = me->def->body.orig_me, *new_me;
+	    me->def->body.orig_me = NULL;
+	    new_me = ALLOC(rb_method_entry_t);
+	    *new_me = *me;
+	    st_add_direct(tbl, key, (st_data_t) new_me);
+	    *me = *orig_me;
+	    xfree(orig_me);
+	    return ST_CONTINUE;
+	}
+	else {
+	    st_add_direct(tbl, key, (st_data_t) me);
+	    return ST_DELETE;
+	}
+    }
+    else {
+	return ST_CONTINUE;
+    }
+}
+
+void
+rb_prepend_module(VALUE klass, VALUE module)
+{
+    void rb_vm_check_redefinition_by_prepend(VALUE klass);
+    VALUE origin;
+    int changed = 0;
+
+    rb_frozen_class_p(klass);
+    if (!OBJ_UNTRUSTED(klass)) {
+	rb_secure(4);
+    }
+
+    Check_Type(module, T_MODULE);
+
+    OBJ_INFECT(klass, module);
+
+    origin = RCLASS_ORIGIN(klass);
+    if (origin == klass) {
+	origin = class_alloc(T_ICLASS, klass);
+	RCLASS_SUPER(origin) = RCLASS_SUPER(klass);
+	RCLASS_SUPER(klass) = origin;
+	RCLASS_ORIGIN(klass) = origin;
+	RCLASS_M_TBL(origin) = RCLASS_M_TBL(klass);
+	RCLASS_M_TBL(klass) = st_init_numtable();
+	st_foreach(RCLASS_M_TBL(origin), move_refined_method,
+		   (st_data_t) RCLASS_M_TBL(klass));
+    }
+    changed = include_modules_at(klass, klass, module);
+    if (changed < 0)
+	rb_raise(rb_eArgError, "cyclic prepend detected");
+    if (changed) {
+	rb_clear_cache();
+	rb_vm_check_redefinition_by_prepend(klass);
+    }
 }
 
 /*
@@ -727,10 +855,13 @@ rb_mod_included_modules(VALUE mod)
 {
     VALUE ary = rb_ary_new();
     VALUE p;
+    VALUE origin = RCLASS_ORIGIN(mod);
 
     for (p = RCLASS_SUPER(mod); p; p = RCLASS_SUPER(p)) {
-	if (BUILTIN_TYPE(p) == T_ICLASS) {
-	    rb_ary_push(ary, RBASIC(p)->klass);
+	if (p != origin && BUILTIN_TYPE(p) == T_ICLASS) {
+	    VALUE m = RBASIC(p)->klass;
+	    if (RB_TYPE_P(m, T_MODULE))
+		rb_ary_push(ary, m);
 	}
     }
     return ary;
@@ -796,7 +927,7 @@ rb_mod_ancestors(VALUE mod)
 	if (BUILTIN_TYPE(p) == T_ICLASS) {
 	    rb_ary_push(ary, RBASIC(p)->klass);
 	}
-	else {
+	else if (p == RCLASS_ORIGIN(p)) {
 	    rb_ary_push(ary, p);
 	}
     }
@@ -858,10 +989,10 @@ method_entry_i(st_data_t key, st_data_t value, st_data_t data)
     st_table *list = (st_table *)data;
     long type;
 
-    if ((ID)key == ID_ALLOCATOR) {
-	return ST_CONTINUE;
+    if (me && me->def->type == VM_METHOD_TYPE_REFINED) {
+	me = rb_resolve_refined_method(Qnil, me, NULL);
+	if (!me) return ST_CONTINUE;
     }
-
     if (!st_lookup(list, key, 0)) {
 	if (UNDEFINED_METHOD_ENTRY_P(me)) {
 	    type = -1; /* none */
@@ -878,7 +1009,7 @@ static VALUE
 class_instance_method_list(int argc, VALUE *argv, VALUE mod, int obj, int (*func) (st_data_t, st_data_t, st_data_t))
 {
     VALUE ary;
-    int recur;
+    int recur, prepended = 0;
     st_table *list;
 
     if (argc == 0) {
@@ -890,10 +1021,15 @@ class_instance_method_list(int argc, VALUE *argv, VALUE mod, int obj, int (*func
 	recur = RTEST(r);
     }
 
+    if (!recur && RCLASS_ORIGIN(mod) != mod) {
+	mod = RCLASS_ORIGIN(mod);
+	prepended = 1;
+    }
+
     list = st_init_numtable();
     for (; mod; mod = RCLASS_SUPER(mod)) {
-	st_foreach(RCLASS_M_TBL(mod), method_entry_i, (st_data_t)list);
-	if (BUILTIN_TYPE(mod) == T_ICLASS) continue;
+	if (RCLASS_M_TBL(mod)) st_foreach(RCLASS_M_TBL(mod), method_entry_i, (st_data_t)list);
+	if (BUILTIN_TYPE(mod) == T_ICLASS && !prepended) continue;
 	if (obj && FL_TEST(mod, FL_SINGLETON)) continue;
 	if (!recur) break;
     }
@@ -992,11 +1128,13 @@ rb_class_public_instance_methods(int argc, VALUE *argv, VALUE mod)
 
 /*
  *  call-seq:
- *     obj.methods    -> array
+ *     obj.methods(all=true)    -> array
  *
  *  Returns a list of the names of public and protected methods of
  *  <i>obj</i>. This will include all the methods accessible in
  *  <i>obj</i>'s ancestors.
+ *  If the <i>all</i> parameter is set to <code>false</code>, only those methods
+ *  in the receiver will be listed.
  *
  *     class Klass
  *       def klass_method()
@@ -1014,9 +1152,6 @@ rb_obj_methods(int argc, VALUE *argv, VALUE obj)
 {
   retry:
     if (argc == 0) {
-	VALUE args[1];
-
-	args[0] = Qtrue;
 	return class_instance_method_list(argc, argv, CLASS_OF(obj), 1, ins_methods_i);
     }
     else {
@@ -1124,12 +1259,14 @@ rb_obj_singleton_methods(int argc, VALUE *argv, VALUE obj)
     klass = CLASS_OF(obj);
     list = st_init_numtable();
     if (klass && FL_TEST(klass, FL_SINGLETON)) {
-	st_foreach(RCLASS_M_TBL(klass), method_entry_i, (st_data_t)list);
+	if (RCLASS_M_TBL(klass))
+	    st_foreach(RCLASS_M_TBL(klass), method_entry_i, (st_data_t)list);
 	klass = RCLASS_SUPER(klass);
     }
     if (RTEST(recur)) {
-	while (klass && (FL_TEST(klass, FL_SINGLETON) || TYPE(klass) == T_ICLASS)) {
-	    st_foreach(RCLASS_M_TBL(klass), method_entry_i, (st_data_t)list);
+	while (klass && (FL_TEST(klass, FL_SINGLETON) || RB_TYPE_P(klass, T_ICLASS))) {
+	    if (RCLASS_M_TBL(klass))
+		st_foreach(RCLASS_M_TBL(klass), method_entry_i, (st_data_t)list);
 	    klass = RCLASS_SUPER(klass);
 	}
     }
@@ -1241,6 +1378,20 @@ rb_undef_method(VALUE klass, const char *name)
     }\
 } while (0)
 
+static inline VALUE
+special_singleton_class_of(VALUE obj)
+{
+    SPECIAL_SINGLETON(Qnil, rb_cNilClass);
+    SPECIAL_SINGLETON(Qfalse, rb_cFalseClass);
+    SPECIAL_SINGLETON(Qtrue, rb_cTrueClass);
+    return Qnil;
+}
+
+VALUE
+rb_special_singleton_class(VALUE obj)
+{
+    return special_singleton_class_of(obj);
+}
 
 /*!
  * \internal
@@ -1256,14 +1407,20 @@ singleton_class_of(VALUE obj)
 {
     VALUE klass;
 
-    if (FIXNUM_P(obj) || SYMBOL_P(obj)) {
+    if (FIXNUM_P(obj) || FLONUM_P(obj) || SYMBOL_P(obj)) {
 	rb_raise(rb_eTypeError, "can't define singleton");
     }
-    if (rb_special_const_p(obj)) {
-	SPECIAL_SINGLETON(Qnil, rb_cNilClass);
-	SPECIAL_SINGLETON(Qfalse, rb_cFalseClass);
-	SPECIAL_SINGLETON(Qtrue, rb_cTrueClass);
-	rb_bug("unknown immediate %p", (void *)obj);
+    if (SPECIAL_CONST_P(obj)) {
+	klass = special_singleton_class_of(obj);
+	if (NIL_P(klass))
+	    rb_bug("unknown immediate %p", (void *)obj);
+	return klass;
+    }
+    else {
+	enum ruby_value_type type = BUILTIN_TYPE(obj);
+	if (type == T_FLOAT || type == T_BIGNUM) {
+           rb_raise(rb_eTypeError, "can't define singleton");
+	}
     }
 
     if (FL_TEST(RBASIC(obj)->klass, FL_SINGLETON) &&
@@ -1315,7 +1472,7 @@ rb_singleton_class(VALUE obj)
     VALUE klass = singleton_class_of(obj);
 
     /* ensures an exposed class belongs to its own eigenclass */
-    if (TYPE(obj) == T_CLASS) (void)ENSURE_EIGENCLASS(klass);
+    if (RB_TYPE_P(obj, T_CLASS)) (void)ENSURE_EIGENCLASS(klass);
 
     return klass;
 }
@@ -1400,7 +1557,7 @@ rb_define_attr(VALUE klass, const char *name, int read, int write)
 int
 rb_obj_basic_to_s_p(VALUE obj)
 {
-    const rb_method_entry_t *me = rb_method_entry(CLASS_OF(obj), rb_intern("to_s"));
+    const rb_method_entry_t *me = rb_method_entry(CLASS_OF(obj), rb_intern("to_s"), 0);
     if (me && me->def && me->def->type == VM_METHOD_TYPE_CFUNC &&
 	me->def->body.cfunc.func == rb_any_to_s)
 	return 1;
@@ -1473,7 +1630,7 @@ rb_scan_args(int argc, const VALUE *argv, const char *fmt, ...)
 		argc--;
 	}
 	else {
-	    hash = rb_check_convert_type(last, T_HASH, "Hash", "to_hash");
+	    hash = rb_check_hash_type(last);
 	    if (!NIL_P(hash))
 		argc--;
 	}
@@ -1531,18 +1688,12 @@ rb_scan_args(int argc, const VALUE *argv, const char *fmt, ...)
     }
     va_end(vargs);
 
-    if (argi < argc)
-	goto argc_error;
+    if (argi < argc) {
+      argc_error:
+	rb_error_arity(argc, n_mand, f_var ? UNLIMITED_ARGUMENTS : n_mand + n_opt);
+    }
 
     return argc;
-
-  argc_error:
-    if (0 < n_opt)
-	rb_raise(rb_eArgError, "wrong number of arguments (%d for %d..%d%s)",
-		 argc, n_mand, n_mand + n_opt, f_var ? "+" : "");
-    else
-	rb_raise(rb_eArgError, "wrong number of arguments (%d for %d%s)",
-		 argc, n_mand, f_var ? "+" : "");
 }
 
 /*!
