@@ -2,7 +2,7 @@
 
   cont.c -
 
-  $Author: naruse $
+  $Author: nagachika $
   created at: Thu May 23 09:03:43 2007
 
   Copyright (C) 2007 Koichi Sasada
@@ -15,9 +15,6 @@
 #include "gc.h"
 #include "eval_intern.h"
 
-#if ((defined(_WIN32) && _WIN32_WINNT >= 0x0400) || (defined(HAVE_GETCONTEXT) && defined(HAVE_SETCONTEXT))) && !defined(__NetBSD__) && !defined(sun) && !defined(FIBER_USE_NATIVE)
-#define FIBER_USE_NATIVE 1
-
 /* FIBER_USE_NATIVE enables Fiber performance improvement using system
  * dependent method such as make/setcontext on POSIX system or
  * CreateFiber() API on Windows.
@@ -29,12 +26,45 @@
  * in Proc. of 51th Programming Symposium, pp.21--28 (2010) (in Japanese).
  */
 
+#if !defined(FIBER_USE_NATIVE)
+# if defined(HAVE_GETCONTEXT) && defined(HAVE_SETCONTEXT)
+#   if 0
+#   elif defined(__NetBSD__)
 /* On our experience, NetBSD doesn't support using setcontext() and pthread
  * simultaneously.  This is because pthread_self(), TLS and other information
  * are represented by stack pointer (higher bits of stack pointer).
  * TODO: check such constraint on configure.
  */
-#elif !defined(FIBER_USE_NATIVE)
+#     define FIBER_USE_NATIVE 0
+#   elif defined(__sun)
+/* On Solaris because resuming any Fiber caused SEGV, for some reason.
+ */
+#     define FIBER_USE_NATIVE 0
+#   elif defined(__ia64)
+/* At least, Linux/ia64's getcontext(3) doesn't save register window.
+ */
+#     define FIBER_USE_NATIVE 0
+#   elif defined(__GNU__)
+/* GNU/Hurd doesn't fully support getcontext, setcontext, makecontext
+ * and swapcontext functions. Disabling their usage till support is
+ * implemented. More info at
+ * http://darnassus.sceen.net/~hurd-web/open_issues/glibc/#getcontext
+ */
+#     define FIBER_USE_NATIVE 0
+#   else
+#     define FIBER_USE_NATIVE 1
+#   endif
+# elif defined(_WIN32)
+#   if _WIN32_WINNT >= 0x0400
+/* only when _WIN32_WINNT >= 0x0400 on Windows because Fiber APIs are
+ * supported only such building (and running) environments.
+ * [ruby-dev:41192]
+ */
+#     define FIBER_USE_NATIVE 1
+#   endif
+# endif
+#endif
+#if !defined(FIBER_USE_NATIVE)
 #define FIBER_USE_NATIVE 0
 #endif
 
@@ -47,8 +77,7 @@
 #define RB_PAGE_SIZE (pagesize)
 #define RB_PAGE_MASK (~(RB_PAGE_SIZE - 1))
 static long pagesize;
-#define FIBER_MACHINE_STACK_ALLOCATION_SIZE  (0x10000)
-#endif
+#endif /*FIBER_USE_NATIVE*/
 
 #define CAPTURE_JUST_VALID_VM_STACK 1
 
@@ -103,6 +132,12 @@ typedef struct rb_fiber_struct {
     enum fiber_status status;
     struct rb_fiber_struct *prev_fiber;
     struct rb_fiber_struct *next_fiber;
+    /* If a fiber invokes "transfer",
+     * then this fiber can't "resume" any more after that.
+     * You shouldn't mix "transfer" and "resume".
+     */
+    int transfered;
+
 #if FIBER_USE_NATIVE
 #ifdef _WIN32
     void *fib_handle;
@@ -123,7 +158,7 @@ static VALUE rb_eFiberError;
 #define GetFiberPtr(obj, ptr)  do {\
     TypedData_Get_Struct((obj), rb_fiber_t, &fiber_data_type, (ptr)); \
     if (!(ptr)) rb_raise(rb_eFiberError, "uninitialized fiber"); \
-} while(0)
+} while (0)
 
 NOINLINE(static VALUE cont_capture(volatile int *stat));
 
@@ -324,6 +359,17 @@ fiber_memsize(const void *ptr)
     return size;
 }
 
+VALUE
+rb_obj_is_fiber(VALUE obj)
+{
+    if (rb_typeddata_is_kind_of(obj, &fiber_data_type)) {
+	return Qtrue;
+    }
+    else {
+	return Qfalse;
+    }
+}
+
 static void
 cont_save_machine_stack(rb_thread_t *th, rb_context_t *cont)
 {
@@ -448,7 +494,7 @@ cont_capture(volatile int *stat)
     }
     else {
 	*stat = 0;
-	return cont->self;
+	return contval;
     }
 }
 
@@ -496,6 +542,8 @@ cont_restore_thread(rb_context_t *cont)
     th->protect_tag = sth->protect_tag;
     th->errinfo = sth->errinfo;
     th->first_proc = sth->first_proc;
+    th->root_lep = sth->root_lep;
+    th->root_svar = sth->root_svar;
 }
 
 #if FIBER_USE_NATIVE
@@ -609,7 +657,7 @@ fiber_setcontext(rb_fiber_t *newfib, rb_fiber_t *oldfib)
     rb_thread_t *th = GET_THREAD(), *sth = &newfib->cont.saved_thread;
 
     if (newfib->status != RUNNING) {
-	fiber_initialize_machine_stack_context(newfib, FIBER_MACHINE_STACK_ALLOCATION_SIZE);
+	fiber_initialize_machine_stack_context(newfib, th->vm->default_params.fiber_machine_stack_size);
     }
 
     /* restore thread context */
@@ -646,17 +694,6 @@ fiber_setcontext(rb_fiber_t *newfib, rb_fiber_t *oldfib)
     /* swap machine context */
 #ifdef _WIN32
     SwitchToFiber(newfib->fib_handle);
-#elif defined(__FreeBSD__) /* FreeBSD 9 doesn't work with swapcontext */
-    if (!ruby_setjmp(oldfib->cont.jmpbuf)) {
-	if (newfib->status != RUNNING) {
-	    if (setcontext(&newfib->context) < 0) {
-		rb_bug("context switch between fiber failed");
-	    }
-	}
-	else {
-	    ruby_longjmp(newfib->cont.jmpbuf, 1);
-	}
-    }
 #else
     swapcontext(&oldfib->context, &newfib->context);
 #endif
@@ -782,17 +819,17 @@ cont_restore_0(rb_context_t *cont, VALUE *addr_in_prev_frame)
     cont_restore_1(cont);
 }
 #ifdef __ia64
-#define cont_restore_0(cont, vp) register_stack_extend((cont), (vp), (VALUE*)rb_ia64_bsp());
+#define cont_restore_0(cont, vp) register_stack_extend((cont), (vp), (VALUE*)rb_ia64_bsp())
 #endif
 
 /*
  *  Document-class: Continuation
  *
- *  Continuation objects are generated by <code>Kernel#callcc</code>,
- *  after having <code>require</code>d <i>continuation</i>. They hold
+ *  Continuation objects are generated by Kernel#callcc,
+ *  after having +require+d <i>continuation</i>. They hold
  *  a return address and execution context, allowing a nonlocal return
  *  to the end of the <code>callcc</code> block from anywhere within a
- *  program.  Continuations are somewhat analogous to a structured
+ *  program. Continuations are somewhat analogous to a structured
  *  version of C's <code>setjmp/longjmp</code> (although they contain
  *  more state, so you might consider them closer to threads).
  *
@@ -838,15 +875,15 @@ cont_restore_0(rb_context_t *cont, VALUE *addr_in_prev_frame)
  *  call-seq:
  *     callcc {|cont| block }   ->  obj
  *
- *  Generates a <code>Continuation</code> object, which it passes to
+ *  Generates a Continuation object, which it passes to
  *  the associated block. You need to <code>require
  *  'continuation'</code> before using this method. Performing a
- *  <em>cont</em><code>.call</code> will cause the <code>callcc</code>
+ *  <em>cont</em><code>.call</code> will cause the #callcc
  *  to return (as will falling through the end of the block). The
- *  value returned by the <code>callcc</code> is the value of the
+ *  value returned by the #callcc is the value of the
  *  block, or the value passed to <em>cont</em><code>.call</code>. See
- *  class <code>Continuation</code> for more details. Also see
- *  <code>Kernel::throw</code> for an alternative mechanism for
+ *  class Continuation for more details. Also see
+ *  Kernel#throw for an alternative mechanism for
  *  unwinding a call stack.
  */
 
@@ -867,7 +904,7 @@ rb_callcc(VALUE self)
 static VALUE
 make_passing_arg(int argc, VALUE *argv)
 {
-    switch(argc) {
+    switch (argc) {
       case 0:
 	return Qnil;
       case 1:
@@ -918,6 +955,9 @@ rb_cont_call(int argc, VALUE *argv, VALUE contval)
     cont->argc = argc;
     cont->value = make_passing_arg(argc, argv);
 
+    /* restore `tracing' context. see [Feature #4347] */
+    th->trace_arg = cont->saved_thread.trace_arg;
+
     cont_restore_0(cont, &contval);
     return Qnil; /* unreachable */
 }
@@ -988,8 +1028,6 @@ rb_cont_call(int argc, VALUE *argv, VALUE contval)
  *
  */
 
-#define FIBER_VM_STACK_SIZE (4 * 1024)
-
 static const rb_data_type_t fiber_data_type = {
     "fiber",
     {fiber_mark, fiber_free, fiber_memsize,},
@@ -1040,18 +1078,20 @@ fiber_init(VALUE fibval, VALUE proc)
 
     fiber_link_join(fib);
 
-    th->stack_size = FIBER_VM_STACK_SIZE;
+    th->stack_size = th->vm->default_params.fiber_vm_stack_size / sizeof(VALUE);
     th->stack = ALLOC_N(VALUE, th->stack_size);
 
     th->cfp = (void *)(th->stack + th->stack_size);
     th->cfp--;
     th->cfp->pc = 0;
     th->cfp->sp = th->stack + 1;
-    th->cfp->bp = 0;
-    th->cfp->lfp = th->stack;
-    *th->cfp->lfp = 0;
-    th->cfp->dfp = th->stack;
+#if VM_DEBUG_BP_CHECK
+    th->cfp->bp_check = 0;
+#endif
+    th->cfp->ep = th->stack;
+    *th->cfp->ep = VM_ENVVAL_BLOCK_PTR(0);
     th->cfp->self = Qnil;
+    th->cfp->klass = Qnil;
     th->cfp->flag = 0;
     th->cfp->iseq = 0;
     th->cfp->proc = 0;
@@ -1087,20 +1127,19 @@ return_fiber(void)
 {
     rb_fiber_t *fib;
     VALUE curr = rb_fiber_current();
+    VALUE prev;
     GetFiberPtr(curr, fib);
 
-    if (fib->prev == Qnil) {
-	rb_thread_t *th = GET_THREAD();
+    prev = fib->prev;
+    if (NIL_P(prev)) {
+	const VALUE root_fiber = GET_THREAD()->root_fiber;
 
-	if (th->root_fiber != curr) {
-	    return th->root_fiber;
-	}
-	else {
+	if (root_fiber == curr) {
 	    rb_raise(rb_eFiberError, "can't yield from root fiber");
 	}
+	return root_fiber;
     }
     else {
-	VALUE prev = fib->prev;
 	fib->prev = Qnil;
 	return prev;
     }
@@ -1145,21 +1184,22 @@ rb_fiber_start(void)
 	argv = (argc = cont->argc) > 1 ? RARRAY_PTR(args) : &args;
 	cont->value = Qnil;
 	th->errinfo = Qnil;
-	th->local_lfp = proc->block.lfp;
-	th->local_svar = Qnil;
+	th->root_lep = rb_vm_ep_local_ep(proc->block.ep);
+	th->root_svar = Qnil;
 
 	fib->status = RUNNING;
-	cont->value = rb_vm_invoke_proc(th, proc, proc->block.self, argc, argv, 0);
+	cont->value = rb_vm_invoke_proc(th, proc, argc, argv, 0);
     }
     TH_POP_TAG();
 
     if (state) {
-	if (state == TAG_RAISE) {
-	    th->thrown_errinfo = th->errinfo;
+	if (state == TAG_RAISE || state == TAG_FATAL) {
+	    rb_threadptr_pending_interrupt_enque(th, th->errinfo);
 	}
 	else {
-	    th->thrown_errinfo =
-	      rb_vm_make_jump_tag_but_local_jump(state, th->errinfo);
+	    VALUE err = rb_vm_make_jump_tag_but_local_jump(state, th->errinfo);
+	    if (!NIL_P(err))
+		rb_threadptr_pending_interrupt_enque(th, err);
 	}
 	RUBY_VM_SET_INTERRUPT(th);
     }
@@ -1216,10 +1256,10 @@ fiber_store(rb_fiber_t *next_fib)
 
 #if !FIBER_USE_NATIVE
     cont_save_machine_stack(th, &fib->cont);
+#endif
 
-    if (ruby_setjmp(fib->cont.jmpbuf)) {
-#else /* FIBER_USE_NATIVE */
-    {
+    if (FIBER_USE_NATIVE || ruby_setjmp(fib->cont.jmpbuf)) {
+#if FIBER_USE_NATIVE
 	fiber_setcontext(next_fib, fib);
 #ifndef _WIN32
 	if (terminated_machine_stack.ptr) {
@@ -1264,6 +1304,13 @@ fiber_switch(VALUE fibval, int argc, VALUE *argv, int is_resume)
     GetFiberPtr(fibval, fib);
     cont = &fib->cont;
 
+    if (th->fiber == fibval) {
+	/* ignore fiber context switch
+         * because destination fiber is same as current fiber
+	 */
+	return make_passing_arg(argc, argv);
+    }
+
     if (cont->saved_thread.self != th->self) {
 	rb_raise(rb_eFiberError, "fiber called across threads");
     }
@@ -1301,6 +1348,10 @@ fiber_switch(VALUE fibval, int argc, VALUE *argv, int is_resume)
     if (is_resume) {
 	fib->prev = rb_fiber_current();
     }
+    else {
+	/* restore `tracing' context. see [Feature #4347] */
+	th->trace_arg = cont->saved_thread.trace_arg;
+    }
 
     cont->argc = argc;
     cont->value = make_passing_arg(argc, argv);
@@ -1312,7 +1363,7 @@ fiber_switch(VALUE fibval, int argc, VALUE *argv, int is_resume)
 	rb_bug("rb_fiber_resume: unreachable");
     }
 #endif
-    RUBY_VM_CHECK_INTS();
+    RUBY_VM_CHECK_INTS(th);
 
     return value;
 }
@@ -1332,6 +1383,9 @@ rb_fiber_resume(VALUE fibval, int argc, VALUE *argv)
     if (fib->prev != Qnil || fib->cont.type == ROOT_FIBER_CONTEXT) {
 	rb_raise(rb_eFiberError, "double resume");
     }
+    if (fib->transfered != 0) {
+	rb_raise(rb_eFiberError, "cannot resume transferred Fiber");
+    }
 
     return fiber_switch(fibval, argc, argv, 1);
 }
@@ -1342,6 +1396,19 @@ rb_fiber_yield(int argc, VALUE *argv)
     return rb_fiber_transfer(return_fiber(), argc, argv);
 }
 
+void
+rb_fiber_reset_root_local_storage(VALUE thval)
+{
+    rb_thread_t *th;
+    rb_fiber_t	*fib;
+
+    GetThreadPtr(thval, th);
+    if (th->root_fiber && th->root_fiber != th->fiber) {
+	GetFiberPtr(th->root_fiber, fib);
+	th->local_storage = fib->cont.saved_thread.local_storage;
+    }
+}
+
 /*
  *  call-seq:
  *     fiber.alive? -> true or false
@@ -1397,11 +1464,41 @@ rb_fiber_m_resume(int argc, VALUE *argv, VALUE fib)
  *  You cannot resume a fiber that transferred control to another one.
  *  This will cause a double resume error. You need to transfer control
  *  back to this fiber before it can yield and resume.
+ *
+ *  Example:
+ *
+ *    fiber1 = Fiber.new do
+ *      puts "In Fiber 1"
+ *      Fiber.yield
+ *    end
+ *
+ *    fiber2 = Fiber.new do
+ *      puts "In Fiber 2"
+ *      fiber1.transfer
+ *      puts "Never see this message"
+ *    end
+ *
+ *    fiber3 = Fiber.new do
+ *      puts "In Fiber 3"
+ *    end
+ *
+ *    fiber2.resume
+ *    fiber3.resume
+ *
+ *    <em>produces</em>
+ *
+ *    In fiber 2
+ *    In fiber 1
+ *    In fiber 3
+ *
  */
 static VALUE
-rb_fiber_m_transfer(int argc, VALUE *argv, VALUE fib)
+rb_fiber_m_transfer(int argc, VALUE *argv, VALUE fibval)
 {
-    return rb_fiber_transfer(fib, argc, argv);
+    rb_fiber_t *fib;
+    GetFiberPtr(fibval, fib);
+    fib->transfered = 1;
+    return rb_fiber_transfer(fibval, argc, argv);
 }
 
 /*
