@@ -30,12 +30,7 @@
 # define NAMLEN(dirent) strlen((dirent)->d_name)
 #else
 # define dirent direct
-# if !defined __NeXT__
-#  define NAMLEN(dirent) (dirent)->d_namlen
-# else
-#  /* On some versions of NextStep, d_namlen is always zero, so avoid it. */
-#  define NAMLEN(dirent) strlen((dirent)->d_name)
-# endif
+# define NAMLEN(dirent) (dirent)->d_namlen
 # if HAVE_SYS_NDIR_H
 #  include <sys/ndir.h>
 # endif
@@ -49,6 +44,10 @@
 #  include "win32/dir.h"
 # endif
 #endif
+#if defined(__native_client__) && defined(NACL_NEWLIB)
+# include "nacl/dirent.h"
+# include "nacl/stat.h"
+#endif
 
 #include <errno.h>
 
@@ -86,6 +85,7 @@ char *strchr(char*,char);
 #define FNM_PATHNAME	0x02
 #define FNM_DOTMATCH	0x04
 #define FNM_CASEFOLD	0x08
+#define FNM_EXTGLOB	0x10
 #if CASEFOLD_FILESYSTEM
 #define FNM_SYSCASE	FNM_CASEFOLD
 #else
@@ -137,7 +137,7 @@ bracket(
 	    p = t2 + (r2 = rb_enc_mbclen(t2, pend, enc));
 	    if (ok) continue;
 	    if ((r <= (send-s) && memcmp(t1, s, r) == 0) ||
-		(r2 <= (send-s) && memcmp(t2, s, r) == 0)) {
+		(r2 <= (send-s) && memcmp(t2, s, r2) == 0)) {
 		ok = 1;
 		continue;
 	    }
@@ -397,7 +397,7 @@ dir_initialize(int argc, VALUE *argv, VALUE dir)
     }
     fsenc = rb_filesystem_encoding();
 
-    argc = rb_scan_args(argc, argv, "1:", &dirname, &opt);
+    rb_scan_args(argc, argv, "1:", &dirname, &opt);
 
     if (!NIL_P(opt)) {
 	VALUE enc = rb_hash_aref(opt, sym_enc);
@@ -522,7 +522,7 @@ dir_path(VALUE dir)
 #if defined HAVE_READDIR_R
 # define READDIR(dir, enc, entry, dp) (readdir_r((dir), (entry), &(dp)) == 0 && (dp) != 0)
 #elif defined _WIN32
-# define READDIR(dir, enc, entry, dp) (((dp) = rb_w32_readdir_with_enc((dir), (enc))) != 0)
+# define READDIR(dir, enc, entry, dp) (((dp) = rb_w32_readdir((dir), (enc))) != 0)
 #else
 # define READDIR(dir, enc, entry, dp) (((dp) = readdir(dir)) != 0)
 #endif
@@ -589,13 +589,10 @@ dir_read(VALUE dir)
     if (READDIR(dirp->dir, dirp->enc, &STRUCT_DIRENT(entry), dp)) {
 	return rb_external_str_new_with_enc(dp->d_name, NAMLEN(dp), dirp->enc);
     }
-    else if (errno == 0) {	/* end of stream */
-	return Qnil;
-    }
     else {
-	rb_sys_fail(0);
+	if (errno != 0) rb_sys_fail(0);
+	return Qnil;		/* end of stream */
     }
-    return Qnil;		/* not reached */
 }
 
 /*
@@ -692,6 +689,7 @@ dir_seek(VALUE dir, VALUE pos)
 #define dir_seek rb_f_notimplement
 #endif
 
+#ifdef HAVE_SEEKDIR
 /*
  *  call-seq:
  *     dir.pos( integer ) -> integer
@@ -712,6 +710,9 @@ dir_set_pos(VALUE dir, VALUE pos)
     dir_seek(dir, pos);
     return pos;
 }
+#else
+#define dir_set_pos rb_f_notimplement
+#endif
 
 /*
  *  call-seq:
@@ -1135,10 +1136,9 @@ find_dirsep(const char *p, const char *pend, int flags, rb_encoding *enc)
 }
 
 /* Remove escaping backslashes */
-static void
-remove_backslashes(char *p, rb_encoding *enc)
+static char *
+remove_backslashes(char *p, register const char *pend, rb_encoding *enc)
 {
-    register const char *pend = p + strlen(p);
     char *t = p;
     char *s = p;
 
@@ -1157,6 +1157,8 @@ remove_backslashes(char *p, rb_encoding *enc)
 
     if (t != s)
 	memmove(t, s, p - s); /* move '\0' too */
+
+    return p;
 }
 
 /* Globing pattern */
@@ -1175,6 +1177,7 @@ glob_make_pattern(const char *p, const char *e, int flags, rb_encoding *enc)
 {
     struct glob_pattern *list, *tmp, **tail = &list;
     int dirsep = 0; /* pattern is terminated with '/' */
+    int recursive = 0;
 
     while (p < e && *p) {
 	tmp = GLOB_ALLOC(struct glob_pattern);
@@ -1185,13 +1188,14 @@ glob_make_pattern(const char *p, const char *e, int flags, rb_encoding *enc)
 	    tmp->type = RECURSIVE;
 	    tmp->str = 0;
 	    dirsep = 1;
+	    recursive = 1;
 	}
 	else {
 	    const char *m = find_dirsep(p, e, flags, enc);
 	    int magic = has_magic(p, m, flags, enc);
 	    char *buf;
 
-	    if (!magic && *m) {
+	    if (!magic && !recursive && *m) {
 		const char *m2;
 		while (!has_magic(m+1, m2 = find_dirsep(m+1, e, flags, enc), flags, enc) &&
 		       *m2) {
@@ -1248,23 +1252,22 @@ glob_free_pattern(struct glob_pattern *list)
 }
 
 static char *
-join_path(const char *path, int dirsep, const char *name)
+join_path(const char *path, int dirsep, const char *name, size_t namlen)
 {
     long len = strlen(path);
-    long len2 = strlen(name)+(dirsep?1:0)+1;
-    char *buf = GLOB_ALLOC_N(char, len+len2);
+    char *buf = GLOB_ALLOC_N(char, len+namlen+(dirsep?1:0)+1);
 
     if (!buf) return 0;
     memcpy(buf, path, len);
     if (dirsep) {
 	buf[len++] = '/';
     }
-    buf[len] = '\0';
-    strlcat(buf+len, name, len2);
+    memcpy(buf+len, name, namlen);
+    buf[len+namlen] = '\0';
     return buf;
 }
 
-enum answer { YES, NO, UNKNOWN };
+enum answer {UNKNOWN = -1, NO, YES};
 
 #ifndef S_ISDIR
 #   define S_ISDIR(m) (((m) & S_IFMT) == S_IFDIR)
@@ -1365,7 +1368,7 @@ glob_helper(
 	    if (status) return status;
 	}
 	if (match_dir && isdir == YES) {
-	    char *tmp = join_path(path, dirsep, "");
+	    char *tmp = join_path(path, dirsep, "", 0);
 	    if (!tmp) return -1;
 	    status = glob_call_func(func, tmp, arg, enc);
 	    GLOB_FREE(tmp);
@@ -1383,15 +1386,30 @@ glob_helper(
 	if (dirp == NULL) return 0;
 
 	while (READDIR(dirp, enc, &STRUCT_DIRENT(entry), dp)) {
-	    char *buf = join_path(path, dirsep, dp->d_name);
+	    char *buf;
 	    enum answer new_isdir = UNKNOWN;
+	    int dotfile = 0;
 
+	    if (recursive && dp->d_name[0] == '.') {
+		++dotfile;
+		if (!dp->d_name[1]) {
+		    /* unless DOTMATCH, skip current directories not to recurse infinitely */
+		    if (!(flags & FNM_DOTMATCH)) continue;
+		    ++dotfile;
+		}
+		else if (dp->d_name[1] == '.' && !dp->d_name[2]) {
+		    /* always skip parent directories not to recurse infinitely */
+		    continue;
+		}
+	    }
+
+	    buf = join_path(path, dirsep, dp->d_name, NAMLEN(dp));
 	    if (!buf) {
 		status = -1;
 		break;
 	    }
-	    if (recursive && strcmp(dp->d_name, ".") != 0 && strcmp(dp->d_name, "..") != 0
-		&& fnmatch("*", rb_usascii_encoding(), dp->d_name, flags) == 0) {
+	    if (recursive && dotfile < ((flags & FNM_DOTMATCH) ? 2 : 1)) {
+		/* RECURSIVE never match dot files unless FNM_DOTMATCH is set */
 #ifndef _WIN32
 		if (do_lstat(buf, &st, flags) == 0)
 		    new_isdir = S_ISDIR(st.st_mode) ? YES : S_ISLNK(st.st_mode) ? UNKNOWN : NO;
@@ -1450,7 +1468,8 @@ glob_helper(
 		    break;
 		}
 		memcpy(name, (*cur)->str, len);
-		if (escape) remove_backslashes(name, enc);
+		if (escape)
+		    len = remove_backslashes(name, name+len-1, enc) - name;
 
 		new_beg = new_end = GLOB_ALLOC_N(struct glob_pattern *, end - beg);
 		if (!new_beg) {
@@ -1466,7 +1485,7 @@ glob_helper(
 		    }
 		}
 
-		buf = join_path(path, dirsep, name);
+		buf = join_path(path, dirsep, name, len);
 		GLOB_FREE(name);
 		if (!buf) {
 		    GLOB_FREE(new_beg);
@@ -1732,11 +1751,9 @@ dir_globs(long argc, VALUE *argv, int flags)
 
 /*
  *  call-seq:
- *     Dir[ array ]                 -> array
  *     Dir[ string [, string ...] ] -> array
  *
  *  Equivalent to calling
- *  <code>Dir.glob(</code><i>array,</i><code>0)</code> and
  *  <code>Dir.glob([</code><i>string,...</i><code>],0)</code>.
  *
  */
@@ -1775,7 +1792,7 @@ dir_s_aref(int argc, VALUE *argv, VALUE obj)
  *                          <code>/ .* /x</code> in regexp. Note, this
  *                          will not match Unix-like hidden files (dotfiles).
  *                          In order to include those in the match results,
- *                          you must use something like "{*,.*}".
+ *                          you must use something like <code>"{*,.*}"</code>.
  *  <code>**</code>::       Matches directories recursively.
  *  <code>?</code>::        Matches any one character. Equivalent to
  *                          <code>/.{1}/</code> in regexp.
@@ -1789,10 +1806,11 @@ dir_s_aref(int argc, VALUE *argv, VALUE obj)
  *                          More than two literals may be specified.
  *                          Equivalent to pattern alternation in
  *                          regexp.
- *  <code>\</code>::        Escapes the next metacharacter.
- *                          Note that this means you cannot use backslash in windows
- *                          as part of a glob, i.e. Dir["c:\\foo*"] will not work
- *                          use Dir["c:/foo*"] instead
+ *  <code> \\ </code>::     Escapes the next metacharacter.
+ *                          Note that this means you cannot use backslash
+ *                          in windows as part of a glob,
+ *                          i.e. <code>Dir["c:\\foo*"]</code> will not work,
+ *                          use <code>Dir["c:/foo*"]</code> instead.
  *
  *     Dir["config.?"]                     #=> ["config.h"]
  *     Dir.glob("config.?")                #=> ["config.h"]
@@ -1847,9 +1865,8 @@ static VALUE
 dir_open_dir(int argc, VALUE *argv)
 {
     VALUE dir = rb_funcall2(rb_cDir, rb_intern("open"), argc, argv);
-    struct dir_data *dirp;
 
-    TypedData_Get_Struct(dir, struct dir_data, &dir_data_type, dirp);
+    rb_check_typeddata(dir, &dir_data_type);
     return dir;
 }
 
@@ -1905,6 +1922,32 @@ dir_entries(int argc, VALUE *argv, VALUE io)
     return rb_ensure(rb_Array, dir, dir_close, dir);
 }
 
+static int
+fnmatch_brace(const char *pattern, VALUE val, void *enc)
+{
+    struct brace_args *arg = (struct brace_args *)val;
+    VALUE path = arg->value;
+    rb_encoding *enc_pattern = enc;
+    rb_encoding *enc_path = rb_enc_get(path);
+
+    if (enc_pattern != enc_path) {
+	if (!rb_enc_asciicompat(enc_pattern))
+	    return FNM_NOMATCH;
+	if (!rb_enc_asciicompat(enc_path))
+	    return FNM_NOMATCH;
+	if (!rb_enc_str_asciionly_p(path)) {
+	    int cr = ENC_CODERANGE_7BIT;
+	    long len = strlen(pattern);
+	    if (rb_str_coderange_scan_restartable(pattern, pattern + len,
+						  enc_pattern, &cr) != len)
+		return FNM_NOMATCH;
+	    if (cr != ENC_CODERANGE_7BIT)
+		return FNM_NOMATCH;
+	}
+    }
+    return (fnmatch(pattern, enc, RSTRING_PTR(path), arg->flags) == 0);
+}
+
 /*
  *  call-seq:
  *     File.fnmatch( pattern, path, [flags] ) -> (true or false)
@@ -1933,7 +1976,7 @@ dir_entries(int argc, VALUE *argv, VALUE io)
  *                          Behaves exactly like character sets in
  *                          Regexp, including set negation
  *                          (<code>[^a-z]</code>).
- *  <code>\</code>::        Escapes the next metacharacter.
+ *  <code> \ </code>::      Escapes the next metacharacter.
  *
  *  <i>flags</i> is a bitwise OR of the <code>FNM_xxx</code>
  *  parameters. The same glob pattern and flags are used by
@@ -1959,7 +2002,7 @@ dir_entries(int argc, VALUE *argv, VALUE io)
  *
  *     File.fnmatch('\?',   '?')                       #=> true  # escaped wildcard becomes ordinary
  *     File.fnmatch('\a',   'a')                       #=> true  # escaped ordinary remains ordinary
- *     File.fnmatch('\a',   '\a', File::FNM_NOESCAPE)  #=> true  # FNM_NOESACPE makes '\' ordinary
+ *     File.fnmatch('\a',   '\a', File::FNM_NOESCAPE)  #=> true  # FNM_NOESCAPE makes '\' ordinary
  *     File.fnmatch('[\?]', '?')                       #=> true  # can escape inside bracket expression
  *
  *     File.fnmatch('*',   '.profile')                      #=> false # wildcard doesn't match leading
@@ -2001,9 +2044,22 @@ file_s_fnmatch(int argc, VALUE *argv, VALUE obj)
     StringValue(pattern);
     FilePathStringValue(path);
 
-    if (fnmatch(RSTRING_PTR(pattern), rb_enc_get(pattern), RSTRING_PTR(path),
-		flags) == 0)
-	return Qtrue;
+    if (flags & FNM_EXTGLOB) {
+	struct brace_args args;
+
+	args.value = path;
+	args.flags = flags;
+	if (ruby_brace_expand(RSTRING_PTR(pattern), flags, fnmatch_brace,
+			      (VALUE)&args, rb_enc_get(pattern)) > 0)
+	    return Qtrue;
+    }
+    else {
+	rb_encoding *enc = rb_enc_compatible(pattern, path);
+	if (!enc) return Qfalse;
+	if (fnmatch(RSTRING_PTR(pattern), enc, RSTRING_PTR(path), flags) == 0)
+	    return Qtrue;
+    }
+    RB_GC_GUARD(pattern);
 
     return Qfalse;
 }
@@ -2030,6 +2086,22 @@ dir_s_home(int argc, VALUE *argv, VALUE obj)
     return rb_home_dir(u, rb_str_new(0, 0));
 }
 
+#if 0
+/*
+ * call-seq:
+ *   Dir.exist?(file_name)   ->  true or false
+ *   Dir.exists?(file_name)   ->  true or false
+ *
+ * Returns <code>true</code> if the named file is a directory,
+ * <code>false</code> otherwise.
+ *
+ */
+VALUE
+rb_file_directory_p()
+{
+}
+#endif
+
 /*
  *  Objects of class <code>Dir</code> are directory streams representing
  *  directories in the underlying file system. They provide a variety of
@@ -2078,8 +2150,8 @@ Init_Dir(void)
 
     rb_define_singleton_method(rb_cDir,"glob", dir_s_glob, -1);
     rb_define_singleton_method(rb_cDir,"[]", dir_s_aref, -1);
-    rb_define_singleton_method(rb_cDir,"exist?", rb_file_directory_p, 1); /* in file.c */
-    rb_define_singleton_method(rb_cDir,"exists?", rb_file_directory_p, 1); /* in file.c */
+    rb_define_singleton_method(rb_cDir,"exist?", rb_file_directory_p, 1);
+    rb_define_singleton_method(rb_cDir,"exists?", rb_file_directory_p, 1);
 
     rb_define_singleton_method(rb_cFile,"fnmatch", file_s_fnmatch, -1);
     rb_define_singleton_method(rb_cFile,"fnmatch?", file_s_fnmatch, -1);
@@ -2088,5 +2160,6 @@ Init_Dir(void)
     rb_file_const("FNM_PATHNAME", INT2FIX(FNM_PATHNAME));
     rb_file_const("FNM_DOTMATCH", INT2FIX(FNM_DOTMATCH));
     rb_file_const("FNM_CASEFOLD", INT2FIX(FNM_CASEFOLD));
+    rb_file_const("FNM_EXTGLOB", INT2FIX(FNM_EXTGLOB));
     rb_file_const("FNM_SYSCASE", INT2FIX(FNM_SYSCASE));
 }
