@@ -1,6 +1,9 @@
 ENC_DEFINE("ASCII-8BIT");
 ENC_DEFINE("Big5");
+ENC_DEFINE("Big5-HKSCS");
+ENC_DEFINE("Big5-UAO");
 ENC_DEFINE("CP949");
+ENC_DEFINE("Emacs-Mule");
 ENC_DEFINE("EUC-JP");
 ENC_DEFINE("EUC-KR");
 ENC_DEFINE("EUC-TW");
@@ -30,6 +33,7 @@ ENC_DEFINE("UTF-16BE");
 ENC_DEFINE("UTF-16LE");
 ENC_DEFINE("UTF-32BE");
 ENC_DEFINE("UTF-32LE");
+ENC_DEFINE("Windows-31J");
 ENC_DEFINE("Windows-1251");
 ENC_ALIAS("BINARY", "ASCII-8BIT");
 ENC_REPLICATE("IBM437", "ASCII-8BIT");
@@ -75,11 +79,18 @@ ENC_REPLICATE("macRomania", "ASCII-8BIT");
 ENC_REPLICATE("macThai", "ASCII-8BIT");
 ENC_REPLICATE("macTurkish", "ASCII-8BIT");
 ENC_REPLICATE("macUkraine", "ASCII-8BIT");
-ENC_ALIAS("CP950", "BIG5");
+ENC_REPLICATE("CP950", "Big5");
+ENC_SET_BASE("Big5-HKSCS", "Big5");
+ENC_ALIAS("Big5-HKSCS:2008", "Big5-HKSCS");
+ENC_REPLICATE("CP951", "Big5-HKSCS");
+ENC_SET_BASE("Big5-UAO", "Big5");
+ENC_REPLICATE("stateless-ISO-2022-JP", "Emacs-Mule");
 ENC_ALIAS("eucJP", "EUC-JP") /* UI-OSF Application Platform Profile for Japanese Environment Version 1.1 */;
 ENC_REPLICATE("eucJP-ms", "EUC-JP") /* TOG/JVC CDE/Motif Technical WG */;
-ENC_ALIAS("euc-jp-ms", "EUC-JP");
+ENC_ALIAS("euc-jp-ms", "eucJP-ms");
 ENC_REPLICATE("CP51932", "EUC-JP");
+ENC_REPLICATE("EUC-JP-2004", "EUC-JP") /* defined at JIS X 0213:2004 */;
+ENC_ALIAS("EUC-JISX0213", "EUC-JP-2004") /* defined at JIS X 0213:2000, and obsolete at JIS X 0213:2004 */;
 ENC_ALIAS("eucKR", "EUC-KR");
 ENC_ALIAS("eucTW", "EUC-TW");
 ENC_ALIAS("EUC-CN", "GB2312");
@@ -90,6 +101,8 @@ ENC_DUMMY("ISO-2022-JP");
 ENC_ALIAS("ISO2022-JP", "ISO-2022-JP");
 ENC_REPLICATE("ISO-2022-JP-2", "ISO-2022-JP");
 ENC_ALIAS("ISO2022-JP2", "ISO-2022-JP-2");
+ENC_REPLICATE("CP50220", "ISO-2022-JP");
+ENC_REPLICATE("CP50221", "ISO-2022-JP");
 ENC_ALIAS("ISO8859-1", "ISO-8859-1");
 ENC_REPLICATE("Windows-1252", "ISO-8859-1");
 ENC_ALIAS("CP1252", "Windows-1252");
@@ -123,10 +136,6 @@ ENC_ALIAS("ISO8859-14", "ISO-8859-14");
 ENC_ALIAS("ISO8859-15", "ISO-8859-15");
 ENC_ALIAS("ISO8859-16", "ISO-8859-16");
 ENC_ALIAS("CP878", "KOI8-R");
-ENC_ALIAS("SJIS", "Shift_JIS");
-ENC_REPLICATE("Windows-31J", "Shift_JIS");
-ENC_ALIAS("CP932", "Windows-31J");
-ENC_ALIAS("csWindows31J", "Windows-31J") /* IANA.  IE6 don't accept Windows-31J but csWindows31J. */;
 ENC_REPLICATE("MacJapanese", "Shift_JIS");
 ENC_ALIAS("MacJapan", "MacJapanese");
 ENC_ALIAS("ASCII", "US-ASCII");
@@ -137,9 +146,24 @@ ENC_ALIAS("CP65000", "UTF-7");
 ENC_ALIAS("CP65001", "UTF-8");
 ENC_REPLICATE("UTF8-MAC", "UTF-8");
 ENC_ALIAS("UTF-8-MAC", "UTF8-MAC");
+ENC_ALIAS("UTF-8-HFS", "UTF8-MAC") /* Emacs 23.2 */;
+ENC_DUMMY_UNICODE("UTF-16");
+ENC_DUMMY_UNICODE("UTF-32");
 ENC_ALIAS("UCS-2BE", "UTF-16BE");
 ENC_ALIAS("UCS-4BE", "UTF-32BE");
 ENC_ALIAS("UCS-4LE", "UTF-32LE");
+ENC_ALIAS("CP932", "Windows-31J");
+ENC_ALIAS("csWindows31J", "Windows-31J") /* IANA.  IE6 don't accept Windows-31J but csWindows31J. */;
+ENC_ALIAS("SJIS", "Windows-31J");
+ENC_ALIAS("PCK", "Windows-31J");
 ENC_ALIAS("CP1251", "Windows-1251");
+ENC_REPLICATE("UTF8-DoCoMo", "UTF-8");
+ENC_REPLICATE("SJIS-DoCoMo", "Windows-31J");
+ENC_REPLICATE("UTF8-KDDI", "UTF-8");
+ENC_REPLICATE("SJIS-KDDI", "Windows-31J");
+ENC_REPLICATE("ISO-2022-JP-KDDI", "ISO-2022-JP");
+ENC_REPLICATE("stateless-ISO-2022-JP-KDDI", "stateless-ISO-2022-JP");
+ENC_REPLICATE("UTF8-SoftBank", "UTF-8");
+ENC_REPLICATE("SJIS-SoftBank", "Windows-31J");
 
-#define ENCODING_COUNT 81
+#define ENCODING_COUNT 100
