@@ -2,7 +2,7 @@
 
   encoding.c -
 
-  $Author: yugui $
+  $Author: usa $
   created at: Thu May 24 17:23:27 JST 2007
 
   Copyright (C) 2007 Yukihiro Matsumoto
@@ -11,6 +11,7 @@
 
 #include "ruby/ruby.h"
 #include "ruby/encoding.h"
+#include "internal.h"
 #include "regenc.h"
 #include <ctype.h>
 #ifndef NO_LOCALE_CHARMAP
@@ -23,6 +24,18 @@
 #endif
 #include "ruby/util.h"
 
+#if defined __GNUC__ && __GNUC__ >= 4
+#pragma GCC visibility push(default)
+int rb_enc_register(const char *name, rb_encoding *encoding);
+void rb_enc_set_base(const char *name, const char *orig);
+void rb_encdb_declare(const char *name);
+int rb_encdb_replicate(const char *name, const char *orig);
+int rb_encdb_dummy(const char *name);
+int rb_encdb_alias(const char *alias, const char *orig);
+void rb_encdb_set_unicode(int index);
+#pragma GCC visibility pop
+#endif
+
 static ID id_encoding;
 VALUE rb_cEncoding;
 static VALUE rb_encoding_list;
@@ -45,39 +58,54 @@ void rb_enc_init(void);
 #define ENCODING_COUNT ENCINDEX_BUILTIN_MAX
 #define UNSPECIFIED_ENCODING INT_MAX
 
+#define ENCODING_NAMELEN_MAX 63
+#define valid_encoding_name_p(name) ((name) && strlen(name) <= ENCODING_NAMELEN_MAX)
+
 #define enc_autoload_p(enc) (!rb_enc_mbmaxlen(enc))
 
 static int load_encoding(const char *name);
 
-static void
-enc_mark(void *ptr)
+static size_t
+enc_memsize(const void *p)
 {
+    return 0;
 }
 
+static const rb_data_type_t encoding_data_type = {
+    "encoding",
+    {0, 0, enc_memsize,},
+};
+
+#define is_data_encoding(obj) (RTYPEDDATA_P(obj) && RTYPEDDATA_TYPE(obj) == &encoding_data_type)
+
 static VALUE
 enc_new(rb_encoding *encoding)
 {
-    return Data_Wrap_Struct(rb_cEncoding, enc_mark, 0, encoding);
+    return TypedData_Wrap_Struct(rb_cEncoding, &encoding_data_type, encoding);
+}
+
+static VALUE
+rb_enc_from_encoding_index(int idx)
+{
+    VALUE list, enc;
+
+    if (!(list = rb_encoding_list)) {
+	rb_bug("rb_enc_from_encoding_index(%d): no rb_encoding_list", idx);
+    }
+    enc = rb_ary_entry(list, idx);
+    if (NIL_P(enc)) {
+	rb_bug("rb_enc_from_encoding_index(%d): not created yet", idx);
+    }
+    return enc;
 }
 
 VALUE
 rb_enc_from_encoding(rb_encoding *encoding)
 {
-    VALUE list, enc;
     int idx;
-
     if (!encoding) return Qnil;
     idx = ENC_TO_ENCINDEX(encoding);
-    if (!(list = rb_encoding_list)) {
-	rb_bug("rb_enc_from_encoding(%d\"%s\"): no rb_encoding_list",
-	       idx, rb_enc_name(encoding));
-    }
-    enc = rb_ary_entry(list, idx);
-    if (NIL_P(enc)) {
-	rb_bug("rb_enc_from_encoding(%d\"%s\"): not created yet",
-	       idx, rb_enc_name(encoding));
-    }
-    return enc;
+    return rb_enc_from_encoding_index(idx);
 }
 
 static int enc_autoload(rb_encoding *);
@@ -97,8 +125,7 @@ check_encoding(rb_encoding *enc)
 static int
 enc_check_encoding(VALUE obj)
 {
-    if (SPECIAL_CONST_P(obj) || BUILTIN_TYPE(obj) != T_DATA ||
-	RDATA(obj)->dmark != enc_mark) {
+    if (SPECIAL_CONST_P(obj) || !rb_typeddata_is_kind_of(obj, &encoding_data_type)) {
 	return -1;
     }
     return check_encoding(RDATA(obj)->data);
@@ -133,8 +160,9 @@ rb_to_encoding_index(VALUE enc)
     return rb_enc_find_index(StringValueCStr(enc));
 }
 
-static rb_encoding *
-to_encoding(VALUE enc)
+/* Returns encoding index or UNSPECIFIED_ENCODING */
+static int
+str_find_encindex(VALUE enc)
 {
     int idx;
 
@@ -143,17 +171,40 @@ to_encoding(VALUE enc)
 	rb_raise(rb_eArgError, "invalid name encoding (non ASCII)");
     }
     idx = rb_enc_find_index(StringValueCStr(enc));
+    return idx;
+}
+
+static int
+str_to_encindex(VALUE enc)
+{
+    int idx = str_find_encindex(enc);
     if (idx < 0) {
 	rb_raise(rb_eArgError, "unknown encoding name - %s", RSTRING_PTR(enc));
     }
-    return rb_enc_from_index(idx);
+    return idx;
+}
+
+static rb_encoding *
+str_to_encoding(VALUE enc)
+{
+    return rb_enc_from_index(str_to_encindex(enc));
 }
 
 rb_encoding *
 rb_to_encoding(VALUE enc)
 {
     if (enc_check_encoding(enc) >= 0) return RDATA(enc)->data;
-    return to_encoding(enc);
+    return str_to_encoding(enc);
+}
+
+rb_encoding *
+rb_find_encoding(VALUE enc)
+{
+    int idx;
+    if (enc_check_encoding(enc) >= 0) return RDATA(enc)->data;
+    idx = str_find_encindex(enc);
+    if (idx < 0) return NULL;
+    return rb_enc_from_index(idx);
 }
 
 void
@@ -183,6 +234,7 @@ enc_register_at(int index, const char *name, rb_encoding *encoding)
     struct rb_encoding_entry *ent = &enc_table.list[index];
     VALUE list;
 
+    if (!valid_encoding_name_p(name)) return -1;
     if (!ent->name) {
 	ent->name = name = strdup(name);
     }
@@ -275,6 +327,18 @@ set_base_encoding(int index, rb_encoding *base)
     return enc;
 }
 
+/* for encdb.h
+ * Set base encoding for encodings which are not replicas
+ * but not in their own files.
+ */
+void
+rb_enc_set_base(const char *name, const char *orig)
+{
+    int idx = rb_enc_registered(name);
+    int origidx = rb_enc_registered(orig);
+    set_base_encoding(idx, rb_enc_from_index(origidx));
+}
+
 int
 rb_enc_replicate(const char *name, rb_encoding *encoding)
 {
@@ -287,8 +351,25 @@ rb_enc_replicate(const char *name, rb_encoding *encoding)
     return idx;
 }
 
+/*
+ * call-seq:
+ *   enc.replicate(name) -> encoding
+ *
+ * Returns a replicated encoding of _enc_ whose name is _name_.
+ * The new encoding should have the same byte structure of _enc_.
+ * If _name_ is used by another encoding, raise ArgumentError.
+ *
+ */
+static VALUE
+enc_replicate(VALUE encoding, VALUE name)
+{
+    return rb_enc_from_encoding_index(
+	rb_enc_replicate(StringValueCStr(name),
+			 rb_to_encoding(encoding)));
+}
+
 static int
-enc_replicate(int idx, const char *name, rb_encoding *origenc)
+enc_replicate_with_index(const char *name, rb_encoding *origenc, int idx)
 {
     if (idx < 0) {
 	idx = enc_register(name, origenc);
@@ -312,7 +393,7 @@ rb_encdb_replicate(const char *name, const char *orig)
     if (origidx < 0) {
 	origidx = enc_register(orig, 0);
     }
-    return enc_replicate(idx, name, rb_enc_from_index(origidx));
+    return enc_replicate_with_index(name, rb_enc_from_index(origidx), idx);
 }
 
 int
@@ -328,8 +409,8 @@ rb_define_dummy_encoding(const char *name)
 int
 rb_encdb_dummy(const char *name)
 {
-    int index = enc_replicate(rb_enc_registered(name), name,
-			      rb_ascii8bit_encoding());
+    int index = enc_replicate_with_index(name, rb_ascii8bit_encoding(),
+					 rb_enc_registered(name));
     rb_encoding *enc = enc_table.list[index].enc;
 
     ENC_SET_DUMMY(enc);
@@ -338,7 +419,7 @@ rb_encdb_dummy(const char *name)
 
 /*
  * call-seq:
- *   enc.dummy? => true or false
+ *   enc.dummy? -> true or false
  *
  * Returns true for dummy encodings.
  * A dummy encoding is an encoding for which character handling is not properly
@@ -355,19 +436,54 @@ enc_dummy_p(VALUE enc)
     return ENC_DUMMY_P(enc_table.list[must_encoding(enc)].enc) ? Qtrue : Qfalse;
 }
 
-static const char *
+/*
+ * call-seq:
+ *   enc.ascii_compatible? -> true or false
+ *
+ * Returns whether ASCII-compatible or not.
+ *
+ *   Encoding::UTF_8.ascii_compatible?     #=> true
+ *   Encoding::UTF_16BE.ascii_compatible?  #=> false
+ *
+ */
+static VALUE
+enc_ascii_compatible_p(VALUE enc)
+{
+    return rb_enc_asciicompat(enc_table.list[must_encoding(enc)].enc) ? Qtrue : Qfalse;
+}
+
+/*
+ * Returns 1 when the encoding is Unicode series other than UTF-7 else 0.
+ */
+int
+rb_enc_unicode_p(rb_encoding *enc)
+{
+    return ONIGENC_IS_UNICODE(enc);
+}
+
+static st_data_t
+enc_dup_name(st_data_t name)
+{
+    return (st_data_t)strdup((const char *)name);
+}
+
+/*
+ * Returns copied alias name when the key is added for st_table,
+ * else returns NULL.
+ */
+static int
 enc_alias_internal(const char *alias, int idx)
 {
-    alias = strdup(alias);
-    st_insert(enc_table.names, (st_data_t)alias, (st_data_t)idx);
-    return alias;
+    return st_insert2(enc_table.names, (st_data_t)alias, (st_data_t)idx,
+		      enc_dup_name);
 }
 
 static int
 enc_alias(const char *alias, int idx)
 {
-    alias = enc_alias_internal(alias, idx);
-    set_encoding_const(alias, rb_enc_from_index(idx));
+    if (!valid_encoding_name_p(alias)) return -1;
+    if (!enc_alias_internal(alias, idx))
+	set_encoding_const(alias, rb_enc_from_index(idx));
     return idx;
 }
 
@@ -397,6 +513,12 @@ rb_encdb_alias(const char *alias, const char *orig)
     return enc_alias(alias, idx);
 }
 
+void
+rb_encdb_set_unicode(int index)
+{
+    rb_enc_from_index(index)->flags |= ONIGENC_FLAG_UNICODE;
+}
+
 enum {
     ENCINDEX_ASCII,
     ENCINDEX_UTF_8,
@@ -450,31 +572,35 @@ rb_enc_registered(const char *name)
 static VALUE
 require_enc(VALUE enclib)
 {
-    return rb_require_safe(enclib, rb_safe_level());
+    int safe = rb_safe_level();
+    return rb_require_safe(enclib, safe > 3 ? 3 : safe);
 }
 
 static int
 load_encoding(const char *name)
 {
-    VALUE enclib = rb_sprintf("enc/%s", name);
+    VALUE enclib = rb_sprintf("enc/%s.so", name);
     VALUE verbose = ruby_verbose;
     VALUE debug = ruby_debug;
+    VALUE errinfo;
     VALUE loaded;
-    char *s = RSTRING_PTR(enclib) + 4, *e = RSTRING_END(enclib);
+    char *s = RSTRING_PTR(enclib) + 4, *e = RSTRING_END(enclib) - 3;
     int idx;
 
     while (s < e) {
 	if (!ISALNUM(*s)) *s = '_';
-	else if (ISUPPER(*s)) *s = TOLOWER(*s);
+	else if (ISUPPER(*s)) *s = (char)TOLOWER(*s);
 	++s;
     }
+    FL_UNSET(enclib, FL_TAINT|FL_UNTRUSTED);
     OBJ_FREEZE(enclib);
     ruby_verbose = Qfalse;
     ruby_debug = Qfalse;
+    errinfo = rb_errinfo();
     loaded = rb_protect(require_enc, enclib, 0);
     ruby_verbose = verbose;
     ruby_debug = debug;
-    rb_set_errinfo(Qnil);
+    rb_set_errinfo(errinfo);
     if (NIL_P(loaded)) return -1;
     if ((idx = rb_enc_registered(name)) < 0) return -1;
     if (enc_autoload_p(enc_table.list[idx].enc)) return -1;
@@ -504,6 +630,7 @@ enc_autoload(rb_encoding *enc)
     return i;
 }
 
+/* Return encoding index or UNSPECIFIED_ENCODING from encoding name */
 int
 rb_enc_find_index(const char *name)
 {
@@ -539,16 +666,16 @@ rb_enc_find(const char *name)
 static inline int
 enc_capable(VALUE obj)
 {
-    if (SPECIAL_CONST_P(obj)) return Qfalse;
+    if (SPECIAL_CONST_P(obj)) return SYMBOL_P(obj);
     switch (BUILTIN_TYPE(obj)) {
       case T_STRING:
       case T_REGEXP:
       case T_FILE:
-	return Qtrue;
+	return TRUE;
       case T_DATA:
-	if (RDATA(obj)->dmark == enc_mark) return Qtrue;
+	if (is_data_encoding(obj)) return TRUE;
       default:
-	return Qfalse;
+	return FALSE;
     }
 }
 
@@ -565,34 +692,42 @@ rb_enc_get_index(VALUE obj)
     int i = -1;
     VALUE tmp;
 
+    if (SPECIAL_CONST_P(obj)) {
+	if (!SYMBOL_P(obj)) return -1;
+	obj = rb_id2str(SYM2ID(obj));
+    }
     switch (BUILTIN_TYPE(obj)) {
-        default:
-	case T_STRING:
-	case T_REGEXP:
-	    i = ENCODING_GET_INLINED(obj);
-	    if (i == ENCODING_INLINE_MAX) {
-		VALUE iv;
+      as_default:
+      default:
+      case T_STRING:
+      case T_REGEXP:
+	i = ENCODING_GET_INLINED(obj);
+	if (i == ENCODING_INLINE_MAX) {
+	    VALUE iv;
 
-		iv = rb_ivar_get(obj, rb_id_encoding());
-		i = NUM2INT(iv);
-	    }
-	    break;
-	case T_FILE:
-	    tmp = rb_funcall(obj, rb_intern("internal_encoding"), 0, 0);
-	    if (NIL_P(tmp)) obj = rb_funcall(obj, rb_intern("external_encoding"), 0, 0);
-	    else obj = tmp;
-	    if (NIL_P(obj)) break;
-	case T_DATA:
-	    if (RDATA(obj)->dmark == enc_mark) {
-		i = enc_check_encoding(obj);
-	    }
-	    break;
+	    iv = rb_ivar_get(obj, rb_id_encoding());
+	    i = NUM2INT(iv);
+	}
+	break;
+      case T_FILE:
+	tmp = rb_funcall(obj, rb_intern("internal_encoding"), 0, 0);
+	if (NIL_P(tmp)) obj = rb_funcall(obj, rb_intern("external_encoding"), 0, 0);
+	else obj = tmp;
+	if (NIL_P(obj)) break;
+      case T_DATA:
+	if (is_data_encoding(obj)) {
+	    i = enc_check_encoding(obj);
+	}
+	else {
+	    goto as_default;
+	}
+	break;
     }
     return i;
 }
 
-void
-rb_enc_set_index(VALUE obj, int idx)
+static void
+enc_set_index(VALUE obj, int idx)
 {
     if (idx < ENCODING_INLINE_MAX) {
 	ENCODING_SET_INLINED(obj, idx);
@@ -600,20 +735,30 @@ rb_enc_set_index(VALUE obj, int idx)
     }
     ENCODING_SET_INLINED(obj, ENCODING_INLINE_MAX);
     rb_ivar_set(obj, rb_id_encoding(), INT2NUM(idx));
-    return;
+}
+
+void
+rb_enc_set_index(VALUE obj, int idx)
+{
+    rb_check_frozen(obj);
+    enc_set_index(obj, idx);
 }
 
 VALUE
 rb_enc_associate_index(VALUE obj, int idx)
 {
 /*    enc_check_capable(obj);*/
+    rb_check_frozen(obj);
     if (rb_enc_get_index(obj) == idx)
-    	return obj;
+	return obj;
+    if (SPECIAL_CONST_P(obj)) {
+	rb_raise(rb_eArgError, "cannot set encoding");
+    }
     if (!ENC_CODERANGE_ASCIIONLY(obj) ||
 	!rb_enc_asciicompat(rb_enc_from_index(idx))) {
 	ENC_CODERANGE_CLEAR(obj);
     }
-    rb_enc_set_index(obj, idx);
+    enc_set_index(obj, idx);
     return obj;
 }
 
@@ -645,6 +790,7 @@ rb_enc_compatible(VALUE str1, VALUE str2)
 {
     int idx1, idx2;
     rb_encoding *enc1, *enc2;
+    int isstr1, isstr2;
 
     idx1 = rb_enc_get_index(str1);
     idx2 = rb_enc_get_index(str2);
@@ -658,33 +804,38 @@ rb_enc_compatible(VALUE str1, VALUE str2)
     enc1 = rb_enc_from_index(idx1);
     enc2 = rb_enc_from_index(idx2);
 
-    if (TYPE(str2) == T_STRING && RSTRING_LEN(str2) == 0)
+    isstr2 = RB_TYPE_P(str2, T_STRING);
+    if (isstr2 && RSTRING_LEN(str2) == 0)
 	return enc1;
-    if (TYPE(str1) == T_STRING && RSTRING_LEN(str1) == 0)
-	return enc2;
+    isstr1 = RB_TYPE_P(str1, T_STRING);
+    if (isstr1 && RSTRING_LEN(str1) == 0)
+	return (rb_enc_asciicompat(enc1) && rb_enc_str_asciionly_p(str2)) ? enc1 : enc2;
     if (!rb_enc_asciicompat(enc1) || !rb_enc_asciicompat(enc2)) {
 	return 0;
     }
 
     /* objects whose encoding is the same of contents */
-    if (BUILTIN_TYPE(str2) != T_STRING && idx2 == ENCINDEX_US_ASCII)
+    if (!isstr2 && idx2 == ENCINDEX_US_ASCII)
 	return enc1;
-    if (BUILTIN_TYPE(str1) != T_STRING && idx1 == ENCINDEX_US_ASCII)
+    if (!isstr1 && idx1 == ENCINDEX_US_ASCII)
 	return enc2;
 
-    if (BUILTIN_TYPE(str1) != T_STRING) {
+    if (!isstr1) {
 	VALUE tmp = str1;
 	int idx0 = idx1;
 	str1 = str2;
 	str2 = tmp;
 	idx1 = idx2;
 	idx2 = idx0;
+	idx0 = isstr1;
+	isstr1 = isstr2;
+	isstr2 = idx0;
     }
-    if (BUILTIN_TYPE(str1) == T_STRING) {
+    if (isstr1) {
 	int cr1, cr2;
 
 	cr1 = rb_enc_str_coderange(str1);
-	if (BUILTIN_TYPE(str2) == T_STRING) {
+	if (isstr2) {
 	    cr2 = rb_enc_str_coderange(str2);
 	    if (cr1 != cr2) {
 		/* may need to handle ENC_CODERANGE_BROKEN */
@@ -692,7 +843,6 @@ rb_enc_compatible(VALUE str1, VALUE str2)
 		if (cr2 == ENC_CODERANGE_7BIT) return enc1;
 	    }
 	    if (cr2 == ENC_CODERANGE_7BIT) {
-		if (idx1 == ENCINDEX_ASCII) return enc2;
 		return enc1;
 	    }
 	}
@@ -711,7 +861,7 @@ rb_enc_copy(VALUE obj1, VALUE obj2)
 
 /*
  *  call-seq:
- *     obj.encoding   => encoding
+ *     obj.encoding   -> encoding
  *
  *  Returns the Encoding object that represents the encoding of obj.
  */
@@ -719,11 +869,17 @@ rb_enc_copy(VALUE obj1, VALUE obj2)
 VALUE
 rb_obj_encoding(VALUE obj)
 {
-    rb_encoding *enc = rb_enc_get(obj);
-    if (!enc) {
+    int idx = rb_enc_get_index(obj);
+    if (idx < 0) {
 	rb_raise(rb_eTypeError, "unknown encoding");
     }
-    return rb_enc_from_encoding(enc);
+    return rb_enc_from_encoding_index(idx);
+}
+
+int
+rb_enc_fast_mbclen(const char *p, const char *e, rb_encoding *enc)
+{
+    return ONIGENC_MBC_ENC_LEN(enc, (UChar*)p, (UChar*)e);
 }
 
 int
@@ -734,7 +890,7 @@ rb_enc_mbclen(const char *p, const char *e, rb_encoding *enc)
         return MBCLEN_CHARFOUND_LEN(n);
     else {
         int min = rb_enc_mbminlen(enc);
-        return min <= e-p ? min : e-p;
+        return min <= e-p ? min : (int)(e-p);
     }
 }
 
@@ -746,7 +902,7 @@ rb_enc_precise_mbclen(const char *p, const char *e, rb_encoding *enc)
         return ONIGENC_CONSTRUCT_MBCLEN_NEEDMORE(1);
     n = ONIGENC_PRECISE_MBC_ENC_LEN(enc, (UChar*)p, (UChar*)e);
     if (e-p < n)
-        return ONIGENC_CONSTRUCT_MBCLEN_NEEDMORE(n-(e-p));
+        return ONIGENC_CONSTRUCT_MBCLEN_NEEDMORE(n-(int)(e-p));
     return n;
 }
 
@@ -774,16 +930,24 @@ rb_enc_ascget(const char *p, const char *e, int *len, rb_encoding *enc)
 }
 
 unsigned int
-rb_enc_codepoint(const char *p, const char *e, rb_encoding *enc)
+rb_enc_codepoint_len(const char *p, const char *e, int *len_p, rb_encoding *enc)
 {
     int r;
     if (e <= p)
         rb_raise(rb_eArgError, "empty string");
     r = rb_enc_precise_mbclen(p, e, enc);
-    if (MBCLEN_CHARFOUND_P(r))
-        return rb_enc_mbc_to_codepoint(p, e, enc);
-    else
+    if (!MBCLEN_CHARFOUND_P(r)) {
 	rb_raise(rb_eArgError, "invalid byte sequence in %s", rb_enc_name(enc));
+    }
+    if (len_p) *len_p = MBCLEN_CHARFOUND_LEN(r);
+    return rb_enc_mbc_to_codepoint(p, e, enc);
+}
+
+#undef rb_enc_codepoint
+unsigned int
+rb_enc_codepoint(const char *p, const char *e, rb_encoding *enc)
+{
+    return rb_enc_codepoint_len(p, e, 0, enc);
 }
 
 int
@@ -810,7 +974,7 @@ rb_enc_tolower(int c, rb_encoding *enc)
 
 /*
  * call-seq:
- *   enc.inspect => string
+ *   enc.inspect -> string
  *
  * Returns a string which represents the encoding for programmers.
  *
@@ -829,11 +993,11 @@ enc_inspect(VALUE self)
 
 /*
  * call-seq:
- *   enc.name => string
+ *   enc.name -> string
  *
  * Returns the name of the encoding.
  *
- *   Encoding::UTF_8.name       => "UTF-8"
+ *   Encoding::UTF_8.name      #=> "UTF-8"
  */
 static VALUE
 enc_name(VALUE self)
@@ -856,11 +1020,11 @@ enc_names_i(st_data_t name, st_data_t idx, st_data_t args)
 
 /*
  * call-seq:
- *   enc.names => array
+ *   enc.names -> array
  *
  * Returns the list of name and aliases of the encoding.
  *
- *   Encoding::WINDOWS_31J.names => ["Windows-31J", "CP932", "csWindows31J"]
+ *   Encoding::WINDOWS_31J.names  #=> ["Windows-31J", "CP932", "csWindows31J"]
  */
 static VALUE
 enc_names(VALUE self)
@@ -875,20 +1039,20 @@ enc_names(VALUE self)
 
 /*
  * call-seq:
- *   Encoding.list => [enc1, enc2, ...]
+ *   Encoding.list -> [enc1, enc2, ...]
  *
  * Returns the list of loaded encodings.
  *
  *   Encoding.list
- *   => [#<Encoding:ASCII-8BIT>, #<Encoding:UTF-8>,
- *       #<Encoding:ISO-2022-JP (dummy)>]
+ *   #=> [#<Encoding:ASCII-8BIT>, #<Encoding:UTF-8>,
+ *         #<Encoding:ISO-2022-JP (dummy)>]
  *
  *   Encoding.find("US-ASCII")
- *   => #<Encoding:US-ASCII>
+ *   #=> #<Encoding:US-ASCII>
  *
  *   Encoding.list
- *   => [#<Encoding:ASCII-8BIT>, #<Encoding:UTF-8>,
- *       #<Encoding:US-ASCII>, #<Encoding:ISO-2022-JP (dummy)>]
+ *   #=> [#<Encoding:ASCII-8BIT>, #<Encoding:UTF-8>,
+ *         #<Encoding:US-ASCII>, #<Encoding:ISO-2022-JP (dummy)>]
  *
  */
 static VALUE
@@ -901,41 +1065,59 @@ enc_list(VALUE klass)
 
 /*
  * call-seq:
- *   Encoding.find(string) => enc
- *   Encoding.find(symbol) => enc
+ *   Encoding.find(string) -> enc
  *
  * Search the encoding with specified <i>name</i>.
- * <i>name</i> should be a string or symbol.
+ * <i>name</i> should be a string.
  *
- *   Encoding.find("US-ASCII")  => #<Encoding:US-ASCII>
- *   Encoding.find(:Shift_JIS)  => #<Encoding:Shift_JIS>
+ *   Encoding.find("US-ASCII")  #=> #<Encoding:US-ASCII>
+ *
+ * Names which this method accept are encoding names and aliases
+ * including following special aliases
+ *
+ * "external"::   default external encoding
+ * "internal"::   default internal encoding
+ * "locale"::     locale encoding
+ * "filesystem":: filesystem encoding
  *
  * An ArgumentError is raised when no encoding with <i>name</i>.
- * Only +Encoding.find("internal")+ however returns nil when no encoding named "internal",
- * in other words, when Ruby has no default internal encoding.
+ * Only <code>Encoding.find("internal")</code> however returns nil
+ * when no encoding named "internal", in other words, when Ruby has no
+ * default internal encoding.
  */
 static VALUE
 enc_find(VALUE klass, VALUE enc)
 {
-    return rb_enc_from_encoding(to_encoding(enc));
+    int idx;
+    if (RB_TYPE_P(enc, T_DATA) && is_data_encoding(enc))
+	return enc;
+    idx = str_to_encindex(enc);
+    if (idx == UNSPECIFIED_ENCODING) return Qnil;
+    return rb_enc_from_encoding_index(idx);
 }
 
 /*
  * call-seq:
- *   Encoding.compatible?(str1, str2) => enc or nil
+ *   Encoding.compatible?(obj1, obj2) -> enc or nil
  *
- * Checks the compatibility of two strings.
- * If they are compatible, means concatenatable, 
- * returns an encoding which the concatinated string will be.
- * If they are not compatible, nil is returned.
+ * Checks the compatibility of two objects.
+ *
+ * If the objects are both strings they are compatible when they are
+ * concatenatable.  The encoding of the concatenated string will be returned
+ * if they are compatible, nil if they are not.
  *
  *   Encoding.compatible?("\xa1".force_encoding("iso-8859-1"), "b")
- *   => #<Encoding:ISO-8859-1>
+ *   #=> #<Encoding:ISO-8859-1>
  *
  *   Encoding.compatible?(
  *     "\xa1".force_encoding("iso-8859-1"),
  *     "\xa1\xa1".force_encoding("euc-jp"))
- *   => nil
+ *   #=> nil
+ *
+ * If the objects are non-strings their encodings are compatible when they
+ * have an encoding and:
+ * * Either encoding is US-ASCII compatible
+ * * One of the encodings is a 7-bit encoding
  *
  */
 static VALUE
@@ -1010,8 +1192,8 @@ rb_usascii_encindex(void)
     return ENCINDEX_US_ASCII;
 }
 
-rb_encoding *
-rb_locale_encoding(void)
+int
+rb_locale_encindex(void)
 {
     VALUE charmap = rb_locale_charmap(rb_cEncoding);
     int idx;
@@ -1023,25 +1205,47 @@ rb_locale_encoding(void)
 
     if (rb_enc_registered("locale") < 0) enc_alias_internal("locale", idx);
 
-    return rb_enc_from_index(idx);
+    return idx;
+}
+
+rb_encoding *
+rb_locale_encoding(void)
+{
+    return rb_enc_from_index(rb_locale_encindex());
+}
+
+static int
+enc_set_filesystem_encoding(void)
+{
+    int idx;
+#if defined NO_LOCALE_CHARMAP
+    idx = rb_enc_to_index(rb_default_external_encoding());
+#elif defined _WIN32 || defined __CYGWIN__
+    char cp[sizeof(int) * 8 / 3 + 4];
+    snprintf(cp, sizeof cp, "CP%d", AreFileApisANSI() ? GetACP() : GetOEMCP());
+    idx = rb_enc_find_index(cp);
+    if (idx < 0) idx = rb_ascii8bit_encindex();
+#else
+    idx = rb_enc_to_index(rb_default_external_encoding());
+#endif
+
+    enc_alias_internal("filesystem", idx);
+    return idx;
+}
+
+int
+rb_filesystem_encindex(void)
+{
+    int idx = rb_enc_registered("filesystem");
+    if (idx < 0)
+	idx = rb_ascii8bit_encindex();
+    return idx;
 }
 
 rb_encoding *
 rb_filesystem_encoding(void)
 {
-    rb_encoding *enc;
-#if defined NO_LOCALE_CHARMAP
-    enc = rb_default_external_encoding();
-#elif defined _WIN32 || defined __CYGWIN__
-    char cp[sizeof(int) * 8 / 3 + 4];
-    snprintf(cp, sizeof cp, "CP%d", AreFileApisANSI() ? GetACP() : GetOEMCP());
-    enc = rb_enc_find(cp);
-#elif defined __APPLE__
-    enc = rb_enc_find("UTF8-MAC");
-#else
-    enc = rb_default_external_encoding();
-#endif
-    return enc;
+    return rb_enc_from_index(rb_filesystem_encindex());
 }
 
 struct default_encoding {
@@ -1049,13 +1253,16 @@ struct default_encoding {
     rb_encoding *enc;
 };
 
+static struct default_encoding default_external = {0};
+
 static int
 enc_set_default_encoding(struct default_encoding *def, VALUE encoding, const char *name)
 {
-    int overridden = Qfalse;
+    int overridden = FALSE;
+
     if (def->index != -2)
 	/* Already set */
-	overridden = Qtrue;
+	overridden = TRUE;
 
     if (NIL_P(encoding)) {
 	def->index = -1;
@@ -1069,11 +1276,12 @@ enc_set_default_encoding(struct default_encoding *def, VALUE encoding, const cha
 	enc_alias_internal(name, def->index);
     }
 
+    if (def == &default_external)
+	enc_set_filesystem_encoding();
+
     return overridden;
 }
 
-static struct default_encoding default_external = {0};
-
 rb_encoding *
 rb_default_external_encoding(void)
 {
@@ -1096,11 +1304,29 @@ rb_enc_default_external(void)
 
 /*
  * call-seq:
- *   Encoding.default_external => enc
+ *   Encoding.default_external -> enc
  *
  * Returns default external encoding.
  *
- * It is initialized by the locale or -E option.
+ * The default external encoding is used by default for strings created from
+ * the following locations:
+ *
+ * * CSV
+ * * File data read from disk
+ * * SDBM
+ * * StringIO
+ * * Zlib::GzipReader
+ * * Zlib::GzipWriter
+ * * String#inspect
+ * * Regexp#inspect
+ *
+ * While strings created from these locations will have this encoding, the
+ * encoding may not be valid.  Be sure to check String#valid_encoding?.
+ *
+ * File data written to disk will be transcoded to the default external
+ * encoding when written.
+ *
+ * The default external encoding is initialized by the locale or -E option.
  */
 static VALUE
 get_default_external(VALUE klass)
@@ -1122,7 +1348,14 @@ rb_enc_set_default_external(VALUE encoding)
  * call-seq:
  *   Encoding.default_external = enc
  *
- * Sets default external encoding.
+ * Sets default external encoding.  You should not set
+ * Encoding::default_external in ruby code as strings created before changing
+ * the value may have a different encoding from strings created after the value
+ * was changed., instead you should use <tt>ruby -E</tt> to invoke ruby with
+ * the correct default_external.
+ *
+ * See Encoding::default_external for information on how the default external
+ * encoding is used.
  */
 static VALUE
 set_default_external(VALUE klass, VALUE encoding)
@@ -1152,11 +1385,34 @@ rb_enc_default_internal(void)
 
 /*
  * call-seq:
- *   Encoding.default_internal => enc
+ *   Encoding.default_internal -> enc
  *
- * Returns default internal encoding.
+ * Returns default internal encoding.  Strings will be transcoded to the
+ * default internal encoding in the following places if the default internal
+ * encoding is not nil:
  *
- * It is initialized by the source internal_encoding or -E option.
+ * * CSV
+ * * Etc.sysconfdir and Etc.systmpdir
+ * * File data read from disk
+ * * File names from Dir
+ * * Integer#chr
+ * * String#inspect and Regexp#inspect
+ * * Strings returned from Curses
+ * * Strings returned from Readline
+ * * Strings returned from SDBM
+ * * Time#zone
+ * * Values from ENV
+ * * Values in ARGV including $PROGRAM_NAME
+ * * __FILE__
+ *
+ * Additionally String#encode and String#encode! use the default internal
+ * encoding if no encoding is given.
+ *
+ * The locale encoding (__ENCODING__), not default_internal, is used as the
+ * encoding of created strings.
+ *
+ * Encoding::default_internal is initialized by the source file's
+ * internal_encoding or -E option.
  */
 static VALUE
 get_default_internal(VALUE klass)
@@ -1175,8 +1431,14 @@ rb_enc_set_default_internal(VALUE encoding)
  * call-seq:
  *   Encoding.default_internal = enc or nil
  *
- * Sets default internal encoding.
- * Or removes default internal encoding when passed nil.
+ * Sets default internal encoding or removes default internal encoding when
+ * passed nil.  You should not set Encoding::default_internal in ruby code as
+ * strings created before changing the value may have a different encoding
+ * from strings created after the change.  Instead you should use
+ * <tt>ruby -E</tt> to invoke ruby with the correct default_internal.
+ *
+ * See Encoding::default_internal for information on how the default internal
+ * encoding is used.
  */
 static VALUE
 set_default_internal(VALUE klass, VALUE encoding)
@@ -1188,23 +1450,24 @@ set_default_internal(VALUE klass, VALUE encoding)
 
 /*
  * call-seq:
- *   Encoding.locale_charmap => string
+ *   Encoding.locale_charmap -> string
  *
  * Returns the locale charmap name.
+ * It returns nil if no appropriate information.
  *
  *   Debian GNU/Linux
  *     LANG=C
- *       Encoding.locale_charmap  => "ANSI_X3.4-1968"
+ *       Encoding.locale_charmap  #=> "ANSI_X3.4-1968"
  *     LANG=ja_JP.EUC-JP
- *       Encoding.locale_charmap  => "EUC-JP"
+ *       Encoding.locale_charmap  #=> "EUC-JP"
  *
  *   SunOS 5
  *     LANG=C
- *       Encoding.locale_charmap  => "646"
+ *       Encoding.locale_charmap  #=> "646"
  *     LANG=ja
- *       Encoding.locale_charmap  => "eucJP"
+ *       Encoding.locale_charmap  #=> "eucJP"
  *
- * The result is higly platform dependent.
+ * The result is highly platform dependent.
  * So Encoding.find(Encoding.locale_charmap) may cause an error.
  * If you need some encoding object even for unknown locale,
  * Encoding.find("locale") can be used.
@@ -1216,11 +1479,16 @@ rb_locale_charmap(VALUE klass)
 #if defined NO_LOCALE_CHARMAP
     return rb_usascii_str_new2("ASCII-8BIT");
 #elif defined _WIN32 || defined __CYGWIN__
-    const char *nl_langinfo_codeset(void);
-    const char *codeset = nl_langinfo_codeset();
+    const char *codeset = 0;
     char cp[sizeof(int) * 3 + 4];
+# ifdef __CYGWIN__
+    const char *nl_langinfo_codeset(void);
+    codeset = nl_langinfo_codeset();
+# endif
     if (!codeset) {
-	snprintf(cp, sizeof(cp), "CP%d", GetConsoleCP());
+	UINT codepage = GetConsoleCP();
+	if (!codepage) codepage = GetACP();
+	snprintf(cp, sizeof(cp), "CP%d", codepage);
 	codeset = cp;
     }
     return rb_usascii_str_new2(codeset);
@@ -1248,17 +1516,22 @@ set_encoding_const(const char *name, rb_encoding *enc)
 	}
     }
     if (!*s) {
+	if (s - name > ENCODING_NAMELEN_MAX) return;
 	valid = 1;
 	rb_define_const(rb_cEncoding, name, encoding);
     }
     if (!valid || haslower) {
-	int len = strlen(name) + 1;
+	size_t len = s - name;
+	if (len > ENCODING_NAMELEN_MAX) return;
 	if (!haslower || !hasupper) {
 	    do {
 		if (ISLOWER(*s)) haslower = 1;
 		if (ISUPPER(*s)) hasupper = 1;
 	    } while (*++s && (!haslower || !hasupper));
+	    len = s - name;
 	}
+	len += strlen(s);
+	if (len++ > ENCODING_NAMELEN_MAX) return;
 	MEMCPY(s = ALLOCA_N(char, len), name, char, len);
 	name = s;
 	if (!valid) {
@@ -1291,17 +1564,15 @@ rb_enc_name_list_i(st_data_t name, st_data_t idx, st_data_t arg)
 
 /*
  * call-seq:
- *   Encoding.name_list => ["enc1", "enc2", ...]
+ *   Encoding.name_list -> ["enc1", "enc2", ...]
  *
  * Returns the list of available encoding names.
  *
  *   Encoding.name_list
- *   => ["US-ASCII", "ASCII-8BIT", "UTF-8",
- *       "ISO-8859-1", "Shift_JIS", "EUC-JP",
- *       "Windows-31J",
- *       "BINARY", "CP932", "eucJP"]
- *
- * This list doesn't include dummy encodings.
+ *   #=> ["US-ASCII", "ASCII-8BIT", "UTF-8",
+ *         "ISO-8859-1", "Shift_JIS", "EUC-JP",
+ *         "Windows-31J",
+ *         "BINARY", "CP932", "eucJP"]
  *
  */
 
@@ -1340,13 +1611,13 @@ rb_enc_aliases_enc_i(st_data_t name, st_data_t orig, st_data_t arg)
 
 /*
  * call-seq:
- *   Encoding.aliases => {"alias1" => "orig1", "alias2" => "orig2", ...}
+ *   Encoding.aliases -> {"alias1" => "orig1", "alias2" => "orig2", ...}
  *
  * Returns the hash of available encoding alias and original encoding name.
  *
  *   Encoding.aliases
- *   => {"BINARY"=>"ASCII-8BIT", "ASCII"=>"US-ASCII", "ANSI_X3.4-1986"=>"US-ASCII",
- *       "SJIS"=>"Shift_JIS", "eucJP"=>"EUC-JP", "CP932"=>"Windows-31J"}
+ *   #=> {"BINARY"=>"ASCII-8BIT", "ASCII"=>"US-ASCII", "ANSI_X3.4-1986"=>"US-ASCII",
+ *         "SJIS"=>"Shift_JIS", "eucJP"=>"EUC-JP", "CP932"=>"Windows-31J"}
  *
  */
 
@@ -1360,6 +1631,206 @@ rb_enc_aliases(VALUE klass)
     return aliases[0];
 }
 
+/*
+ * An Encoding instance represents a character encoding usable in Ruby. It is
+ * defined as a constant under the Encoding namespace. It has a name and
+ * optionally, aliases:
+ *
+ *   Encoding::ISO_8859_1.name
+ *   #=> #<Encoding:ISO-8859-1>
+ *
+ *   Encoding::ISO_8859_1.names
+ *   #=> ["ISO-8859-1", "ISO8859-1"]
+ *
+ * Ruby methods dealing with encodings return or accept Encoding instances as
+ * arguments (when a method accepts an Encoding instance as an argument, it
+ * can be passed an Encoding name or alias instead).
+ *
+ *   "some string".encoding
+ *   #=> #<Encoding:UTF-8>
+ *
+ *   string = "some string".encode(Encoding::ISO_8859_1)
+ *   #=> "some string"
+ *   string.encoding
+ *   #=> #<Encoding:ISO-8859-1>
+ *
+ *   "some string".encode "ISO-8859-1"
+ *   #=> "some string"
+ *
+ * <code>Encoding::ASCII_8BIT</code> is a special encoding that is usually
+ * used for a byte string, not a character string. But as the name insists,
+ * its characters in the range of ASCII are considered as ASCII characters.
+ * This is useful when you use ASCII-8BIT characters with other ASCII
+ * compatible characters.
+ *
+ * == Changing an encoding
+ *
+ * The associated Encoding of a String can be changed in two different ways.
+ *
+ * First, it is possible to set the Encoding of a string to a new Encoding
+ * without changing the internal byte representation of the string, with
+ * String#force_encoding. This is how you can tell Ruby the correct encoding
+ * of a string.
+ *
+ *   string
+ *   #=> "R\xC3\xA9sum\xC3\xA9"
+ *   string.encoding
+ *   #=> #<Encoding:ISO-8859-1>
+ *   string.force_encoding(Encoding::UTF-8)
+ *   #=> "R\u00E9sum\u00E9"
+ *
+ * Second, it is possible to transcode a string, i.e. translate its internal
+ * byte representation to another encoding. Its associated encoding is also
+ * set to the other encoding. See String#encode for the various forms of
+ * transcoding, and the Encoding::Converter class for additional control over
+ * the transcoding process.
+ *
+ *   string
+ *   #=> "R\u00E9sum\u00E9"
+ *   string.encoding
+ *   #=> #<Encoding:UTF-8>
+ *   string = string.encode!(Encoding::ISO_8859_1)
+ *   #=> "R\xE9sum\xE9"
+ *   string.encoding
+ *   #=> #<Encoding::ISO-8859-1>
+ *
+ * == Script encoding
+ *
+ * All Ruby script code has an associated Encoding which any String literal
+ * created in the source code will be associated to.
+ *
+ * The default script encoding is <code>Encoding::US-ASCII</code>, but it can
+ * be changed by a magic comment on the first line of the source code file (or
+ * second line, if there is a shebang line on the first). The comment must
+ * contain the word <code>coding</code> or <code>encoding</code>, followed
+ * by a colon, space and the Encoding name or alias:
+ *
+ *   # encoding: UTF-8
+ *
+ *   "some string".encoding
+ *   #=> #<Encoding:UTF-8>
+ *
+ * The <code>__ENCODING__</code> keyword returns the script encoding of the file
+ * which the keyword is written:
+ *
+ *   # encoding: ISO-8859-1
+ *
+ *   __ENCODING__
+ *   #=> #<Encoding:ISO-8859-1>
+ *
+ * <code>ruby -K</code> will change the default locale encoding, but this is
+ * not recommended. Ruby source files should declare its script encoding by a
+ * magic comment even when they only depend on US-ASCII strings or regular
+ * expressions.
+ *
+ * == Locale encoding
+ *
+ * The default encoding of the environment. Usually derived from locale.
+ *
+ * see Encoding.locale_charmap, Encoding.find('locale')
+ *
+ * == Filesystem encoding
+ *
+ * The default encoding of strings from the filesystem of the environment.
+ * This is used for strings of file names or paths.
+ *
+ * see Encoding.find('filesystem')
+ *
+ * == External encoding
+ *
+ * Each IO object has an external encoding which indicates the encoding that
+ * Ruby will use to read its data. By default Ruby sets the external encoding
+ * of an IO object to the default external encoding. The default external
+ * encoding is set by locale encoding or the interpreter <code>-E</code> option.
+ * Encoding.default_external returns the current value of the external
+ * encoding.
+ *
+ *   ENV["LANG"]
+ *   #=> "UTF-8"
+ *   Encoding.default_external
+ *   #=> #<Encoding:UTF-8>
+ *
+ *   $ ruby -E ISO-8859-1 -e "p Encoding.default_external"
+ *   #<Encoding:ISO-8859-1>
+ *
+ *   $ LANG=C ruby -e 'p Encoding.default_external'
+ *   #<Encoding:US-ASCII>
+ *
+ * The default external encoding may also be set through
+ * Encoding.default_external=, but you should not do this as strings created
+ * before and after the change will have inconsistent encodings.  Instead use
+ * <code>ruby -E</code> to invoke ruby with the correct external encoding.
+ *
+ * When you know that the actual encoding of the data of an IO object is not
+ * the default external encoding, you can reset its external encoding with
+ * IO#set_encoding or set it at IO object creation (see IO.new options).
+ *
+ * == Internal encoding
+ *
+ * To process the data of an IO object which has an encoding different
+ * from its external encoding, you can set its internal encoding. Ruby will use
+ * this internal encoding to transcode the data when it is read from the IO
+ * object.
+ *
+ * Conversely, when data is written to the IO object it is transcoded from the
+ * internal encoding to the external encoding of the IO object.
+ *
+ * The internal encoding of an IO object can be set with
+ * IO#set_encoding or at IO object creation (see IO.new options).
+ *
+ * The internal encoding is optional and when not set, the Ruby default
+ * internal encoding is used. If not explicitly set this default internal
+ * encoding is +nil+ meaning that by default, no transcoding occurs.
+ *
+ * The default internal encoding can be set with the interpreter option
+ * <code>-E</code>. Encoding.default_internal returns the current internal
+ * encoding.
+ *
+ *    $ ruby -e 'p Encoding.default_internal'
+ *    nil
+ *
+ *    $ ruby -E ISO-8859-1:UTF-8 -e "p [Encoding.default_external, \
+ *      Encoding.default_internal]"
+ *    [#<Encoding:ISO-8859-1>, #<Encoding:UTF-8>]
+ *
+ * The default internal encoding may also be set through
+ * Encoding.default_internal=, but you should not do this as strings created
+ * before and after the change will have inconsistent encodings.  Instead use
+ * <code>ruby -E</code> to invoke ruby with the correct internal encoding.
+ *
+ * == IO encoding example
+ *
+ * In the following example a UTF-8 encoded string "R\u00E9sum\u00E9" is transcoded for
+ * output to ISO-8859-1 encoding, then read back in and transcoded to UTF-8:
+ *
+ *   string = "R\u00E9sum\u00E9"
+ *
+ *   open("transcoded.txt", "w:ISO-8859-1") do |io|
+ *     io.write(string)
+ *   end
+ *
+ *   puts "raw text:"
+ *   p File.binread("transcoded.txt")
+ *   puts
+ *
+ *   open("transcoded.txt", "r:ISO-8859-1:UTF-8") do |io|
+ *     puts "transcoded text:"
+ *     p io.read
+ *   end
+ *
+ * While writing the file, the internal encoding is not specified as it is
+ * only necessary for reading.  While reading the file both the internal and
+ * external encoding must be specified to obtain the correct result.
+ *
+ *   $ ruby t.rb
+ *   raw text:
+ *   "R\xE9sum\xE9"
+ *
+ *   transcoded text:
+ *   "R\u00E9sum\u00E9"
+ *
+ */
+
 void
 Init_Encoding(void)
 {
@@ -1370,11 +1841,14 @@ Init_Encoding(void)
 
     rb_cEncoding = rb_define_class("Encoding", rb_cObject);
     rb_undef_alloc_func(rb_cEncoding);
+    rb_undef_method(CLASS_OF(rb_cEncoding), "new");
     rb_define_method(rb_cEncoding, "to_s", enc_name, 0);
     rb_define_method(rb_cEncoding, "inspect", enc_inspect, 0);
     rb_define_method(rb_cEncoding, "name", enc_name, 0);
     rb_define_method(rb_cEncoding, "names", enc_names, 0);
     rb_define_method(rb_cEncoding, "dummy?", enc_dummy_p, 0);
+    rb_define_method(rb_cEncoding, "ascii_compatible?", enc_ascii_compatible_p, 0);
+    rb_define_method(rb_cEncoding, "replicate", enc_replicate, 1);
     rb_define_singleton_method(rb_cEncoding, "list", enc_list, 0);
     rb_define_singleton_method(rb_cEncoding, "name_list", rb_enc_name_list, 0);
     rb_define_singleton_method(rb_cEncoding, "aliases", rb_enc_aliases, 0);
@@ -1400,10 +1874,10 @@ Init_Encoding(void)
     }
 }
 
-/* locale insensitive functions */
+/* locale insensitive ctype functions */
 
 #define ctype_test(c, ctype) \
-    (rb_isascii(c) && ONIGENC_IS_ASCII_CODE_CTYPE((c), ctype))
+    (rb_isascii(c) && ONIGENC_IS_ASCII_CODE_CTYPE((c), (ctype)))
 
 int rb_isalnum(int c) { return ctype_test(c, ONIGENC_CTYPE_ALNUM); }
 int rb_isalpha(int c) { return ctype_test(c, ONIGENC_CTYPE_ALPHA); }
