@@ -2,7 +2,7 @@
 
   error.c -
 
-  $Author: yugui $
+  $Author: usa $
   created at: Mon Aug  9 16:11:34 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -12,6 +12,7 @@
 #include "ruby/ruby.h"
 #include "ruby/st.h"
 #include "ruby/encoding.h"
+#include "internal.h"
 #include "vm_core.h"
 
 #include <stdio.h>
@@ -20,17 +21,42 @@
 #include <stdlib.h>
 #endif
 #include <errno.h>
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
+#define numberof(array) (int)(sizeof(array) / sizeof((array)[0]))
 
 #ifndef EXIT_SUCCESS
 #define EXIT_SUCCESS 0
 #endif
 
+#ifndef WIFEXITED
+#define WIFEXITED(status) 1
+#endif
+
+#ifndef WEXITSTATUS
+#define WEXITSTATUS(status) (status)
+#endif
+
 extern const char ruby_description[];
 
+static const char REPORTBUG_MSG[] =
+	"[NOTE]\n" \
+	"You may have encountered a bug in the Ruby interpreter" \
+	" or extension libraries.\n" \
+	"Bug reports are welcome.\n" \
+	""
+#if defined __APPLE__
+	"Don't forget to include the above Crash Report log file.\n"
+#endif
+	"For details: http://www.ruby-lang.org/bugreport.html\n\n" \
+    ;
+
 static const char *
 rb_strerrno(int err)
 {
-#define defined_error(name, num) if (err == num) return name;
+#define defined_error(name, num) if (err == (num)) return (name);
 #define undefined_error(name)
 #include "known_errors.inc"
 #undef defined_error
@@ -52,84 +78,110 @@ err_position_0(char *buf, long len, const char *file, int line)
     }
 }
 
-static int
-err_position(char *buf, long len)
+static VALUE
+compile_snprintf(rb_encoding *enc, const char *pre, const char *file, int line, const char *fmt, va_list args)
 {
-    return err_position_0(buf, len, rb_sourcefile(), rb_sourceline());
+    VALUE str = rb_enc_str_new(0, 0, enc);
+
+    if (file) {
+	rb_str_cat2(str, file);
+	if (line) rb_str_catf(str, ":%d", line);
+	rb_str_cat2(str, ": ");
+    }
+    if (pre) rb_str_cat2(str, pre);
+    rb_str_vcatf(str, fmt, args);
+    return str;
 }
 
 static void
-err_snprintf(char *buf, long len, const char *fmt, va_list args)
+compile_err_append(VALUE mesg)
 {
-    long n;
+    rb_thread_t *th = GET_THREAD();
+    VALUE err = th->errinfo;
+    rb_block_t *prev_base_block = th->base_block;
+    th->base_block = 0;
+    /* base_block should be zero while normal Ruby execution */
+    /* after this line, any Ruby code *can* run */
 
-    n = err_position(buf, len);
-    if (len > n) {
-	vsnprintf((char*)buf+n, len-n, fmt, args);
+    if (th->mild_compile_error) {
+	if (RTEST(err)) {
+	    VALUE str = rb_obj_as_string(err);
+
+	    rb_str_cat2(str, "\n");
+	    rb_str_append(str, mesg);
+	    mesg = str;
+	}
+	err = rb_exc_new3(rb_eSyntaxError, mesg);
+	th->errinfo = err;
     }
+    else {
+	if (!RTEST(err)) {
+	    err = rb_exc_new2(rb_eSyntaxError, "compile error");
+	    th->errinfo = err;
+	}
+	rb_str_cat2(mesg, "\n");
+	rb_write_error_str(mesg);
+    }
+
+    /* returned to the parser world */
+    th->base_block = prev_base_block;
 }
 
-static void
-compile_snprintf(char *buf, long len, const char *file, int line, const char *fmt, va_list args)
+void
+rb_compile_error_with_enc(const char *file, int line, void *enc, const char *fmt, ...)
 {
-    long n;
+    va_list args;
+    VALUE str;
 
-    n = err_position_0(buf, len, file, line);
-    if (len > n) {
-	vsnprintf((char*)buf+n, len-n, fmt, args);
-    }
+    va_start(args, fmt);
+    str = compile_snprintf(enc, NULL, file, line, fmt, args);
+    va_end(args);
+    compile_err_append(str);
 }
 
-static void err_append(const char*);
-
 void
 rb_compile_error(const char *file, int line, const char *fmt, ...)
 {
     va_list args;
-    char buf[BUFSIZ];
+    VALUE str;
 
     va_start(args, fmt);
-    compile_snprintf(buf, BUFSIZ, file, line, fmt, args);
+    str = compile_snprintf(NULL, NULL, file, line, fmt, args);
     va_end(args);
-    err_append(buf);
+    compile_err_append(str);
 }
 
 void
 rb_compile_error_append(const char *fmt, ...)
 {
     va_list args;
-    char buf[BUFSIZ];
+    VALUE str;
 
     va_start(args, fmt);
-    vsnprintf(buf, BUFSIZ, fmt, args);
+    str = rb_vsprintf(fmt, args);
     va_end(args);
-    err_append(buf);
+    compile_err_append(str);
 }
 
 static void
 compile_warn_print(const char *file, int line, const char *fmt, va_list args)
 {
-    char buf[BUFSIZ];
-    int len;
+    VALUE str;
 
-    compile_snprintf(buf, BUFSIZ, file, line, fmt, args);
-    len = (int)strlen(buf);
-    buf[len++] = '\n';
-    rb_write_error2(buf, len);
+    str = compile_snprintf(NULL, "warning: ", file, line, fmt, args);
+    rb_str_cat2(str, "\n");
+    rb_write_error_str(str);
 }
 
 void
 rb_compile_warn(const char *file, int line, const char *fmt, ...)
 {
-    char buf[BUFSIZ];
     va_list args;
 
     if (NIL_P(ruby_verbose)) return;
 
-    snprintf(buf, BUFSIZ, "warning: %s", fmt);
-
     va_start(args, fmt);
-    compile_warn_print(file, line, buf, args);
+    compile_warn_print(file, line, fmt, args);
     va_end(args);
 }
 
@@ -137,42 +189,43 @@ rb_compile_warn(const char *file, int line, const char *fmt, ...)
 void
 rb_compile_warning(const char *file, int line, const char *fmt, ...)
 {
-    char buf[BUFSIZ];
     va_list args;
 
     if (!RTEST(ruby_verbose)) return;
 
-    snprintf(buf, BUFSIZ, "warning: %s", fmt);
-
     va_start(args, fmt);
-    compile_warn_print(file, line, buf, args);
+    compile_warn_print(file, line, fmt, args);
     va_end(args);
 }
 
 static void
 warn_print(const char *fmt, va_list args)
 {
-    char buf[BUFSIZ];
-    int len;
+    VALUE str = rb_str_new(0, 0);
+    VALUE file = rb_sourcefilename();
 
-    err_snprintf(buf, BUFSIZ, fmt, args);
-    len = (int)strlen(buf);
-    buf[len++] = '\n';
-    rb_write_error2(buf, len);
+    if (!NIL_P(file)) {
+	int line = rb_sourceline();
+	str = rb_str_append(str, file);
+	if (line) rb_str_catf(str, ":%d", line);
+	rb_str_cat2(str, ": ");
+    }
+
+    rb_str_cat2(str, "warning: ");
+    rb_str_vcatf(str, fmt, args);
+    rb_str_cat2(str, "\n");
+    rb_write_error_str(str);
 }
 
 void
 rb_warn(const char *fmt, ...)
 {
-    char buf[BUFSIZ];
     va_list args;
 
     if (NIL_P(ruby_verbose)) return;
 
-    snprintf(buf, BUFSIZ, "warning: %s", fmt);
-
     va_start(args, fmt);
-    warn_print(buf, args);
+    warn_print(fmt, args);
     va_end(args);
 }
 
@@ -180,60 +233,61 @@ rb_warn(const char *fmt, ...)
 void
 rb_warning(const char *fmt, ...)
 {
-    char buf[BUFSIZ];
     va_list args;
 
     if (!RTEST(ruby_verbose)) return;
 
-    snprintf(buf, BUFSIZ, "warning: %s", fmt);
-
     va_start(args, fmt);
-    warn_print(buf, args);
+    warn_print(fmt, args);
     va_end(args);
 }
 
 /*
  * call-seq:
- *    warn(msg)   -> nil
+ *    warn(msg, ...)   -> nil
  *
- * Display the given message (followed by a newline) on STDERR unless
- * warnings are disabled (for example with the <code>-W0</code> flag).
+ * Displays each of the given messages followed by a record separator on
+ * STDERR unless warnings have been disabled (for example with the
+ * <code>-W0</code> flag).
+ *
+ *    warn("warning 1", "warning 2")
+ *
+ *  <em>produces:</em>
+ *
+ *    warning 1
+ *    warning 2
  */
 
 static VALUE
-rb_warn_m(VALUE self, VALUE mesg)
+rb_warn_m(int argc, VALUE *argv, VALUE exc)
 {
-    if (!NIL_P(ruby_verbose)) {
-	rb_io_write(rb_stderr, mesg);
-	rb_io_write(rb_stderr, rb_default_rs);
+    if (!NIL_P(ruby_verbose) && argc > 0) {
+	rb_io_puts(argc, argv, rb_stderr);
     }
     return Qnil;
 }
 
-void rb_vm_bugreport(void);
-
 static void
 report_bug(const char *file, int line, const char *fmt, va_list args)
 {
-    char buf[BUFSIZ];
+    /* SIGSEGV handler might have a very small stack. Thus we need to use it carefully. */
+    char buf[256];
     FILE *out = stderr;
-    int len = err_position_0(buf, BUFSIZ, file, line);
+    int len = err_position_0(buf, 256, file, line);
 
     if ((ssize_t)fwrite(buf, 1, len, out) == (ssize_t)len ||
 	(ssize_t)fwrite(buf, 1, len, (out = stdout)) == (ssize_t)len) {
 
 	fputs("[BUG] ", out);
-	vfprintf(out, fmt, args);
-	fprintf(out, "\n%s\n\n", ruby_description);
+	vsnprintf(buf, 256, fmt, args);
+	fputs(buf, out);
+	snprintf(buf, 256, "\n%s\n\n", ruby_description);
+	fputs(buf, out);
+
 
 	rb_vm_bugreport();
 
-	fprintf(out,
-		"[NOTE]\n"
-		"You may have encountered a bug in the Ruby interpreter"
-		" or extension libraries.\n"
-		"Bug reports are welcome.\n"
-		"For details: http://www.ruby-lang.org/bugreport.html\n\n");
+	fprintf(out, REPORTBUG_MSG);
     }
 }
 
@@ -241,12 +295,19 @@ void
 rb_bug(const char *fmt, ...)
 {
     va_list args;
+    const char *file = NULL;
+    int line = 0;
+
+    if (GET_THREAD()) {
+	file = rb_sourcefile();
+	line = rb_sourceline();
+    }
 
     va_start(args, fmt);
-    report_bug(rb_sourcefile(), rb_sourceline(), fmt, args);
+    report_bug(file, line, fmt, args);
     va_end(args);
 
-#if defined(_WIN32) && defined(RT_VER) && RT_VER >= 80
+#if defined(_WIN32) && defined(RUBY_MSVCRT_VERSION) && RUBY_MSVCRT_VERSION >= 80
     _set_abort_behavior( 0, _CALL_REPORTFAULT);
 #endif
 
@@ -267,6 +328,37 @@ rb_bug_errno(const char *mesg, int errno_arg)
     }
 }
 
+/*
+ * this is safe to call inside signal handler and timer thread
+ * (which isn't a Ruby Thread object)
+ */
+#define write_or_abort(fd, str, len) (write((fd), (str), (len)) < 0 ? abort() : (void)0)
+#define WRITE_CONST(fd,str) write_or_abort((fd),(str),sizeof(str) - 1)
+
+void
+rb_async_bug_errno(const char *mesg, int errno_arg)
+{
+    WRITE_CONST(2, "[ASYNC BUG] ");
+    write_or_abort(2, mesg, strlen(mesg));
+    WRITE_CONST(2, "\n");
+
+    if (errno_arg == 0) {
+	WRITE_CONST(2, "errno == 0 (NOERROR)\n");
+    }
+    else {
+	const char *errno_str = rb_strerrno(errno_arg);
+
+	if (!errno_str)
+	    errno_str = "undefined errno";
+	write_or_abort(2, errno_str, strlen(errno_str));
+    }
+    WRITE_CONST(2, "\n\n");
+    write_or_abort(2, ruby_description, strlen(ruby_description));
+    WRITE_CONST(2, "\n\n");
+    WRITE_CONST(2, REPORTBUG_MSG);
+    abort();
+}
+
 void
 rb_compile_bug(const char *file, int line, const char *fmt, ...)
 {
@@ -279,41 +371,79 @@ rb_compile_bug(const char *file, int line, const char *fmt, ...)
     abort();
 }
 
-static const struct types {
-    int type;
-    const char *name;
-} builtin_types[] = {
-    {T_NIL,	"nil"},
-    {T_OBJECT,	"Object"},
-    {T_CLASS,	"Class"},
-    {T_ICLASS,	"iClass"},	/* internal use: mixed-in module holder */
-    {T_MODULE,	"Module"},
-    {T_FLOAT,	"Float"},
-    {T_STRING,	"String"},
-    {T_REGEXP,	"Regexp"},
-    {T_ARRAY,	"Array"},
-    {T_FIXNUM,	"Fixnum"},
-    {T_HASH,	"Hash"},
-    {T_STRUCT,	"Struct"},
-    {T_BIGNUM,	"Bignum"},
-    {T_FILE,	"File"},
-    {T_RATIONAL,"Rational"},
-    {T_COMPLEX, "Complex"},
-    {T_TRUE,	"true"},
-    {T_FALSE,	"false"},
-    {T_SYMBOL,	"Symbol"},	/* :symbol */
-    {T_DATA,	"Data"},	/* internal use: wrapped C pointers */
-    {T_MATCH,	"MatchData"},	/* data of $~ */
-    {T_NODE,	"Node"},	/* internal use: syntax tree node */
-    {T_UNDEF,	"undef"},	/* internal use: #undef; should not happen */
+static const char builtin_types[][10] = {
+    "", 			/* 0x00, */
+    "Object",
+    "Class",
+    "Module",
+    "Float",
+    "String",
+    "Regexp",
+    "Array",
+    "Hash",
+    "Struct",
+    "Bignum",
+    "File",
+    "Data",			/* internal use: wrapped C pointers */
+    "MatchData",		/* data of $~ */
+    "Complex",
+    "Rational",
+    "",				/* 0x10 */
+    "nil",
+    "true",
+    "false",
+    "Symbol",			/* :symbol */
+    "Fixnum",
+    "",				/* 0x16 */
+    "",				/* 0x17 */
+    "",				/* 0x18 */
+    "",				/* 0x19 */
+    "",				/* 0x1a */
+    "undef",			/* internal use: #undef; should not happen */
+    "Node",			/* internal use: syntax tree node */
+    "iClass",			/* internal use: mixed-in module holder */
 };
 
+const char *
+rb_builtin_type_name(int t)
+{
+    const char *name;
+    if ((unsigned int)t >= numberof(builtin_types)) return 0;
+    name = builtin_types[t];
+    if (*name) return name;
+    return 0;
+}
+
+#define builtin_class_name rb_builtin_class_name
+const char *
+rb_builtin_class_name(VALUE x)
+{
+    const char *etype;
+
+    if (NIL_P(x)) {
+	etype = "nil";
+    }
+    else if (FIXNUM_P(x)) {
+	etype = "Fixnum";
+    }
+    else if (SYMBOL_P(x)) {
+	etype = "Symbol";
+    }
+    else if (RB_TYPE_P(x, T_TRUE)) {
+	etype = "true";
+    }
+    else if (RB_TYPE_P(x, T_FALSE)) {
+	etype = "false";
+    }
+    else {
+	etype = rb_obj_classname(x);
+    }
+    return etype;
+}
+
 void
 rb_check_type(VALUE x, int t)
 {
-    const struct types *type = builtin_types;
-    const struct types *const typeend = builtin_types +
-	sizeof(builtin_types) / sizeof(builtin_types[0]);
     int xt;
 
     if (x == Qundef) {
@@ -322,39 +452,33 @@ rb_check_type(VALUE x, int t)
 
     xt = TYPE(x);
     if (xt != t || (xt == T_DATA && RTYPEDDATA_P(x))) {
-	while (type < typeend) {
-	    if (type->type == t) {
-		const char *etype;
-
-		if (NIL_P(x)) {
-		    etype = "nil";
-		}
-		else if (FIXNUM_P(x)) {
-		    etype = "Fixnum";
-		}
-		else if (SYMBOL_P(x)) {
-		    etype = "Symbol";
-		}
-		else if (rb_special_const_p(x)) {
-		    etype = RSTRING_PTR(rb_obj_as_string(x));
-		}
-		else {
-		    etype = rb_obj_classname(x);
-		}
-		rb_raise(rb_eTypeError, "wrong argument type %s (expected %s)",
-			 etype, type->name);
-	    }
-	    type++;
+	const char *tname = rb_builtin_type_name(t);
+	if (tname) {
+	    rb_raise(rb_eTypeError, "wrong argument type %s (expected %s)",
+		     builtin_class_name(x), tname);
 	}
-	rb_bug("unknown type 0x%x (0x%x given)", t, TYPE(x));
+	if (xt > T_MASK && xt <= 0x3f) {
+	    rb_fatal("unknown type 0x%x (0x%x given, probably comes from extension library for ruby 1.8)", t, xt);
+	}
+	rb_bug("unknown type 0x%x (0x%x given)", t, xt);
     }
 }
 
+int
+rb_typeddata_inherited_p(const rb_data_type_t *child, const rb_data_type_t *parent)
+{
+    while (child) {
+	if (child == parent) return 1;
+	child = child->parent;
+    }
+    return 0;
+}
+
 int
 rb_typeddata_is_kind_of(VALUE obj, const rb_data_type_t *data_type)
 {
-    if (SPECIAL_CONST_P(obj) || BUILTIN_TYPE(obj) != T_DATA ||
-	!RTYPEDDATA_P(obj) || RTYPEDDATA_TYPE(obj) != data_type) {
+    if (!RB_TYPE_P(obj, T_DATA) ||
+	!RTYPEDDATA_P(obj) || !rb_typeddata_inherited_p(RTYPEDDATA_TYPE(obj), data_type)) {
 	return 0;
     }
     return 1;
@@ -366,14 +490,15 @@ rb_check_typeddata(VALUE obj, const rb_data_type_t *data_type)
     const char *etype;
     static const char mesg[] = "wrong argument type %s (expected %s)";
 
-    if (SPECIAL_CONST_P(obj) || BUILTIN_TYPE(obj) != T_DATA) {
-	Check_Type(obj, T_DATA);
+    if (!RB_TYPE_P(obj, T_DATA)) {
+	etype = builtin_class_name(obj);
+	rb_raise(rb_eTypeError, mesg, etype, data_type->wrap_struct_name);
     }
     if (!RTYPEDDATA_P(obj)) {
 	etype = rb_obj_classname(obj);
 	rb_raise(rb_eTypeError, mesg, etype, data_type->wrap_struct_name);
     }
-    else if (RTYPEDDATA_TYPE(obj) != data_type) {
+    else if (!rb_typeddata_inherited_p(RTYPEDDATA_TYPE(obj), data_type)) {
 	etype = RTYPEDDATA_TYPE(obj)->wrap_struct_name;
 	rb_raise(rb_eTypeError, mesg, etype, data_type->wrap_struct_name);
     }
@@ -489,10 +614,11 @@ static VALUE
 exc_to_s(VALUE exc)
 {
     VALUE mesg = rb_attr_get(exc, rb_intern("mesg"));
+    VALUE r = Qnil;
 
     if (NIL_P(mesg)) return rb_class_name(CLASS_OF(exc));
-    if (OBJ_TAINTED(exc)) OBJ_TAINT(mesg);
-    return mesg;
+    r = rb_String(mesg);
+    return r;
 }
 
 /*
@@ -515,7 +641,7 @@ exc_message(VALUE exc)
  * call-seq:
  *   exception.inspect   -> string
  *
- * Return this exception's class name an message
+ * Return this exception's class name and message
  */
 
 static VALUE
@@ -572,9 +698,17 @@ static VALUE
 exc_backtrace(VALUE exc)
 {
     ID bt;
+    VALUE obj;
 
     CONST_ID(bt, "bt");
-    return rb_attr_get(exc, bt);
+    obj = rb_attr_get(exc, bt);
+
+    if (rb_backtrace_p(obj)) {
+	obj = rb_backtrace_to_str_ary(obj);
+	/* rb_iv_set(exc, "bt", obj); */
+    }
+
+    return obj;
 }
 
 VALUE
@@ -584,14 +718,13 @@ rb_check_backtrace(VALUE bt)
     static const char err[] = "backtrace must be Array of String";
 
     if (!NIL_P(bt)) {
-	int t = TYPE(bt);
-
-	if (t == T_STRING) return rb_ary_new3(1, bt);
-	if (t != T_ARRAY) {
+	if (RB_TYPE_P(bt, T_STRING)) return rb_ary_new3(1, bt);
+	if (rb_backtrace_p(bt)) return bt;
+	if (!RB_TYPE_P(bt, T_ARRAY)) {
 	    rb_raise(rb_eTypeError, err);
 	}
 	for (i=0;i<RARRAY_LEN(bt);i++) {
-	    if (TYPE(RARRAY_PTR(bt)[i]) != T_STRING) {
+	    if (!RB_TYPE_P(RARRAY_PTR(bt)[i], T_STRING)) {
 		rb_raise(rb_eTypeError, err);
 	    }
 	}
@@ -601,11 +734,11 @@ rb_check_backtrace(VALUE bt)
 
 /*
  *  call-seq:
- *     exc.set_backtrace(array)   ->  array
+ *     exc.set_backtrace(backtrace)   ->  array
  *
- *  Sets the backtrace information associated with <i>exc</i>. The
- *  argument must be an array of <code>String</code> objects in the
- *  format described in <code>Exception#backtrace</code>.
+ *  Sets the backtrace information associated with +exc+. The +backtrace+ must
+ *  be an array of String objects or a single String in the format described
+ *  in Exception#backtrace.
  *
  */
 
@@ -615,6 +748,20 @@ exc_set_backtrace(VALUE exc, VALUE bt)
     return rb_iv_set(exc, "bt", rb_check_backtrace(bt));
 }
 
+VALUE
+rb_exc_set_backtrace(VALUE exc, VALUE bt)
+{
+    return exc_set_backtrace(exc, bt);
+}
+
+static VALUE
+try_convert_to_exception(VALUE obj)
+{
+    ID id_exception;
+    CONST_ID(id_exception, "exception");
+    return rb_check_funcall(obj, id_exception, 0, 0);
+}
+
 /*
  *  call-seq:
  *     exc == obj   -> true or false
@@ -634,10 +781,17 @@ exc_equal(VALUE exc, VALUE obj)
     CONST_ID(id_mesg, "mesg");
 
     if (rb_obj_class(exc) != rb_obj_class(obj)) {
+	int status = 0;
 	ID id_message, id_backtrace;
 	CONST_ID(id_message, "message");
 	CONST_ID(id_backtrace, "backtrace");
 
+	obj = rb_protect(try_convert_to_exception, obj, &status);
+	if (status || obj == Qundef) {
+	    rb_set_errinfo(Qnil);
+	    return Qfalse;
+	}
+	if (rb_obj_class(exc) != rb_obj_class(obj)) return Qfalse;
 	mesg = rb_check_funcall(obj, id_message, 0, 0);
 	if (mesg == Qundef) return Qfalse;
 	backtrace = rb_check_funcall(obj, id_backtrace, 0, 0);
@@ -657,18 +811,52 @@ exc_equal(VALUE exc, VALUE obj)
 
 /*
  * call-seq:
- *   SystemExit.new(status=0)   -> system_exit
+ *   SystemExit.new              -> system_exit
+ *   SystemExit.new(status)      -> system_exit
+ *   SystemExit.new(status, msg) -> system_exit
+ *   SystemExit.new(msg)         -> system_exit
  *
- * Create a new +SystemExit+ exception with the given status.
+ * Create a new +SystemExit+ exception with the given status and message.
+ * Status is true, false, or an integer.
+ * If status is not given, true is used.
  */
 
 static VALUE
 exit_initialize(int argc, VALUE *argv, VALUE exc)
 {
-    VALUE status = INT2FIX(EXIT_SUCCESS);
-    if (argc > 0 && FIXNUM_P(argv[0])) {
-	status = *argv++;
-	--argc;
+    VALUE status;
+    if (argc > 0) {
+	status = *argv;
+
+	switch (status) {
+	  case Qtrue:
+	    status = INT2FIX(EXIT_SUCCESS);
+	    ++argv;
+	    --argc;
+	    break;
+	  case Qfalse:
+	    status = INT2FIX(EXIT_FAILURE);
+	    ++argv;
+	    --argc;
+	    break;
+	  default:
+	    status = rb_check_to_int(status);
+	    if (NIL_P(status)) {
+		status = INT2FIX(EXIT_SUCCESS);
+	    }
+	    else {
+#if EXIT_SUCCESS != 0
+		if (status == INT2FIX(0))
+		    status = INT2FIX(EXIT_SUCCESS);
+#endif
+		++argv;
+		--argc;
+	    }
+	    break;
+	}
+    }
+    else {
+	status = INT2FIX(EXIT_SUCCESS);
     }
     rb_call_super(argc, argv);
     rb_iv_set(exc, "status", status);
@@ -700,9 +888,15 @@ exit_status(VALUE exc)
 static VALUE
 exit_success_p(VALUE exc)
 {
-    VALUE status = rb_attr_get(exc, rb_intern("status"));
-    if (NIL_P(status)) return Qtrue;
-    if (status == INT2FIX(EXIT_SUCCESS)) return Qtrue;
+    VALUE status_val = rb_attr_get(exc, rb_intern("status"));
+    int status;
+
+    if (NIL_P(status_val))
+	return Qtrue;
+    status = NUM2INT(status_val);
+    if (WIFEXITED(status) && WEXITSTATUS(status) == EXIT_SUCCESS)
+	return Qtrue;
+
     return Qfalse;
 }
 
@@ -721,6 +915,21 @@ rb_name_error(ID id, const char *fmt, ...)
     rb_exc_raise(exc);
 }
 
+void
+rb_name_error_str(VALUE str, const char *fmt, ...)
+{
+    VALUE exc, argv[2];
+    va_list args;
+
+    va_start(args, fmt);
+    argv[0] = rb_vsprintf(fmt, args);
+    va_end(args);
+
+    argv[1] = str;
+    exc = rb_class_new_instance(2, argv, rb_eNameError);
+    rb_exc_raise(exc);
+}
+
 /*
  * call-seq:
  *   NameError.new(msg [, name])  -> name_error
@@ -769,11 +978,7 @@ name_err_to_s(VALUE exc)
 
     if (NIL_P(mesg)) return rb_class_name(CLASS_OF(exc));
     StringValue(str);
-    if (str != mesg) {
-	rb_iv_set(exc, "mesg", mesg = str);
-    }
-    if (OBJ_TAINTED(exc)) OBJ_TAINT(mesg);
-    return mesg;
+    return str;
 }
 
 /*
@@ -815,9 +1020,11 @@ name_err_mesg_memsize(const void *p)
 
 static const rb_data_type_t name_err_mesg_data_type = {
     "name_err_mesg",
-    name_err_mesg_mark,
-    name_err_mesg_free,
-    name_err_mesg_memsize,
+    {
+	name_err_mesg_mark,
+	name_err_mesg_free,
+	name_err_mesg_memsize,
+    },
 };
 
 /* :nodoc: */
@@ -869,20 +1076,23 @@ name_err_mesg_to_str(VALUE obj)
     else {
 	const char *desc = 0;
 	VALUE d = 0, args[NAME_ERR_MESG_COUNT];
+	int state = 0;
 
 	obj = ptr[1];
-	switch (TYPE(obj)) {
-	  case T_NIL:
+	switch (obj) {
+	  case Qnil:
 	    desc = "nil";
 	    break;
-	  case T_TRUE:
+	  case Qtrue:
 	    desc = "true";
 	    break;
-	  case T_FALSE:
+	  case Qfalse:
 	    desc = "false";
 	    break;
 	  default:
-	    d = rb_protect(rb_inspect, obj, 0);
+	    d = rb_protect(rb_inspect, obj, &state);
+	    if (state)
+		rb_set_errinfo(Qnil);
 	    if (NIL_P(d) || RSTRING_LEN(d) > 65) {
 		d = rb_any_to_s(obj);
 	    }
@@ -892,17 +1102,23 @@ name_err_mesg_to_str(VALUE obj)
 	if (desc && desc[0] != '#') {
 	    d = d ? rb_str_dup(d) : rb_str_new2(desc);
 	    rb_str_cat2(d, ":");
-	    rb_str_cat2(d, rb_obj_classname(obj));
+	    rb_str_append(d, rb_class_name(CLASS_OF(obj)));
 	}
 	args[0] = mesg;
 	args[1] = ptr[2];
 	args[2] = d;
 	mesg = rb_f_sprintf(NAME_ERR_MESG_COUNT, args);
     }
-    if (OBJ_TAINTED(obj)) OBJ_TAINT(mesg);
     return mesg;
 }
 
+/* :nodoc: */
+static VALUE
+name_err_mesg_dump(VALUE obj, VALUE limit)
+{
+    return name_err_mesg_to_str(obj);
+}
+
 /* :nodoc: */
 static VALUE
 name_err_mesg_load(VALUE klass, VALUE str)
@@ -927,9 +1143,9 @@ nometh_err_args(VALUE self)
 void
 rb_invalid_str(const char *str, const char *type)
 {
-    volatile VALUE s = rb_str_inspect(rb_str_new2(str));
+    VALUE s = rb_str_new2(str);
 
-    rb_raise(rb_eArgError, "invalid value for %s: %s", type, RSTRING_PTR(s));
+    rb_raise(rb_eArgError, "invalid value for %s: %+"PRIsVALUE, type, s);
 }
 
 /*
@@ -968,7 +1184,7 @@ static st_table *syserr_tbl;
 static VALUE
 set_syserr(int n, const char *name)
 {
-    VALUE error;
+    st_data_t error;
 
     if (!st_lookup(syserr_tbl, n, &error)) {
 	error = rb_define_class_under(rb_mErrno, name, rb_eSystemCallError);
@@ -984,7 +1200,7 @@ set_syserr(int n, const char *name)
 static VALUE
 get_syserr(int n)
 {
-    VALUE error;
+    st_data_t error;
 
     if (!st_lookup(syserr_tbl, n, &error)) {
 	char name[8];	/* some Windows' errno have 5 digits. */
@@ -1017,13 +1233,15 @@ syserr_initialize(int argc, VALUE *argv, VALUE self)
     VALUE klass = rb_obj_class(self);
 
     if (klass == rb_eSystemCallError) {
+	st_data_t data = (st_data_t)klass;
 	rb_scan_args(argc, argv, "11", &mesg, &error);
 	if (argc == 1 && FIXNUM_P(mesg)) {
 	    error = mesg; mesg = Qnil;
 	}
-	if (!NIL_P(error) && st_lookup(syserr_tbl, NUM2LONG(error), &klass)) {
+	if (!NIL_P(error) && st_lookup(syserr_tbl, NUM2LONG(error), &data)) {
+	    klass = (VALUE)data;
 	    /* change class */
-	    if (TYPE(self) != T_OBJECT) { /* insurance to avoid type crash */
+	    if (!RB_TYPE_P(self, T_OBJECT)) { /* insurance to avoid type crash */
 		rb_raise(rb_eTypeError, "invalid instance type");
 	    }
 	    RBASIC(self)->klass = klass;
@@ -1036,16 +1254,21 @@ syserr_initialize(int argc, VALUE *argv, VALUE self)
     if (!NIL_P(error)) err = strerror(NUM2INT(error));
     else err = "unknown error";
     if (!NIL_P(mesg)) {
-	VALUE str = mesg;
+	rb_encoding *le = rb_locale_encoding();
+	VALUE str = StringValue(mesg);
+	rb_encoding *me = rb_enc_get(mesg);
 
-	StringValue(str);
-	mesg = rb_sprintf("%s - %.*s", err,
-			  (int)RSTRING_LEN(str), RSTRING_PTR(str));
+	mesg = rb_sprintf("%s - %"PRIsVALUE, err, mesg);
+	if (le != me && rb_enc_asciicompat(me)) {
+	    le = me;
+	}/* else assume err is non ASCII string. */
+	OBJ_INFECT(mesg, str);
+	rb_enc_associate(mesg, le);
     }
     else {
 	mesg = rb_str_new2(err);
+	rb_enc_associate(mesg, rb_locale_encoding());
     }
-    rb_enc_associate(mesg, rb_locale_encoding());
     rb_call_super(1, &mesg);
     rb_iv_set(self, "errno", error);
     return self;
@@ -1131,6 +1354,7 @@ syserr_eqq(VALUE self, VALUE exc)
  *
  *     begin
  *       Process.kill('HUP',Process.pid)
+ *       sleep # wait for receiver to handle signal sent by Process.kill
  *     rescue SignalException => e
  *       puts "received Exception #{e}"
  *     end
@@ -1172,7 +1396,7 @@ syserr_eqq(VALUE self, VALUE exc)
  *
  *  <em>raises the exception:</em>
  *
- *     TypeError: can't convert String into Integer
+ *     TypeError: no implicit conversion of String into Integer
  *
  */
 
@@ -1377,6 +1601,12 @@ syserr_eqq(VALUE self, VALUE exc)
  *     Errno::ENOENT: No such file or directory - does/not/exist
  */
 
+/*
+ * Document-class: EncodingError
+ *
+ * EncodingError is the base class for encoding errors.
+ */
+
 /*
  *  Document-class: Encoding::CompatibilityError
  *
@@ -1385,14 +1615,88 @@ syserr_eqq(VALUE self, VALUE exc)
  */
 
 /*
- *  Descendants of class <code>Exception</code> are used to communicate
- *  between <code>raise</code> methods and <code>rescue</code>
- *  statements in <code>begin/end</code> blocks. <code>Exception</code>
- *  objects carry information about the exception---its type (the
- *  exception's class name), an optional descriptive string, and
- *  optional traceback information. Programs may subclass
- *  <code>Exception</code>, or more typically <code>StandardError</code>
- *  to provide custom classes and add additional information.
+ * Document-class: fatal
+ *
+ * fatal is an Exception that is raised when ruby has encountered a fatal
+ * error and must exit.  You are not able to rescue fatal.
+ */
+
+/*
+ * Document-class: NameError::message
+ * :nodoc:
+ */
+
+/*
+ *  Descendants of class Exception are used to communicate between
+ *  Kernel#raise and +rescue+ statements in <code>begin ... end</code> blocks.
+ *  Exception objects carry information about the exception -- its type (the
+ *  exception's class name), an optional descriptive string, and optional
+ *  traceback information.  Exception subclasses may add additional
+ *  information like NameError#name.
+ *
+ *  Programs may make subclasses of Exception, typically of StandardError or
+ *  RuntimeError, to provide custom classes and add additional information.
+ *  See the subclass list below for defaults for +raise+ and +rescue+.
+ *
+ *  When an exception has been raised but not yet handled (in +rescue+,
+ *  +ensure+, +at_exit+ and +END+ blocks) the global variable <code>$!</code>
+ *  will contain the current exception and <code>$@</code> contains the
+ *  current exception's backtrace.
+ *
+ *  It is recommended that a library should have one subclass of StandardError
+ *  or RuntimeError and have specific exception types inherit from it.  This
+ *  allows the user to rescue a generic exception type to catch all exceptions
+ *  the library may raise even if future versions of the library add new
+ *  exception subclasses.
+ *
+ *  For example:
+ *
+ *    class MyLibrary
+ *      class Error < RuntimeError
+ *      end
+ *
+ *      class WidgetError < Error
+ *      end
+ *
+ *      class FrobError < Error
+ *      end
+ *
+ *    end
+ *
+ *  To handle both WidgetError and FrobError the library user can rescue
+ *  MyLibrary::Error.
+ *
+ *  The built-in subclasses of Exception are:
+ *
+ *  * NoMemoryError
+ *  * ScriptError
+ *    * LoadError
+ *    * NotImplementedError
+ *    * SyntaxError
+ *  * SignalException
+ *    * Interrupt
+ *  * StandardError -- default for +rescue+
+ *    * ArgumentError
+ *    * IndexError
+ *      * StopIteration
+ *    * IOError
+ *      * EOFError
+ *    * LocalJumpError
+ *    * NameError
+ *      * NoMethodError
+ *    * RangeError
+ *      * FloatDomainError
+ *    * RegexpError
+ *    * RuntimeError -- default for +raise+
+ *    * SecurityError
+ *    * SystemCallError
+ *      * Errno::*
+ *    * SystemStackError
+ *    * ThreadError
+ *    * TypeError
+ *    * ZeroDivisionError
+ *  * SystemExit
+ *  * fatal -- impossible to rescue
  */
 
 void
@@ -1427,7 +1731,10 @@ Init_Exception(void)
 
     rb_eScriptError = rb_define_class("ScriptError", rb_eException);
     rb_eSyntaxError = rb_define_class("SyntaxError", rb_eScriptError);
+
     rb_eLoadError   = rb_define_class("LoadError", rb_eScriptError);
+    rb_attr(rb_eLoadError, rb_intern("path"), 1, 0, Qfalse);
+
     rb_eNotImpError = rb_define_class("NotImplementedError", rb_eScriptError);
 
     rb_eNameError     = rb_define_class("NameError", rb_eStandardError);
@@ -1438,7 +1745,7 @@ Init_Exception(void)
     rb_define_singleton_method(rb_cNameErrorMesg, "!", rb_name_err_mesg_new, NAME_ERR_MESG_COUNT);
     rb_define_method(rb_cNameErrorMesg, "==", name_err_mesg_equal, 1);
     rb_define_method(rb_cNameErrorMesg, "to_str", name_err_mesg_to_str, 0);
-    rb_define_method(rb_cNameErrorMesg, "_dump", name_err_mesg_to_str, 1);
+    rb_define_method(rb_cNameErrorMesg, "_dump", name_err_mesg_dump, 1);
     rb_define_singleton_method(rb_cNameErrorMesg, "_load", name_err_mesg_load, 1);
     rb_eNoMethodError = rb_define_class("NoMethodError", rb_eNameError);
     rb_define_method(rb_eNoMethodError, "initialize", nometh_err_initialize, -1);
@@ -1458,7 +1765,20 @@ Init_Exception(void)
 
     rb_mErrno = rb_define_module("Errno");
 
-    rb_define_global_function("warn", rb_warn_m, 1);
+    rb_define_global_function("warn", rb_warn_m, -1);
+}
+
+void
+rb_enc_raise(rb_encoding *enc, VALUE exc, const char *fmt, ...)
+{
+    va_list args;
+    VALUE mesg;
+
+    va_start(args, fmt);
+    mesg = rb_enc_vsprintf(enc, fmt, args);
+    va_end(args);
+
+    rb_exc_raise(rb_exc_new3(exc, mesg));
 }
 
 void
@@ -1473,6 +1793,16 @@ rb_raise(VALUE exc, const char *fmt, ...)
     rb_exc_raise(rb_exc_new3(exc, mesg));
 }
 
+NORETURN(static void raise_loaderror(VALUE path, VALUE mesg));
+
+static void
+raise_loaderror(VALUE path, VALUE mesg)
+{
+    VALUE err = rb_exc_new3(rb_eLoadError, mesg);
+    rb_ivar_set(err, rb_intern("@path"), path);
+    rb_exc_raise(err);
+}
+
 void
 rb_loaderror(const char *fmt, ...)
 {
@@ -1482,7 +1812,19 @@ rb_loaderror(const char *fmt, ...)
     va_start(args, fmt);
     mesg = rb_enc_vsprintf(rb_locale_encoding(), fmt, args);
     va_end(args);
-    rb_exc_raise(rb_exc_new3(rb_eLoadError, mesg));
+    raise_loaderror(Qnil, mesg);
+}
+
+void
+rb_loaderror_with_path(VALUE path, const char *fmt, ...)
+{
+    va_list args;
+    VALUE mesg;
+
+    va_start(args, fmt);
+    mesg = rb_enc_vsprintf(rb_locale_encoding(), fmt, args);
+    va_end(args);
+    raise_loaderror(path, mesg);
 }
 
 void
@@ -1510,23 +1852,66 @@ static VALUE
 make_errno_exc(const char *mesg)
 {
     int n = errno;
-    VALUE arg;
 
     errno = 0;
     if (n == 0) {
 	rb_bug("rb_sys_fail(%s) - errno == 0", mesg ? mesg : "");
     }
+    return rb_syserr_new(n, mesg);
+}
 
+static VALUE
+make_errno_exc_str(VALUE mesg)
+{
+    int n = errno;
+
+    errno = 0;
+    if (!mesg) mesg = Qnil;
+    if (n == 0) {
+	const char *s = !NIL_P(mesg) ? RSTRING_PTR(mesg) : "";
+	rb_bug("rb_sys_fail_str(%s) - errno == 0", s);
+    }
+    return rb_syserr_new_str(n, mesg);
+}
+
+VALUE
+rb_syserr_new(int n, const char *mesg)
+{
+    VALUE arg;
     arg = mesg ? rb_str_new2(mesg) : Qnil;
+    return rb_syserr_new_str(n, arg);
+}
+
+VALUE
+rb_syserr_new_str(int n, VALUE arg)
+{
     return rb_class_new_instance(1, &arg, get_syserr(n));
 }
 
+void
+rb_syserr_fail(int e, const char *mesg)
+{
+    rb_exc_raise(rb_syserr_new(e, mesg));
+}
+
+void
+rb_syserr_fail_str(int e, VALUE mesg)
+{
+    rb_exc_raise(rb_syserr_new_str(e, mesg));
+}
+
 void
 rb_sys_fail(const char *mesg)
 {
     rb_exc_raise(make_errno_exc(mesg));
 }
 
+void
+rb_sys_fail_str(VALUE mesg)
+{
+    rb_exc_raise(make_errno_exc_str(mesg));
+}
+
 void
 rb_mod_sys_fail(VALUE mod, const char *mesg)
 {
@@ -1535,6 +1920,30 @@ rb_mod_sys_fail(VALUE mod, const char *mesg)
     rb_exc_raise(exc);
 }
 
+void
+rb_mod_sys_fail_str(VALUE mod, VALUE mesg)
+{
+    VALUE exc = make_errno_exc_str(mesg);
+    rb_extend_object(exc, mod);
+    rb_exc_raise(exc);
+}
+
+void
+rb_mod_syserr_fail(VALUE mod, int e, const char *mesg)
+{
+    VALUE exc = rb_syserr_new(e, mesg);
+    rb_extend_object(exc, mod);
+    rb_exc_raise(exc);
+}
+
+void
+rb_mod_syserr_fail_str(VALUE mod, int e, VALUE mesg)
+{
+    VALUE exc = rb_syserr_new_str(e, mesg);
+    rb_extend_object(exc, mod);
+    rb_exc_raise(exc);
+}
+
 void
 rb_sys_warning(const char *fmt, ...)
 {
@@ -1556,9 +1965,12 @@ rb_sys_warning(const char *fmt, ...)
 }
 
 void
-rb_load_fail(const char *path)
+rb_load_fail(VALUE path, const char *err)
 {
-    rb_loaderror("%s -- %s", strerror(errno), path);
+    VALUE mesg = rb_str_buf_new_cstr(err);
+    rb_str_cat2(mesg, " -- ");
+    rb_str_append(mesg, path);	/* should be ASCII compatible */
+    raise_loaderror(path, mesg);
 }
 
 void
@@ -1567,48 +1979,51 @@ rb_error_frozen(const char *what)
     rb_raise(rb_eRuntimeError, "can't modify frozen %s", what);
 }
 
+#undef rb_check_frozen
 void
 rb_check_frozen(VALUE obj)
 {
-    if (OBJ_FROZEN(obj)) rb_error_frozen(rb_obj_classname(obj));
+    rb_check_frozen_internal(obj);
+}
+
+void
+rb_error_untrusted(VALUE obj)
+{
+    if (rb_safe_level() >= 4) {
+	rb_raise(rb_eSecurityError, "Insecure: can't modify %s",
+		 rb_obj_classname(obj));
+    }
+}
+
+#undef rb_check_trusted
+void
+rb_check_trusted(VALUE obj)
+{
+    rb_check_trusted_internal(obj);
+}
+
+void
+rb_check_copyable(VALUE obj, VALUE orig)
+{
+    if (!FL_ABLE(obj)) return;
+    rb_check_frozen_internal(obj);
+    rb_check_trusted_internal(obj);
+    if (!FL_ABLE(orig)) return;
+    if ((~RBASIC(obj)->flags & RBASIC(orig)->flags) & (FL_UNTRUSTED|FL_TAINT)) {
+	if (rb_safe_level() > 0) {
+	    rb_raise(rb_eSecurityError, "Insecure: can't modify %"PRIsVALUE,
+		     RBASIC(obj)->klass);
+	}
+    }
 }
 
 void
 Init_syserr(void)
 {
     rb_eNOERROR = set_syserr(0, "NOERROR");
-#define defined_error(name, num) set_syserr(num, name);
-#define undefined_error(name) set_syserr(0, name);
+#define defined_error(name, num) set_syserr((num), (name));
+#define undefined_error(name) set_syserr(0, (name));
 #include "known_errors.inc"
 #undef defined_error
 #undef undefined_error
 }
-
-static void
-err_append(const char *s)
-{
-    rb_thread_t *th = GET_THREAD();
-    VALUE err = th->errinfo;
-
-    if (th->mild_compile_error) {
-	if (!RTEST(err)) {
-	    err = rb_exc_new2(rb_eSyntaxError, s);
-	    th->errinfo = err;
-	}
-	else {
-	    VALUE str = rb_obj_as_string(err);
-
-	    rb_str_cat2(str, "\n");
-	    rb_str_cat2(str, s);
-	    th->errinfo = rb_exc_new3(rb_eSyntaxError, str);
-	}
-    }
-    else {
-	if (!RTEST(err)) {
-	    err = rb_exc_new2(rb_eSyntaxError, "compile error");
-	    th->errinfo = err;
-	}
-	rb_write_error(s);
-	rb_write_error("\n");
-    }
-}
