@@ -2,7 +2,7 @@
 
   eval.c -
 
-  $Author: usa $
+  $Author: nobu $
   created at: Thu Jun 10 14:22:17 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -18,11 +18,14 @@
 #include "ruby/encoding.h"
 #include "internal.h"
 #include "vm_core.h"
+#include "probes_helper.h"
 
 #define numberof(array) (int)(sizeof(array) / sizeof((array)[0]))
 
 NORETURN(void rb_raise_jump(VALUE));
 
+NODE *rb_vm_get_cref(const rb_iseq_t *, const VALUE *);
+
 VALUE rb_eLocalJumpError;
 VALUE rb_eSysStackError;
 
@@ -31,16 +34,18 @@ VALUE rb_eSysStackError;
 #include "eval_error.c"
 #include "eval_jump.c"
 
-/* initialize ruby */
-
-void
-ruby_init(void)
+/* Initializes the Ruby VM and builtin libraries.
+ * @retval 0 if succeeded.
+ * @retval non-zero an error occured.
+ */
+int
+ruby_setup(void)
 {
     static int initialized = 0;
     int state;
 
     if (initialized)
-	return;
+	return 0;
     initialized = 1;
 
     ruby_init_stack((void *)&state);
@@ -51,16 +56,37 @@ ruby_init(void)
     if ((state = EXEC_TAG()) == 0) {
 	rb_call_inits();
 	ruby_prog_init();
+	GET_VM()->running = 1;
     }
     POP_TAG();
 
+    return state;
+}
+
+/* Calls ruby_setup() and check error.
+ *
+ * Prints errors and calls exit(3) if an error occured.
+ */
+void
+ruby_init(void)
+{
+    int state = ruby_setup();
     if (state) {
 	error_print();
 	exit(EXIT_FAILURE);
     }
-    GET_VM()->running = 1;
 }
 
+/*! Processes command line arguments and compiles the Ruby source to execute.
+ *
+ * This function does:
+ * \li  Processses the given command line flags and arguments for ruby(1)
+ * \li compiles the source code from the given argument, -e or stdin, and
+ * \li returns the compiled source as an opaque pointer to an internal data structure
+ *
+ * @return an opaque pointer to the compiled source or an internal special value.
+ * @sa ruby_executable_node().
+ */
 void *
 ruby_options(int argc, char **argv)
 {
@@ -101,6 +127,13 @@ ruby_finalize_1(void)
     rb_gc_call_finalizer_at_exit();
 }
 
+/** Runs the VM finalization processes.
+ *
+ * <code>END{}</code> and procs registered by <code>Kernel.#at_exit</code> are
+ * executed here. See the Ruby language spec for more details.
+ *
+ * @note This function is allowed to raise an exception if an error occurred.
+ */
 void
 ruby_finalize(void)
 {
@@ -108,6 +141,16 @@ ruby_finalize(void)
     ruby_finalize_1();
 }
 
+/** Destructs the VM.
+ *
+ * Runs the VM finalization processes as well as ruby_finalize(), and frees
+ * resources used by the VM.
+ *
+ * @param ex Default value to the return value.
+ * @return If an error occured returns a non-zero. If otherwise, returns the
+ *         given ex.
+ * @note This function does not raise any exception.
+ */
 int
 ruby_cleanup(volatile int ex)
 {
@@ -120,7 +163,7 @@ ruby_cleanup(volatile int ex)
     rb_threadptr_check_signal(th);
     PUSH_TAG();
     if ((state = EXEC_TAG()) == 0) {
-	SAVE_ROOT_JMPBUF(th, { RUBY_VM_CHECK_INTS(); });
+	SAVE_ROOT_JMPBUF(th, { RUBY_VM_CHECK_INTS(th); });
     }
     POP_TAG();
 
@@ -134,6 +177,9 @@ ruby_cleanup(volatile int ex)
     }
     POP_TAG();
 
+    /* protect from Thread#raise */
+    th->status = THREAD_KILLED;
+
     errs[0] = th->errinfo;
     PUSH_TAG();
     if ((state = EXEC_TAG()) == 0) {
@@ -169,7 +215,7 @@ ruby_cleanup(volatile int ex)
 	if (!RTEST(err)) continue;
 
 	/* th->errinfo contains a NODE while break'ing */
-	if (TYPE(err) == T_NODE) continue;
+	if (RB_TYPE_P(err, T_NODE)) continue;
 
 	if (rb_obj_is_kind_of(err, rb_eSystemExit)) {
 	    ex = sysexit_status(err);
@@ -210,12 +256,25 @@ ruby_exec_internal(void *n)
     return state;
 }
 
+/*! Calls ruby_cleanup() and exits the process */
 void
 ruby_stop(int ex)
 {
     exit(ruby_cleanup(ex));
 }
 
+/*! Checks the return value of ruby_options().
+ * @param n return value of ruby_options().
+ * @param status pointer to the exit status of this process.
+ *
+ * ruby_options() sometimes returns a special value to indicate this process
+ * should immediately exit. This function checks if the case. Also stores the
+ * exit status that the caller have to pass to exit(3) into
+ * <code>*status</code>.
+ *
+ * @retval non-zero if the given opaque pointer is actually a compiled source.
+ * @retval 0 if the given value is such a special value.
+ */
 int
 ruby_executable_node(void *n, int *status)
 {
@@ -233,6 +292,10 @@ ruby_executable_node(void *n, int *status)
     return FALSE;
 }
 
+/*! Runs the given compiled source and exits this process.
+ * @retval 0 if successfully run thhe source
+ * @retval non-zero if an error occurred.
+*/
 int
 ruby_run_node(void *n)
 {
@@ -244,6 +307,7 @@ ruby_run_node(void *n)
     return ruby_cleanup(ruby_exec_node(n));
 }
 
+/*! Runs the given compiled source */
 int
 ruby_exec_node(void *n)
 {
@@ -313,8 +377,8 @@ rb_mod_s_constants(int argc, VALUE *argv, VALUE mod)
     VALUE cbase = 0;
     void *data = 0;
 
-    if (argc > 0 || mod != rb_cModule) {
-	return rb_mod_constants(argc, argv, mod);
+    if (argc > 0) {
+	return rb_mod_constants(argc, argv, rb_cModule);
     }
 
     while (cref) {
@@ -387,7 +451,7 @@ setup_exception(rb_thread_t *th, int tag, volatile VALUE mesg)
 	else {
 	    at = get_backtrace(mesg);
 	    if (NIL_P(at)) {
-		at = rb_make_backtrace();
+		at = rb_vm_backtrace_object();
 		if (OBJ_FROZEN(mesg)) {
 		    mesg = rb_obj_dup(mesg);
 		}
@@ -438,10 +502,13 @@ setup_exception(rb_thread_t *th, int tag, volatile VALUE mesg)
 	JUMP_TAG(TAG_FATAL);
     }
 
-    rb_trap_restore_mask();
-
     if (tag != TAG_FATAL) {
-	EXEC_EVENT_HOOK(th, RUBY_EVENT_RAISE, th->cfp->self, 0, 0);
+	if (RUBY_DTRACE_RAISE_ENABLED()) {
+	    RUBY_DTRACE_RAISE(rb_obj_classname(th->errinfo),
+			      rb_sourcefile(),
+			      rb_sourceline());
+	}
+	EXEC_EVENT_HOOK(th, RUBY_EVENT_RAISE, th->cfp->self, 0, 0, mesg);
     }
 }
 
@@ -518,7 +585,8 @@ rb_f_raise(int argc, VALUE *argv)
 	}
     }
     rb_raise_jump(rb_make_exception(argc, argv));
-    return Qnil;		/* not reached */
+
+    UNREACHABLE;
 }
 
 static VALUE
@@ -557,7 +625,7 @@ make_exception(int argc, VALUE *argv, int isstr)
 	}
 	break;
       default:
-	rb_raise(rb_eArgError, "wrong number of arguments (%d for 0..3)", argc);
+	rb_check_arity(argc, 0, 3);
 	break;
     }
     if (argc > 0) {
@@ -589,7 +657,7 @@ rb_raise_jump(VALUE mesg)
 
     setup_exception(th, TAG_RAISE, mesg);
 
-    EXEC_EVENT_HOOK(th, RUBY_EVENT_C_RETURN, self, mid, klass);
+    EXEC_EVENT_HOOK(th, RUBY_EVENT_C_RETURN, self, mid, klass, Qnil);
     rb_thread_raised_clear(th);
     JUMP_TAG(TAG_RAISE);
 }
@@ -605,8 +673,7 @@ rb_block_given_p(void)
 {
     rb_thread_t *th = GET_THREAD();
 
-    if ((th->cfp->lfp[0] & 0x02) == 0 &&
-	GC_GUARDED_PTR_REF(th->cfp->lfp[0])) {
+    if (rb_vm_control_frame_block_ptr(th->cfp)) {
 	return TRUE;
     }
     else {
@@ -641,8 +708,8 @@ rb_rescue2(VALUE (* b_proc) (ANYARGS), VALUE data1,
     volatile VALUE e_info = th->errinfo;
     va_list args;
 
-    PUSH_TAG();
-    if ((state = EXEC_TAG()) == 0) {
+    TH_PUSH_TAG(th);
+    if ((state = TH_EXEC_TAG()) == 0) {
       retry_entry:
 	result = (*b_proc) (data1);
     }
@@ -685,7 +752,7 @@ rb_rescue2(VALUE (* b_proc) (ANYARGS), VALUE data1,
 	    }
 	}
     }
-    POP_TAG();
+    TH_POP_TAG();
     if (state)
 	JUMP_TAG(state);
 
@@ -712,15 +779,15 @@ rb_protect(VALUE (* proc) (VALUE), VALUE data, int * state)
 
     protect_tag.prev = th->protect_tag;
 
-    PUSH_TAG();
+    TH_PUSH_TAG(th);
     th->protect_tag = &protect_tag;
     MEMCPY(&org_jmpbuf, &(th)->root_jmpbuf, rb_jmpbuf_t, 1);
-    if ((status = EXEC_TAG()) == 0) {
+    if ((status = TH_EXEC_TAG()) == 0) {
 	SAVE_ROOT_JMPBUF(th, result = (*proc) (data));
     }
     MEMCPY(&(th)->root_jmpbuf, &org_jmpbuf, rb_jmpbuf_t, 1);
     th->protect_tag = protect_tag.prev;
-    POP_TAG();
+    TH_POP_TAG();
 
     if (state) {
 	*state = status;
@@ -738,6 +805,8 @@ rb_ensure(VALUE (*b_proc)(ANYARGS), VALUE data1, VALUE (*e_proc)(ANYARGS), VALUE
 {
     int state;
     volatile VALUE result = Qnil;
+    volatile VALUE errinfo;
+    rb_thread_t *const th = GET_THREAD();
 
     PUSH_TAG();
     if ((state = EXEC_TAG()) == 0) {
@@ -746,7 +815,9 @@ rb_ensure(VALUE (*b_proc)(ANYARGS), VALUE data1, VALUE (*e_proc)(ANYARGS), VALUE
     POP_TAG();
     /* TODO: fix me */
     /* retval = prot_tag ? prot_tag->retval : Qnil; */     /* save retval */
+    errinfo = th->errinfo;
     (*e_proc) (data2);
+    th->errinfo = errinfo;
     if (state)
 	JUMP_TAG(state);
     return result;
@@ -797,27 +868,66 @@ frame_func_id(rb_control_frame_t *cfp)
     return 0;
 }
 
+static ID
+frame_called_id(rb_control_frame_t *cfp)
+{
+    const rb_method_entry_t *me_local;
+    rb_iseq_t *iseq = cfp->iseq;
+    if (cfp->me) {
+	return cfp->me->called_id;
+    }
+    while (iseq) {
+	if (RUBY_VM_IFUNC_P(iseq)) {
+	    NODE *ifunc = (NODE *)iseq;
+	    if (ifunc->nd_aid) return ifunc->nd_aid;
+	    return rb_intern("<ifunc>");
+	}
+	me_local = method_entry_of_iseq(cfp, iseq);
+	if (me_local) {
+	    cfp->me = me_local;
+	    return me_local->called_id;
+	}
+	if (iseq->defined_method_id) {
+	    return iseq->defined_method_id;
+	}
+	if (iseq->local_iseq == iseq) {
+	    break;
+	}
+	iseq = iseq->parent_iseq;
+    }
+    return 0;
+}
+
 ID
 rb_frame_this_func(void)
 {
     return frame_func_id(GET_THREAD()->cfp);
 }
 
-ID
-rb_frame_callee(void)
+static rb_control_frame_t *
+previous_frame(rb_thread_t *th)
 {
-    return frame_func_id(GET_THREAD()->cfp);
-}
-
-static ID
-rb_frame_caller(void)
-{
-    rb_thread_t *th = GET_THREAD();
     rb_control_frame_t *prev_cfp = RUBY_VM_PREVIOUS_CONTROL_FRAME(th->cfp);
     /* check if prev_cfp can be accessible */
     if ((void *)(th->stack + th->stack_size) == (void *)(prev_cfp)) {
         return 0;
     }
+    return prev_cfp;
+}
+
+ID
+rb_frame_callee(void)
+{
+    rb_control_frame_t *prev_cfp = previous_frame(GET_THREAD());
+    if (!prev_cfp) return 0;
+    return frame_called_id(prev_cfp);
+}
+
+static ID
+rb_frame_caller(void)
+{
+    rb_control_frame_t *prev_cfp = previous_frame(GET_THREAD());
+    if (!prev_cfp) return 0;
     return frame_func_id(prev_cfp);
 }
 
@@ -867,16 +977,252 @@ static VALUE
 rb_mod_include(int argc, VALUE *argv, VALUE module)
 {
     int i;
+    ID id_append_features, id_included;
+
+    CONST_ID(id_append_features, "append_features");
+    CONST_ID(id_included, "included");
 
     for (i = 0; i < argc; i++)
 	Check_Type(argv[i], T_MODULE);
     while (argc--) {
-	rb_funcall(argv[argc], rb_intern("append_features"), 1, module);
-	rb_funcall(argv[argc], rb_intern("included"), 1, module);
+	rb_funcall(argv[argc], id_append_features, 1, module);
+	rb_funcall(argv[argc], id_included, 1, module);
     }
     return module;
 }
 
+/*
+ *  call-seq:
+ *     prepend_features(mod)   -> mod
+ *
+ *  When this module is prepended in another, Ruby calls
+ *  <code>prepend_features</code> in this module, passing it the
+ *  receiving module in _mod_. Ruby's default implementation is
+ *  to overlay the constants, methods, and module variables of this module
+ *  to _mod_ if this module has not already been added to
+ *  _mod_ or one of its ancestors. See also <code>Module#prepend</code>.
+ */
+
+static VALUE
+rb_mod_prepend_features(VALUE module, VALUE prepend)
+{
+    switch (TYPE(prepend)) {
+      case T_CLASS:
+      case T_MODULE:
+	break;
+      default:
+	Check_Type(prepend, T_CLASS);
+	break;
+    }
+    rb_prepend_module(prepend, module);
+
+    return module;
+}
+
+/*
+ *  call-seq:
+ *     prepend(module, ...)    -> self
+ *
+ *  Invokes <code>Module.prepend_features</code> on each parameter in reverse order.
+ */
+
+static VALUE
+rb_mod_prepend(int argc, VALUE *argv, VALUE module)
+{
+    int i;
+    ID id_prepend_features, id_prepended;
+
+    CONST_ID(id_prepend_features, "prepend_features");
+    CONST_ID(id_prepended, "prepended");
+    for (i = 0; i < argc; i++)
+	Check_Type(argv[i], T_MODULE);
+    while (argc--) {
+	rb_funcall(argv[argc], id_prepend_features, 1, module);
+	rb_funcall(argv[argc], id_prepended, 1, module);
+    }
+    return module;
+}
+
+static void
+warn_refinements_once()
+{
+    static int warned = 0;
+
+    if (warned)
+	return;
+    rb_warn("Refinements are experimental, and the behavior may change in future versions of Ruby!");
+    warned = 1;
+}
+
+static VALUE
+hidden_identity_hash_new()
+{
+    VALUE hash = rb_hash_new();
+
+    rb_funcall(hash, rb_intern("compare_by_identity"), 0);
+    RBASIC(hash)->klass = 0;  /* hide from ObjectSpace */
+    return hash;
+}
+
+void
+rb_using_refinement(NODE *cref, VALUE klass, VALUE module)
+{
+    VALUE iclass, c, superclass = klass;
+
+    Check_Type(klass, T_CLASS);
+    Check_Type(module, T_MODULE);
+    if (NIL_P(cref->nd_refinements)) {
+	cref->nd_refinements = hidden_identity_hash_new();
+    }
+    else {
+	if (cref->flags & NODE_FL_CREF_OMOD_SHARED) {
+	    cref->nd_refinements = rb_hash_dup(cref->nd_refinements);
+	    cref->flags &= ~NODE_FL_CREF_OMOD_SHARED;
+	}
+	if (!NIL_P(c = rb_hash_lookup(cref->nd_refinements, klass))) {
+	    superclass = c;
+	    while (c && RB_TYPE_P(c, T_ICLASS)) {
+		if (RBASIC(c)->klass == module) {
+		    /* already used refinement */
+		    return;
+		}
+		c = RCLASS_SUPER(c);
+	    }
+	}
+    }
+    FL_SET(module, RMODULE_IS_OVERLAID);
+    c = iclass = rb_include_class_new(module, superclass);
+    RCLASS_REFINED_CLASS(c) = klass;
+    RCLASS_M_TBL(c) = RCLASS_M_TBL(module);
+    module = RCLASS_SUPER(module);
+    while (module && module != klass) {
+	FL_SET(module, RMODULE_IS_OVERLAID);
+	c = RCLASS_SUPER(c) = rb_include_class_new(module, RCLASS_SUPER(c));
+	RCLASS_REFINED_CLASS(c) = klass;
+	module = RCLASS_SUPER(module);
+    }
+    rb_hash_aset(cref->nd_refinements, klass, iclass);
+}
+
+static int
+using_refinement(VALUE klass, VALUE module, VALUE arg)
+{
+    NODE *cref = (NODE *) arg;
+
+    rb_using_refinement(cref, klass, module);
+    return ST_CONTINUE;
+}
+
+void
+rb_using_module(NODE *cref, VALUE module)
+{
+    ID id_refinements;
+    VALUE refinements;
+
+    Check_Type(module, T_MODULE);
+    CONST_ID(id_refinements, "__refinements__");
+    refinements = rb_attr_get(module, id_refinements);
+    if (NIL_P(refinements)) return;
+    rb_hash_foreach(refinements, using_refinement, (VALUE) cref);
+}
+
+VALUE rb_refinement_module_get_refined_class(VALUE module)
+{
+    ID id_refined_class;
+
+    CONST_ID(id_refined_class, "__refined_class__");
+    return rb_attr_get(module, id_refined_class);
+}
+
+static void
+add_activated_refinement(VALUE activated_refinements,
+			 VALUE klass, VALUE refinement)
+{
+    VALUE iclass, c, superclass = klass;
+
+    if (!NIL_P(c = rb_hash_lookup(activated_refinements, klass))) {
+	superclass = c;
+	while (c && RB_TYPE_P(c, T_ICLASS)) {
+	    if (RBASIC(c)->klass == refinement) {
+		/* already used refinement */
+		return;
+	    }
+	    c = RCLASS_SUPER(c);
+	}
+    }
+    FL_SET(refinement, RMODULE_IS_OVERLAID);
+    c = iclass = rb_include_class_new(refinement, superclass);
+    RCLASS_REFINED_CLASS(c) = klass;
+    refinement = RCLASS_SUPER(refinement);
+    while (refinement) {
+	FL_SET(refinement, RMODULE_IS_OVERLAID);
+	c = RCLASS_SUPER(c) =
+	    rb_include_class_new(refinement, RCLASS_SUPER(c));
+	RCLASS_REFINED_CLASS(c) = klass;
+	refinement = RCLASS_SUPER(refinement);
+    }
+    rb_hash_aset(activated_refinements, klass, iclass);
+}
+
+VALUE rb_yield_refine_block(VALUE refinement, VALUE refinements);
+
+/*
+ *  call-seq:
+ *     refine(klass) { block }   -> module
+ *
+ *  Refine <i>klass</i> in the receiver.
+ *
+ *  Returns an overlaid module.
+ */
+
+static VALUE
+rb_mod_refine(VALUE module, VALUE klass)
+{
+    VALUE refinement;
+    ID id_refinements, id_activated_refinements,
+       id_refined_class, id_defined_at;
+    VALUE refinements, activated_refinements;
+    rb_thread_t *th = GET_THREAD();
+    rb_block_t *block = rb_vm_control_frame_block_ptr(th->cfp);
+
+    warn_refinements_once();
+    if (!block) {
+        rb_raise(rb_eArgError, "no block given");
+    }
+    if (block->proc) {
+        rb_raise(rb_eArgError,
+		 "can't pass a Proc as a block to Module#refine");
+    }
+    Check_Type(klass, T_CLASS);
+    CONST_ID(id_refinements, "__refinements__");
+    refinements = rb_attr_get(module, id_refinements);
+    if (NIL_P(refinements)) {
+	refinements = hidden_identity_hash_new();
+	rb_ivar_set(module, id_refinements, refinements);
+    }
+    CONST_ID(id_activated_refinements, "__activated_refinements__");
+    activated_refinements = rb_attr_get(module, id_activated_refinements);
+    if (NIL_P(activated_refinements)) {
+	activated_refinements = hidden_identity_hash_new();
+	rb_ivar_set(module, id_activated_refinements,
+		    activated_refinements);
+    }
+    refinement = rb_hash_lookup(refinements, klass);
+    if (NIL_P(refinement)) {
+	refinement = rb_module_new();
+	RCLASS_SUPER(refinement) = klass;
+	FL_SET(refinement, RMODULE_IS_REFINEMENT);
+	CONST_ID(id_refined_class, "__refined_class__");
+	rb_ivar_set(refinement, id_refined_class, klass);
+	CONST_ID(id_defined_at, "__defined_at__");
+	rb_ivar_set(refinement, id_defined_at, module);
+	rb_hash_aset(refinements, klass, refinement);
+	add_activated_refinement(activated_refinements, klass, refinement);
+    }
+    rb_yield_refine_block(refinement, activated_refinements);
+    return refinement;
+}
+
 void
 rb_obj_call_init(VALUE obj, int argc, VALUE *argv)
 {
@@ -953,15 +1299,17 @@ static VALUE
 rb_obj_extend(int argc, VALUE *argv, VALUE obj)
 {
     int i;
+    ID id_extend_object, id_extended;
 
-    if (argc == 0) {
-	rb_raise(rb_eArgError, "wrong number of arguments (at least 1)");
-    }
+    CONST_ID(id_extend_object, "extend_object");
+    CONST_ID(id_extended, "extended");
+
+    rb_check_arity(argc, 1, UNLIMITED_ARGUMENTS);
     for (i = 0; i < argc; i++)
 	Check_Type(argv[i], T_MODULE);
     while (argc--) {
-	rb_funcall(argv[argc], rb_intern("extend_object"), 1, obj);
-	rb_funcall(argv[argc], rb_intern("extended"), 1, obj);
+	rb_funcall(argv[argc], id_extend_object, 1, obj);
+	rb_funcall(argv[argc], id_extended, 1, obj);
     }
     return obj;
 }
@@ -982,13 +1330,36 @@ top_include(int argc, VALUE *argv, VALUE self)
 
     rb_secure(4);
     if (th->top_wrapper) {
-	rb_warning
-	    ("main#include in the wrapped load is effective only in wrapper module");
+	rb_warning("main.include in the wrapped load is effective only in wrapper module");
 	return rb_mod_include(argc, argv, th->top_wrapper);
     }
     return rb_mod_include(argc, argv, rb_cObject);
 }
 
+/*
+ *  call-seq:
+ *     using(module)    -> self
+ *
+ *  Import class refinements from <i>module</i> into the scope where
+ *  <code>using</code> is called.
+ */
+
+static VALUE
+top_using(VALUE self, VALUE module)
+{
+    NODE *cref = rb_vm_cref();
+    rb_control_frame_t *prev_cfp = previous_frame(GET_THREAD());
+
+    warn_refinements_once();
+    if (cref->nd_next || (prev_cfp && prev_cfp->me)) {
+	rb_raise(rb_eRuntimeError, "using is permitted only at toplevel");
+    }
+    Check_Type(module, T_MODULE);
+    rb_using_module(cref, module);
+    rb_clear_cache();
+    return self;
+}
+
 static VALUE *
 errinfo_place(rb_thread_t *th)
 {
@@ -998,12 +1369,12 @@ errinfo_place(rb_thread_t *th)
     while (RUBY_VM_VALID_CONTROL_FRAME_P(cfp, end_cfp)) {
 	if (RUBY_VM_NORMAL_ISEQ_P(cfp->iseq)) {
 	    if (cfp->iseq->type == ISEQ_TYPE_RESCUE) {
-		return &cfp->dfp[-2];
+		return &cfp->ep[-2];
 	    }
 	    else if (cfp->iseq->type == ISEQ_TYPE_ENSURE &&
-		     TYPE(cfp->dfp[-2]) != T_NODE &&
-		     !FIXNUM_P(cfp->dfp[-2])) {
-		return &cfp->dfp[-2];
+		     !RB_TYPE_P(cfp->ep[-2], T_NODE) &&
+		     !FIXNUM_P(cfp->ep[-2])) {
+		return &cfp->ep[-2];
 	    }
 	}
 	cfp = RUBY_VM_PREVIOUS_CONTROL_FRAME(cfp);
@@ -1121,6 +1492,40 @@ rb_f_method_name(void)
     }
 }
 
+static VALUE
+rb_f_callee_name(void)
+{
+    ID fname = rb_frame_callee(); /* need *callee* ID */
+
+    if (fname) {
+	return ID2SYM(fname);
+    }
+    else {
+	return Qnil;
+    }
+}
+
+/*
+ *  call-seq:
+ *     __dir__         -> string
+ *
+ *  Returns the canonicalized absolute path of the directory of the file from
+ *  which this method is called. It means symlinks in the path is resolved.
+ *  If <code>__FILE__</code> is <code>nil</code>, it returns <code>nil</code>.
+ *  The return value equals to <code>File.dirname(File.realpath(__FILE__))</code>.
+ *
+ */
+static VALUE
+f_current_dirname(void)
+{
+    VALUE base = rb_current_realfilepath();
+    if (NIL_P(base)) {
+	return Qnil;
+    }
+    base = rb_file_dirname(base);
+    return base;
+}
+
 void
 Init_eval(void)
 {
@@ -1133,11 +1538,16 @@ Init_eval(void)
     rb_define_global_function("global_variables", rb_f_global_variables, 0);	/* in variable.c */
 
     rb_define_global_function("__method__", rb_f_method_name, 0);
-    rb_define_global_function("__callee__", rb_f_method_name, 0);
+    rb_define_global_function("__callee__", rb_f_callee_name, 0);
+    rb_define_global_function("__dir__", f_current_dirname, 0);
 
     rb_define_private_method(rb_cModule, "append_features", rb_mod_append_features, 1);
     rb_define_private_method(rb_cModule, "extend_object", rb_mod_extend_object, 1);
     rb_define_private_method(rb_cModule, "include", rb_mod_include, -1);
+    rb_define_private_method(rb_cModule, "prepend_features", rb_mod_prepend_features, 1);
+    rb_define_private_method(rb_cModule, "prepend", rb_mod_prepend, -1);
+    rb_define_private_method(rb_cModule, "refine", rb_mod_refine, 1);
+    rb_undef_method(rb_cClass, "refine");
 
     rb_undef_method(rb_cClass, "module_function");
 
@@ -1147,7 +1557,10 @@ Init_eval(void)
     rb_define_singleton_method(rb_cModule, "nesting", rb_mod_nesting, 0);
     rb_define_singleton_method(rb_cModule, "constants", rb_mod_s_constants, -1);
 
-    rb_define_singleton_method(rb_vm_top_self(), "include", top_include, -1);
+    rb_define_private_method(rb_singleton_class(rb_vm_top_self()),
+			     "include", top_include, -1);
+    rb_define_private_method(rb_singleton_class(rb_vm_top_self()),
+			     "using", top_using, 1);
 
     rb_define_method(rb_mKernel, "extend", rb_obj_extend, -1);
 
