@@ -6,17 +6,18 @@
 static void
 warn_printf(const char *fmt, ...)
 {
-    char buf[BUFSIZ];
+    VALUE str;
     va_list args;
 
     va_init_list(args, fmt);
-    vsnprintf(buf, BUFSIZ, fmt, args);
+    str = rb_vsprintf(fmt, args);
     va_end(args);
-    rb_write_error(buf);
+    rb_write_error_str(str);
 }
 
 #define warn_print(x) rb_write_error(x)
-#define warn_print2(x,l) rb_write_error2(x,l)
+#define warn_print2(x,l) rb_write_error2((x),(l))
+#define warn_print_str(x) rb_write_error_str(x)
 
 static void
 error_pos(void)
@@ -38,8 +39,6 @@ error_pos(void)
     }
 }
 
-VALUE rb_check_backtrace(VALUE);
-
 static VALUE
 get_backtrace(VALUE info)
 {
@@ -57,9 +56,22 @@ rb_get_backtrace(VALUE info)
     return get_backtrace(info);
 }
 
+VALUE rb_exc_set_backtrace(VALUE exc, VALUE bt);
+
 static void
 set_backtrace(VALUE info, VALUE bt)
 {
+    ID set_backtrace = rb_intern("set_backtrace");
+
+    if (rb_backtrace_p(bt)) {
+	if (rb_method_basic_definition_p(CLASS_OF(info), set_backtrace)) {
+	    rb_exc_set_backtrace(info, bt);
+	    return;
+	}
+	else {
+	    bt = rb_backtrace_to_str_ary(bt);
+	}
+    }
     rb_funcall(info, rb_intern("set_backtrace"), 1, bt);
 }
 
@@ -78,14 +90,14 @@ error_print(void)
 	return;
     rb_thread_raised_clear(th);
 
-    PUSH_TAG();
-    if (EXEC_TAG() == 0) {
+    TH_PUSH_TAG(th);
+    if (TH_EXEC_TAG() == 0) {
 	errat = get_backtrace(errinfo);
     }
     else {
 	errat = Qnil;
     }
-    if (EXEC_TAG())
+    if (TH_EXEC_TAG())
 	goto error;
     if (NIL_P(errat)) {
 	const char *file = rb_sourcefile();
@@ -106,12 +118,12 @@ error_print(void)
 	if (NIL_P(mesg))
 	    error_pos();
 	else {
-	    warn_print2(RSTRING_PTR(mesg), RSTRING_LEN(mesg));
+	    warn_print_str(mesg);
 	}
     }
 
     eclass = CLASS_OF(errinfo);
-    if (EXEC_TAG() == 0) {
+    if (TH_EXEC_TAG() == 0) {
 	e = rb_funcall(errinfo, rb_intern("message"), 0, 0);
 	StringValue(e);
 	einfo = RSTRING_PTR(e);
@@ -121,7 +133,7 @@ error_print(void)
 	einfo = "";
 	elen = 0;
     }
-    if (EXEC_TAG())
+    if (TH_EXEC_TAG())
 	goto error;
     if (eclass == rb_eRuntimeError && elen == 0) {
 	warn_print(": unhandled exception\n");
@@ -132,7 +144,7 @@ error_print(void)
 	epath = rb_class_name(eclass);
 	if (elen == 0) {
 	    warn_print(": ");
-	    warn_print2(RSTRING_PTR(epath), RSTRING_LEN(epath));
+	    warn_print_str(epath);
 	    warn_print("\n");
 	}
 	else {
@@ -149,7 +161,7 @@ error_print(void)
 	    warn_print2(einfo, len);
 	    if (epath) {
 		warn_print(" (");
-		warn_print2(RSTRING_PTR(epath), RSTRING_LEN(epath));
+		warn_print_str(epath);
 		warn_print(")\n");
 	    }
 	    if (tail) {
@@ -170,8 +182,8 @@ error_print(void)
 #define TRACE_TAIL 5
 
 	for (i = 1; i < len; i++) {
-	    if (TYPE(ptr[i]) == T_STRING) {
-		warn_printf("\tfrom %s\n", RSTRING_PTR(ptr[i]));
+	    if (RB_TYPE_P(ptr[i], T_STRING)) {
+		warn_printf("\tfrom %"PRIsVALUE"\n", ptr[i]);
 	    }
 	    if (skip && i == TRACE_HEAD && len > TRACE_MAX) {
 		warn_printf("\t ... %ld levels...\n",
@@ -181,7 +193,7 @@ error_print(void)
 	}
     }
   error:
-    POP_TAG();
+    TH_POP_TAG();
     rb_thread_raised_set(th, raised_flag);
 }
 
@@ -202,10 +214,19 @@ rb_print_undef(VALUE klass, ID id, int scope)
       case NOEX_PRIVATE: v = " private"; break;
       case NOEX_PROTECTED: v = " protected"; break;
     }
-    rb_name_error(id, "undefined%s method `%s' for %s `%s'", v,
-		  rb_id2name(id),
-		  (TYPE(klass) == T_MODULE) ? "module" : "class",
-		  rb_class2name(klass));
+    rb_name_error(id, "undefined%s method `%"PRIsVALUE"' for %s `%"PRIsVALUE"'", v,
+		  QUOTE_ID(id),
+		  (RB_TYPE_P(klass, T_MODULE)) ? "module" : "class",
+		  rb_class_name(klass));
+}
+
+void
+rb_print_undef_str(VALUE klass, VALUE name)
+{
+    rb_name_error_str(name, "undefined method `%"PRIsVALUE"' for %s `%"PRIsVALUE"'",
+		      QUOTE(name),
+		      (RB_TYPE_P(klass, T_MODULE)) ? "module" : "class",
+		      rb_class_name(klass));
 }
 
 static int
