@@ -5,7 +5,7 @@
 #include "vm_core.h"
 
 #define PASS_PASSED_BLOCK_TH(th) do { \
-    (th)->passed_block = GC_GUARDED_PTR_REF((rb_block_t *)(th)->cfp->lfp[0]); \
+    (th)->passed_block = rb_vm_control_frame_block_ptr(th->cfp); \
     (th)->cfp->flag |= VM_FRAME_FLAG_PASSED; \
 } while (0)
 
@@ -28,29 +28,12 @@
 #include <setjmp.h>
 
 #ifdef __APPLE__
-#include <crt_externs.h>
-#endif
-
-/* Make alloca work the best possible way.  */
-#ifdef __GNUC__
-# ifndef atarist
-#  ifndef alloca
-#   define alloca __builtin_alloca
-#  endif
-# endif	/* atarist */
-#else
-# ifdef HAVE_ALLOCA_H
-#  include <alloca.h>
+# ifdef HAVE_CRT_EXTERNS_H
+#  include <crt_externs.h>
 # else
-#  ifdef _AIX
-#pragma alloca
-#  else
-#   ifndef alloca		/* predefined by HP cc +Olibcalls */
-void *alloca();
-#   endif
-#  endif /* AIX */
-# endif	/* HAVE_ALLOCA_H */
-#endif /* __GNUC__ */
+#  include "missing/crt_externs.h"
+# endif
+#endif
 
 #ifndef HAVE_STRING_H
 char *strrchr(const char *, const char);
@@ -65,7 +48,7 @@ char *strrchr(const char *, const char);
 #endif
 
 #define ruby_setjmp(env) RUBY_SETJMP(env)
-#define ruby_longjmp(env,val) RUBY_LONGJMP(env,val)
+#define ruby_longjmp(env,val) RUBY_LONGJMP((env),(val))
 #ifdef __CYGWIN__
 # ifndef _setjmp
 int _setjmp(jmp_buf);
@@ -90,7 +73,8 @@ NORETURN(void _longjmp(jmp_buf, int));
   So following definition is required to use select_large_fdset.
 */
 #ifdef HAVE_SELECT_LARGE_FDSET
-#define select(n, r, w, e, t) select_large_fdset(n, r, w, e, t)
+#define select(n, r, w, e, t) select_large_fdset((n), (r), (w), (e), (t))
+extern int select_large_fdset(int, fd_set *, fd_set *, fd_set *, struct timeval *);
 #endif
 
 #ifdef HAVE_SYS_PARAM_H
@@ -99,16 +83,44 @@ NORETURN(void _longjmp(jmp_buf, int));
 
 #include <sys/stat.h>
 
+#ifdef _MSC_VER
+#define SAVE_ROOT_JMPBUF_BEFORE_STMT \
+    __try {
+#define SAVE_ROOT_JMPBUF_AFTER_STMT \
+    } \
+    __except (GetExceptionCode() == EXCEPTION_STACK_OVERFLOW ? \
+	      (rb_thread_raised_set(GET_THREAD(), RAISED_STACKOVERFLOW), \
+	       raise(SIGSEGV), \
+	       EXCEPTION_EXECUTE_HANDLER) : \
+	      EXCEPTION_CONTINUE_SEARCH) { \
+	/* never reaches here */ \
+    }
+#elif defined(__MINGW32__)
+LONG WINAPI rb_w32_stack_overflow_handler(struct _EXCEPTION_POINTERS *);
+#define SAVE_ROOT_JMPBUF_BEFORE_STMT \
+    do { \
+	PVOID _handler = AddVectoredExceptionHandler(1, rb_w32_stack_overflow_handler);
+
+#define SAVE_ROOT_JMPBUF_AFTER_STMT \
+	RemoveVectoredExceptionHandler(_handler); \
+    } while (0);
+#else
+#define SAVE_ROOT_JMPBUF_BEFORE_STMT
+#define SAVE_ROOT_JMPBUF_AFTER_STMT
+#endif
+
 #define SAVE_ROOT_JMPBUF(th, stmt) do \
   if (ruby_setjmp((th)->root_jmpbuf) == 0) { \
+      SAVE_ROOT_JMPBUF_BEFORE_STMT \
       stmt; \
+      SAVE_ROOT_JMPBUF_AFTER_STMT \
   } \
   else { \
       rb_fiber_start(); \
   } while (0)
 
 #define TH_PUSH_TAG(th) do { \
-  rb_thread_t * const _th = th; \
+  rb_thread_t * const _th = (th); \
   struct rb_vm_tag _tag; \
   _tag.tag = 0; \
   _tag.prev = _th->tag; \
@@ -130,10 +142,12 @@ NORETURN(void _longjmp(jmp_buf, int));
   TH_EXEC_TAG()
 
 #define TH_JUMP_TAG(th, st) do { \
-  ruby_longjmp(th->tag->buf,(st)); \
+  ruby_longjmp((th)->tag->buf,(st)); \
 } while (0)
 
-#define JUMP_TAG(st) TH_JUMP_TAG(GET_THREAD(), st)
+#define JUMP_TAG(st) TH_JUMP_TAG(GET_THREAD(), (st))
+
+#define INTERNAL_EXCEPTION_P(exc) FIXNUM_P(exc)
 
 enum ruby_tag_type {
     RUBY_TAG_RETURN	= 0x1,
@@ -171,12 +185,6 @@ enum ruby_tag_type {
 #define SCOPE_CHECK(f) (rb_vm_cref()->nd_visi == (f))
 #define SCOPE_SET(f)   (rb_vm_cref()->nd_visi = (f))
 
-#define CHECK_STACK_OVERFLOW(cfp, margin) do \
-  if ((VALUE *)((char *)(((VALUE *)(cfp)->sp) + (margin)) + sizeof(rb_control_frame_t)) >= ((VALUE *)cfp)) { \
-      rb_exc_raise(sysstack_error); \
-  } \
-while (0)
-
 void rb_thread_cleanup(void);
 void rb_thread_wait_other_threads(void);
 
@@ -195,26 +203,27 @@ int rb_threadptr_reset_raised(rb_thread_t *th);
 VALUE rb_f_eval(int argc, VALUE *argv, VALUE self);
 VALUE rb_make_exception(int argc, VALUE *argv);
 
+NORETURN(void rb_method_name_error(VALUE, VALUE));
+
 NORETURN(void rb_fiber_start(void));
 
 NORETURN(void rb_print_undef(VALUE, ID, int));
+NORETURN(void rb_print_undef_str(VALUE, VALUE));
 NORETURN(void rb_vm_localjump_error(const char *,VALUE, int));
-NORETURN(void rb_vm_jump_tag_but_local_jump(int, VALUE));
+NORETURN(void rb_vm_jump_tag_but_local_jump(int));
 NORETURN(void rb_raise_method_missing(rb_thread_t *th, int argc, VALUE *argv,
 				      VALUE obj, int call_status));
 
 VALUE rb_vm_make_jump_tag_but_local_jump(int state, VALUE val);
 NODE *rb_vm_cref(void);
-VALUE rb_vm_call_cfunc(VALUE recv, VALUE (*func)(VALUE), VALUE arg, const rb_block_t *blockptr, VALUE filename, VALUE filepath);
+VALUE rb_vm_call_cfunc(VALUE recv, VALUE (*func)(VALUE), VALUE arg, const rb_block_t *blockptr, VALUE filename);
 void rb_vm_set_progname(VALUE filename);
 void rb_thread_terminate_all(void);
 VALUE rb_vm_top_self();
 VALUE rb_vm_cbase(void);
-int rb_vm_get_sourceline(const rb_control_frame_t *);
-void rb_trap_restore_mask(void);
 
 #ifndef CharNext		/* defined as CharNext[AW] on Windows. */
-#define CharNext(p) ((p) + mblen(p, RUBY_MBCHAR_MAXSIZE))
+#define CharNext(p) ((p) + mblen((p), RUBY_MBCHAR_MAXSIZE))
 #endif
 
 #if defined DOSISH || defined __CYGWIN__
