@@ -1,5 +1,5 @@
 /* -*- C -*-
- * $Id: cptr.c 32983 2011-08-16 00:51:58Z drbrain $
+ * $Id: cptr.c 46508 2014-06-23 07:52:37Z usa $
  */
 
 #include <ruby/ruby.h>
@@ -10,19 +10,36 @@
 VALUE rb_cDLCPtr;
 
 static inline freefunc_t
-get_freefunc(VALUE func)
+get_freefunc(VALUE func, volatile VALUE *wrap)
 {
+    VALUE addrnum;
     if (NIL_P(func)) {
+	*wrap = 0;
 	return NULL;
     }
     if (rb_dlcfunc_kind_p(func)) {
+	*wrap = func;
 	return (freefunc_t)(VALUE)RCFUNC_DATA(func)->ptr;
     }
-    return (freefunc_t)(VALUE)NUM2PTR(rb_Integer(func));
+    addrnum = rb_Integer(func);
+    *wrap = (addrnum != func) ? func : 0;
+    return (freefunc_t)(VALUE)NUM2PTR(addrnum);
 }
 
 static ID id_to_ptr;
 
+static void
+dlptr_mark(void *ptr)
+{
+    struct ptr_data *data = ptr;
+    if (data->wrap[0]) {
+	rb_gc_mark(data->wrap[0]);
+    }
+    if (data->wrap[1]) {
+	rb_gc_mark(data->wrap[1]);
+    }
+}
+
 static void
 dlptr_free(void *ptr)
 {
@@ -32,6 +49,7 @@ dlptr_free(void *ptr)
 	    (*(data->free))(data->ptr);
 	}
     }
+    xfree(ptr);
 }
 
 static size_t
@@ -43,18 +61,9 @@ dlptr_memsize(const void *ptr)
 
 static const rb_data_type_t dlptr_data_type = {
     "dl/ptr",
-    {0, dlptr_free, dlptr_memsize,},
+    {dlptr_mark, dlptr_free, dlptr_memsize,},
 };
 
-void
-dlptr_init(VALUE val)
-{
-    struct ptr_data *data;
-
-    TypedData_Get_Struct(val, struct ptr_data, &dlptr_data_type, data);
-    OBJ_TAINT(val);
-}
-
 VALUE
 rb_dlptr_new2(VALUE klass, void *ptr, long size, freefunc_t func)
 {
@@ -66,7 +75,7 @@ rb_dlptr_new2(VALUE klass, void *ptr, long size, freefunc_t func)
     data->ptr = ptr;
     data->free = func;
     data->size = size;
-    dlptr_init(val);
+    OBJ_TAINT(val);
 
     return val;
 }
@@ -135,27 +144,22 @@ rb_dlptr_s_allocate(VALUE klass)
 static VALUE
 rb_dlptr_initialize(int argc, VALUE argv[], VALUE self)
 {
-    VALUE ptr, sym, size;
+    VALUE ptr, sym, size, wrap = 0, funcwrap = 0;
     struct ptr_data *data;
     void *p = NULL;
     freefunc_t f = NULL;
     long s = 0;
 
-    switch (rb_scan_args(argc, argv, "12", &ptr, &size, &sym)) {
-      case 1:
-	p = (void*)(NUM2PTR(rb_Integer(ptr)));
-	break;
-      case 2:
-	p = (void*)(NUM2PTR(rb_Integer(ptr)));
+    if (rb_scan_args(argc, argv, "12", &ptr, &size, &sym) >= 1) {
+	VALUE addrnum = rb_Integer(ptr);
+	if (addrnum != ptr) wrap = ptr;
+	p = NUM2PTR(addrnum);
+    }
+    if (argc >= 2) {
 	s = NUM2LONG(size);
-	break;
-      case 3:
-	p = (void*)(NUM2PTR(rb_Integer(ptr)));
-	s = NUM2LONG(size);
-	f = get_freefunc(sym);
-	break;
-      default:
-	rb_bug("rb_dlptr_initialize");
+    }
+    if (argc >= 3) {
+	f = get_freefunc(sym, &funcwrap);
     }
 
     if (p) {
@@ -164,6 +168,8 @@ rb_dlptr_initialize(int argc, VALUE argv[], VALUE self)
 	    /* Free previous memory. Use of inappropriate initialize may cause SEGV. */
 	    (*(data->free))(data->ptr);
 	}
+	data->wrap[0] = wrap;
+	data->wrap[1] = funcwrap;
 	data->ptr  = p;
 	data->size = s;
 	data->free = f;
@@ -185,7 +191,7 @@ rb_dlptr_initialize(int argc, VALUE argv[], VALUE self)
 static VALUE
 rb_dlptr_s_malloc(int argc, VALUE argv[], VALUE klass)
 {
-    VALUE size, sym, obj;
+    VALUE size, sym, obj, wrap = 0;
     long s;
     freefunc_t f;
 
@@ -196,13 +202,14 @@ rb_dlptr_s_malloc(int argc, VALUE argv[], VALUE klass)
 	break;
       case 2:
 	s = NUM2LONG(size);
-	f = get_freefunc(sym);
+	f = get_freefunc(sym, &wrap);
 	break;
       default:
 	rb_bug("rb_dlptr_s_malloc");
     }
 
     obj = rb_dlptr_malloc(s,f);
+    if (wrap) RPTR_DATA(obj)->wrap[1] = wrap;
 
     return obj;
 }
@@ -289,7 +296,7 @@ rb_dlptr_free_set(VALUE self, VALUE val)
     struct ptr_data *data;
 
     TypedData_Get_Struct(self, struct ptr_data, &dlptr_data_type, data);
-    data->free = get_freefunc(val);
+    data->free = get_freefunc(val, &data->wrap[1]);
 
     return Qnil;
 }
@@ -385,12 +392,10 @@ static VALUE
 rb_dlptr_inspect(VALUE self)
 {
     struct ptr_data *data;
-    char str[1024];
 
     TypedData_Get_Struct(self, struct ptr_data, &dlptr_data_type, data);
-    snprintf(str, 1023, "#<%s:%p ptr=%p size=%ld free=%p>",
-	     rb_class2name(CLASS_OF(self)), data, data->ptr, data->size, data->free);
-    return rb_str_new2(str);
+    return rb_sprintf("#<%"PRIsVALUE":%p ptr=%p size=%ld free=%p>",
+		      rb_obj_class(self), data, data->ptr, data->size, data->free);
 }
 
 /*
@@ -487,16 +492,19 @@ rb_dlptr_aref(int argc, VALUE argv[], VALUE self)
     VALUE arg0, arg1;
     VALUE retval = Qnil;
     size_t offset, len;
+    struct ptr_data *data;
 
+    TypedData_Get_Struct(self, struct ptr_data, &dlptr_data_type, data);
+    if (!data->ptr) rb_raise(rb_eDLError, "NULL pointer dereference");
     switch( rb_scan_args(argc, argv, "11", &arg0, &arg1) ){
       case 1:
 	offset = NUM2ULONG(arg0);
-	retval = INT2NUM(*((char*)RPTR_DATA(self)->ptr + offset));
+	retval = INT2NUM(*((char *)data->ptr + offset));
 	break;
       case 2:
 	offset = NUM2ULONG(arg0);
 	len    = NUM2ULONG(arg1);
-	retval = rb_tainted_str_new((char *)RPTR_DATA(self)->ptr + offset, len);
+	retval = rb_tainted_str_new((char *)data->ptr + offset, len);
 	break;
       default:
 	rb_bug("rb_dlptr_aref()");
@@ -520,17 +528,20 @@ rb_dlptr_aset(int argc, VALUE argv[], VALUE self)
     VALUE retval = Qnil;
     size_t offset, len;
     void *mem;
+    struct ptr_data *data;
 
+    TypedData_Get_Struct(self, struct ptr_data, &dlptr_data_type, data);
+    if (!data->ptr) rb_raise(rb_eDLError, "NULL pointer dereference");
     switch( rb_scan_args(argc, argv, "21", &arg0, &arg1, &arg2) ){
       case 2:
 	offset = NUM2ULONG(arg0);
-	((char*)RPTR_DATA(self)->ptr)[offset] = NUM2UINT(arg1);
+	((char*)data->ptr)[offset] = NUM2UINT(arg1);
 	retval = arg1;
 	break;
       case 3:
 	offset = NUM2ULONG(arg0);
 	len    = NUM2ULONG(arg1);
-	if( TYPE(arg2) == T_STRING ){
+	if (RB_TYPE_P(arg2, T_STRING)) {
 	    mem = StringValuePtr(arg2);
 	}
 	else if( rb_obj_is_kind_of(arg2, rb_cDLCPtr) ){
@@ -539,7 +550,7 @@ rb_dlptr_aset(int argc, VALUE argv[], VALUE self)
 	else{
 	    mem    = NUM2PTR(arg2);
 	}
-	memcpy((char *)RPTR_DATA(self)->ptr + offset, mem, len);
+	memcpy((char *)data->ptr + offset, mem, len);
 	retval = arg2;
 	break;
       default:
@@ -582,7 +593,7 @@ rb_dlptr_size_get(VALUE self)
 static VALUE
 rb_dlptr_s_to_ptr(VALUE self, VALUE val)
 {
-    VALUE ptr;
+    VALUE ptr, wrap = val, vptr;
 
     if (RTEST(rb_obj_is_kind_of(val, rb_cIO))){
 	rb_io_t *fptr;
@@ -595,20 +606,22 @@ rb_dlptr_s_to_ptr(VALUE self, VALUE val)
 	char *str = StringValuePtr(val);
 	ptr = rb_dlptr_new(str, RSTRING_LEN(val), NULL);
     }
-    else if (rb_respond_to(val, id_to_ptr)){
-	VALUE vptr = rb_funcall(val, id_to_ptr, 0);
+    else if ((vptr = rb_check_funcall(val, id_to_ptr, 0, 0)) != Qundef){
 	if (rb_obj_is_kind_of(vptr, rb_cDLCPtr)){
 	    ptr = vptr;
+	    wrap = 0;
 	}
 	else{
 	    rb_raise(rb_eDLError, "to_ptr should return a CPtr object");
 	}
     }
     else{
-	ptr = rb_dlptr_new(NUM2PTR(rb_Integer(val)), 0, NULL);
+	VALUE num = rb_Integer(val);
+	if (num == val) wrap = 0;
+	ptr = rb_dlptr_new(NUM2PTR(num), 0, NULL);
     }
     OBJ_INFECT(ptr, val);
-    rb_iv_set(ptr, "wrapping", val);
+    if (wrap) RPTR_DATA(ptr)->wrap[0] = wrap;
     return ptr;
 }
 
