@@ -2,7 +2,7 @@
 
   objspace.c - ObjectSpace extender for MRI.
 
-  $Author: usa $
+  $Author: nagachika $
   created at: Wed Jun 17 07:39:17 2009
 
   NOTE: This extension library is not expected to exist except C Ruby.
@@ -60,7 +60,9 @@ memsize_of(VALUE obj)
 	break;
       case T_MODULE:
       case T_CLASS:
-	size += st_memsize(RCLASS_M_TBL(obj));
+	if (RCLASS_M_TBL(obj)) {
+	    size += st_memsize(RCLASS_M_TBL(obj));
+	}
 	if (RCLASS_IV_TBL(obj)) {
 	    size += st_memsize(RCLASS_IV_TBL(obj));
 	}
@@ -97,7 +99,7 @@ memsize_of(VALUE obj)
       case T_MATCH:
 	if (RMATCH(obj)->rmatch) {
             struct rmatch *rm = RMATCH(obj)->rmatch;
-	    size += sizeof(struct re_registers); /* TODO: onig_region_memsize(&rm->regs); */
+	    size += onig_region_memsize(&rm->regs);
 	    size += sizeof(struct rmatch_offset) * rm->char_offset_num_allocated;
 	    size += sizeof(struct rmatch);
 	}
@@ -223,7 +225,7 @@ total_i(void *vstart, void *vend, size_t stride, void *ptr)
  *
  *  def memsize_of_all klass = false
  *    total = 0
- *    ObjectSpace.each_objects{|e|
+ *    ObjectSpace.each_object{|e|
  *      total += ObjectSpace.memsize_of(e) if klass == false || e.kind_of?(klass)
  *    }
  *    total
@@ -268,6 +270,43 @@ cos_i(void *vstart, void *vend, size_t stride, void *data)
     return 0;
 }
 
+static VALUE
+type2sym(enum ruby_value_type i)
+{
+    VALUE type;
+    switch (i) {
+#define CASE_TYPE(t) case t: type = ID2SYM(rb_intern(#t)); break;
+	CASE_TYPE(T_NONE);
+	CASE_TYPE(T_OBJECT);
+	CASE_TYPE(T_CLASS);
+	CASE_TYPE(T_MODULE);
+	CASE_TYPE(T_FLOAT);
+	CASE_TYPE(T_STRING);
+	CASE_TYPE(T_REGEXP);
+	CASE_TYPE(T_ARRAY);
+	CASE_TYPE(T_HASH);
+	CASE_TYPE(T_STRUCT);
+	CASE_TYPE(T_BIGNUM);
+	CASE_TYPE(T_FILE);
+	CASE_TYPE(T_DATA);
+	CASE_TYPE(T_MATCH);
+	CASE_TYPE(T_COMPLEX);
+	CASE_TYPE(T_RATIONAL);
+	CASE_TYPE(T_NIL);
+	CASE_TYPE(T_TRUE);
+	CASE_TYPE(T_FALSE);
+	CASE_TYPE(T_SYMBOL);
+	CASE_TYPE(T_FIXNUM);
+	CASE_TYPE(T_UNDEF);
+	CASE_TYPE(T_NODE);
+	CASE_TYPE(T_ICLASS);
+	CASE_TYPE(T_ZOMBIE);
+#undef CASE_TYPE
+      default: rb_bug("type2sym: unknown type (%d)", i);
+    }
+    return type;
+}
+
 /*
  *  call-seq:
  *    ObjectSpace.count_objects_size([result_hash]) -> hash
@@ -296,11 +335,11 @@ count_objects_size(int argc, VALUE *argv, VALUE os)
 {
     size_t counts[T_MASK+1];
     size_t total = 0;
-    size_t i;
+    enum ruby_value_type i;
     VALUE hash;
 
     if (rb_scan_args(argc, argv, "01", &hash) == 1) {
-        if (TYPE(hash) != T_HASH)
+        if (!RB_TYPE_P(hash, T_HASH))
             rb_raise(rb_eTypeError, "non-hash given");
     }
 
@@ -319,37 +358,7 @@ count_objects_size(int argc, VALUE *argv, VALUE os)
 
     for (i = 0; i <= T_MASK; i++) {
 	if (counts[i]) {
-	    VALUE type;
-	    switch (i) {
-#define COUNT_TYPE(t) case t: type = ID2SYM(rb_intern(#t)); break;
-		COUNT_TYPE(T_NONE);
-		COUNT_TYPE(T_OBJECT);
-		COUNT_TYPE(T_CLASS);
-		COUNT_TYPE(T_MODULE);
-		COUNT_TYPE(T_FLOAT);
-		COUNT_TYPE(T_STRING);
-		COUNT_TYPE(T_REGEXP);
-		COUNT_TYPE(T_ARRAY);
-		COUNT_TYPE(T_HASH);
-		COUNT_TYPE(T_STRUCT);
-		COUNT_TYPE(T_BIGNUM);
-		COUNT_TYPE(T_FILE);
-		COUNT_TYPE(T_DATA);
-		COUNT_TYPE(T_MATCH);
-		COUNT_TYPE(T_COMPLEX);
-		COUNT_TYPE(T_RATIONAL);
-		COUNT_TYPE(T_NIL);
-		COUNT_TYPE(T_TRUE);
-		COUNT_TYPE(T_FALSE);
-		COUNT_TYPE(T_SYMBOL);
-		COUNT_TYPE(T_FIXNUM);
-		COUNT_TYPE(T_UNDEF);
-		COUNT_TYPE(T_NODE);
-		COUNT_TYPE(T_ICLASS);
-		COUNT_TYPE(T_ZOMBIE);
-#undef COUNT_TYPE
-	      default: type = INT2NUM(i); break;
-	    }
+	    VALUE type = type2sym(i);
 	    total += counts[i];
 	    rb_hash_aset(hash, type, SIZET2NUM(counts[i]));
 	}
@@ -404,7 +413,7 @@ count_nodes(int argc, VALUE *argv, VALUE os)
     VALUE hash;
 
     if (rb_scan_args(argc, argv, "01", &hash) == 1) {
-        if (TYPE(hash) != T_HASH)
+        if (!RB_TYPE_P(hash, T_HASH))
             rb_raise(rb_eTypeError, "non-hash given");
     }
 
@@ -460,6 +469,7 @@ count_nodes(int argc, VALUE *argv, VALUE os)
 		COUNT_NODE(NODE_OP_ASGN2);
 		COUNT_NODE(NODE_OP_ASGN_AND);
 		COUNT_NODE(NODE_OP_ASGN_OR);
+		COUNT_NODE(NODE_OP_CDECL);
 		COUNT_NODE(NODE_CALL);
 		COUNT_NODE(NODE_FCALL);
 		COUNT_NODE(NODE_VCALL);
@@ -493,6 +503,7 @@ count_nodes(int argc, VALUE *argv, VALUE os)
 		COUNT_NODE(NODE_ARGS);
 		COUNT_NODE(NODE_ARGS_AUX);
 		COUNT_NODE(NODE_OPT_ARG);
+		COUNT_NODE(NODE_KW_ARG);
 		COUNT_NODE(NODE_POSTARG);
 		COUNT_NODE(NODE_ARGSCAT);
 		COUNT_NODE(NODE_ARGSPUSH);
@@ -510,6 +521,7 @@ count_nodes(int argc, VALUE *argv, VALUE os)
 		COUNT_NODE(NODE_SCLASS);
 		COUNT_NODE(NODE_COLON2);
 		COUNT_NODE(NODE_COLON3);
+		COUNT_NODE(NODE_CREF);
 		COUNT_NODE(NODE_DOT2);
 		COUNT_NODE(NODE_DOT3);
 		COUNT_NODE(NODE_FLIP2);
@@ -529,7 +541,6 @@ count_nodes(int argc, VALUE *argv, VALUE os)
 		COUNT_NODE(NODE_ATTRASGN);
 		COUNT_NODE(NODE_PRELUDE);
 		COUNT_NODE(NODE_LAMBDA);
-		COUNT_NODE(NODE_OPTBLOCK);
 #undef COUNT_NODE
 	      default: node = INT2FIX(i);
 	    }
@@ -609,7 +620,7 @@ count_tdata_objects(int argc, VALUE *argv, VALUE self)
     VALUE hash;
 
     if (rb_scan_args(argc, argv, "01", &hash) == 1) {
-        if (TYPE(hash) != T_HASH)
+        if (!RB_TYPE_P(hash, T_HASH))
             rb_raise(rb_eTypeError, "non-hash given");
     }
 
@@ -625,6 +636,149 @@ count_tdata_objects(int argc, VALUE *argv, VALUE self)
     return hash;
 }
 
+static void
+iow_mark(void *ptr)
+{
+    rb_gc_mark((VALUE)ptr);
+}
+
+static size_t
+iow_size(const void *ptr)
+{
+    VALUE obj = (VALUE)ptr;
+    return memsize_of(obj);
+}
+
+static const rb_data_type_t iow_data_type = {
+    "ObjectSpace::InternalObjectWrapper",
+    {iow_mark, 0, iow_size,},
+};
+
+static VALUE rb_mInternalObjectWrapper;
+
+static VALUE
+iow_newobj(VALUE obj)
+{
+    return rb_data_typed_object_alloc(rb_mInternalObjectWrapper, (void *)obj, &iow_data_type);
+}
+
+static VALUE
+iow_type(VALUE self)
+{
+    VALUE obj = (VALUE)DATA_PTR(self);
+    return type2sym(BUILTIN_TYPE(obj));
+}
+
+static VALUE
+iow_inspect(VALUE self)
+{
+    VALUE obj = (VALUE)DATA_PTR(self);
+    VALUE type = type2sym(BUILTIN_TYPE(obj));
+
+    return rb_sprintf("#<InternalObject:%p %s>", (void *)obj, rb_id2name(SYM2ID(type)));
+}
+
+static VALUE
+iow_internal_object_id(VALUE self)
+{
+    VALUE obj = (VALUE)DATA_PTR(self);
+    return rb_obj_id(obj);
+}
+
+struct rof_data {
+    st_table *refs;
+    VALUE internals;
+};
+
+static void
+reachable_object_from_i(VALUE obj, void *data_ptr)
+{
+    struct rof_data *data = (struct rof_data *)data_ptr;
+    VALUE key = obj;
+    VALUE val = obj;
+
+    if (rb_objspace_markable_object_p(obj)) {
+	if (rb_objspace_internal_object_p(obj)) {
+	    val = iow_newobj(obj);
+	    rb_ary_push(data->internals, val);
+	}
+	st_insert(data->refs, key, val);
+    }
+}
+
+static int
+collect_values(st_data_t key, st_data_t value, st_data_t data)
+{
+    VALUE ary = (VALUE)data;
+    rb_ary_push(ary, (VALUE)value);
+    return ST_CONTINUE;
+}
+
+/*
+ *  call-seq:
+ *     ObjectSpace.reachable_objects_from(obj) -> array or nil
+ *
+ *  [MRI specific feature] Return all reachable objects from `obj'.
+ *
+ *  This method returns all reachable objects from `obj'.
+ *  If `obj' has references two or more references to same object `x',
+ *  them returned array only include one `x' object.
+ *  If `obj' is non-markable (non-heap management) object such as
+ *  true, false, nil, symbols and Fixnums (and Flonum) them it simply
+ *  returns nil.
+ *
+ *  If `obj' has references to internal object, then it returns
+ *  instances of `ObjectSpace::InternalObjectWrapper' class.
+ *  This object contains a reference to an internal object and
+ *  you can check the type of internal object with `type' method.
+ *
+ *  If `obj' is instance of `ObjectSpace::InternalObjectWrapper'
+ *  class, then this method returns all reachable object from
+ *  an internal object, which is pointed by `obj'.
+ *
+ *  With this method, you can find memory leaks.
+ *
+ *  This method is not expected to work except C Ruby.
+ *
+ *  Example:
+ *    ObjectSpace.reachable_objects_from(['a', 'b', 'c'])
+ *    #=> [Array, 'a', 'b', 'c']
+ *
+ *    ObjectSpace.reachable_objects_from(['a', 'a', 'a'])
+ *    #=> [Array, 'a', 'a', 'a'] # all 'a' strings have different object id
+ *
+ *    ObjectSpace.reachable_objects_from([v = 'a', v, v])
+ *    #=> [Array, 'a']
+ *
+ *    ObjectSpace.reachable_objects_from(1)
+ *    #=> nil # 1 is not markable (heap managed) object
+ *
+ */
+
+static VALUE
+reachable_objects_from(VALUE self, VALUE obj)
+{
+    if (rb_objspace_markable_object_p(obj)) {
+	VALUE ret = rb_ary_new();
+	struct rof_data data;
+
+	if (rb_typeddata_is_kind_of(obj, &iow_data_type)) {
+	    obj = (VALUE)DATA_PTR(obj);
+	}
+
+	data.refs = st_init_numtable();
+	data.internals = rb_ary_new();
+
+	rb_objspace_reachable_objects_from(obj, reachable_object_from_i, &data);
+
+	st_foreach(data.refs, collect_values, (st_data_t)ret);
+	return ret;
+    }
+    else {
+	return Qnil;
+    }
+}
+
 /* objspace library extends ObjectSpace module and add several
  * methods to get internal statistic information about
  * object/memory management.
@@ -641,10 +795,16 @@ Init_objspace(void)
     VALUE rb_mObjSpace = rb_const_get(rb_cObject, rb_intern("ObjectSpace"));
 
     rb_define_module_function(rb_mObjSpace, "memsize_of", memsize_of_m, 1);
-    rb_define_module_function(rb_mObjSpace, "memsize_of_all",
-			      memsize_of_all_m, -1);
+    rb_define_module_function(rb_mObjSpace, "memsize_of_all", memsize_of_all_m, -1);
 
     rb_define_module_function(rb_mObjSpace, "count_objects_size", count_objects_size, -1);
     rb_define_module_function(rb_mObjSpace, "count_nodes", count_nodes, -1);
     rb_define_module_function(rb_mObjSpace, "count_tdata_objects", count_tdata_objects, -1);
+
+    rb_define_module_function(rb_mObjSpace, "reachable_objects_from", reachable_objects_from, 1);
+
+    rb_mInternalObjectWrapper = rb_define_class_under(rb_mObjSpace, "InternalObjectWrapper", rb_cObject);
+    rb_define_method(rb_mInternalObjectWrapper, "type", iow_type, 0);
+    rb_define_method(rb_mInternalObjectWrapper, "inspect", iow_inspect, 0);
+    rb_define_method(rb_mInternalObjectWrapper, "internal_object_id", iow_internal_object_id, 0);
 }
