@@ -1,3 +1,4 @@
+# -*- coding: us-ascii -*-
 =begin
 = $RCSfile$ -- Generator for Makefile
 
@@ -11,34 +12,31 @@
   (See the file 'LICENCE'.)
 
 = Version
-  $Id: extconf.rb 32230 2011-06-26 01:32:03Z emboss $
+  $Id: extconf.rb 38492 2012-12-20 07:42:56Z emboss $
 =end
 
 require "mkmf"
+require File.expand_path('../deprecation', __FILE__)
 
 dir_config("openssl")
 dir_config("kerberos")
 
-message "=== OpenSSL for Ruby configurator ===\n"
+Logging::message "=== OpenSSL for Ruby configurator ===\n"
 
 ##
-# Adds -Wall -DOSSL_DEBUG for compilation and some more targets when GCC is used
+# Adds -DOSSL_DEBUG for compilation and some more targets when GCC is used
 # To turn it on, use: --with-debug or --enable-debug
 #
 if with_config("debug") or enable_config("debug")
   $defs.push("-DOSSL_DEBUG") unless $defs.include? "-DOSSL_DEBUG"
-
-  if CONFIG['GCC'] == 'yes'
-    $CPPFLAGS += " -Wall" unless $CPPFLAGS.split.include? "-Wall"
-  end
 end
 
-message "=== Checking for system dependent stuff... ===\n"
+Logging::message "=== Checking for system dependent stuff... ===\n"
 have_library("nsl", "t_open")
 have_library("socket", "socket")
 have_header("assert.h")
 
-message "=== Checking for required stuff... ===\n"
+Logging::message "=== Checking for required stuff... ===\n"
 if $mingw
   have_library("wsock32")
   have_library("gdi32")
@@ -51,20 +49,20 @@ unless result
   result &&= %w[crypto libeay32].any? {|lib| have_library(lib, "OpenSSL_add_all_digests")}
   result &&= %w[ssl ssleay32].any? {|lib| have_library(lib, "SSL_library_init")}
   unless result
-    message "=== Checking for required stuff failed. ===\n"
-    message "Makefile wasn't created. Fix the errors above.\n"
+    Logging::message "=== Checking for required stuff failed. ===\n"
+    Logging::message "Makefile wasn't created. Fix the errors above.\n"
     exit 1
   end
 end
 
 unless have_header("openssl/conf_api.h")
-  message "OpenSSL 0.9.6 or later required.\n"
-  exit 1
+  raise "OpenSSL 0.9.6 or later required."
+end
+unless OpenSSL.check_func("SSL_library_init()", "openssl/ssl.h")
+  raise "Ignore OpenSSL broken by Apple.\nPlease use another openssl. (e.g. using `configure --with-openssl-dir=/path/to/openssl')"
 end
 
-%w"rb_str_set_len rb_block_call".each {|func| have_func(func, "ruby.h")}
-
-message "=== Checking for OpenSSL features... ===\n"
+Logging::message "=== Checking for OpenSSL features... ===\n"
 have_func("ERR_peek_last_error")
 have_func("ASN1_put_eoc")
 have_func("BN_mod_add")
@@ -105,6 +103,13 @@ have_func("OPENSSL_cleanse")
 have_func("SSLv2_method")
 have_func("SSLv2_server_method")
 have_func("SSLv2_client_method")
+have_func("TLSv1_1_method")
+have_func("TLSv1_1_server_method")
+have_func("TLSv1_1_client_method")
+have_func("TLSv1_2_method")
+have_func("TLSv1_2_server_method")
+have_func("TLSv1_2_client_method")
+have_macro("OPENSSL_NPN_NEGOTIATED", ['openssl/ssl.h']) && $defs.push("-DHAVE_OPENSSL_NPN_NEGOTIATED")
 unless have_func("SSL_set_tlsext_host_name", ['openssl/ssl.h'])
   have_macro("SSL_set_tlsext_host_name", ['openssl/ssl.h']) && $defs.push("-DHAVE_SSL_SET_TLSEXT_HOST_NAME")
 end
@@ -115,6 +120,7 @@ if have_header("openssl/engine.h")
   have_func("ENGINE_get_digest")
   have_func("ENGINE_get_cipher")
   have_func("ENGINE_cleanup")
+  have_func("ENGINE_load_dynamic")
   have_func("ENGINE_load_4758cca")
   have_func("ENGINE_load_aep")
   have_func("ENGINE_load_atalla")
@@ -123,7 +129,16 @@ if have_header("openssl/engine.h")
   have_func("ENGINE_load_nuron")
   have_func("ENGINE_load_sureware")
   have_func("ENGINE_load_ubsec")
+  have_func("ENGINE_load_padlock")
+  have_func("ENGINE_load_capi")
+  have_func("ENGINE_load_gmp")
+  have_func("ENGINE_load_gost")
+  have_func("ENGINE_load_cryptodev")
+  have_func("ENGINE_load_aesni")
 end
+have_func("DH_generate_parameters_ex")
+have_func("DSA_generate_parameters_ex")
+have_func("RSA_generate_key_ex")
 if checking_for('OpenSSL version is 0.9.7 or later') {
     try_static_assert('OPENSSL_VERSION_NUMBER >= 0x00907000L', 'openssl/opensslv.h')
   }
@@ -132,9 +147,11 @@ end
 have_struct_member("EVP_CIPHER_CTX", "flags", "openssl/evp.h")
 have_struct_member("EVP_CIPHER_CTX", "engine", "openssl/evp.h")
 have_struct_member("X509_ATTRIBUTE", "single", "openssl/x509.h")
+have_macro("OPENSSL_FIPS", ['openssl/opensslconf.h']) && $defs.push("-DHAVE_OPENSSL_FIPS")
+have_macro("EVP_CTRL_GCM_GET_TAG", ['openssl/evp.h']) && $defs.push("-DHAVE_AUTHENTICATED_ENCRYPTION")
 
-message "=== Checking done. ===\n"
+Logging::message "=== Checking done. ===\n"
 
 create_header
 create_makefile("openssl")
-message "Done.\n"
+Logging::message "Done.\n"
