@@ -2,13 +2,13 @@
 
   readline.c - GNU Readline module
 
-  $Author: usa $
+  $Author: nagachika $
   created at: Wed Jan 20 13:59:32 JST 1999
 
   Copyright (C) 1997-2008  Shugo Maeda
   Copyright (C) 2008-2009  TAKAO Kouji
 
-  $Id: readline.c 39377 2013-02-22 05:27:47Z usa $
+  $Id: readline.c 45252 2014-03-02 16:33:36Z nagachika $
 
   Contact:
    - TAKAO Kouji <kouji at takao7 dot net> (current maintainer)
@@ -48,7 +48,7 @@ static VALUE mReadline;
 
 #define EDIT_LINE_LIBRARY_VERSION "EditLine wrapper"
 #ifndef USE_INSERT_IGNORE_ESCAPE
-# if !defined(HAVE_EDITLINE_READLINE_H) && defined(HAVE_RL_PROMPT_START_IGNORE) && defined(HAVE_RL_PROMPT_END_IGNORE)
+# if !defined(HAVE_EDITLINE_READLINE_H) && defined(RL_PROMPT_START_IGNORE) && defined(RL_PROMPT_END_IGNORE)
 #  define USE_INSERT_IGNORE_ESCAPE 1
 # else
 #  define USE_INSERT_IGNORE_ESCAPE 0
@@ -61,6 +61,12 @@ static ID completion_proc, completion_case_fold;
 #if USE_INSERT_IGNORE_ESCAPE
 static ID id_orig_prompt, id_last_prompt;
 #endif
+#if defined(HAVE_RL_PRE_INPUT_HOOK)
+static ID id_pre_input_hook;
+#endif
+#if defined(HAVE_RL_SPECIAL_PREFIXES)
+static ID id_special_prefixes;
+#endif
 
 #ifndef HAVE_RL_FILENAME_COMPLETION_FUNCTION
 # define rl_filename_completion_function filename_completion_function
@@ -124,7 +130,6 @@ static char **readline_attempted_completion_function(const char *text,
 
 #if defined HAVE_RL_GETC_FUNCTION
 static VALUE readline_instream;
-static ID id_getbyte;
 
 #ifndef HAVE_RL_GETC
 #define rl_getc(f) EOF
@@ -167,9 +172,21 @@ readline_getc(FILE *input)
         }
     }
 #endif
-    c = rb_funcall(readline_instream, id_getbyte, 0, 0);
+    c = rb_io_getbyte(readline_instream);
     if (NIL_P(c)) return EOF;
-    return NUM2CHR(c);
+#ifdef ESC
+    if (c == INT2FIX(ESC) &&
+	RL_ISSTATE(RL_STATE_ISEARCH) && /* isn't needed in other states? */
+	rb_io_read_pending(ifp)) {
+	int meta = 0;
+	c = rb_io_getbyte(readline_instream);
+	if (FIXNUM_P(c) && isascii(FIX2INT(c))) meta = 1;
+	rb_io_ungetbyte(readline_instream, c);
+	if (meta) rl_execute_next(ESC);
+	return ESC;
+    }
+#endif
+    return FIX2INT(c);
 }
 #elif defined HAVE_RL_EVENT_HOOK
 #define BUSY_WAIT 0
@@ -468,6 +485,108 @@ readline_s_set_output(VALUE self, VALUE output)
     return output;
 }
 
+#if defined(HAVE_RL_PRE_INPUT_HOOK)
+/*
+ * call-seq:
+ *   Readline.pre_input_hook = proc
+ *
+ * Specifies a Proc object +proc+ to call after the first prompt has
+ * been printed and just before readline starts reading input
+ * characters.
+ *
+ * See GNU Readline's rl_pre_input_hook variable.
+ *
+ * Raises ArgumentError if +proc+ does not respond to the call method.
+ *
+ * Raises SecurityError if $SAFE is 4.
+ */
+static VALUE
+readline_s_set_pre_input_hook(VALUE self, VALUE proc)
+{
+    rb_secure(4);
+    if (!NIL_P(proc) && !rb_respond_to(proc, rb_intern("call")))
+	rb_raise(rb_eArgError, "argument must respond to `call'");
+    return rb_ivar_set(mReadline, id_pre_input_hook, proc);
+}
+
+/*
+ * call-seq:
+ *   Readline.pre_input_hook -> proc
+ *
+ * Returns a Proc object +proc+ to call after the first prompt has
+ * been printed and just before readline starts reading input
+ * characters. The default is nil.
+ *
+ * Raises SecurityError if $SAFE is 4.
+ */
+static VALUE
+readline_s_get_pre_input_hook(VALUE self)
+{
+    rb_secure(4);
+    return rb_attr_get(mReadline, id_pre_input_hook);
+}
+
+static int
+readline_pre_input_hook(void)
+{
+    VALUE proc;
+
+    proc = rb_attr_get(mReadline, id_pre_input_hook);
+    if (!NIL_P(proc))
+	rb_funcall(proc, rb_intern("call"), 0);
+    return 0;
+}
+#else
+#define readline_s_set_pre_input_hook rb_f_notimplement
+#define readline_s_get_pre_input_hook rb_f_notimplement
+#endif
+
+#if defined(HAVE_RL_INSERT_TEXT)
+/*
+ * call-seq:
+ *   Readline.insert_text(string) -> self
+ *
+ * Insert text into the line at the current cursor position.
+ *
+ * See GNU Readline's rl_insert_text function.
+ *
+ * Raises SecurityError if $SAFE is 4.
+ */
+static VALUE
+readline_s_insert_text(VALUE self, VALUE str)
+{
+    rb_secure(4);
+    OutputStringValue(str);
+    rl_insert_text(RSTRING_PTR(str));
+    return self;
+}
+#else
+#define readline_s_insert_text rb_f_notimplement
+#endif
+
+#if defined(HAVE_RL_REDISPLAY)
+/*
+ * call-seq:
+ *   Readline.redisplay -> self
+ *
+ * Change what's displayed on the screen to reflect the current
+ * contents.
+ *
+ * See GNU Readline's rl_redisplay function.
+ *
+ * Raises SecurityError if $SAFE is 4.
+ */
+static VALUE
+readline_s_redisplay(VALUE self)
+{
+    rb_secure(4);
+    rl_redisplay();
+    return self;
+}
+#else
+#define readline_s_redisplay rb_f_notimplement
+#endif
+
 /*
  * call-seq:
  *   Readline.completion_proc = proc
@@ -674,7 +793,7 @@ readline_attempted_completion_function(const char *text, int start, int end)
     matches = RARRAY_LEN(ary);
     if (matches == 0) return NULL;
     result = (char**)malloc((matches + 2)*sizeof(char*));
-    if (result == NULL) rb_raise(rb_eNoMemError, "failed to allocate memory");
+    if (result == NULL) rb_memerror();
     enc = rb_locale_encoding();
     encobj = rb_enc_from_encoding(enc);
     for (i = 0; i < matches; i++) {
@@ -713,7 +832,8 @@ readline_attempted_completion_function(const char *text, int start, int end)
 
 	    low = i1;
 	}
-	result[0] = ALLOC_N(char, low + 1);
+	result[0] = (char*)malloc(low + 1);
+	if (result[0]  == NULL) rb_memerror();
 	strncpy(result[0], result[1], low);
 	result[0][low] = '\0';
     }
@@ -958,7 +1078,7 @@ readline_s_get_completion_append_character(VALUE self)
  *
  * Sets the basic list of characters that signal a break between words
  * for the completer routine. The default is the characters which
- * break words for completion in Bash: "\t\n\"\\'`@$><=;|&{(".
+ * break words for completion in Bash: " \t\n\"\\'`@$><=;|&{(".
  *
  * Raises NotImplementedError if the using readline library does not support.
  *
@@ -1073,6 +1193,74 @@ readline_s_get_completer_word_break_characters(VALUE self, VALUE str)
 #define readline_s_get_completer_word_break_characters rb_f_notimplement
 #endif
 
+#if defined(HAVE_RL_SPECIAL_PREFIXES)
+/*
+ * call-seq:
+ *   Readline.special_prefixes = string
+ *
+ * Sets the list of characters that are word break characters, but
+ * should be left in text when it is passed to the completion
+ * function. Programs can use this to help determine what kind of
+ * completing to do. For instance, Bash sets this variable to "$@" so
+ * that it can complete shell variables and hostnames.
+ *
+ * See GNU Readline's rl_special_prefixes variable.
+ *
+ * Raises NotImplementedError if the using readline library does not support.
+ *
+ * Raises SecurityError exception if $SAFE is 4.
+ */
+static VALUE
+readline_s_set_special_prefixes(VALUE self, VALUE str)
+{
+    rb_secure(4);
+    if (!NIL_P(str)) {
+	OutputStringValue(str);
+	str = rb_str_dup_frozen(str);
+	RBASIC(str)->klass = 0;
+    }
+    rb_ivar_set(mReadline, id_special_prefixes, str);
+    if (NIL_P(str)) {
+	rl_special_prefixes = NULL;
+    }
+    else {
+	rl_special_prefixes = RSTRING_PTR(str);
+    }
+    return self;
+}
+
+/*
+ * call-seq:
+ *   Readline.special_prefixes -> string
+ *
+ * Gets the list of characters that are word break characters, but
+ * should be left in text when it is passed to the completion
+ * function.
+ *
+ * See GNU Readline's rl_special_prefixes variable.
+ *
+ * Raises NotImplementedError if the using readline library does not support.
+ *
+ * Raises SecurityError exception if $SAFE is 4.
+ */
+static VALUE
+readline_s_get_special_prefixes(VALUE self)
+{
+    VALUE str;
+    rb_secure(4);
+    if (rl_special_prefixes == NULL) return Qnil;
+    str = rb_ivar_get(mReadline, id_special_prefixes);
+    if (!NIL_P(str)) {
+	str = rb_str_dup_frozen(str);
+	RBASIC(str)->klass = rb_cString;
+    }
+    return str;
+}
+#else
+#define readline_s_set_special_prefixes rb_f_notimplement
+#define readline_s_get_special_prefixes rb_f_notimplement
+#endif
+
 #ifdef HAVE_RL_BASIC_QUOTE_CHARACTERS
 /*
  * call-seq:
@@ -1378,7 +1566,8 @@ rb_remove_history(int index)
     return Qnil;
 #else
     rb_notimplement();
-    return Qnil; /* not reached */
+
+    UNREACHABLE;
 #endif
 }
 
@@ -1526,7 +1715,6 @@ Init_readline()
     /* and using_history() call rl_initialize(). */
     /* This assignment should be placed before using_history() */
     rl_getc_function = readline_getc;
-    id_getbyte = rb_intern_const("getbyte");
 #elif defined HAVE_RL_EVENT_HOOK
     rl_event_hook = readline_event;
 #endif
@@ -1535,6 +1723,12 @@ Init_readline()
 
     completion_proc = rb_intern(COMPLETION_PROC);
     completion_case_fold = rb_intern(COMPLETION_CASE_FOLD);
+#if defined(HAVE_RL_PRE_INPUT_HOOK)
+    id_pre_input_hook = rb_intern("pre_input_hook");
+#endif
+#if defined(HAVE_RL_SPECIAL_PREFIXES)
+    id_special_prefixes = rb_intern("special_prefixes");
+#endif
 
     mReadline = rb_define_module("Readline");
     rb_define_module_function(mReadline, "readline",
@@ -1593,6 +1787,18 @@ Init_readline()
 			       readline_s_get_filename_quote_characters, 0);
     rb_define_singleton_method(mReadline, "refresh_line",
 			       readline_s_refresh_line, 0);
+    rb_define_singleton_method(mReadline, "pre_input_hook=",
+			       readline_s_set_pre_input_hook, 1);
+    rb_define_singleton_method(mReadline, "pre_input_hook",
+			       readline_s_get_pre_input_hook, 0);
+    rb_define_singleton_method(mReadline, "insert_text",
+			       readline_s_insert_text, 1);
+    rb_define_singleton_method(mReadline, "redisplay",
+			       readline_s_redisplay, 0);
+    rb_define_singleton_method(mReadline, "special_prefixes=",
+ 			       readline_s_set_special_prefixes, 1);
+    rb_define_singleton_method(mReadline, "special_prefixes",
+ 			       readline_s_get_special_prefixes, 0);
 
 #if USE_INSERT_IGNORE_ESCAPE
     CONST_ID(id_orig_prompt, "orig_prompt");
@@ -1676,10 +1882,12 @@ Init_readline()
     rb_define_const(mReadline, "VERSION", version);
 
     rl_attempted_completion_function = readline_attempted_completion_function;
+#if defined(HAVE_RL_PRE_INPUT_HOOK)
+    rl_pre_input_hook = (rl_hook_func_t *)readline_pre_input_hook;
+#endif
 #ifdef HAVE_RL_CATCH_SIGNALS
     rl_catch_signals = 0;
-#endif 
-
+#endif
 #ifdef HAVE_RL_CLEAR_SIGNALS
     rl_clear_signals();
 #endif
