@@ -6,7 +6,6 @@ case RUBY_PLATFORM
 when /(ms|bcc)win(32|64)|mingw/
   test_func = "WSACleanup"
   have_library("ws2_32", "WSACleanup")
-  $defs << "-DHAVE_SOCKETPAIR"
 when /cygwin/
   test_func = "socket"
 when /beos/
@@ -42,7 +41,7 @@ if have_header("arpa/inet.h")
 end
 
 ipv6 = false
-default_ipv6 = /mswin|cygwin|beos|haiku/ !~ RUBY_PLATFORM
+default_ipv6 = /cygwin|beos|haiku/ !~ RUBY_PLATFORM
 if enable_config("ipv6", default_ipv6)
   if checking_for("ipv6") {try_link(<<EOF)}
 #include <sys/types.h>
@@ -50,9 +49,10 @@ if enable_config("ipv6", default_ipv6)
 #include <sys/socket.h>
 #endif
 int
-main()
+main(void)
 {
   socket(AF_INET6, SOCK_STREAM, 0);
+  return 0;
 }
 EOF
     $defs << "-DENABLE_IPV6" << "-DINET6"
@@ -62,7 +62,7 @@ end
 
 if ipv6
   if $mingw
-    $CPPFLAGS << " -D_WIN32_WINNT=0x501"
+    $CPPFLAGS << " -D_WIN32_WINNT=0x501" unless $CPPFLAGS.include?("_WIN32_WINNT")
   end
   ipv6lib = nil
   class << (fmt = "unknown")
@@ -270,7 +270,7 @@ getaddr_info_ok = (enable_config("wide-getaddrinfo") && :wide) ||
 #endif
 
 int
-main()
+main(void)
 {
   int passive, gaierr, inet4 = 0, inet6 = 0;
   struct addrinfo hints, *ai, *aitop;
@@ -354,12 +354,12 @@ main()
 
   if (aitop)
     freeaddrinfo(aitop);
-  exit(EXIT_SUCCESS);
+  return EXIT_SUCCESS;
 
  bad:
   if (aitop)
     freeaddrinfo(aitop);
-  exit(EXIT_FAILURE);
+  return EXIT_FAILURE;
 }
 EOF
 if ipv6 and not getaddr_info_ok
@@ -403,6 +403,8 @@ EOF
   end
 end
 
+have_func("accept4")
+
 $objs = [
   "init.#{$OBJEXT}",
   "constants.#{$OBJEXT}",
@@ -434,9 +436,10 @@ if getaddr_info_ok == :wide or
   $defs << "-DGETADDRINFO_EMU"
 end
 
-have_func("inet_ntop") or have_func("inet_ntoa")
-have_func("inet_pton") or have_func("inet_aton")
-have_func("getservbyport")
+have_func('inet_ntop(0, (const void *)0, (char *)0, 0)') or
+  have_func("inet_ntoa(*(struct in_addr *)NULL)")
+have_func('inet_pton(0, "", (void *)0)') or have_func('inet_aton("", (struct in_addr *)0)')
+have_func('getservbyport(0, "")')
 have_header("arpa/nameser.h")
 have_header("resolv.h")
 
@@ -456,11 +459,11 @@ end
 have_header("sys/un.h")
 have_header("sys/uio.h")
 have_type("struct in_pktinfo", headers) {|src|
-  src.sub(%r'^/\*top\*/', '\1'"\n#if defined(IPPROTO_IP) && defined(IP_PKTINFO)") <<
+  src.sub(%r'^/\*top\*/', '\&'"\n#if defined(IPPROTO_IP) && defined(IP_PKTINFO)") <<
   "#else\n" << "#error\n" << ">>>>>> no in_pktinfo <<<<<<\n" << "#endif\n"
 } and have_struct_member("struct in_pktinfo", "ipi_spec_dst", headers)
 have_type("struct in6_pktinfo", headers) {|src|
-  src.sub(%r'^/\*top\*/', '\1'"\n#if defined(IPPROTO_IPV6) && defined(IPV6_PKTINFO)") <<
+  src.sub(%r'^/\*top\*/', '\&'"\n#if defined(IPPROTO_IPV6) && defined(IPV6_PKTINFO)") <<
   "#else\n" << "#error\n" << ">>>>>> no in6_pktinfo <<<<<<\n" << "#endif\n"
 }
 
@@ -486,8 +489,11 @@ $distcleanfiles << "constants.h" << "constdefs.*"
 if have_func(test_func)
   have_func("hsterror")
   have_func("getipnodebyname") or have_func("gethostbyname2")
-  have_func("socketpair") unless $defs.include?("-DHAVE_SOCKETPAIR")
-  unless have_func("gethostname")
+  if !have_func("socketpair(0, 0, 0, 0)") and have_func("rb_w32_socketpair(0, 0, 0, 0)")
+    $defs << "-Dsocketpair(a,b,c,d)=rb_w32_socketpair((a),(b),(c),(d))"
+    $defs << "-DHAVE_SOCKETPAIR"
+  end
+  unless have_func("gethostname((char *)0, 0)")
     have_func("uname")
   end
   if enable_config("socks", ENV["SOCKS_SERVER"])
@@ -497,5 +503,29 @@ if have_func(test_func)
       $defs << "-DSOCKS"
     end
   end
+  hdr = "netinet6/in6.h"
+  if /darwin/ =~ RUBY_PLATFORM and !try_compile(<<"SRC", nil, :werror=>true)
+#include <netinet/in.h>
+int t(struct in6_addr *addr) {return IN6_IS_ADDR_UNSPECIFIED(addr);}
+SRC
+    print "fixing apple's netinet6/in6.rb ..."; $stdout.flush
+    in6 = File.read("/usr/include/#{hdr}")
+    if in6.gsub!(/\*\(const\s+__uint32_t\s+\*\)\(const\s+void\s+\*\)\(&(\(\w+\))->s6_addr\[(\d+)\]\)/) do
+        i, r = $2.to_i.divmod(4)
+        if r.zero?
+          "#$1->__u6_addr.__u6_addr32[#{i}]"
+        else
+          $&
+        end
+      end
+      FileUtils.mkdir_p(File.dirname(hdr))
+      open(hdr, "w") {|f| f.write(in6)}
+      $distcleanfiles << hdr
+      $distcleandirs << File.dirname(hdr)
+      puts "done"
+    else
+      puts "not needed"
+    end
+  end
   create_makefile("socket")
 end
