@@ -1,6 +1,6 @@
 /************************************************
 
-  ainfo.c -
+  raddrinfo.c -
 
   created at: Thu Mar 31 12:21:29 JST 1994
 
@@ -56,7 +56,7 @@ ruby_getaddrinfo(const char *nodename, const char *servname,
 #if defined(_AIX)
 static int
 ruby_getaddrinfo__aix(const char *nodename, const char *servname,
-		      struct addrinfo *hints, struct addrinfo **res)
+		      const struct addrinfo *hints, struct addrinfo **res)
 {
     int error = getaddrinfo(nodename, servname, hints, res);
     struct addrinfo *r;
@@ -100,13 +100,15 @@ ruby_getnameinfo__aix(const struct sockaddr *sa, size_t salen,
 static int str_is_number(const char *);
 
 #if defined(__APPLE__)
-/* fix [ruby-core:29427] */
 static int
 ruby_getaddrinfo__darwin(const char *nodename, const char *servname,
-			 struct addrinfo *hints, struct addrinfo **res)
+			 const struct addrinfo *hints, struct addrinfo **res)
 {
+    /* fix [ruby-core:29427] */
     const char *tmp_servname;
     struct addrinfo tmp_hints;
+    int error;
+
     tmp_servname = servname;
     MEMCPY(&tmp_hints, hints, struct addrinfo, 1);
     if (nodename && servname) {
@@ -117,7 +119,26 @@ ruby_getaddrinfo__darwin(const char *nodename, const char *servname,
 #endif
 	}
     }
-    int error = getaddrinfo(nodename, tmp_servname, &tmp_hints, res);
+
+    error = getaddrinfo(nodename, tmp_servname, &tmp_hints, res);
+    if (error == 0) {
+        /* [ruby-dev:23164] */
+        struct addrinfo *r;
+        r = *res;
+        while (r) {
+            if (! r->ai_socktype) r->ai_socktype = hints->ai_socktype;
+            if (! r->ai_protocol) {
+                if (r->ai_socktype == SOCK_DGRAM) {
+                    r->ai_protocol = IPPROTO_UDP;
+                }
+                else if (r->ai_socktype == SOCK_STREAM) {
+                    r->ai_protocol = IPPROTO_TCP;
+                }
+            }
+            r = r->ai_next;
+        }
+    }
+
     return error;
 }
 #undef getaddrinfo
@@ -133,12 +154,20 @@ struct getaddrinfo_arg
     struct addrinfo **res;
 };
 
-static VALUE
+static void *
 nogvl_getaddrinfo(void *arg)
 {
+    int ret;
     struct getaddrinfo_arg *ptr = arg;
-    return getaddrinfo(ptr->node, ptr->service,
-                       ptr->hints, ptr->res);
+    ret = getaddrinfo(ptr->node, ptr->service, ptr->hints, ptr->res);
+#ifdef __linux__
+    /* On Linux (mainly Ubuntu 13.04) /etc/nsswitch.conf has mdns4 and
+     * it cause getaddrinfo to return EAI_SYSTEM/ENOENT. [ruby-list:49420]
+     */
+    if (ret == EAI_SYSTEM && errno == ENOENT)
+	ret = EAI_NONAME;
+#endif
+    return (void *)(VALUE)ret;
 }
 #endif
 
@@ -157,7 +186,7 @@ rb_getaddrinfo(const char *node, const char *service,
     arg.service = service;
     arg.hints = hints;
     arg.res = res;
-    ret = (int)BLOCKING_REGION(nogvl_getaddrinfo, &arg);
+    ret = (int)(VALUE)rb_thread_call_without_gvl(nogvl_getaddrinfo, &arg, RUBY_UBF_IO, 0);
     return ret;
 #endif
 }
@@ -174,14 +203,14 @@ struct getnameinfo_arg
     int flags;
 };
 
-static VALUE
+static void *
 nogvl_getnameinfo(void *arg)
 {
     struct getnameinfo_arg *ptr = arg;
-    return getnameinfo(ptr->sa, ptr->salen,
-                       ptr->host, (socklen_t)ptr->hostlen,
-                       ptr->serv, (socklen_t)ptr->servlen,
-                       ptr->flags);
+    return (void *)(VALUE)getnameinfo(ptr->sa, ptr->salen,
+				      ptr->host, (socklen_t)ptr->hostlen,
+				      ptr->serv, (socklen_t)ptr->servlen,
+				      ptr->flags);
 }
 #endif
 
@@ -202,33 +231,33 @@ rb_getnameinfo(const struct sockaddr *sa, socklen_t salen,
     arg.serv = serv;
     arg.servlen = servlen;
     arg.flags = flags;
-    ret = (int)BLOCKING_REGION(nogvl_getnameinfo, &arg);
+    ret = (int)(VALUE)rb_thread_call_without_gvl(nogvl_getnameinfo, &arg, RUBY_UBF_IO, 0);
     return ret;
 #endif
 }
 
 static void
-make_ipaddr0(struct sockaddr *addr, char *buf, size_t len)
+make_ipaddr0(struct sockaddr *addr, socklen_t addrlen, char *buf, size_t buflen)
 {
     int error;
 
-    error = rb_getnameinfo(addr, SA_LEN(addr), buf, len, NULL, 0, NI_NUMERICHOST);
+    error = rb_getnameinfo(addr, addrlen, buf, buflen, NULL, 0, NI_NUMERICHOST);
     if (error) {
         rsock_raise_socket_error("getnameinfo", error);
     }
 }
 
 VALUE
-rsock_make_ipaddr(struct sockaddr *addr)
+rsock_make_ipaddr(struct sockaddr *addr, socklen_t addrlen)
 {
     char hbuf[1024];
 
-    make_ipaddr0(addr, hbuf, sizeof(hbuf));
+    make_ipaddr0(addr, addrlen, hbuf, sizeof(hbuf));
     return rb_str_new2(hbuf);
 }
 
 static void
-make_inetaddr(unsigned int host, char *buf, size_t len)
+make_inetaddr(unsigned int host, char *buf, size_t buflen)
 {
     struct sockaddr_in sin;
 
@@ -236,7 +265,7 @@ make_inetaddr(unsigned int host, char *buf, size_t len)
     sin.sin_family = AF_INET;
     SET_SIN_LEN(&sin, sizeof(sin));
     sin.sin_addr.s_addr = host;
-    make_ipaddr0((struct sockaddr*)&sin, buf, len);
+    make_ipaddr0((struct sockaddr*)&sin, sizeof(sin), buf, buflen);
 }
 
 static int
@@ -255,7 +284,7 @@ str_is_number(const char *p)
 }
 
 static char*
-host_str(VALUE host, char *hbuf, size_t len, int *flags_ptr)
+host_str(VALUE host, char *hbuf, size_t hbuflen, int *flags_ptr)
 {
     if (NIL_P(host)) {
         return NULL;
@@ -263,7 +292,7 @@ host_str(VALUE host, char *hbuf, size_t len, int *flags_ptr)
     else if (rb_obj_is_kind_of(host, rb_cInteger)) {
         unsigned int i = NUM2UINT(host);
 
-        make_inetaddr(htonl(i), hbuf, len);
+        make_inetaddr(htonl(i), hbuf, hbuflen);
         if (flags_ptr) *flags_ptr |= AI_NUMERICHOST;
         return hbuf;
     }
@@ -273,14 +302,14 @@ host_str(VALUE host, char *hbuf, size_t len, int *flags_ptr)
         SafeStringValue(host);
         name = RSTRING_PTR(host);
         if (!name || *name == 0 || (name[0] == '<' && strcmp(name, "<any>") == 0)) {
-            make_inetaddr(INADDR_ANY, hbuf, len);
+            make_inetaddr(INADDR_ANY, hbuf, hbuflen);
             if (flags_ptr) *flags_ptr |= AI_NUMERICHOST;
         }
         else if (name[0] == '<' && strcmp(name, "<broadcast>") == 0) {
-            make_inetaddr(INADDR_BROADCAST, hbuf, len);
+            make_inetaddr(INADDR_BROADCAST, hbuf, hbuflen);
             if (flags_ptr) *flags_ptr |= AI_NUMERICHOST;
         }
-        else if (strlen(name) >= len) {
+        else if (strlen(name) >= hbuflen) {
             rb_raise(rb_eArgError, "hostname too long (%"PRIuSIZE")",
                 strlen(name));
         }
@@ -292,13 +321,13 @@ host_str(VALUE host, char *hbuf, size_t len, int *flags_ptr)
 }
 
 static char*
-port_str(VALUE port, char *pbuf, size_t len, int *flags_ptr)
+port_str(VALUE port, char *pbuf, size_t pbuflen, int *flags_ptr)
 {
     if (NIL_P(port)) {
         return 0;
     }
     else if (FIXNUM_P(port)) {
-        snprintf(pbuf, len, "%ld", FIX2LONG(port));
+        snprintf(pbuf, pbuflen, "%ld", FIX2LONG(port));
 #ifdef AI_NUMERICSERV
         if (flags_ptr) *flags_ptr |= AI_NUMERICSERV;
 #endif
@@ -309,7 +338,7 @@ port_str(VALUE port, char *pbuf, size_t len, int *flags_ptr)
 
         SafeStringValue(port);
         serv = RSTRING_PTR(port);
-        if (strlen(serv) >= len) {
+        if (strlen(serv) >= pbuflen) {
             rb_raise(rb_eArgError, "service name too long (%"PRIuSIZE")",
                 strlen(serv));
         }
@@ -343,25 +372,6 @@ rsock_getaddrinfo(VALUE host, VALUE port, struct addrinfo *hints, int socktype_h
         rsock_raise_socket_error("getaddrinfo", error);
     }
 
-#if defined(__APPLE__) && defined(__MACH__)
-    /* [ruby-dev:23164] */
-    {
-        struct addrinfo *r;
-        r = res;
-        while (r) {
-            if (! r->ai_socktype) r->ai_socktype = hints->ai_socktype;
-            if (! r->ai_protocol) {
-                if (r->ai_socktype == SOCK_DGRAM) {
-                    r->ai_protocol = IPPROTO_UDP;
-                }
-                else if (r->ai_socktype == SOCK_STREAM) {
-                    r->ai_protocol = IPPROTO_TCP;
-                }
-            }
-            r = r->ai_next;
-        }
-    }
-#endif
     return res;
 }
 
@@ -378,7 +388,7 @@ rsock_addrinfo(VALUE host, VALUE port, int socktype, int flags)
 }
 
 VALUE
-rsock_ipaddr(struct sockaddr *sockaddr, int norevlookup)
+rsock_ipaddr(struct sockaddr *sockaddr, socklen_t sockaddrlen, int norevlookup)
 {
     VALUE family, port, addr1, addr2;
     VALUE ary;
@@ -397,13 +407,13 @@ rsock_ipaddr(struct sockaddr *sockaddr, int norevlookup)
 
     addr1 = Qnil;
     if (!norevlookup) {
-        error = rb_getnameinfo(sockaddr, SA_LEN(sockaddr), hbuf, sizeof(hbuf),
+        error = rb_getnameinfo(sockaddr, sockaddrlen, hbuf, sizeof(hbuf),
                                NULL, 0, 0);
         if (! error) {
             addr1 = rb_str_new2(hbuf);
         }
     }
-    error = rb_getnameinfo(sockaddr, SA_LEN(sockaddr), hbuf, sizeof(hbuf),
+    error = rb_getnameinfo(sockaddr, sockaddrlen, hbuf, sizeof(hbuf),
                            pbuf, sizeof(pbuf), NI_NUMERICHOST | NI_NUMERICSERV);
     if (error) {
         rsock_raise_socket_error("getnameinfo", error);
@@ -419,20 +429,46 @@ rsock_ipaddr(struct sockaddr *sockaddr, int norevlookup)
 }
 
 #ifdef HAVE_SYS_UN_H
-const char*
-rsock_unixpath(struct sockaddr_un *sockaddr, socklen_t len)
+VALUE
+rsock_unixpath_str(struct sockaddr_un *sockaddr, socklen_t len)
 {
-    if (sockaddr->sun_path < (char*)sockaddr + len)
-        return sockaddr->sun_path;
+    char *s, *e;
+    s = sockaddr->sun_path;
+    e = (char *)sockaddr + len;
+    while (s < e && *(e-1) == '\0')
+        e--;
+    if (s <= e)
+        return rb_str_new(s, e-s);
     else
-        return "";
+        return rb_str_new2("");
 }
 
 VALUE
 rsock_unixaddr(struct sockaddr_un *sockaddr, socklen_t len)
 {
     return rb_assoc_new(rb_str_new2("AF_UNIX"),
-                        rb_str_new2(rsock_unixpath(sockaddr, len)));
+                        rsock_unixpath_str(sockaddr, len));
+}
+
+socklen_t
+rsock_unix_sockaddr_len(VALUE path)
+{
+#ifdef __linux__
+    if (RSTRING_LEN(path) == 0) {
+	/* autobind; see unix(7) for details. */
+	return (socklen_t) sizeof(sa_family_t);
+    }
+    else if (RSTRING_PTR(path)[0] == '\0') {
+	/* abstract namespace; see unix(7) for details. */
+	return (socklen_t) offsetof(struct sockaddr_un, sun_path) +
+	    RSTRING_LEN(path);
+    }
+    else {
+#endif
+	return (socklen_t) sizeof(struct sockaddr_un);
+#ifdef __linux__
+    }
+#endif
 }
 #endif
 
@@ -465,7 +501,8 @@ make_hostent_internal(struct hostent_arg *arg)
     }
     rb_ary_push(ary, rb_str_new2(hostp));
 
-    if (addr->ai_canonname && (h = gethostbyname(addr->ai_canonname))) {
+    if (addr->ai_canonname && strlen(addr->ai_canonname) < NI_MAXHOST &&
+	(h = gethostbyname(addr->ai_canonname))) {
         names = rb_ary_new();
         if (h->h_aliases != NULL) {
             for (pch = h->h_aliases; *pch; pch++) {
@@ -534,7 +571,7 @@ addrinfo_memsize(const void *ptr)
 
 static const rb_data_type_t addrinfo_type = {
     "socket/addrinfo",
-    addrinfo_mark, addrinfo_free, addrinfo_memsize,
+    {addrinfo_mark, addrinfo_free, addrinfo_memsize,},
 };
 
 static VALUE
@@ -543,7 +580,7 @@ addrinfo_s_allocate(VALUE klass)
     return TypedData_Wrap_Struct(klass, &addrinfo_type, 0);
 }
 
-#define IS_ADDRINFO(obj) rb_typeddata_is_kind_of(obj, &addrinfo_type)
+#define IS_ADDRINFO(obj) rb_typeddata_is_kind_of((obj), &addrinfo_type)
 static inline rb_addrinfo_t *
 check_addrinfo(VALUE self)
 {
@@ -577,7 +614,7 @@ init_addrinfo(rb_addrinfo_t *rai, struct sockaddr *sa, socklen_t len,
               int pfamily, int socktype, int protocol,
               VALUE canonname, VALUE inspectname)
 {
-    if (sizeof(rai->addr) < len)
+    if ((socklen_t)sizeof(rai->addr) < len)
         rb_raise(rb_eArgError, "sockaddr string too big");
     memcpy((void *)&rai->addr, (void *)sa, len);
     rai->sockaddr_len = len;
@@ -659,31 +696,32 @@ make_inspectname(VALUE node, VALUE service, struct addrinfo *res)
     VALUE inspectname = Qnil;
 
     if (res) {
+        /* drop redundant information which also shown in address:port part. */
         char hbuf[NI_MAXHOST], pbuf[NI_MAXSERV];
         int ret;
         ret = rb_getnameinfo(res->ai_addr, res->ai_addrlen, hbuf,
                              sizeof(hbuf), pbuf, sizeof(pbuf),
                              NI_NUMERICHOST|NI_NUMERICSERV);
         if (ret == 0) {
-            if (TYPE(node) == T_STRING && strcmp(hbuf, RSTRING_PTR(node)) == 0)
+            if (RB_TYPE_P(node, T_STRING) && strcmp(hbuf, RSTRING_PTR(node)) == 0)
                 node = Qnil;
-            if (TYPE(service) == T_STRING && strcmp(pbuf, RSTRING_PTR(service)) == 0)
+            if (RB_TYPE_P(service, T_STRING) && strcmp(pbuf, RSTRING_PTR(service)) == 0)
                 service = Qnil;
-            else if (TYPE(service) == T_FIXNUM && atoi(pbuf) == FIX2INT(service))
+            else if (RB_TYPE_P(service, T_FIXNUM) && atoi(pbuf) == FIX2INT(service))
                 service = Qnil;
         }
     }
 
-    if (TYPE(node) == T_STRING) {
+    if (RB_TYPE_P(node, T_STRING)) {
         inspectname = rb_str_dup(node);
     }
-    if (TYPE(service) == T_STRING) {
+    if (RB_TYPE_P(service, T_STRING)) {
         if (NIL_P(inspectname))
             inspectname = rb_sprintf(":%s", StringValueCStr(service));
         else
             rb_str_catf(inspectname, ":%s", StringValueCStr(service));
     }
-    else if (TYPE(service) == T_FIXNUM && FIX2INT(service) != 0)
+    else if (RB_TYPE_P(service, T_FIXNUM) && FIX2INT(service) != 0)
     {
         if (NIL_P(inspectname))
             inspectname = rb_sprintf(":%d", FIX2INT(service));
@@ -761,19 +799,22 @@ static void
 init_unix_addrinfo(rb_addrinfo_t *rai, VALUE path, int socktype)
 {
     struct sockaddr_un un;
+    socklen_t len;
 
     StringValue(path);
 
-    if (sizeof(un.sun_path) <= (size_t)RSTRING_LEN(path))
-        rb_raise(rb_eArgError, "too long unix socket path (max: %dbytes)",
-            (int)sizeof(un.sun_path)-1);
+    if (sizeof(un.sun_path) < (size_t)RSTRING_LEN(path))
+        rb_raise(rb_eArgError,
+            "too long unix socket path (%"PRIuSIZE" bytes given but %"PRIuSIZE" bytes max)",
+            (size_t)RSTRING_LEN(path), sizeof(un.sun_path));
 
     MEMZERO(&un, struct sockaddr_un, 1);
 
     un.sun_family = AF_UNIX;
     memcpy((void*)&un.sun_path, RSTRING_PTR(path), RSTRING_LEN(path));
 
-    init_addrinfo(rai, (struct sockaddr *)&un, (socklen_t)sizeof(un),
+    len = rsock_unix_sockaddr_len(path);
+    init_addrinfo(rai, (struct sockaddr *)&un, len,
 		  PF_UNIX, socktype, 0, Qnil, Qnil);
 }
 #endif
@@ -907,7 +948,7 @@ addrinfo_initialize(int argc, VALUE *argv, VALUE self)
 static int
 get_afamily(struct sockaddr *addr, socklen_t len)
 {
-    if ((char*)&addr->sa_family + sizeof(addr->sa_family) - (char*)addr <= len)
+    if ((socklen_t)((char*)&addr->sa_family + sizeof(addr->sa_family) - (char*)addr) <= len)
         return addr->sa_family;
     else
         return AF_UNSPEC;
@@ -935,7 +976,7 @@ inspect_sockaddr(VALUE addrinfo, VALUE ret)
           {
             struct sockaddr_in *addr;
             int port;
-            if (rai->sockaddr_len < sizeof(struct sockaddr_in)) {
+            if (rai->sockaddr_len < (socklen_t)sizeof(struct sockaddr_in)) {
                 rb_str_cat2(ret, "too-short-AF_INET-sockaddr");
             }
             else {
@@ -948,7 +989,7 @@ inspect_sockaddr(VALUE addrinfo, VALUE ret)
                 port = ntohs(addr->sin_port);
                 if (port)
                     rb_str_catf(ret, ":%d", port);
-                if (sizeof(struct sockaddr_in) < rai->sockaddr_len)
+                if ((socklen_t)sizeof(struct sockaddr_in) < rai->sockaddr_len)
                     rb_str_catf(ret, "(sockaddr %d bytes too long)", (int)(rai->sockaddr_len - sizeof(struct sockaddr_in)));
             }
             break;
@@ -961,7 +1002,7 @@ inspect_sockaddr(VALUE addrinfo, VALUE ret)
             char hbuf[1024];
             int port;
             int error;
-            if (rai->sockaddr_len < sizeof(struct sockaddr_in6)) {
+            if (rai->sockaddr_len < (socklen_t)sizeof(struct sockaddr_in6)) {
                 rb_str_cat2(ret, "too-short-AF_INET6-sockaddr");
             }
             else {
@@ -983,7 +1024,7 @@ inspect_sockaddr(VALUE addrinfo, VALUE ret)
                     port = ntohs(addr->sin6_port);
                     rb_str_catf(ret, "[%s]:%d", hbuf, port);
                 }
-                if (sizeof(struct sockaddr_in6) < rai->sockaddr_len)
+                if ((socklen_t)sizeof(struct sockaddr_in6) < rai->sockaddr_len)
                     rb_str_catf(ret, "(sockaddr %d bytes too long)", (int)(rai->sockaddr_len - sizeof(struct sockaddr_in6)));
             }
             break;
@@ -994,9 +1035,11 @@ inspect_sockaddr(VALUE addrinfo, VALUE ret)
           case AF_UNIX:
           {
             struct sockaddr_un *addr = (struct sockaddr_un *)&rai->addr;
-            char *p, *s, *t, *e;
+            char *p, *s, *e;
             s = addr->sun_path;
             e = (char*)addr + rai->sockaddr_len;
+            while (s < e && *(e-1) == '\0')
+                e--;
             if (e < s)
                 rb_str_cat2(ret, "too-short-AF_UNIX-sockaddr");
             else if (s == e)
@@ -1004,28 +1047,17 @@ inspect_sockaddr(VALUE addrinfo, VALUE ret)
             else {
                 int printable_only = 1;
                 p = s;
-                while (p < e && *p != '\0') {
+                while (p < e) {
                     printable_only = printable_only && ISPRINT(*p) && !ISSPACE(*p);
                     p++;
                 }
-                t = p;
-                while (p < e && *p == '\0')
-                    p++;
-                if (printable_only && /* only printable, no space */
-                    t < e && /* NUL terminated */
-                    p == e) { /* no data after NUL */
-		    if (s == t)
-			rb_str_cat2(ret, "empty-path-AF_UNIX-sockaddr");
-		    else if (s[0] == '/') /* absolute path */
-			rb_str_cat2(ret, s);
-		    else
-                        rb_str_catf(ret, "AF_UNIX %s", s);
+                if (printable_only) { /* only printable, no space */
+                    if (s[0] != '/') /* relative path */
+                        rb_str_cat2(ret, "AF_UNIX ");
+                    rb_str_cat(ret, s, p - s);
                 }
                 else {
                     rb_str_cat2(ret, "AF_UNIX");
-                    e = (char *)addr->sun_path + sizeof(addr->sun_path);
-                    while (s < e && *(e-1) == '\0')
-                        e--;
                     while (s < e)
                         rb_str_catf(ret, ":%02x", (unsigned char)*s++);
                 }
@@ -1058,7 +1090,7 @@ inspect_sockaddr(VALUE addrinfo, VALUE ret)
  *
  * returns a string which shows addrinfo in human-readable form.
  *
- *   Addrinfo.tcp("localhost", 80).inspect #=> "#<Addrinfo: 127.0.0.1:80 TCP (localhost:80)>"
+ *   Addrinfo.tcp("localhost", 80).inspect #=> "#<Addrinfo: 127.0.0.1:80 TCP (localhost)>"
  *   Addrinfo.unix("/tmp/sock").inspect    #=> "#<Addrinfo: /tmp/sock SOCK_STREAM>"
  *
  */
@@ -1199,7 +1231,7 @@ addrinfo_mdump(VALUE self)
         struct sockaddr_un *su = (struct sockaddr_un *)&rai->addr;
         char *s, *e;
         s = su->sun_path;
-        e = (char*)s + sizeof(su->sun_path);
+        e = (char*)su + rai->sockaddr_len;
         while (s < e && *(e-1) == '\0')
             e--;
         sockaddr = rb_str_new(s, e-s);
@@ -1296,12 +1328,14 @@ addrinfo_mload(VALUE self, VALUE ary)
       case AF_UNIX:
       {
         struct sockaddr_un uaddr;
-        memset(&uaddr, 0, sizeof(uaddr));
+        MEMZERO(&uaddr, struct sockaddr_un, 1);
         uaddr.sun_family = AF_UNIX;
 
         StringValue(v);
-        if (sizeof(uaddr.sun_path) <= (size_t)RSTRING_LEN(v))
-            rb_raise(rb_eSocket, "too long AF_UNIX path");
+        if (sizeof(uaddr.sun_path) < (size_t)RSTRING_LEN(v))
+            rb_raise(rb_eSocket,
+                "too long AF_UNIX path (%"PRIuSIZE" bytes given but %"PRIuSIZE" bytes max)",
+                (size_t)RSTRING_LEN(v), sizeof(uaddr.sun_path));
         memcpy(uaddr.sun_path, RSTRING_PTR(v), RSTRING_LEN(v));
         len = (socklen_t)sizeof(uaddr);
         memcpy(&ss, &uaddr, len);
@@ -1401,6 +1435,7 @@ addrinfo_protocol(VALUE self)
 /*
  * call-seq:
  *   addrinfo.to_sockaddr => string
+ *   addrinfo.to_s => string
  *
  * returns the socket address as packed struct sockaddr string.
  *
@@ -1429,7 +1464,7 @@ addrinfo_to_sockaddr(VALUE self)
  * The canonical name is set by Addrinfo.getaddrinfo when AI_CANONNAME is specified.
  *
  *   list = Addrinfo.getaddrinfo("www.ruby-lang.org", 80, :INET, :STREAM, nil, Socket::AI_CANONNAME)
- *   p list[0] #=> #<Addrinfo: 221.186.184.68:80 TCP carbon.ruby-lang.org (www.ruby-lang.org:80)>
+ *   p list[0] #=> #<Addrinfo: 221.186.184.68:80 TCP carbon.ruby-lang.org (www.ruby-lang.org)>
  *   p list[0].canonname #=> "carbon.ruby-lang.org"
  *
  */
@@ -1789,6 +1824,18 @@ addrinfo_ipv6_sitelocal_p(VALUE self)
     return Qfalse;
 }
 
+/*
+ * Returns true for IPv6 unique local address (fc00::/7, RFC4193).
+ * It returns false otherwise.
+ */
+static VALUE
+addrinfo_ipv6_unique_local_p(VALUE self)
+{
+    struct in6_addr *addr = extract_in6_addr(self);
+    if (addr && IN6_IS_ADDR_UNIQUE_LOCAL(addr)) return Qtrue;
+    return Qfalse;
+}
+
 /*
  * Returns true for IPv4-mapped IPv6 address (::ffff:0:0/80).
  * It returns false otherwise.
@@ -1933,9 +1980,12 @@ addrinfo_unix_path(VALUE self)
     s = addr->sun_path;
     e = (char*)addr + rai->sockaddr_len;
     if (e < s)
-        rb_raise(rb_eSocket, "too short AF_UNIX address");
+        rb_raise(rb_eSocket, "too short AF_UNIX address: %"PRIuSIZE" bytes given for minimum %"PRIuSIZE" bytes.",
+            (size_t)rai->sockaddr_len, (size_t)(s - (char *)addr));
     if (addr->sun_path + sizeof(addr->sun_path) < e)
-        rb_raise(rb_eSocket, "too long AF_UNIX address");
+        rb_raise(rb_eSocket,
+            "too long AF_UNIX path (%"PRIuSIZE" bytes given but %"PRIuSIZE" bytes max)",
+            (size_t)(e - addr->sun_path), sizeof(addr->sun_path));
     while (s < e && *(e-1) == '\0')
         e--;
     return rb_str_new(s, e-s);
@@ -1965,15 +2015,24 @@ addrinfo_unix_path(VALUE self)
  *
  * Similarly, PF_INET6 as family restricts for IPv6.
  *
- * flags should be bitwise OR of Socket::AI_??? constants.
+ * flags should be bitwise OR of Socket::AI_??? constants such as follows.
+ * Note that the exact list of the constants depends on OS.
+ *
+ *   AI_PASSIVE      Get address to use with bind()
+ *   AI_CANONNAME    Fill in the canonical name
+ *   AI_NUMERICHOST  Prevent host name resolution
+ *   AI_NUMERICSERV  Prevent service name resolution
+ *   AI_V4MAPPED     Accept IPv4-mapped IPv6 addresses
+ *   AI_ALL          Allow all addresses
+ *   AI_ADDRCONFIG   Accept only if any address is assigned
  *
  * Note that socktype should be specified whenever application knows the usage of the address.
  * Some platform causes an error when socktype is omitted and servname is specified as an integer
  * because some port numbers, 512 for example, are ambiguous without socktype.
  *
  *   Addrinfo.getaddrinfo("www.kame.net", 80, nil, :STREAM)
- *   #=> [#<Addrinfo: 203.178.141.194:80 TCP (www.kame.net:80)>,
- *   #    #<Addrinfo: [2001:200:0:8002:203:47ff:fea5:3085]:80 TCP (www.kame.net:80)>]
+ *   #=> [#<Addrinfo: 203.178.141.194:80 TCP (www.kame.net)>,
+ *   #    #<Addrinfo: [2001:200:dff:fff1:216:3eff:feb1:44d7]:80 TCP (www.kame.net)>]
  *
  */
 static VALUE
@@ -2140,6 +2199,8 @@ rsock_io_socket_addrinfo(VALUE io, struct sockaddr *addr, socklen_t len)
       default:
         rb_raise(rb_eTypeError, "neither IO nor file descriptor");
     }
+
+    UNREACHABLE;
 }
 
 /*
@@ -2148,6 +2209,10 @@ rsock_io_socket_addrinfo(VALUE io, struct sockaddr *addr, socklen_t len)
 void
 rsock_init_addrinfo(void)
 {
+    /*
+     * The Addrinfo class maps <tt>struct addrinfo</tt> to ruby.  This
+     * structure identifies an Internet host and a service.
+     */
     rb_cAddrinfo = rb_define_class("Addrinfo", rb_cData);
     rb_define_alloc_func(rb_cAddrinfo, addrinfo_s_allocate);
     rb_define_method(rb_cAddrinfo, "initialize", addrinfo_initialize, -1);
@@ -2186,6 +2251,7 @@ rsock_init_addrinfo(void)
     rb_define_method(rb_cAddrinfo, "ipv6_multicast?", addrinfo_ipv6_multicast_p, 0);
     rb_define_method(rb_cAddrinfo, "ipv6_linklocal?", addrinfo_ipv6_linklocal_p, 0);
     rb_define_method(rb_cAddrinfo, "ipv6_sitelocal?", addrinfo_ipv6_sitelocal_p, 0);
+    rb_define_method(rb_cAddrinfo, "ipv6_unique_local?", addrinfo_ipv6_unique_local_p, 0);
     rb_define_method(rb_cAddrinfo, "ipv6_v4mapped?", addrinfo_ipv6_v4mapped_p, 0);
     rb_define_method(rb_cAddrinfo, "ipv6_v4compat?", addrinfo_ipv6_v4compat_p, 0);
     rb_define_method(rb_cAddrinfo, "ipv6_mc_nodelocal?", addrinfo_ipv6_mc_nodelocal_p, 0);
