@@ -1,5 +1,5 @@
 /*
-    $Id: strscan.c 25546 2009-10-28 16:11:20Z yugui $
+    $Id: strscan.c 37916 2012-11-28 00:17:33Z ryan $
 
     Copyright (c) 1999-2006 Minero Aoki
 
@@ -11,6 +11,7 @@
 #include "ruby/ruby.h"
 #include "ruby/re.h"
 #include "ruby/encoding.h"
+#include "regint.h"
 
 #define STRSCAN_VERSION "0.7.0"
 
@@ -20,6 +21,7 @@
 
 static VALUE StringScanner;
 static VALUE ScanError;
+static ID id_byteslice;
 
 struct strscanner
 {
@@ -29,7 +31,7 @@ struct strscanner
 
     /* the string to scan */
     VALUE str;
-    
+
     /* scan pointers */
     long prev;   /* legal only when MATCHED_P(s) */
     long curr;   /* always legal */
@@ -51,8 +53,8 @@ struct strscanner
 #define EOS_P(s) ((s)->curr >= RSTRING_LEN(p->str))
 
 #define GET_SCANNER(obj,var) do {\
-    Data_Get_Struct(obj, struct strscanner, var);\
-    if (NIL_P(var->str)) rb_raise(rb_eArgError, "uninitialized StringScanner object");\
+    (var) = check_strscan(obj);\
+    if (NIL_P((var)->str)) rb_raise(rb_eArgError, "uninitialized StringScanner object");\
 } while (0)
 
 /* =======================================================================
@@ -63,9 +65,10 @@ static VALUE infect _((VALUE str, struct strscanner *p));
 static VALUE extract_range _((struct strscanner *p, long beg_i, long end_i));
 static VALUE extract_beg_len _((struct strscanner *p, long beg_i, long len));
 
-void check_strscan _((VALUE obj));
-static void strscan_mark _((struct strscanner *p));
-static void strscan_free _((struct strscanner *p));
+static struct strscanner *check_strscan _((VALUE obj));
+static void strscan_mark _((void *p));
+static void strscan_free _((void *p));
+static size_t strscan_memsize _((const void *p));
 static VALUE strscan_s_allocate _((VALUE klass));
 static VALUE strscan_initialize _((int argc, VALUE *argv, VALUE self));
 static VALUE strscan_init_copy _((VALUE vself, VALUE vorig));
@@ -157,29 +160,47 @@ extract_beg_len(struct strscanner *p, long beg_i, long len)
    ======================================================================= */
 
 static void
-strscan_mark(struct strscanner *p)
+strscan_mark(void *ptr)
 {
+    struct strscanner *p = ptr;
     rb_gc_mark(p->str);
 }
 
 static void
-strscan_free(struct strscanner *p)
+strscan_free(void *ptr)
 {
+    struct strscanner *p = ptr;
     onig_region_free(&(p->regs), 0);
     ruby_xfree(p);
 }
 
+static size_t
+strscan_memsize(const void *ptr)
+{
+    const struct strscanner *p = ptr;
+    size_t size = 0;
+    if (p) {
+	size = sizeof(*p) - sizeof(p->regs) + onig_region_memsize(&p->regs);
+    }
+    return size;
+}
+
+static const rb_data_type_t strscanner_type = {
+    "StringScanner",
+    {strscan_mark, strscan_free, strscan_memsize}
+};
+
 static VALUE
 strscan_s_allocate(VALUE klass)
 {
     struct strscanner *p;
-    
+
     p = ALLOC(struct strscanner);
     MEMZERO(p, struct strscanner, 1);
     CLEAR_MATCH_STATUS(p);
     onig_region_init(&(p->regs));
     p->str = Qnil;
-    return Data_Wrap_Struct(klass, strscan_mark, strscan_free, p);
+    return TypedData_Wrap_Struct(klass, &strscanner_type, p);
 }
 
 /*
@@ -194,7 +215,7 @@ strscan_initialize(int argc, VALUE *argv, VALUE self)
     struct strscanner *p;
     VALUE str, need_dup;
 
-    Data_Get_Struct(self, struct strscanner, p);
+    p = check_strscan(self);
     rb_scan_args(argc, argv, "11", &str, &need_dup);
     StringValue(str);
     p->str = str;
@@ -202,14 +223,10 @@ strscan_initialize(int argc, VALUE *argv, VALUE self)
     return self;
 }
 
-void
+static struct strscanner *
 check_strscan(VALUE obj)
 {
-    if (TYPE(obj) != T_DATA || RDATA(obj)->dmark != (RUBY_DATA_FUNC)strscan_mark) {
-        rb_raise(rb_eTypeError,
-                 "wrong argument type %s (expected StringScanner)",
-                 rb_obj_classname(obj));
-    }
+    return rb_check_typeddata(obj, &strscanner_type);
 }
 
 /*
@@ -224,9 +241,8 @@ strscan_init_copy(VALUE vself, VALUE vorig)
 {
     struct strscanner *self, *orig;
 
-    Data_Get_Struct(vself, struct strscanner, self);
-    check_strscan(vorig);
-    Data_Get_Struct(vorig, struct strscanner, orig);
+    self = check_strscan(vself);
+    orig = check_strscan(vorig);
     if (self != orig) {
 	self->flags = orig->flags;
 	self->str = orig->str;
@@ -317,9 +333,8 @@ strscan_get_string(VALUE self)
 static VALUE
 strscan_set_string(VALUE self, VALUE str)
 {
-    struct strscanner *p;
+    struct strscanner *p = check_strscan(self);
 
-    Data_Get_Struct(self, struct strscanner, p);
     StringValue(str);
     p->str = str;
     p->curr = 0;
@@ -353,11 +368,11 @@ strscan_concat(VALUE self, VALUE str)
 }
 
 /*
- * Returns the position of the scan pointer.  In the 'reset' position, this
+ * Returns the byte position of the scan pointer.  In the 'reset' position, this
  * value is zero.  In the 'terminated' position (i.e. the string is exhausted),
- * this value is the length of the string.
+ * this value is the bytesize of the string.
  *
- * In short, it's a 0-based index into the string.
+ * In short, it's a 0-based index into bytes of the string.
  *
  *   s = StringScanner.new('test string')
  *   s.pos               # -> 0
@@ -375,10 +390,36 @@ strscan_get_pos(VALUE self)
     return INT2FIX(p->curr);
 }
 
+/*
+ * Returns the character position of the scan pointer.  In the 'reset' position, this
+ * value is zero.  In the 'terminated' position (i.e. the string is exhausted),
+ * this value is the size of the string.
+ *
+ * In short, it's a 0-based index into the string.
+ *
+ *   s = StringScanner.new("abcädeföghi")
+ *   s.charpos           # -> 0
+ *   s.scan_until(/ä/)   # -> "abcä"
+ *   s.pos               # -> 5
+ *   s.charpos           # -> 4
+ */
+static VALUE
+strscan_get_charpos(VALUE self)
+{
+    struct strscanner *p;
+    VALUE substr;
+
+    GET_SCANNER(self, p);
+
+    substr = rb_funcall(p->str, id_byteslice, 2, INT2FIX(0), INT2NUM(p->curr));
+
+    return rb_str_length(substr);
+}
+
 /*
  * call-seq: pos=(n)
  *
- * Modify the scan pointer.
+ * Set the byte position of the scan pointer.
  *
  *   s = StringScanner.new('test string')
  *   s.pos = 7            # -> 7
@@ -405,7 +446,7 @@ strscan_do_scan(VALUE self, VALUE regex, int succptr, int getstr, int headonly)
     regex_t *rb_reg_prepare_re(VALUE re, VALUE str);
     struct strscanner *p;
     regex_t *re;
-    int ret;
+    long ret;
     int tmpreg;
 
     Check_Type(regex, T_REGEXP);
@@ -609,7 +650,7 @@ strscan_exist_p(VALUE self, VALUE re)
  *
  *   s = StringScanner.new("Fri Dec 12 1975 14:39")
  *   s.skip_until /12/           # -> 10
- *   s                           # 
+ *   s                           #
  */
 static VALUE
 strscan_skip_until(VALUE self, VALUE re)
@@ -655,7 +696,7 @@ static void
 adjust_registers_to_matched(struct strscanner *p)
 {
     onig_region_clear(&(p->regs));
-    onig_region_set(&(p->regs), 0, 0, p->curr - p->prev);
+    onig_region_set(&(p->regs), 0, 0, (int)(p->curr - p->prev));
 }
 
 /*
@@ -893,7 +934,7 @@ strscan_matched_p(VALUE self)
 
 /*
  * Returns the last matched string.
- * 
+ *
  *   s = StringScanner.new('test string')
  *   s.match?(/\w+/)     # -> 4
  *   s.matched           # -> "test"
@@ -929,17 +970,6 @@ strscan_matched_size(VALUE self)
     return INT2NUM(p->regs.end[0] - p->regs.beg[0]);
 }
 
-/*
- * Equivalent to #matched_size.
- * This method is obsolete; use #matched_size instead.
- */
-static VALUE
-strscan_matchedsize(VALUE self)
-{
-    rb_warning("StringScanner#matchedsize is obsolete; use #matched_size instead");
-    return strscan_matched_size(self);
-}
-
 /*
  * call-seq: [](n)
  *
@@ -962,7 +992,7 @@ strscan_aref(VALUE self, VALUE idx)
 
     GET_SCANNER(self, p);
     if (! MATCHED_P(p))        return Qnil;
-    
+
     i = NUM2LONG(idx);
     if (i < 0)
         i += p->regs.num_regs;
@@ -1078,7 +1108,7 @@ strscan_inspect(VALUE self)
     long len;
     VALUE a, b;
 
-    Data_Get_Struct(self, struct strscanner, p);
+    p = check_strscan(self);
     if (NIL_P(p->str)) {
         len = snprintf(buf, BUFSIZE, "#<%s (uninitialized)>",
                        rb_class2name(CLASS_OF(self)));
@@ -1152,20 +1182,20 @@ inspect2(struct strscanner *p)
 
 /*
  * Document-class: StringScanner
- * 
+ *
  * StringScanner provides for lexical scanning operations on a String.  Here is
  * an example of its usage:
  *
  *   s = StringScanner.new('This is an example string')
  *   s.eos?               # -> false
- *   
+ *
  *   p s.scan(/\w+/)      # -> "This"
  *   p s.scan(/\w+/)      # -> nil
  *   p s.scan(/\s+/)      # -> " "
  *   p s.scan(/\s+/)      # -> nil
  *   p s.scan(/\w+/)      # -> "is"
  *   s.eos?               # -> false
- *   
+ *
  *   p s.scan(/\s+/)      # -> " "
  *   p s.scan(/\w+/)      # -> "an"
  *   p s.scan(/\s+/)      # -> " "
@@ -1173,7 +1203,7 @@ inspect2(struct strscanner *p)
  *   p s.scan(/\s+/)      # -> " "
  *   p s.scan(/\w+/)      # -> "string"
  *   s.eos?               # -> true
- *   
+ *
  *   p s.scan(/\s+/)      # -> nil
  *   p s.scan(/\w+/)      # -> nil
  *
@@ -1202,7 +1232,7 @@ inspect2(struct strscanner *p)
  * the string without actually scanning.  You can access the most recent match.
  * You can modify the string being scanned, reset or terminate the scanner,
  * find out or change the position of the scan pointer, skip ahead, and so on.
- * 
+ *
  * === Advancing the Scan Pointer
  *
  * - #getch
@@ -1233,7 +1263,7 @@ inspect2(struct strscanner *p)
  * - #reset
  * - #terminate
  * - #pos=
- * 
+ *
  * === Match Data
  *
  * - #matched
@@ -1259,6 +1289,8 @@ Init_strscan()
     ID id_scanerr = rb_intern("ScanError");
     VALUE tmp;
 
+    id_byteslice = rb_intern("byteslice");
+
     StringScanner = rb_define_class("StringScanner", rb_cObject);
     ScanError = rb_define_class_under(StringScanner, "Error", rb_eStandardError);
     if (!rb_const_defined(rb_cObject, id_scanerr)) {
@@ -1267,10 +1299,10 @@ Init_strscan()
     tmp = rb_str_new2(STRSCAN_VERSION);
     rb_obj_freeze(tmp);
     rb_const_set(StringScanner, rb_intern("Version"), tmp);
-    tmp = rb_str_new2("$Id: strscan.c 25546 2009-10-28 16:11:20Z yugui $");
+    tmp = rb_str_new2("$Id: strscan.c 37916 2012-11-28 00:17:33Z ryan $");
     rb_obj_freeze(tmp);
     rb_const_set(StringScanner, rb_intern("Id"), tmp);
-    
+
     rb_define_alloc_func(StringScanner, strscan_s_allocate);
     rb_define_private_method(StringScanner, "initialize", strscan_initialize, -1);
     rb_define_private_method(StringScanner, "initialize_copy", strscan_init_copy, 1);
@@ -1284,6 +1316,7 @@ Init_strscan()
     rb_define_method(StringScanner, "<<",          strscan_concat,      1);
     rb_define_method(StringScanner, "pos",         strscan_get_pos,     0);
     rb_define_method(StringScanner, "pos=",        strscan_set_pos,     1);
+    rb_define_method(StringScanner, "charpos",     strscan_get_charpos, 0);
     rb_define_method(StringScanner, "pointer",     strscan_get_pos,     0);
     rb_define_method(StringScanner, "pointer=",    strscan_set_pos,     1);
 
@@ -1316,7 +1349,6 @@ Init_strscan()
     rb_define_method(StringScanner, "matched?",    strscan_matched_p,   0);
     rb_define_method(StringScanner, "matched",     strscan_matched,     0);
     rb_define_method(StringScanner, "matched_size", strscan_matched_size, 0);
-    rb_define_method(StringScanner, "matchedsize", strscan_matchedsize, 0);
     rb_define_method(StringScanner, "[]",          strscan_aref,        1);
     rb_define_method(StringScanner, "pre_match",   strscan_pre_match,   0);
     rb_define_method(StringScanner, "post_match",  strscan_post_match,  0);
