@@ -13,7 +13,7 @@ module TkCanvasItemConfig
   def __item_strval_optkeys(id)
     # maybe need to override
     super(id) + [
-      'fill', 'activefill', 'disabledfill', 
+      'fill', 'activefill', 'disabledfill',
       'outline', 'activeoutline', 'disabledoutline'
     ]
   end
@@ -25,7 +25,10 @@ module TkCanvasItemConfig
   private :__item_methodcall_optkeys
 
   def __item_val2ruby_optkeys(id)  # { key=>proc, ... }
-    super(id).update('window'=>proc{|i, v| window(v)})
+    super(id).update('window'=>proc{|i, v| window(v)},
+                     'tags'=>proc{|i, v|
+                       simplelist(v).collect{|tag| TkcTag.id2obj(self, tag)}
+                     })
   end
   private :__item_val2ruby_optkeys
 
@@ -45,7 +48,7 @@ class Tk::Canvas<TkWindow
 
   TkCommandNames = ['canvas'.freeze].freeze
   WidgetClassName = 'Canvas'.freeze
-  WidgetClassNames[WidgetClassName] = self
+  WidgetClassNames[WidgetClassName] ||= self
 
   def __destroy_hook__
     TkcItem::CItemID_TBL.delete(@path)
@@ -82,10 +85,16 @@ class Tk::Canvas<TkWindow
 
   # create a canvas item without creating a TkcItem object
   def create(type, *args)
+    if type.kind_of?(Class) && type < TkcItem
+      # do nothing
+    elsif TkcItem.type2class(type.to_s)
+      type = TkcItem.type2class(type.to_s)
+    else
+      fail ArgumentError, "type must a subclass of TkcItem class, or a string in CItemTypeToClass"
+    end
     type.create(self, *args)
   end
 
-
   def addtag(tag, mode, *args)
     mode = mode.to_s
     if args[0] && mode =~ /^(above|below|with(tag)?)$/
@@ -117,7 +126,7 @@ class Tk::Canvas<TkWindow
   end
 
   def bbox(tagOrId, *tags)
-    list(tk_send_without_enc('bbox', tagid(tagOrId), 
+    list(tk_send_without_enc('bbox', tagid(tagOrId),
                              *tags.collect{|t| tagid(t)}))
   end
 
@@ -168,9 +177,11 @@ class Tk::Canvas<TkWindow
     #tk_tcl2ruby(tk_send_without_enc('canvasy', screen_y, *args))
     number(tk_send_without_enc('canvasy', screen_y, *args))
   end
+  alias canvas_x canvasx
+  alias canvas_y canvasy
 
   def coords(tag, *args)
-    if args == []
+    if args.empty?
       tk_split_list(tk_send_without_enc('coords', tagid(tag)))
     else
       tk_send_without_enc('coords', tagid(tag), *(args.flatten))
@@ -179,17 +190,23 @@ class Tk::Canvas<TkWindow
   end
 
   def dchars(tag, first, last=None)
-    tk_send_without_enc('dchars', tagid(tag), 
+    tk_send_without_enc('dchars', tagid(tag),
                         _get_eval_enc_str(first), _get_eval_enc_str(last))
     self
   end
 
   def delete(*args)
-    if TkcItem::CItemID_TBL[self.path]
+    tbl = nil
+    TkcItem::CItemID_TBL.mutex.synchronize{
+      tbl = TkcItem::CItemID_TBL[self.path]
+    }
+    if tbl
       args.each{|tag|
         find('withtag', tag).each{|item|
           if item.kind_of?(TkcItem)
-            TkcItem::CItemID_TBL[self.path].delete(item.id)
+            TkcItem::CItemID_TBL.mutex.synchronize{
+              tbl.delete(item.id)
+            }
           end
         }
       }
@@ -206,7 +223,7 @@ class Tk::Canvas<TkWindow
   alias deltag dtag
 
   def find(mode, *args)
-    list(tk_send_without_enc('find', mode, *args)).collect!{|id| 
+    list(tk_send_without_enc('find', mode, *args)).collect!{|id|
       TkcItem.id2obj(self, id)
     }
   end
@@ -257,12 +274,18 @@ class Tk::Canvas<TkWindow
     self
   end
 
+  def imove(tagOrId, idx, x, y)
+    tk_send_without_enc('imove', tagid(tagOrId), idx, x, y)
+    self
+  end
+  alias i_move imove
+
   def index(tagOrId, idx)
     number(tk_send_without_enc('index', tagid(tagOrId), idx))
   end
 
   def insert(tagOrId, index, string)
-    tk_send_without_enc('insert', tagid(tagOrId), index, 
+    tk_send_without_enc('insert', tagid(tagOrId), index,
                         _get_eval_enc_str(string))
     self
   end
@@ -292,7 +315,7 @@ class Tk::Canvas<TkWindow
         fnt
       end
     else
-      tk_tcl2ruby(_fromUTF8(tk_send_without_enc('itemcget', tagid(tagOrId), 
+      tk_tcl2ruby(_fromUTF8(tk_send_without_enc('itemcget', tagid(tagOrId),
                                                 "-#{option}")))
     end
   end
@@ -307,16 +330,16 @@ class Tk::Canvas<TkWindow
           || key['latinfont'] || key['asciifont'] )
         tagfont_configure(tagid(tagOrId), key.dup)
       else
-        _fromUTF8(tk_send_without_enc('itemconfigure', tagid(tagOrId), 
+        _fromUTF8(tk_send_without_enc('itemconfigure', tagid(tagOrId),
                                       *hash_kv(key, true)))
       end
 
     else
       if ( key == 'coords' || key == :coords )
         self.coords(tagOrId, value)
-      elsif ( key == 'font' || key == :font || 
-              key == 'kanjifont' || key == :kanjifont || 
-              key == 'latinfont' || key == :latinfont || 
+      elsif ( key == 'font' || key == :font ||
+              key == 'kanjifont' || key == :kanjifont ||
+              key == 'latinfont' || key == :latinfont ||
               key == 'asciifont' || key == :asciifont )
         if value == None
           tagfontobj(tagid(tagOrId))
@@ -324,7 +347,7 @@ class Tk::Canvas<TkWindow
           tagfont_configure(tagid(tagOrId), {key=>value})
         end
       else
-        _fromUTF8(tk_send_without_enc('itemconfigure', tagid(tagOrId), 
+        _fromUTF8(tk_send_without_enc('itemconfigure', tagid(tagOrId),
                                       "-#{key}", _get_eval_enc_str(value)))
       end
     end
@@ -381,16 +404,16 @@ class Tk::Canvas<TkWindow
           else
             if conf[3]
               if conf[3].index('{')
-                conf[3] = tk_split_list(conf[3]) 
+                conf[3] = tk_split_list(conf[3])
               else
-                conf[3] = tk_tcl2ruby(conf[3]) 
+                conf[3] = tk_tcl2ruby(conf[3])
               end
             end
             if conf[4]
               if conf[4].index('{')
-                conf[4] = tk_split_list(conf[4]) 
+                conf[4] = tk_split_list(conf[4])
               else
-                conf[4] = tk_tcl2ruby(conf[4]) 
+                conf[4] = tk_tcl2ruby(conf[4])
               end
             end
           end
@@ -413,8 +436,8 @@ class Tk::Canvas<TkWindow
         when 'coords'
           {'coords' => ['', '', '', self.coords(tagOrId)]}
         when 'dash', 'activedash', 'disableddash'
-          conf = tk_split_simplelist(tk_send_without_enc('itemconfigure', 
-                                                         tagid(tagOrId), 
+          conf = tk_split_simplelist(tk_send_without_enc('itemconfigure',
+                                                         tagid(tagOrId),
                                                          "-#{key}"))
           if conf[3] && conf[3] =~ /^[0-9]/
             conf[3] = list(conf[3])
@@ -449,16 +472,16 @@ class Tk::Canvas<TkWindow
           else
             if conf[2]
               if conf[2].index('{')
-                conf[2] = tk_split_list(conf[2]) 
+                conf[2] = tk_split_list(conf[2])
               else
-                conf[2] = tk_tcl2ruby(conf[2]) 
+                conf[2] = tk_tcl2ruby(conf[2])
               end
             end
             if conf[3]
               if conf[3].index('{')
-                conf[3] = tk_split_list(conf[3]) 
+                conf[3] = tk_split_list(conf[3])
               else
-                conf[3] = tk_tcl2ruby(conf[3]) 
+                conf[3] = tk_tcl2ruby(conf[3])
               end
             end
           end
@@ -515,11 +538,18 @@ class Tk::Canvas<TkWindow
     self
   end
 
-  def move(tag, x, y)
-    tk_send_without_enc('move', tagid(tag), x, y)
+  def move(tag, dx, dy)
+    tk_send_without_enc('move', tagid(tag), dx, dy)
     self
   end
 
+  def moveto(tag, x, y)
+    # Tcl/Tk 8.6 or later
+    tk_send_without_enc('moveto', tagid(tag), x, y)
+    self
+  end
+  alias move_to moveto
+
   def postscript(keys)
     tk_send("postscript", *hash_kv(keys))
   end
@@ -533,6 +563,15 @@ class Tk::Canvas<TkWindow
     self
   end
 
+  def rchars(tag, first, last, str_or_coords)
+    # Tcl/Tk 8.6 or later
+    str_or_coords = str_or_coords.flatten if str_or_coords.kinad_of? Array
+    tk_send_without_enc('rchars', tagid(tag), first, last, str_or_coords)
+    self
+  end
+  alias replace_chars rchars
+  alias replace_coords rchars
+
   def scale(tag, x, y, xs, ys)
     tk_send_without_enc('scale', tagid(tag), x, y, xs, ys)
     self
@@ -570,10 +609,35 @@ class Tk::Canvas<TkWindow
   def itemtype(tag)
     TkcItem.type2class(tk_send('type', tagid(tag)))
   end
+
+  def create_itemobj_from_id(idnum)
+    id = TkcItem.id2obj(self, idnum.to_i)
+    return id if id.kind_of?(TkcItem)
+
+    typename = tk_send('type', id)
+    unless type = TkcItem.type2class(typename)
+      (itemclass = typename.dup)[0,1] = typename[0,1].upcase
+      type = TkcItem.const_set(itemclass, Class.new(TkcItem))
+      type.const_set("CItemTypeName", typename.freeze)
+      TkcItem::CItemTypeToClass[typename] = type
+    end
+
+    canvas = self
+    (obj = type.allocate).instance_eval{
+      @parent = @c = canvas
+      @path = canvas.path
+      @id = id
+      TkcItem::CItemID_TBL.mutex.synchronize{
+        TkcItem::CItemID_TBL[@path] = {} unless TkcItem::CItemID_TBL[@path]
+        TkcItem::CItemID_TBL[@path][@id] = self
+      }
+    }
+  end
 end
 
 #TkCanvas = Tk::Canvas unless Object.const_defined? :TkCanvas
-Tk.__set_toplevel_aliases__(:Tk, Tk::Canvas, :TkCanvas)
+#Tk.__set_toplevel_aliases__(:Tk, Tk::Canvas, :TkCanvas)
+Tk.__set_loaded_toplevel_aliases__('tk/canvas.rb', :Tk, Tk::Canvas, :TkCanvas)
 
 
 class TkcItem<TkObject
@@ -584,9 +648,12 @@ class TkcItem<TkObject
 
   CItemTypeName = nil
   CItemTypeToClass = {}
+
   CItemID_TBL = TkCore::INTERP.create_table
 
-  TkCore::INTERP.init_ip_env{ CItemID_TBL.clear }
+  TkCore::INTERP.init_ip_env{
+    CItemID_TBL.mutex.synchronize{ CItemID_TBL.clear }
+  }
 
   def TkcItem.type2class(type)
     CItemTypeToClass[type]
@@ -594,8 +661,13 @@ class TkcItem<TkObject
 
   def TkcItem.id2obj(canvas, id)
     cpath = canvas.path
-    return id unless CItemID_TBL[cpath]
-    CItemID_TBL[cpath][id]? CItemID_TBL[cpath][id]: id
+    CItemID_TBL.mutex.synchronize{
+      if CItemID_TBL[cpath]
+        CItemID_TBL[cpath][id]? CItemID_TBL[cpath][id]: id
+      else
+        id
+      end
+    }
   end
 
   ########################################
@@ -628,6 +700,13 @@ class TkcItem<TkObject
         fontkeys[fkey] = keys.delete(fkey) if keys.key?(fkey)
       }
 
+      __item_optkey_aliases(nil).each{|alias_name, real_name|
+        alias_name = alias_name.to_s
+        if keys.has_key?(alias_name)
+          keys[real_name.to_s] = keys.delete(alias_name)
+        end
+      }
+
       __item_methodcall_optkeys(nil).each{|key|
         key = key.to_s
         methodkeys[key] = keys.delete(key) if keys.key?(key)
@@ -644,7 +723,7 @@ class TkcItem<TkObject
       args = args.flatten
     end
 
-    [args, fontkeys]
+    [args, fontkeys, methodkeys]
   end
   private_class_method :_parse_create_args
 
@@ -652,10 +731,11 @@ class TkcItem<TkObject
     unless self::CItemTypeName
       fail RuntimeError, "#{self} is an abstract class"
     end
-    args, fontkeys = _parse_create_args(args)
-    idnum = tk_call_without_enc(canvas.path, 'create', 
+    args, fontkeys, methodkeys = _parse_create_args(args)
+    idnum = tk_call_without_enc(canvas.path, 'create',
                                 self::CItemTypeName, *args)
     canvas.itemconfigure(idnum, fontkeys) unless fontkeys.empty?
+    canvas.itemconfigure(idnum, methodkeys) unless methodkeys.empty?
     idnum.to_i  # 'canvas item id' is an integer number
   end
   ########################################
@@ -668,8 +748,10 @@ class TkcItem<TkObject
     @path = parent.path
 
     @id = create_self(*args) # an integer number as 'canvas item id'
-    CItemID_TBL[@path] = {} unless CItemID_TBL[@path]
-    CItemID_TBL[@path][@id] = self
+    CItemID_TBL.mutex.synchronize{
+      CItemID_TBL[@path] = {} unless CItemID_TBL[@path]
+      CItemID_TBL[@path][@id] = self
+    }
   end
   def create_self(*args)
     self.class.create(@c, *args) # return an integer number as 'canvas item id'
@@ -690,7 +772,9 @@ class TkcItem<TkObject
 
   def delete
     @c.delete @id
-    CItemID_TBL[@path].delete(@id) if CItemID_TBL[@path]
+    CItemID_TBL.mutex.synchronize{
+      CItemID_TBL[@path].delete(@id) if CItemID_TBL[@path]
+    }
     self
   end
   alias remove  delete
