@@ -115,6 +115,10 @@ module FileUtils
   #
   #   FileUtils.cd('/', :verbose => true)   # chdir and report it
   #
+  #   FileUtils.cd('/') do  # chdir
+  #     [...]               # do something
+  #   end                   # return to original directory
+  #
   def cd(dir, options = {}, &block) # :yield: dir
     fu_check_options options, OPT_TABLE['cd']
     fu_output_message "cd #{dir}" if options[:verbose]
@@ -138,9 +142,7 @@ module FileUtils
   #   FileUtils.uptodate?('hello.o', %w(hello.c hello.h)) or \
   #       system 'make hello.o'
   #
-  def uptodate?(new, old_list, options = nil)
-    raise ArgumentError, 'uptodate? does not accept any option' if options
-
+  def uptodate?(new, old_list)
     return false unless File.exist?(new)
     new_time = File.mtime(new)
     old_list.each do |old|
@@ -198,7 +200,7 @@ module FileUtils
     fu_output_message "mkdir -p #{options[:mode] ? ('-m %03o ' % options[:mode]) : ''}#{list.join ' '}" if options[:verbose]
     return *list if options[:noop]
 
-    list.map {|path| path.sub(%r</\z>, '') }.each do |path|
+    list.map {|path| path.chomp(?/) }.each do |path|
       # optimize for the most common case
       begin
         fu_mkdir path, options[:mode]
@@ -235,7 +237,7 @@ module FileUtils
   OPT_TABLE['makedirs'] = [:mode, :noop, :verbose]
 
   def fu_mkdir(path, mode)   #:nodoc:
-    path = path.sub(%r</\z>, '')
+    path = path.chomp(?/)
     if mode
       Dir.mkdir path, mode
       File.chmod mode, path
@@ -263,7 +265,7 @@ module FileUtils
     return if options[:noop]
     list.each do |dir|
       begin
-        Dir.rmdir(dir = dir.sub(%r</\z>, ''))
+        Dir.rmdir(dir = dir.chomp(?/))
         if parents
           until (parent = File.dirname(dir)) == '.' or parent == dir
             Dir.rmdir(dir)
@@ -359,7 +361,7 @@ module FileUtils
   # Options: noop verbose
   #
   # Same as
-  #   #ln_s(src, dest, :force)
+  #   #ln_s(src, dest, :force => true)
   #
   def ln_sf(src, dest, options = {})
     fu_check_options options, OPT_TABLE['ln_sf']
@@ -420,7 +422,7 @@ module FileUtils
   #   # If you want to copy all contents of a directory instead of the
   #   # directory itself, c.f. src/x -> dest/x, src/y -> dest/y,
   #   # use following code.
-  #   FileUtils.cp_r 'src/.', 'dest'     # cp_r('src', 'dest') makes src/dest,
+  #   FileUtils.cp_r 'src/.', 'dest'     # cp_r('src', 'dest') makes dest/src,
   #                                      # but this doesn't.
   #
   def cp_r(src, dest, options = {})
@@ -455,12 +457,14 @@ module FileUtils
   # If +remove_destination+ is true, this method removes each destination file before copy.
   #
   def copy_entry(src, dest, preserve = false, dereference_root = false, remove_destination = false)
-    Entry_.new(src, nil, dereference_root).traverse do |ent|
+    Entry_.new(src, nil, dereference_root).wrap_traverse(proc do |ent|
       destent = Entry_.new(dest, ent.rel, false)
       File.unlink destent.path if remove_destination && File.file?(destent.path)
       ent.copy destent.path
+    end, proc do |ent|
+      destent = Entry_.new(dest, ent.rel, false)
       ent.copy_metadata destent.path if preserve
-    end
+    end)
   end
   module_function :copy_entry
 
@@ -739,11 +743,11 @@ module FileUtils
   end
   module_function :remove_entry_secure
 
-  def fu_have_symlink?   #:nodoc
+  def fu_have_symlink?   #:nodoc:
     File.symlink nil, nil
   rescue NotImplementedError
     return false
-  rescue
+  rescue TypeError
     return true
   end
   private_module_function :fu_have_symlink?
@@ -819,16 +823,13 @@ module FileUtils
   #
   def compare_stream(a, b)
     bsize = fu_stream_blksize(a, b)
-    sa = sb = nil
-    while sa == sb
-      sa = a.read(bsize)
-      sb = b.read(bsize)
-      unless sa and sb
-        if sa.nil? and sb.nil?
-          return true
-        end
-      end
-    end
+    sa = ""
+    sb = ""
+    begin
+      a.read(bsize, sa)
+      b.read(bsize, sb)
+      return true if sa.empty? && sb.empty?
+    end while sa == sb
     false
   end
   module_function :compare_stream
@@ -860,23 +861,117 @@ module FileUtils
 
   OPT_TABLE['install'] = [:mode, :preserve, :noop, :verbose]
 
+  def user_mask(target)  #:nodoc:
+    mask = 0
+    target.each_byte do |byte_chr|
+      case byte_chr.chr
+        when "u"
+          mask |= 04700
+        when "g"
+          mask |= 02070
+        when "o"
+          mask |= 01007
+        when "a"
+          mask |= 07777
+      end
+    end
+    mask
+  end
+  private_module_function :user_mask
+
+  def mode_mask(mode, path)  #:nodoc:
+    mask = 0
+    mode.each_byte do |byte_chr|
+      case byte_chr.chr
+        when "r"
+          mask |= 0444
+        when "w"
+          mask |= 0222
+        when "x"
+          mask |= 0111
+        when "X"
+          mask |= 0111 if FileTest::directory? path
+        when "s"
+          mask |= 06000
+        when "t"
+          mask |= 01000
+      end
+    end
+    mask
+  end
+  private_module_function :mode_mask
+
+  def symbolic_modes_to_i(modes, path)  #:nodoc:
+    current_mode = (File.stat(path).mode & 07777)
+    modes.split(/,/).inject(0) do |mode, mode_sym|
+      mode_sym = "a#{mode_sym}" if mode_sym =~ %r!^[=+-]!
+      target, mode = mode_sym.split %r![=+-]!
+      user_mask = user_mask(target)
+      mode_mask = mode_mask(mode ? mode : "", path)
+
+      case mode_sym
+        when /=/
+          current_mode &= ~(user_mask)
+          current_mode |= user_mask & mode_mask
+        when /\+/
+          current_mode |= user_mask & mode_mask
+        when /-/
+          current_mode &= ~(user_mask & mode_mask)
+      end
+    end
+  end
+  private_module_function :symbolic_modes_to_i
+
+  def fu_mode(mode, path)  #:nodoc:
+    mode.is_a?(String) ? symbolic_modes_to_i(mode, path) : mode
+  end
+  private_module_function :fu_mode
+
+  def mode_to_s(mode)  #:nodoc:
+    mode.is_a?(String) ? mode : "%o" % mode
+  end
+  private_module_function :mode_to_s
+
   #
   # Options: noop verbose
   #
   # Changes permission bits on the named files (in +list+) to the bit pattern
   # represented by +mode+.
   #
+  # +mode+ is the symbolic and absolute mode can be used.
+  #
+  # Absolute mode is
   #   FileUtils.chmod 0755, 'somecommand'
   #   FileUtils.chmod 0644, %w(my.rb your.rb his.rb her.rb)
   #   FileUtils.chmod 0755, '/usr/bin/ruby', :verbose => true
   #
+  # Symbolic mode is
+  #   FileUtils.chmod "u=wrx,go=rx", 'somecommand'
+  #   FileUtils.chmod "u=wr,go=rr", %w(my.rb your.rb his.rb her.rb)
+  #   FileUtils.chmod "u=wrx,go=rx", '/usr/bin/ruby', :verbose => true
+  #
+  # "a" :: is user, group, other mask.
+  # "u" :: is user's mask.
+  # "g" :: is group's mask.
+  # "o" :: is other's mask.
+  # "w" :: is write permission.
+  # "r" :: is read permission.
+  # "x" :: is execute permission.
+  # "X" ::
+  #   is execute permission for directories only, must be used in conjunction with "+"
+  # "s" :: is uid, gid.
+  # "t" :: is sticky bit.
+  # "+" :: is added to a class given the specified mode.
+  # "-" :: Is removed from a given class given mode.
+  # "=" :: Is the exact nature of the class will be given a specified mode.
+
   def chmod(mode, list, options = {})
     fu_check_options options, OPT_TABLE['chmod']
     list = fu_list(list)
-    fu_output_message sprintf('chmod %o %s', mode, list.join(' ')) if options[:verbose]
+    fu_output_message sprintf('chmod %s %s', mode_to_s(mode), list.join(' ')) if options[:verbose]
     return if options[:noop]
     list.each do |path|
-      Entry_.new(path).chmod mode
+      Entry_.new(path).chmod(fu_mode(mode, path))
     end
   end
   module_function :chmod
@@ -890,18 +985,19 @@ module FileUtils
   # to the bit pattern represented by +mode+.
   #
   #   FileUtils.chmod_R 0700, "/tmp/app.#{$$}"
+  #   FileUtils.chmod_R "u=wrx", "/tmp/app.#{$$}"
   #
   def chmod_R(mode, list, options = {})
     fu_check_options options, OPT_TABLE['chmod_R']
     list = fu_list(list)
-    fu_output_message sprintf('chmod -R%s %o %s',
+    fu_output_message sprintf('chmod -R%s %s %s',
                               (options[:force] ? 'f' : ''),
-                              mode, list.join(' ')) if options[:verbose]
+                              mode_to_s(mode), list.join(' ')) if options[:verbose]
     return if options[:noop]
     list.each do |root|
       Entry_.new(root).traverse do |ent|
         begin
-          ent.chmod mode
+          ent.chmod(fu_mode(mode, ent.path))
         rescue
           raise unless options[:force]
         end
@@ -927,8 +1023,8 @@ module FileUtils
   def chown(user, group, list, options = {})
     fu_check_options options, OPT_TABLE['chown']
     list = fu_list(list)
-    fu_output_message sprintf('chown %s%s',
-                              [user,group].compact.join(':') + ' ',
+    fu_output_message sprintf('chown %s %s',
+                              (group ? "#{user}:#{group}" : user || ':'),
                               list.join(' ')) if options[:verbose]
     return if options[:noop]
     uid = fu_get_uid(user)
@@ -956,9 +1052,9 @@ module FileUtils
   def chown_R(user, group, list, options = {})
     fu_check_options options, OPT_TABLE['chown_R']
     list = fu_list(list)
-    fu_output_message sprintf('chown -R%s %s%s',
+    fu_output_message sprintf('chown -R%s %s %s',
                               (options[:force] ? 'f' : ''),
-                              [user,group].compact.join(':') + ' ',
+                              (group ? "#{user}:#{group}" : user || ':'),
                               list.join(' ')) if options[:verbose]
     return if options[:noop]
     uid = fu_get_uid(user)
@@ -1036,7 +1132,7 @@ module FileUtils
     created = nocreate = options[:nocreate]
     t = options[:mtime]
     if options[:verbose]
-      fu_output_message "touch #{nocreate ? ' -c' : ''}#{t ? t.strftime(' -t %Y%m%d%H%M.%S') : ''}#{list.join ' '}"
+      fu_output_message "touch #{nocreate ? '-c ' : ''}#{t ? t.strftime('-t %Y%m%d%H%M.%S ') : ''}#{list.join ' '}"
     end
     return if options[:noop]
     list.each do |path|
@@ -1182,7 +1278,7 @@ module FileUtils
 
     def entries
       opts = {}
-      opts[:encoding] = "UTF-8" if /mswin|mignw/ =~ RUBY_PLATFORM
+      opts[:encoding] = ::Encoding::UTF_8 if fu_windows?
       Dir.entries(path(), opts)\
           .reject {|n| n == '.' or n == '..' }\
           .map {|n| Entry_.new(prefix(), join(rel(), n.untaint)) }
@@ -1276,7 +1372,7 @@ module FileUtils
 
     def copy_file(dest)
       File.open(path()) do |s|
-        File.open(dest, 'wb') do |f|
+        File.open(dest, 'wb', s.stat.mode) do |f|
           IO.copy_stream(s, f)
         end
       end
@@ -1284,14 +1380,37 @@ module FileUtils
 
     def copy_metadata(path)
       st = lstat()
-      File.utime st.atime, st.mtime, path
+      if !st.symlink?
+        File.utime st.atime, st.mtime, path
+      end
       begin
-        File.chown st.uid, st.gid, path
+        if st.symlink?
+          begin
+            File.lchown st.uid, st.gid, path
+          rescue NotImplementedError
+          end
+        else
+          File.chown st.uid, st.gid, path
+        end
       rescue Errno::EPERM
         # clear setuid/setgid
-        File.chmod st.mode & 01777, path
+        if st.symlink?
+          begin
+            File.lchmod st.mode & 01777, path
+          rescue NotImplementedError
+          end
+        else
+          File.chmod st.mode & 01777, path
+        end
       else
-        File.chmod st.mode, path
+        if st.symlink?
+          begin
+            File.lchmod st.mode, path
+          rescue NotImplementedError
+          end
+        else
+          File.chmod st.mode, path
+        end
       end
     end
 
@@ -1305,7 +1424,7 @@ module FileUtils
 
     def remove_dir1
       platform_support {
-        Dir.rmdir path().sub(%r</\z>, '')
+        Dir.rmdir path().chomp(?/)
       }
     end
 
@@ -1356,6 +1475,16 @@ module FileUtils
       yield self
     end
 
+    def wrap_traverse(pre, post)
+      pre.call self
+      if directory?
+        entries.each do |ent|
+          ent.wrap_traverse pre, post
+        end
+      end
+      post.call self
+    end
+
     private
 
     $fileutils_rb_have_lchmod = nil
@@ -1524,6 +1653,12 @@ module FileUtils
     OPT_TABLE.keys.select {|m| OPT_TABLE[m].include?(opt) }
   end
 
+  LOW_METHODS = singleton_methods(false) - collect_method(:noop).map(&:intern)
+  module LowMethods
+    module_eval("private\n" + ::FileUtils::LOW_METHODS.map {|name| "def #{name}(*)end"}.join("\n"),
+                __FILE__, __LINE__)
+  end
+
   METHODS = singleton_methods() - [:private_module_function,
       :commands, :options, :have_option?, :options_of, :collect_method]
 
@@ -1559,6 +1694,7 @@ module FileUtils
   #
   module NoWrite
     include FileUtils
+    include LowMethods
     @fileutils_output  = $stderr
     @fileutils_label   = ''
     ::FileUtils.collect_method(:noop).each do |name|
@@ -1585,6 +1721,7 @@ module FileUtils
   #
   module DryRun
     include FileUtils
+    include LowMethods
     @fileutils_output  = $stderr
     @fileutils_label   = ''
     ::FileUtils.collect_method(:noop).each do |name|
