@@ -13,6 +13,7 @@
 #include "ruby/encoding.h"
 #include "ruby/util.h"
 #include "internal.h"
+#include "id.h"
 #include <ctype.h>
 #include <math.h>
 #include <stdio.h>
@@ -29,6 +30,14 @@
 #include <ieeefp.h>
 #endif
 
+#if !defined HAVE_ISFINITE && !defined isfinite
+#if defined HAVE_FINITE && !defined finite && !defined _WIN32
+extern int finite(double);
+# define HAVE_ISFINITE 1
+# define isfinite(x) finite(x)
+#endif
+#endif
+
 /* use IEEE 64bit values if not defined */
 #ifndef FLT_RADIX
 #define FLT_RADIX 2
@@ -66,16 +75,16 @@
 
 #ifdef HAVE_INFINITY
 #elif !defined(WORDS_BIGENDIAN) /* BYTE_ORDER == LITTLE_ENDIAN */
-const unsigned char rb_infinity[] = "\x00\x00\x80\x7f";
+const union bytesequence4_or_float rb_infinity = {{0x00, 0x00, 0x80, 0x7f}};
 #else
-const unsigned char rb_infinity[] = "\x7f\x80\x00\x00";
+const union bytesequence4_or_float rb_infinity = {{0x7f, 0x80, 0x00, 0x00}};
 #endif
 
 #ifdef HAVE_NAN
 #elif !defined(WORDS_BIGENDIAN) /* BYTE_ORDER == LITTLE_ENDIAN */
-const unsigned char rb_nan[] = "\x00\x00\xc0\x7f";
+const union bytesequence4_or_float rb_nan = {{0x00, 0x00, 0xc0, 0x7f}};
 #else
-const unsigned char rb_nan[] = "\x7f\xc0\x00\x00";
+const union bytesequence4_or_float rb_nan = {{0x7f, 0xc0, 0x00, 0x00}};
 #endif
 
 #ifndef HAVE_ROUND
@@ -100,7 +109,7 @@ static VALUE fix_uminus(VALUE num);
 static VALUE fix_mul(VALUE x, VALUE y);
 static VALUE int_pow(long x, unsigned long y);
 
-static ID id_coerce, id_to_i, id_eq;
+static ID id_coerce, id_to_i, id_eq, id_div;
 
 VALUE rb_cNumeric;
 VALUE rb_cFloat;
@@ -150,6 +159,46 @@ rb_num_to_uint(VALUE val, unsigned int *ret)
     return NUMERR_TYPE;
 }
 
+#define method_basic_p(klass) rb_method_basic_definition_p(klass, mid)
+
+static inline int
+positive_int_p(VALUE num)
+{
+    const ID mid = '>';
+
+    if (FIXNUM_P(num)) {
+	if (method_basic_p(rb_cFixnum))
+	    return (SIGNED_VALUE)num > 0;
+    }
+    else if (RB_TYPE_P(num, T_BIGNUM)) {
+	if (method_basic_p(rb_cBignum))
+	    return RBIGNUM_POSITIVE_P(num);
+    }
+    return RTEST(rb_funcall(num, mid, 1, INT2FIX(0)));
+}
+
+static inline int
+negative_int_p(VALUE num)
+{
+    const ID mid = '<';
+
+    if (FIXNUM_P(num)) {
+	if (method_basic_p(rb_cFixnum))
+	    return (SIGNED_VALUE)num < 0;
+    }
+    else if (RB_TYPE_P(num, T_BIGNUM)) {
+	if (method_basic_p(rb_cBignum))
+	    return RBIGNUM_NEGATIVE_P(num);
+    }
+    return RTEST(rb_funcall(num, mid, 1, INT2FIX(0)));
+}
+
+int
+rb_num_negative_p(VALUE num)
+{
+    return negative_int_p(num);
+}
+
 /*
  *  call-seq:
  *     num.coerce(numeric)  ->  array
@@ -182,16 +231,19 @@ coerce_body(VALUE *x)
     return rb_funcall(x[1], id_coerce, 1, x[0]);
 }
 
+NORETURN(static void coerce_failed(VALUE x, VALUE y));
+static void
+coerce_failed(VALUE x, VALUE y)
+{
+    rb_raise(rb_eTypeError, "%"PRIsVALUE" can't be coerced into %"PRIsVALUE,
+	     (rb_special_const_p(y)? rb_inspect(y) : rb_obj_class(y)),
+	     rb_obj_class(x));
+}
+
 static VALUE
 coerce_rescue(VALUE *x)
 {
-    volatile VALUE v = rb_inspect(x[1]);
-
-    rb_raise(rb_eTypeError, "%s can't be coerced into %s",
-	     rb_special_const_p(x[1])?
-	     RSTRING_PTR(v):
-	     rb_obj_classname(x[1]),
-	     rb_obj_classname(x[0]));
+    coerce_failed(x[0], x[1]);
     return Qnil;		/* dummy */
 }
 
@@ -203,8 +255,15 @@ do_coerce(VALUE *x, VALUE *y, int err)
 
     a[0] = *x; a[1] = *y;
 
-    ary = rb_rescue(coerce_body, (VALUE)a, err?coerce_rescue:0, (VALUE)a);
-    if (TYPE(ary) != T_ARRAY || RARRAY_LEN(ary) != 2) {
+    if (!rb_respond_to(*y, id_coerce)) {
+	if (err) {
+	    coerce_rescue(a);
+	}
+	return FALSE;
+    }
+
+    ary = rb_rescue(coerce_body, (VALUE)a, err ? coerce_rescue : 0, (VALUE)a);
+    if (!RB_TYPE_P(ary, T_ARRAY) || RARRAY_LEN(ary) != 2) {
 	if (err) {
 	    rb_raise(rb_eTypeError, "coerce must return [x, y]");
 	}
@@ -257,10 +316,11 @@ num_sadded(VALUE x, VALUE name)
     /* Numerics should be values; singleton_methods should not be added to them */
     rb_remove_method_id(rb_singleton_class(x), mid);
     rb_raise(rb_eTypeError,
-	     "can't define singleton method \"%s\" for %s",
-	     rb_id2name(mid),
-	     rb_obj_classname(x));
-    return Qnil;		/* not reached */
+	     "can't define singleton method \"%"PRIsVALUE"\" for %"PRIsVALUE,
+	     rb_id2str(mid),
+	     rb_obj_class(x));
+
+    UNREACHABLE;
 }
 
 /* :nodoc: */
@@ -268,8 +328,9 @@ static VALUE
 num_init_copy(VALUE x, VALUE y)
 {
     /* Numerics are immutable values, which should not be copied */
-    rb_raise(rb_eTypeError, "can't copy %s", rb_obj_classname(x));
-    return Qnil;		/* not reached */
+    rb_raise(rb_eTypeError, "can't copy %"PRIsVALUE, rb_obj_class(x));
+
+    UNREACHABLE;
 }
 
 /*
@@ -403,10 +464,10 @@ num_remainder(VALUE x, VALUE y)
     VALUE z = rb_funcall(x, '%', 1, y);
 
     if ((!rb_equal(z, INT2FIX(0))) &&
-	((RTEST(rb_funcall(x, '<', 1, INT2FIX(0))) &&
-	  RTEST(rb_funcall(y, '>', 1, INT2FIX(0)))) ||
-	 (RTEST(rb_funcall(x, '>', 1, INT2FIX(0))) &&
-	  RTEST(rb_funcall(y, '<', 1, INT2FIX(0)))))) {
+	((negative_int_p(x) &&
+	  positive_int_p(y)) ||
+	 (positive_int_p(x) &&
+	  negative_int_p(y)))) {
 	return rb_funcall(z, '-', 1, y);
     }
     return z;
@@ -477,8 +538,10 @@ num_real_p(VALUE num)
  *  call-seq:
  *     num.integer?  ->  true or false
  *
- *  Returns <code>true</code> if <i>num</i> is an <code>Integer</code>
- *  (including <code>Fixnum</code> and <code>Bignum</code>).
+ *  Returns +true+ if +num+ is an Integer (including Fixnum and Bignum).
+ *
+ *      (1.0).integer? #=> false
+ *      (1).integer?   #=> true
  */
 
 static VALUE
@@ -502,7 +565,7 @@ num_int_p(VALUE num)
 static VALUE
 num_abs(VALUE num)
 {
-    if (RTEST(rb_funcall(num, '<', 1, INT2FIX(0)))) {
+    if (negative_int_p(num)) {
 	return rb_funcall(num, rb_intern("-@"), 0);
     }
     return num;
@@ -551,8 +614,11 @@ num_nonzero_p(VALUE num)
  *  call-seq:
  *     num.to_int  ->  integer
  *
- *  Invokes the child class's <code>to_i</code> method to convert
- *  <i>num</i> to an integer.
+ *  Invokes the child class's +to_i+ method to convert +num+ to an integer.
+ *
+ *      1.0.class => Float
+ *      1.0.to_int.class => Fixnum
+ *      1.0.to_i.class => Fixnum
  */
 
 static VALUE
@@ -574,17 +640,17 @@ num_to_int(VALUE num)
  *  So you should know its esoteric system. see following:
  *
  *  - http://docs.sun.com/source/806-3568/ncg_goldberg.html
- *  - http://wiki.github.com/rdp/ruby_tutorials_core/ruby-talk-faq#floats_imprecise
+ *  - http://wiki.github.com/rdp/ruby_tutorials_core/ruby-talk-faq#wiki-floats_imprecise
  *  - http://en.wikipedia.org/wiki/Floating_point#Accuracy_problems
  */
 
 VALUE
-rb_float_new(double d)
+rb_float_new_in_heap(double d)
 {
-    NEWOBJ(flt, struct RFloat);
-    OBJSETUP(flt, rb_cFloat, T_FLOAT);
+    NEWOBJ_OF(flt, struct RFloat, rb_cFloat, T_FLOAT);
 
     flt->float_value = d;
+    OBJ_FREEZE(flt);
     return (VALUE)flt;
 }
 
@@ -626,7 +692,7 @@ flo_to_s(VALUE flt)
 	    buf[decpt] = '.';
 	    rb_str_cat(s, buf, digs + 1);
 	}
-	else if (decpt - digs < float_dig) {
+	else if (decpt <= DBL_DIG) {
 	    long len;
 	    char *ptr;
 	    rb_str_cat(s, buf, digs);
@@ -814,7 +880,7 @@ flodivmod(double x, double y, double *divp, double *modp)
     double div, mod;
 
     if (y == 0.0) rb_num_zerodiv();
-    if((x == 0.0) || (isinf(y) && !isinf(x)))
+    if ((x == 0.0) || (isinf(y) && !isinf(x)))
         mod = x;
     else {
 #ifdef HAVE_FMOD
@@ -843,7 +909,9 @@ flodivmod(double x, double y, double *divp, double *modp)
  * An error will be raised if y == 0.
  */
 
-double ruby_float_mod(double x, double y) {
+double
+ruby_float_mod(double x, double y)
+{
     double mod;
     flodivmod(x, y, 0, &mod);
     return mod;
@@ -852,10 +920,10 @@ double ruby_float_mod(double x, double y) {
 
 /*
  *  call-seq:
- *     flt % other        ->  float
- *     flt.modulo(other)  ->  float
+ *     float % other        ->  float
+ *     float.modulo(other)  ->  float
  *
- *  Return the modulo after division of <code>flt</code> by <code>other</code>.
+ *  Return the modulo after division of +float+ by +other+.
  *
  *     6543.21.modulo(137)      #=> 104.21
  *     6543.21.modulo(137.24)   #=> 92.9299999999996
@@ -894,9 +962,12 @@ dbl2ival(double d)
 
 /*
  *  call-seq:
- *     flt.divmod(numeric)  ->  array
+ *     float.divmod(numeric)  ->  array
  *
- *  See <code>Numeric#divmod</code>.
+ *  See Numeric#divmod.
+ *
+ *      42.0.divmod 6 #=> [7, 0.0]
+ *      42.0.divmod 5 #=> [8, 2.0]
  */
 
 static VALUE
@@ -977,10 +1048,10 @@ num_eql(VALUE x, VALUE y)
 
 /*
  *  call-seq:
- *     num <=> other  ->  0 or nil
+ *     number <=> other  ->  0 or nil
  *
- *  Returns zero if <i>num</i> equals <i>other</i>, <code>nil</code>
- *  otherwise.
+ *  Returns zero if +number+ equals +other+, otherwise +nil+ is returned if the
+ *  two values are incomparable.
  */
 
 static VALUE
@@ -1004,6 +1075,8 @@ num_equal(VALUE x, VALUE y)
  *  Returns <code>true</code> only if <i>obj</i> has the same value
  *  as <i>flt</i>. Contrast this with <code>Float#eql?</code>, which
  *  requires <i>obj</i> to be a <code>Float</code>.
+ *  The result of <code>NaN == NaN</code> is undefined, so the
+ *  implementation-dependent value is returned.
  *
  *     1.0 == 1   #=> true
  *
@@ -1016,11 +1089,8 @@ flo_eq(VALUE x, VALUE y)
 
     switch (TYPE(y)) {
       case T_FIXNUM:
-	b = (double)FIX2LONG(y);
-	break;
       case T_BIGNUM:
-	b = rb_big2dbl(y);
-	break;
+        return rb_integer_float_eq(y, x);
       case T_FLOAT:
 	b = RFLOAT_VALUE(y);
 #if defined(_MSC_VER) && _MSC_VER < 1300
@@ -1069,11 +1139,15 @@ rb_dbl_cmp(double a, double b)
 
 /*
  *  call-seq:
- *     flt <=> real  ->  -1, 0, +1 or nil
+ *     float <=> real  ->  -1, 0, +1 or nil
  *
- *  Returns -1, 0, +1 or nil depending on whether <i>flt</i> is less
- *  than, equal to, or greater than <i>real</i>. This is the basis for
- *  the tests in <code>Comparable</code>.
+ *  Returns -1, 0, +1 or nil depending on whether +float+ is less than, equal
+ *  to, or greater than +real+. This is the basis for the tests in Comparable.
+ *
+ *  The result of <code>NaN <=> NaN</code> is undefined, so the
+ *  implementation-dependent value is returned.
+ *
+ *  +nil+ is returned if the two values are incomparable.
  */
 
 static VALUE
@@ -1086,16 +1160,13 @@ flo_cmp(VALUE x, VALUE y)
     if (isnan(a)) return Qnil;
     switch (TYPE(y)) {
       case T_FIXNUM:
-	b = (double)FIX2LONG(y);
-	break;
-
       case T_BIGNUM:
-	if (isinf(a)) {
-	    if (a > 0.0) return INT2FIX(1);
-	    else return INT2FIX(-1);
-	}
-	b = rb_big2dbl(y);
-	break;
+      {
+        VALUE rel = rb_integer_float_cmp(y, x);
+        if (FIXNUM_P(rel))
+            return INT2FIX(-FIX2INT(rel));
+        return rel;
+      }
 
       case T_FLOAT:
 	b = RFLOAT_VALUE(y);
@@ -1121,6 +1192,8 @@ flo_cmp(VALUE x, VALUE y)
  *   flt > real  ->  true or false
  *
  * <code>true</code> if <code>flt</code> is greater than <code>real</code>.
+ * The result of <code>NaN > NaN</code> is undefined, so the
+ * implementation-dependent value is returned.
  */
 
 static VALUE
@@ -1131,12 +1204,13 @@ flo_gt(VALUE x, VALUE y)
     a = RFLOAT_VALUE(x);
     switch (TYPE(y)) {
       case T_FIXNUM:
-	b = (double)FIX2LONG(y);
-	break;
-
       case T_BIGNUM:
-	b = rb_big2dbl(y);
-	break;
+      {
+        VALUE rel = rb_integer_float_cmp(y, x);
+        if (FIXNUM_P(rel))
+            return -FIX2INT(rel) > 0 ? Qtrue : Qfalse;
+        return Qfalse;
+      }
 
       case T_FLOAT:
 	b = RFLOAT_VALUE(y);
@@ -1160,6 +1234,8 @@ flo_gt(VALUE x, VALUE y)
  *
  * <code>true</code> if <code>flt</code> is greater than
  * or equal to <code>real</code>.
+ * The result of <code>NaN >= NaN</code> is undefined, so the
+ * implementation-dependent value is returned.
  */
 
 static VALUE
@@ -1170,12 +1246,13 @@ flo_ge(VALUE x, VALUE y)
     a = RFLOAT_VALUE(x);
     switch (TYPE(y)) {
       case T_FIXNUM:
-	b = (double)FIX2LONG(y);
-	break;
-
       case T_BIGNUM:
-	b = rb_big2dbl(y);
-	break;
+      {
+        VALUE rel = rb_integer_float_cmp(y, x);
+        if (FIXNUM_P(rel))
+            return -FIX2INT(rel) >= 0 ? Qtrue : Qfalse;
+        return Qfalse;
+      }
 
       case T_FLOAT:
 	b = RFLOAT_VALUE(y);
@@ -1198,6 +1275,8 @@ flo_ge(VALUE x, VALUE y)
  *   flt < real  ->  true or false
  *
  * <code>true</code> if <code>flt</code> is less than <code>real</code>.
+ * The result of <code>NaN < NaN</code> is undefined, so the
+ * implementation-dependent value is returned.
  */
 
 static VALUE
@@ -1208,12 +1287,13 @@ flo_lt(VALUE x, VALUE y)
     a = RFLOAT_VALUE(x);
     switch (TYPE(y)) {
       case T_FIXNUM:
-	b = (double)FIX2LONG(y);
-	break;
-
       case T_BIGNUM:
-	b = rb_big2dbl(y);
-	break;
+      {
+        VALUE rel = rb_integer_float_cmp(y, x);
+        if (FIXNUM_P(rel))
+            return -FIX2INT(rel) < 0 ? Qtrue : Qfalse;
+        return Qfalse;
+      }
 
       case T_FLOAT:
 	b = RFLOAT_VALUE(y);
@@ -1237,6 +1317,8 @@ flo_lt(VALUE x, VALUE y)
  *
  * <code>true</code> if <code>flt</code> is less than
  * or equal to <code>real</code>.
+ * The result of <code>NaN <= NaN</code> is undefined, so the
+ * implementation-dependent value is returned.
  */
 
 static VALUE
@@ -1247,12 +1329,13 @@ flo_le(VALUE x, VALUE y)
     a = RFLOAT_VALUE(x);
     switch (TYPE(y)) {
       case T_FIXNUM:
-	b = (double)FIX2LONG(y);
-	break;
-
       case T_BIGNUM:
-	b = rb_big2dbl(y);
-	break;
+      {
+        VALUE rel = rb_integer_float_cmp(y, x);
+        if (FIXNUM_P(rel))
+            return -FIX2INT(rel) <= 0 ? Qtrue : Qfalse;
+        return Qfalse;
+      }
 
       case T_FLOAT:
 	b = RFLOAT_VALUE(y);
@@ -1277,6 +1360,8 @@ flo_le(VALUE x, VALUE y)
  *  Returns <code>true</code> only if <i>obj</i> is a
  *  <code>Float</code> with the same value as <i>flt</i>. Contrast this
  *  with <code>Float#==</code>, which performs type conversions.
+ *  The result of <code>NaN.eql?(NaN)</code> is undefined, so the
+ *  implementation-dependent value is returned.
  *
  *     1.0.eql?(1)   #=> false
  */
@@ -1284,7 +1369,7 @@ flo_le(VALUE x, VALUE y)
 static VALUE
 flo_eql(VALUE x, VALUE y)
 {
-    if (TYPE(y) == T_FLOAT) {
+    if (RB_TYPE_P(y, T_FLOAT)) {
 	double a = RFLOAT_VALUE(x);
 	double b = RFLOAT_VALUE(y);
 #if defined(_MSC_VER) && _MSC_VER < 1300
@@ -1405,8 +1490,8 @@ flo_is_finite_p(VALUE num)
 {
     double value = RFLOAT_VALUE(num);
 
-#if HAVE_FINITE
-    if (!finite(value))
+#if HAVE_ISFINITE
+    if (!isfinite(value))
 	return Qfalse;
 #else
     if (isinf(value) || isnan(value))
@@ -1478,7 +1563,7 @@ int_round_0(VALUE num, int ndigits)
     ID op;
     /* If 10**N / 2 > num, then return 0 */
     /* We have log_256(10) > 0.415241 and log_256(1/2) = -0.125, so */
-    bytes = FIXNUM_P(num) ? sizeof(long) : rb_funcall(num, rb_intern("size"), 0);
+    bytes = FIXNUM_P(num) ? sizeof(long) : rb_funcall(num, idSize, 0);
     if (-0.415241 * ndigits - 0.125 > bytes ) {
 	return INT2FIX(0);
     }
@@ -1492,14 +1577,14 @@ int_round_0(VALUE num, int ndigits)
 	if (neg) x = -x;
 	return LONG2NUM(x);
     }
-    if (TYPE(f) == T_FLOAT) {
+    if (RB_TYPE_P(f, T_FLOAT)) {
 	/* then int_pow overflow */
 	return INT2FIX(0);
     }
     h = rb_funcall(f, '/', 1, INT2FIX(2));
     r = rb_funcall(num, '%', 1, f);
     n = rb_funcall(num, '-', 1, r);
-    op = RTEST(rb_funcall(num, '<', 1, INT2FIX(0))) ? rb_intern("<=") : '<';
+    op = negative_int_p(num) ? rb_intern("<=") : '<';
     if (!RTEST(rb_funcall(r, op, 1, h))) {
 	n = rb_funcall(n, '+', 1, f);
     }
@@ -1684,28 +1769,50 @@ num_truncate(VALUE num)
     return flo_truncate(rb_Float(num));
 }
 
+static double
+ruby_float_step_size(double beg, double end, double unit, int excl)
+{
+    const double epsilon = DBL_EPSILON;
+    double n = (end - beg)/unit;
+    double err = (fabs(beg) + fabs(end) + fabs(end-beg)) / fabs(unit) * epsilon;
+
+    if (isinf(unit)) {
+	return unit > 0 ? beg <= end : beg >= end;
+    }
+    if (err>0.5) err=0.5;
+    if (excl) {
+	if (n<=0) return 0;
+	if (n<1)
+	    n = 0;
+	else
+	    n = floor(n - err);
+    }
+    else {
+	if (n<0) return 0;
+	n = floor(n + err);
+    }
+    return n+1;
+}
 
 int
 ruby_float_step(VALUE from, VALUE to, VALUE step, int excl)
 {
-    if (TYPE(from) == T_FLOAT || TYPE(to) == T_FLOAT || TYPE(step) == T_FLOAT) {
-	const double epsilon = DBL_EPSILON;
+    if (RB_TYPE_P(from, T_FLOAT) || RB_TYPE_P(to, T_FLOAT) || RB_TYPE_P(step, T_FLOAT)) {
 	double beg = NUM2DBL(from);
 	double end = NUM2DBL(to);
 	double unit = NUM2DBL(step);
-	double n = (end - beg)/unit;
-	double err = (fabs(beg) + fabs(end) + fabs(end-beg)) / fabs(unit) * epsilon;
+	double n = ruby_float_step_size(beg, end, unit, excl);
 	long i;
 
 	if (isinf(unit)) {
-	    if (unit > 0 ? beg <= end : beg >= end) rb_yield(DBL2NUM(beg));
+	    /* if unit is infinity, i*unit+beg is NaN */
+	    if (n) rb_yield(DBL2NUM(beg));
 	}
 	else {
-	    if (err>0.5) err=0.5;
-	    n = floor(n + err);
-	    if (!excl || ((long)n)*unit+beg < end) n++;
 	    for (i=0; i<n; i++) {
-		rb_yield(DBL2NUM(i*unit+beg));
+		double d = i*unit+beg;
+		if (unit >= 0 ? end < d : d < end) d = end;
+		rb_yield(DBL2NUM(d));
 	    }
 	}
 	return TRUE;
@@ -1713,6 +1820,53 @@ ruby_float_step(VALUE from, VALUE to, VALUE step, int excl)
     return FALSE;
 }
 
+VALUE
+num_interval_step_size(VALUE from, VALUE to, VALUE step, int excl)
+{
+    if (FIXNUM_P(from) && FIXNUM_P(to) && FIXNUM_P(step)) {
+	long delta, diff;
+
+	diff = FIX2LONG(step);
+	if (!diff) rb_num_zerodiv();
+	delta = FIX2LONG(to) - FIX2LONG(from);
+	if (diff < 0) {
+	    diff = -diff;
+	    delta = -delta;
+	}
+	if (excl) {
+	    delta--;
+	}
+	if (delta < 0) {
+	    return INT2FIX(0);
+	}
+	return ULONG2NUM(delta / diff + 1UL);
+    }
+    else if (RB_TYPE_P(from, T_FLOAT) || RB_TYPE_P(to, T_FLOAT) || RB_TYPE_P(step, T_FLOAT)) {
+	double n = ruby_float_step_size(NUM2DBL(from), NUM2DBL(to), NUM2DBL(step), excl);
+
+	if (isinf(n)) return DBL2NUM(n);
+	if (POSFIXABLE(n)) return LONG2FIX(n);
+	return rb_dbl2big(n);
+    }
+    else {
+	VALUE result;
+	ID cmp = RTEST(rb_funcall(step, '>', 1, INT2FIX(0))) ? '>' : '<';
+	if (RTEST(rb_funcall(from, cmp, 1, to))) return INT2FIX(0);
+	result = rb_funcall(rb_funcall(to, '-', 1, from), id_div, 1, step);
+	if (!excl || RTEST(rb_funcall(rb_funcall(from, '+', 1, rb_funcall(result, '*', 1, step)), cmp, 1, to))) {
+	    result = rb_funcall(result, '+', 1, INT2FIX(1));
+	}
+	return result;
+    }
+}
+
+static VALUE
+num_step_size(VALUE from, VALUE args)
+{
+    VALUE to = RARRAY_PTR(args)[0];
+    VALUE step = (RARRAY_LEN(args) > 1) ? RARRAY_PTR(args)[1] : INT2FIX(1);
+    return num_interval_step_size(from, to, step, FALSE);
+}
 /*
  *  call-seq:
  *     num.step(limit[, step]) {|i| block }  ->  self
@@ -1748,19 +1902,15 @@ num_step(int argc, VALUE *argv, VALUE from)
 {
     VALUE to, step;
 
-    RETURN_ENUMERATOR(from, argc, argv);
+    RETURN_SIZED_ENUMERATOR(from, argc, argv, num_step_size);
     if (argc == 1) {
 	to = argv[0];
 	step = INT2FIX(1);
     }
     else {
-	if (argc == 2) {
-	    to = argv[0];
-	    step = argv[1];
-	}
-	else {
-	    rb_raise(rb_eArgError, "wrong number of arguments (%d for 1..2)", argc);
-	}
+	rb_check_arity(argc, 1, 2);
+	to = argv[0];
+	step = argv[1];
 	if (rb_equal(step, INT2FIX(0))) {
 	    rb_raise(rb_eArgError, "step can't be 0");
 	}
@@ -1790,7 +1940,7 @@ num_step(int argc, VALUE *argv, VALUE from)
 	VALUE i = from;
 	ID cmp;
 
-	if (RTEST(rb_funcall(step, '>', 1, INT2FIX(0)))) {
+	if (positive_int_p(step)) {
 	    cmp = '>';
 	}
 	else {
@@ -1894,13 +2044,13 @@ check_int(SIGNED_VALUE num)
 }
 
 static void
-check_uint(VALUE num, VALUE sign)
+check_uint(VALUE num, int sign)
 {
     static const VALUE mask = ~(VALUE)UINT_MAX;
 
-    if (RTEST(sign)) {
+    if (sign) {
 	/* minus */
-	if ((num & mask) != mask || (num & ~mask) <= INT_MAX + 1UL)
+	if ((num & mask) != mask || (num & ~mask) <= INT_MAX)
 #define VALUE_MSBMASK   ((VALUE)1 << ((sizeof(VALUE) * CHAR_BIT) - 1))
 	    rb_raise(rb_eRangeError, "integer %"PRIdVALUE " too small to convert to `unsigned int'", num|VALUE_MSBMASK);
     }
@@ -1934,7 +2084,7 @@ rb_num2uint(VALUE val)
 {
     VALUE num = rb_num2ulong(val);
 
-    check_uint(num, rb_funcall(val, '<', 1, INT2FIX(0)));
+    check_uint(num, negative_int_p(val));
     return (unsigned long)num;
 }
 
@@ -1948,7 +2098,7 @@ rb_fix2uint(VALUE val)
     }
     num = FIX2ULONG(val);
 
-    check_uint(num, rb_funcall(val, '<', 1, INT2FIX(0)));
+    check_uint(num, negative_int_p(val));
     return num;
 }
 #else
@@ -1965,6 +2115,80 @@ rb_fix2int(VALUE val)
 }
 #endif
 
+void
+rb_out_of_short(SIGNED_VALUE num)
+{
+    rb_raise(rb_eRangeError, "integer %"PRIdVALUE " too %s to convert to `short'",
+	     num, num < 0 ? "small" : "big");
+}
+
+static void
+check_short(SIGNED_VALUE num)
+{
+    if ((SIGNED_VALUE)(short)num != num) {
+	rb_out_of_short(num);
+    }
+}
+
+static void
+check_ushort(VALUE num, int sign)
+{
+    static const VALUE mask = ~(VALUE)USHRT_MAX;
+
+    if (sign) {
+	/* minus */
+	if ((num & mask) != mask || (num & ~mask) <= SHRT_MAX)
+#define VALUE_MSBMASK   ((VALUE)1 << ((sizeof(VALUE) * CHAR_BIT) - 1))
+	    rb_raise(rb_eRangeError, "integer %"PRIdVALUE " too small to convert to `unsigned short'", num|VALUE_MSBMASK);
+    }
+    else {
+	/* plus */
+	if ((num & mask) != 0)
+	    rb_raise(rb_eRangeError, "integer %"PRIuVALUE " too big to convert to `unsigned short'", num);
+    }
+}
+
+short
+rb_num2short(VALUE val)
+{
+    long num = rb_num2long(val);
+
+    check_short(num);
+    return num;
+}
+
+short
+rb_fix2short(VALUE val)
+{
+    long num = FIXNUM_P(val)?FIX2LONG(val):rb_num2long(val);
+
+    check_short(num);
+    return num;
+}
+
+unsigned short
+rb_num2ushort(VALUE val)
+{
+    VALUE num = rb_num2ulong(val);
+
+    check_ushort(num, negative_int_p(val));
+    return (unsigned long)num;
+}
+
+unsigned short
+rb_fix2ushort(VALUE val)
+{
+    unsigned long num;
+
+    if (!FIXNUM_P(val)) {
+	return rb_num2ushort(val);
+    }
+    num = FIX2ULONG(val);
+
+    check_ushort(num, negative_int_p(val));
+    return num;
+}
+
 VALUE
 rb_num2fix(VALUE val)
 {
@@ -2016,17 +2240,19 @@ rb_num2ll(VALUE val)
 
       case T_STRING:
 	rb_raise(rb_eTypeError, "no implicit conversion from string");
-	return Qnil;            /* not reached */
+	break;
 
       case T_TRUE:
       case T_FALSE:
 	rb_raise(rb_eTypeError, "no implicit conversion from boolean");
-	return Qnil;		/* not reached */
+	break;
 
       default:
-	val = rb_to_int(val);
-	return NUM2LL(val);
+	break;
     }
+
+    val = rb_to_int(val);
+    return NUM2LL(val);
 }
 
 unsigned LONG_LONG
@@ -2058,17 +2284,19 @@ rb_num2ull(VALUE val)
 
       case T_STRING:
 	rb_raise(rb_eTypeError, "no implicit conversion from string");
-	return Qnil;            /* not reached */
+	break;
 
       case T_TRUE:
       case T_FALSE:
 	rb_raise(rb_eTypeError, "no implicit conversion from boolean");
-	return Qnil;		/* not reached */
+	break;
 
       default:
-	val = rb_to_int(val);
-	return NUM2ULL(val);
+	break;
     }
+
+    val = rb_to_int(val);
+    return NUM2ULL(val);
 }
 
 #endif  /* HAVE_LONG_LONG */
@@ -2276,7 +2504,7 @@ int_chr(int argc, VALUE *argv, VALUE num)
       case 1:
 	break;
       default:
-	rb_raise(rb_eArgError, "wrong number of arguments (%d for 0..1)", argc);
+	rb_check_arity(argc, 0, 1);
 	break;
     }
     enc = rb_to_encoding(argv[0]);
@@ -2488,7 +2716,6 @@ fix_mul(VALUE x, VALUE y)
 #if SIZEOF_LONG * 2 <= SIZEOF_LONG_LONG
 	LONG_LONG d;
 #else
-	volatile long c;
 	VALUE r;
 #endif
 
@@ -2502,13 +2729,11 @@ fix_mul(VALUE x, VALUE y)
 #else
 	if (FIT_SQRT_LONG(a) && FIT_SQRT_LONG(b))
 	    return LONG2FIX(a*b);
-	c = a * b;
-	r = LONG2FIX(c);
-
 	if (a == 0) return x;
-	if (FIX2LONG(r) != c || c/a != b) {
+        if (MUL_OVERFLOW_FIXNUM_P(a, b))
 	    r = rb_big_mul(rb_int2big(a), rb_int2big(b));
-	}
+        else
+            r = LONG2FIX(a * b);
 	return r;
 #endif
     }
@@ -2730,11 +2955,10 @@ int_pow(long x, unsigned long y)
 	    y >>= 1;
 	}
 	{
-	    volatile long xz = x * z;
-	    if (!POSFIXABLE(xz) || xz / x != z) {
+            if (MUL_OVERFLOW_FIXNUM_P(x, z)) {
 		goto bignum;
 	    }
-	    z = xz;
+	    z = x * z;
 	}
     } while (--y);
     if (neg) z = -z;
@@ -2749,7 +2973,7 @@ int_pow(long x, unsigned long y)
  *  be negative or fractional.
  *
  *    2 ** 3      #=> 8
- *    2 ** -1     #=> 0.5
+ *    2 ** -1     #=> (1/2)
  *    2 ** 0.5    #=> 1.4142135623731
  */
 
@@ -2761,6 +2985,13 @@ fix_pow(VALUE x, VALUE y)
     if (FIXNUM_P(y)) {
 	long b = FIX2LONG(y);
 
+	if (a == 1) return INT2FIX(1);
+	if (a == -1) {
+	    if (b % 2 == 0)
+		return INT2FIX(1);
+	    else
+		return INT2FIX(-1);
+	}
 	if (b < 0)
 	    return rb_funcall(rb_rational_raw1(x), rb_intern("**"), 1, y);
 
@@ -2770,27 +3001,18 @@ fix_pow(VALUE x, VALUE y)
 	    if (b > 0) return INT2FIX(0);
 	    return DBL2NUM(INFINITY);
 	}
-	if (a == 1) return INT2FIX(1);
-	if (a == -1) {
-	    if (b % 2 == 0)
-		return INT2FIX(1);
-	    else
-		return INT2FIX(-1);
-	}
 	return int_pow(a, b);
     }
     switch (TYPE(y)) {
       case T_BIGNUM:
-
-	if (rb_funcall(y, '<', 1, INT2FIX(0)))
-	    return rb_funcall(rb_rational_raw1(x), rb_intern("**"), 1, y);
-
-	if (a == 0) return INT2FIX(0);
 	if (a == 1) return INT2FIX(1);
 	if (a == -1) {
 	    if (int_even_p(y)) return INT2FIX(1);
 	    else return INT2FIX(-1);
 	}
+	if (negative_int_p(y))
+	    return rb_funcall(rb_rational_raw1(x), rb_intern("**"), 1, y);
+	if (a == 0) return INT2FIX(0);
 	x = rb_int2big(FIX2LONG(x));
 	return rb_big_pow(x, y);
       case T_FLOAT:
@@ -2830,7 +3052,7 @@ fix_equal(VALUE x, VALUE y)
       case T_BIGNUM:
 	return rb_big_eq(y, x);
       case T_FLOAT:
-	return (double)FIX2LONG(x) == RFLOAT_VALUE(y) ? Qtrue : Qfalse;
+        return rb_integer_float_eq(x, y);
       default:
 	return num_equal(x, y);
     }
@@ -2840,10 +3062,11 @@ fix_equal(VALUE x, VALUE y)
  *  call-seq:
  *     fix <=> numeric  ->  -1, 0, +1 or nil
  *
- *  Comparison---Returns -1, 0, +1 or nil depending on whether
- *  <i>fix</i> is less than, equal to, or greater than
- *  <i>numeric</i>. This is the basis for the tests in
- *  <code>Comparable</code>.
+ *  Comparison---Returns -1, 0, +1 or nil depending on whether +fix+ is less
+ *  than, equal to, or greater than +numeric+. This is the basis for the tests
+ *  in  Comparable.
+ *
+ *  +nil+ is returned if the two values are incomparable.
  */
 
 static VALUE
@@ -2858,7 +3081,7 @@ fix_cmp(VALUE x, VALUE y)
       case T_BIGNUM:
 	return rb_big_cmp(rb_int2big(FIX2LONG(x)), y);
       case T_FLOAT:
-	return rb_dbl_cmp((double)FIX2LONG(x), RFLOAT_VALUE(y));
+        return rb_integer_float_cmp(x, y);
       default:
 	return rb_num_coerce_cmp(x, y, rb_intern("<=>"));
     }
@@ -2883,7 +3106,7 @@ fix_gt(VALUE x, VALUE y)
       case T_BIGNUM:
 	return FIX2INT(rb_big_cmp(rb_int2big(FIX2LONG(x)), y)) > 0 ? Qtrue : Qfalse;
       case T_FLOAT:
-	return (double)FIX2LONG(x) > RFLOAT_VALUE(y) ? Qtrue : Qfalse;
+        return rb_integer_float_cmp(x, y) == INT2FIX(1) ? Qtrue : Qfalse;
       default:
 	return rb_num_coerce_relop(x, y, '>');
     }
@@ -2908,7 +3131,10 @@ fix_ge(VALUE x, VALUE y)
       case T_BIGNUM:
 	return FIX2INT(rb_big_cmp(rb_int2big(FIX2LONG(x)), y)) >= 0 ? Qtrue : Qfalse;
       case T_FLOAT:
-	return (double)FIX2LONG(x) >= RFLOAT_VALUE(y) ? Qtrue : Qfalse;
+        {
+          VALUE rel = rb_integer_float_cmp(x, y);
+          return rel == INT2FIX(1) || rel == INT2FIX(0) ? Qtrue : Qfalse;
+        }
       default:
 	return rb_num_coerce_relop(x, y, rb_intern(">="));
     }
@@ -2933,7 +3159,7 @@ fix_lt(VALUE x, VALUE y)
       case T_BIGNUM:
 	return FIX2INT(rb_big_cmp(rb_int2big(FIX2LONG(x)), y)) < 0 ? Qtrue : Qfalse;
       case T_FLOAT:
-	return (double)FIX2LONG(x) < RFLOAT_VALUE(y) ? Qtrue : Qfalse;
+        return rb_integer_float_cmp(x, y) == INT2FIX(-1) ? Qtrue : Qfalse;
       default:
 	return rb_num_coerce_relop(x, y, '<');
     }
@@ -2958,7 +3184,10 @@ fix_le(VALUE x, VALUE y)
       case T_BIGNUM:
 	return FIX2INT(rb_big_cmp(rb_int2big(FIX2LONG(x)), y)) <= 0 ? Qtrue : Qfalse;
       case T_FLOAT:
-	return (double)FIX2LONG(x) <= RFLOAT_VALUE(y) ? Qtrue : Qfalse;
+        {
+          VALUE rel = rb_integer_float_cmp(x, y);
+          return rel == INT2FIX(-1) || rel == INT2FIX(0) ? Qtrue : Qfalse;
+        }
       default:
 	return rb_num_coerce_relop(x, y, rb_intern("<="));
     }
@@ -2977,16 +3206,25 @@ fix_rev(VALUE num)
     return ~num | FIXNUM_FLAG;
 }
 
-static VALUE
-bit_coerce(VALUE x)
+static int
+bit_coerce(VALUE *x, VALUE *y, int err)
 {
-    while (!FIXNUM_P(x) && TYPE(x) != T_BIGNUM) {
-	if (TYPE(x) == T_FLOAT) {
-	    rb_raise(rb_eTypeError, "can't convert Float into Integer");
+    if (!FIXNUM_P(*y) && !RB_TYPE_P(*y, T_BIGNUM)) {
+	do_coerce(x, y, err);
+	if (!FIXNUM_P(*x) && !RB_TYPE_P(*x, T_BIGNUM)
+	    && !FIXNUM_P(*y) && !RB_TYPE_P(*y, T_BIGNUM)) {
+	    if (!err) return FALSE;
+	    coerce_failed(*x, *y);
 	}
-	x = rb_to_int(x);
     }
-    return x;
+    return TRUE;
+}
+
+VALUE
+rb_num_coerce_bit(VALUE x, VALUE y, ID func)
+{
+    bit_coerce(&x, &y, TRUE);
+    return rb_funcall(x, func, 1, y);
 }
 
 /*
@@ -2999,13 +3237,17 @@ bit_coerce(VALUE x)
 static VALUE
 fix_and(VALUE x, VALUE y)
 {
-    long val;
+    if (FIXNUM_P(y)) {
+	long val = FIX2LONG(x) & FIX2LONG(y);
+	return LONG2NUM(val);
+    }
 
-    if (!FIXNUM_P(y = bit_coerce(y))) {
+    if (RB_TYPE_P(y, T_BIGNUM)) {
 	return rb_big_and(y, x);
     }
-    val = FIX2LONG(x) & FIX2LONG(y);
-    return LONG2NUM(val);
+
+    bit_coerce(&x, &y, TRUE);
+    return rb_funcall(x, rb_intern("&"), 1, y);
 }
 
 /*
@@ -3018,13 +3260,17 @@ fix_and(VALUE x, VALUE y)
 static VALUE
 fix_or(VALUE x, VALUE y)
 {
-    long val;
+    if (FIXNUM_P(y)) {
+	long val = FIX2LONG(x) | FIX2LONG(y);
+	return LONG2NUM(val);
+    }
 
-    if (!FIXNUM_P(y = bit_coerce(y))) {
+    if (RB_TYPE_P(y, T_BIGNUM)) {
 	return rb_big_or(y, x);
     }
-    val = FIX2LONG(x) | FIX2LONG(y);
-    return LONG2NUM(val);
+
+    bit_coerce(&x, &y, TRUE);
+    return rb_funcall(x, rb_intern("|"), 1, y);
 }
 
 /*
@@ -3037,13 +3283,17 @@ fix_or(VALUE x, VALUE y)
 static VALUE
 fix_xor(VALUE x, VALUE y)
 {
-    long val;
+    if (FIXNUM_P(y)) {
+	long val = FIX2LONG(x) ^ FIX2LONG(y);
+	return LONG2NUM(val);
+    }
 
-    if (!FIXNUM_P(y = bit_coerce(y))) {
+    if (RB_TYPE_P(y, T_BIGNUM)) {
 	return rb_big_xor(y, x);
     }
-    val = FIX2LONG(x) ^ FIX2LONG(y);
-    return LONG2NUM(val);
+
+    bit_coerce(&x, &y, TRUE);
+    return rb_funcall(x, rb_intern("^"), 1, y);
 }
 
 static VALUE fix_lshift(long, unsigned long);
@@ -3217,6 +3467,12 @@ fix_size(VALUE fix)
     return INT2FIX(sizeof(long));
 }
 
+static VALUE
+int_upto_size(VALUE from, VALUE args)
+{
+    return num_interval_step_size(from, RARRAY_PTR(args)[0], INT2FIX(1), FALSE);
+}
+
 /*
  *  call-seq:
  *     int.upto(limit) {|i| block }  ->  self
@@ -3237,7 +3493,7 @@ fix_size(VALUE fix)
 static VALUE
 int_upto(VALUE from, VALUE to)
 {
-    RETURN_ENUMERATOR(from, 1, &to);
+    RETURN_SIZED_ENUMERATOR(from, 1, &to, int_upto_size);
     if (FIXNUM_P(from) && FIXNUM_P(to)) {
 	long i, end;
 
@@ -3258,6 +3514,12 @@ int_upto(VALUE from, VALUE to)
     return from;
 }
 
+static VALUE
+int_downto_size(VALUE from, VALUE args)
+{
+    return num_interval_step_size(from, RARRAY_PTR(args)[0], INT2FIX(-1), FALSE);
+}
+
 /*
  *  call-seq:
  *     int.downto(limit) {|i| block }  ->  self
@@ -3279,7 +3541,7 @@ int_upto(VALUE from, VALUE to)
 static VALUE
 int_downto(VALUE from, VALUE to)
 {
-    RETURN_ENUMERATOR(from, 1, &to);
+    RETURN_SIZED_ENUMERATOR(from, 1, &to, int_downto_size);
     if (FIXNUM_P(from) && FIXNUM_P(to)) {
 	long i, end;
 
@@ -3300,6 +3562,18 @@ int_downto(VALUE from, VALUE to)
     return from;
 }
 
+static VALUE
+int_dotimes_size(VALUE num)
+{
+    if (FIXNUM_P(num)) {
+	if (NUM2LONG(num) <= 0) return INT2FIX(0);
+    }
+    else {
+	if (RTEST(rb_funcall(num, '<', 1, INT2FIX(0)))) return INT2FIX(0);
+    }
+    return num;
+}
+
 /*
  *  call-seq:
  *     int.times {|i| block }  ->  self
@@ -3322,7 +3596,7 @@ int_downto(VALUE from, VALUE to)
 static VALUE
 int_dotimes(VALUE num)
 {
-    RETURN_ENUMERATOR(num, 0, 0);
+    RETURN_SIZED_ENUMERATOR(num, 0, 0, int_dotimes_size);
 
     if (FIXNUM_P(num)) {
 	long i, end;
@@ -3476,6 +3750,7 @@ Init_Numeric(void)
     id_coerce = rb_intern("coerce");
     id_to_i = rb_intern("to_i");
     id_eq = rb_intern("==");
+    id_div = rb_intern("div");
 
     rb_eZeroDivError = rb_define_class("ZeroDivisionError", rb_eStandardError);
     rb_eFloatDomainError = rb_define_class("FloatDomainError", rb_eRangeError);
@@ -3538,6 +3813,7 @@ Init_Numeric(void)
     rb_cFixnum = rb_define_class("Fixnum", rb_cInteger);
 
     rb_define_method(rb_cFixnum, "to_s", fix_to_s, -1);
+    rb_define_alias(rb_cFixnum, "inspect", "to_s");
 
     rb_define_method(rb_cFixnum, "-@", fix_uminus, 0);
     rb_define_method(rb_cFixnum, "+", fix_plus, 1);
@@ -3583,21 +3859,97 @@ Init_Numeric(void)
     rb_undef_alloc_func(rb_cFloat);
     rb_undef_method(CLASS_OF(rb_cFloat), "new");
 
+    /*
+     *  Represents the rounding mode for floating point addition.
+     *
+     *  Usually defaults to 1, rounding to the nearest number.
+     *
+     *  Other modes include:
+     *
+     *  -1::	Indeterminable
+     *	0::	Rounding towards zero
+     *	1::	Rounding to the nearest number
+     *	2::	Rounding towards positive infinity
+     *	3::	Rounding towards negative infinity
+     */
     rb_define_const(rb_cFloat, "ROUNDS", INT2FIX(FLT_ROUNDS));
+    /*
+     *	The base of the floating point, or number of unique digits used to
+     *	represent the number.
+     *
+     *  Usually defaults to 2 on most systems, which would represent a base-10 decimal.
+     */
     rb_define_const(rb_cFloat, "RADIX", INT2FIX(FLT_RADIX));
+    /*
+     * The number of base digits for the +double+ data type.
+     *
+     * Usually defaults to 53.
+     */
     rb_define_const(rb_cFloat, "MANT_DIG", INT2FIX(DBL_MANT_DIG));
+    /*
+     *	The number of decimal digits in a double-precision floating point.
+     *
+     *	Usually defaults to 15.
+     */
     rb_define_const(rb_cFloat, "DIG", INT2FIX(DBL_DIG));
+    /*
+     *	The smallest posable exponent value in a double-precision floating
+     *	point.
+     *
+     *	Usually defaults to -1021.
+     */
     rb_define_const(rb_cFloat, "MIN_EXP", INT2FIX(DBL_MIN_EXP));
+    /*
+     *	The largest possible exponent value in a double-precision floating
+     *	point.
+     *
+     *	Usually defaults to 1024.
+     */
     rb_define_const(rb_cFloat, "MAX_EXP", INT2FIX(DBL_MAX_EXP));
+    /*
+     *	The smallest negative exponent in a double-precision floating point
+     *	where 10 raised to this power minus 1.
+     *
+     *	Usually defaults to -307.
+     */
     rb_define_const(rb_cFloat, "MIN_10_EXP", INT2FIX(DBL_MIN_10_EXP));
+    /*
+     *	The largest positive exponent in a double-precision floating point where
+     *	10 raised to this power minus 1.
+     *
+     *	Usually defaults to 308.
+     */
     rb_define_const(rb_cFloat, "MAX_10_EXP", INT2FIX(DBL_MAX_10_EXP));
+    /*
+     *	The smallest positive integer in a double-precision floating point.
+     *
+     *	Usually defaults to 2.2250738585072014e-308.
+     */
     rb_define_const(rb_cFloat, "MIN", DBL2NUM(DBL_MIN));
+    /*
+     *	The largest possible integer in a double-precision floating point number.
+     *
+     *	Usually defaults to 1.7976931348623157e+308.
+     */
     rb_define_const(rb_cFloat, "MAX", DBL2NUM(DBL_MAX));
+    /*
+     *	The difference between 1 and the smallest double-precision floating
+     *	point number.
+     *
+     *	Usually defaults to 2.2204460492503131e-16.
+     */
     rb_define_const(rb_cFloat, "EPSILON", DBL2NUM(DBL_EPSILON));
+    /*
+     *	An expression representing positive infinity.
+     */
     rb_define_const(rb_cFloat, "INFINITY", DBL2NUM(INFINITY));
+    /*
+     *	An expression representing a value which is "not a number".
+     */
     rb_define_const(rb_cFloat, "NAN", DBL2NUM(NAN));
 
     rb_define_method(rb_cFloat, "to_s", flo_to_s, 0);
+    rb_define_alias(rb_cFloat, "inspect", "to_s");
     rb_define_method(rb_cFloat, "coerce", flo_coerce, 1);
     rb_define_method(rb_cFloat, "-@", flo_uminus, 0);
     rb_define_method(rb_cFloat, "+", flo_plus, 1);
