@@ -2,7 +2,7 @@
 
   random.c -
 
-  $Author: usa $
+  $Author: nagachika $
   created at: Fri Dec 24 16:39:21 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -218,8 +218,6 @@ genrand_real2(struct MT *mt)
 #undef N
 #undef M
 
-/* These real versions are due to Isaku Wada, 2002/01/09 added */
-
 typedef struct {
     VALUE seed;
     struct MT mt;
@@ -327,7 +325,6 @@ int_pair_to_real_inclusive(unsigned int a, unsigned int b)
 }
 
 VALUE rb_cRandom;
-static VALUE rb_Random_DEFAULT;
 #define id_minus '-'
 #define id_plus  '+'
 static ID id_rand, id_bytes;
@@ -453,20 +450,13 @@ rand_init(struct MT *mt, VALUE vseed)
 }
 
 /*
- * call-seq: Random.new([seed]) -> prng
+ * call-seq:
+ *   Random.new(seed = Random.new_seed) -> prng
  *
- * Creates new Mersenne Twister based pseudorandom number generator with
- * seed.  When the argument seed is omitted, the generator is initialized
- * with Random.new_seed.
+ * Creates a new PRNG using +seed+ to set the initial state. If +seed+ is
+ * omitted, the generator is initialized with Random.new_seed.
  *
- * The argument seed is used to ensure repeatable sequences of random numbers
- * between different runs of the program.
- *
- *     prng = Random.new(1234)
- *     [ prng.rand, prng.rand ]   #=> [0.191519450378892, 0.622108771039832]
- *     [ prng.integer(10), prng.integer(1000) ]  #=> [4, 664]
- *     prng = Random.new(1234)
- *     [ prng.rand, prng.rand ]   #=> [0.191519450378892, 0.622108771039832]
+ * See Random.srand for more information on the use of seed values.
  */
 static VALUE
 random_init(int argc, VALUE *argv, VALUE obj)
@@ -475,10 +465,12 @@ random_init(int argc, VALUE *argv, VALUE obj)
     rb_random_t *rnd = get_rnd(obj);
 
     if (argc == 0) {
+	rb_check_frozen(obj);
 	vseed = random_seed();
     }
     else {
 	rb_scan_args(argc, argv, "01", &vseed);
+	rb_check_copyable(obj, vseed);
     }
     rnd->seed = rand_init(&rnd->mt, vseed);
     return obj;
@@ -507,14 +499,14 @@ fill_random_seed(unsigned int seed[DEFAULT_SEED_CNT])
     memset(seed, 0, DEFAULT_SEED_LEN);
 
 #if USE_DEV_URANDOM
-    if ((fd = open("/dev/urandom", O_RDONLY
+    if ((fd = rb_cloexec_open("/dev/urandom", O_RDONLY
 #ifdef O_NONBLOCK
             |O_NONBLOCK
 #endif
 #ifdef O_NOCTTY
             |O_NOCTTY
 #endif
-            )) >= 0) {
+            , 0)) >= 0) {
         rb_update_max_fd(fd);
         if (fstat(fd, &statbuf) == 0 && S_ISCHR(statbuf.st_mode)) {
 	    if (read(fd, seed, DEFAULT_SEED_LEN) < DEFAULT_SEED_LEN) {
@@ -548,8 +540,7 @@ make_seed_value(const void *ptr)
 {
     const long len = DEFAULT_SEED_LEN/SIZEOF_BDIGITS;
     BDIGIT *digits;
-    NEWOBJ(big, struct RBignum);
-    OBJSETUP(big, rb_cBignum, T_BIGNUM);
+    NEWOBJ_OF(big, struct RBignum, rb_cBignum, T_BIGNUM);
 
     RBIGNUM_SET_SIGN(big, 1);
     rb_big_resize((VALUE)big, len + 1);
@@ -572,7 +563,10 @@ make_seed_value(const void *ptr)
 /*
  * call-seq: Random.new_seed -> integer
  *
- * Returns arbitrary value for seed.
+ * Returns an arbitrary seed value. This is used by Random.new
+ * when no seed value is specified as an argument.
+ *
+ *   Random.new_seed  #=> 115032730400174366788466674494640623225
  */
 static VALUE
 random_seed(void)
@@ -585,7 +579,16 @@ random_seed(void)
 /*
  * call-seq: prng.seed -> integer
  *
- * Returns the seed of the generator.
+ * Returns the seed value used to initialize the generator. This may be used to
+ * initialize another generator with the same state at a later time, causing it
+ * to produce the same sequence of numbers.
+ *
+ *   prng1 = Random.new(1234)
+ *   prng1.seed       #=> 1234
+ *   prng1.rand(100)  #=> 47
+ *
+ *   prng2 = Random.new(prng1.seed)
+ *   prng2.rand(100)  #=> 47
  */
 static VALUE
 random_get_seed(VALUE obj)
@@ -597,9 +600,14 @@ random_get_seed(VALUE obj)
 static VALUE
 random_copy(VALUE obj, VALUE orig)
 {
-    rb_random_t *rnd1 = get_rnd(obj);
-    rb_random_t *rnd2 = get_rnd(orig);
-    struct MT *mt = &rnd1->mt;
+    rb_random_t *rnd1, *rnd2;
+    struct MT *mt;
+
+    if (!OBJ_INIT_COPY(obj, orig)) return obj;
+
+    rnd1 = get_rnd(obj);
+    rnd2 = get_rnd(orig);
+    mt = &rnd1->mt;
 
     *rnd1 = *rnd2;
     mt->next = mt->state + numberof(mt->state) - mt->left + 1;
@@ -682,6 +690,7 @@ random_load(VALUE obj, VALUE dump)
     VALUE *ary;
     unsigned long x;
 
+    rb_check_copyable(obj, dump);
     Check_Type(dump, T_ARRAY);
     ary = RARRAY_PTR(dump);
     switch (RARRAY_LEN(dump)) {
@@ -761,17 +770,26 @@ random_load(VALUE obj, VALUE dump)
 }
 
 /*
- *  call-seq:
- *     srand(number=0)    -> old_seed
+ * call-seq:
+ *   srand(number = Random.new_seed) -> old_seed
  *
- *  Seeds the pseudorandom number generator to the value of
- *  <i>number</i>. If <i>number</i> is omitted,
- *  seeds the generator using a combination of the time, the
- *  process id, and a sequence number. (This is also the behavior if
- *  <code>Kernel::rand</code> is called without previously calling
- *  <code>srand</code>, but without the sequence.) By setting the seed
- *  to a known value, scripts can be made deterministic during testing.
- *  The previous seed value is returned. Also see <code>Kernel::rand</code>.
+ * Seeds the system pseudo-random number generator, Random::DEFAULT, with
+ * +number+.  The previous seed value is returned.
+ *
+ * If +number+ is omitted, seeds the generator using a source of entropy
+ * provided by the operating system, if available (/dev/urandom on Unix systems
+ * or the RSA cryptographic provider on Windows), which is then combined with
+ * the time, the process id, and a sequence number.
+ *
+ * srand may be used to ensure repeatable sequences of pseudo-random numbers
+ * between different runs of the program. By setting the seed to a known value,
+ * programs can be made deterministic during testing.
+ *
+ *   srand 1234               # => 268519324636777531569100071560086917274
+ *   [ rand, rand ]           # => [0.1915194503788923, 0.6221087710398319]
+ *   [ rand(10), rand(1000) ] # => [4, 664]
+ *   srand 1234               # => 1234
+ *   [ rand, rand ]           # => [0.1915194503788923, 0.6221087710398319]
  */
 
 static VALUE
@@ -881,9 +899,9 @@ limited_big_rand(struct MT *mt, struct RBignum *limit)
 }
 
 /*
- * Returns random unsigned long value in [0, _limit_].
+ * Returns random unsigned long value in [0, +limit+].
  *
- * Note that _limit_ is included, and the range of the argument and the
+ * Note that +limit+ is included, and the range of the argument and the
  * return value depends on environments.
  */
 unsigned long
@@ -898,7 +916,7 @@ rb_random_int32(VALUE obj)
     rb_random_t *rnd = try_get_rnd(obj);
     if (!rnd) {
 #if SIZEOF_LONG * CHAR_BIT > 32
-	VALUE lim = ULONG2NUM(0x100000000);
+	VALUE lim = ULONG2NUM(0x100000000UL);
 #elif defined HAVE_LONG_LONG
 	VALUE lim = ULL2NUM((LONG_LONG)0xffffffff+1);
 #else
@@ -916,7 +934,10 @@ rb_random_real(VALUE obj)
     if (!rnd) {
 	VALUE v = rb_funcall2(obj, id_rand, 0, 0);
 	double d = NUM2DBL(v);
-	if (d < 0.0 || d >= 1.0) {
+	if (d < 0.0) {
+	    rb_raise(rb_eRangeError, "random number too small %g", d);
+	}
+	else if (d >= 1.0) {
 	    rb_raise(rb_eRangeError, "random number too big %g", d);
 	}
 	return d;
@@ -924,11 +945,46 @@ rb_random_real(VALUE obj)
     return genrand_real(&rnd->mt);
 }
 
+static inline VALUE
+ulong_to_num_plus_1(unsigned long n)
+{
+#if HAVE_LONG_LONG
+    return ULL2NUM((LONG_LONG)n+1);
+#else
+    if (n >= ULONG_MAX) {
+	return rb_big_plus(ULONG2NUM(n), INT2FIX(1));
+    }
+    return ULONG2NUM(n+1);
+#endif
+}
+
+unsigned long
+rb_random_ulong_limited(VALUE obj, unsigned long limit)
+{
+    rb_random_t *rnd = try_get_rnd(obj);
+    if (!rnd) {
+	extern int rb_num_negative_p(VALUE);
+	VALUE lim = ulong_to_num_plus_1(limit);
+	VALUE v = rb_to_int(rb_funcall2(obj, id_rand, 1, &lim));
+	unsigned long r = NUM2ULONG(v);
+	if (rb_num_negative_p(v)) {
+	    rb_raise(rb_eRangeError, "random number too small %ld", r);
+	}
+	if (r > limit) {
+	    rb_raise(rb_eRangeError, "random number too big %ld", r);
+	}
+	return r;
+    }
+    return limited_rand(&rnd->mt, limit);
+}
+
 /*
  * call-seq: prng.bytes(size) -> a_string
  *
- * Returns a random binary string.  The argument size specified the length of
- * the result string.
+ * Returns a random binary string containing +size+ bytes.
+ *
+ *   random_string = Random.new.bytes(10) # => "\xD7:R\xAB?\x83\xCE\xFAkO"
+ *   random_string.size                   # => 10
  */
 static VALUE
 random_bytes(VALUE obj, VALUE len)
@@ -1038,7 +1094,7 @@ rand_range(struct MT* mt, VALUE range)
 
     if ((v = vmax = range_values(range, &beg, &end, &excl)) == Qfalse)
 	return Qfalse;
-    if (TYPE(vmax) != T_FLOAT && (v = rb_check_to_integer(vmax, "to_int"), !NIL_P(v))) {
+    if (!RB_TYPE_P(vmax, T_FLOAT) && (v = rb_check_to_integer(vmax, "to_int"), !NIL_P(v))) {
 	long max;
 	vmax = v;
 	v = Qnil;
@@ -1101,8 +1157,7 @@ rand_range(struct MT* mt, VALUE range)
       case T_FLOAT: {
 	VALUE f = rb_check_to_float(beg);
 	if (!NIL_P(f)) {
-	    RFLOAT_VALUE(v) += RFLOAT_VALUE(f);
-	    return v;
+	    return DBL2NUM(RFLOAT_VALUE(v) + RFLOAT_VALUE(f));
 	}
       }
       default:
@@ -1112,47 +1167,59 @@ rand_range(struct MT* mt, VALUE range)
     return v;
 }
 
+static VALUE rand_random(int argc, VALUE *argv, rb_random_t *rnd);
+
 /*
  * call-seq:
- *     prng.rand -> float
- *     prng.rand(limit) -> number
+ *   prng.rand -> float
+ *   prng.rand(max) -> number
  *
- * When the argument is an +Integer+ or a +Bignum+, it returns a
- * random integer greater than or equal to zero and less than the
- * argument.  Unlike Random.rand, when the argument is a negative
- * integer or zero, it raises an ArgumentError.
+ * When +max+ is an Integer, +rand+ returns a random integer greater than
+ * or equal to zero and less than +max+. Unlike Kernel.rand, when +max+
+ * is a negative integer or zero, +rand+ raises an ArgumentError.
  *
- * When the argument is a +Float+, it returns a random floating point
- * number between 0.0 and _max_, including 0.0 and excluding _max_.
+ *   prng = Random.new
+ *   prng.rand(100)       # => 42
  *
- * When the argument _limit_ is a +Range+, it returns a random
- * number where range.member?(number) == true.
- *     prng.rand(5..9)  #=> one of [5, 6, 7, 8, 9]
- *     prng.rand(5...9) #=> one of [5, 6, 7, 8]
- *     prng.rand(5.0..9.0) #=> between 5.0 and 9.0, including 9.0
- *     prng.rand(5.0...9.0) #=> between 5.0 and 9.0, excluding 9.0
+ * When +max+ is a Float, +rand+ returns a random floating point number
+ * between 0.0 and +max+, including 0.0 and excluding +max+.
  *
- * +begin+/+end+ of the range have to have subtract and add methods.
+ *   prng.rand(1.5)       # => 1.4600282860034115
  *
- * Otherwise, it raises an ArgumentError.
+ * When +max+ is a Range, +rand+ returns a random number where
+ * range.member?(number) == true.
+ *
+ *   prng.rand(5..9)      # => one of [5, 6, 7, 8, 9]
+ *   prng.rand(5...9)     # => one of [5, 6, 7, 8]
+ *   prng.rand(5.0..9.0)  # => between 5.0 and 9.0, including 9.0
+ *   prng.rand(5.0...9.0) # => between 5.0 and 9.0, excluding 9.0
+ *
+ * Both the beginning and ending values of the range must respond to subtract
+ * (<tt>-</tt>) and add (<tt>+</tt>)methods, or rand will raise an
+ * ArgumentError.
  */
 static VALUE
 random_rand(int argc, VALUE *argv, VALUE obj)
 {
-    rb_random_t *rnd = get_rnd(obj);
+    return rand_random(argc, argv, get_rnd(obj));
+}
+
+static VALUE
+rand_random(int argc, VALUE *argv, rb_random_t *rnd)
+{
     VALUE vmax, v;
 
     if (argc == 0) {
 	return rb_float_new(genrand_real(&rnd->mt));
     }
-    else if (argc != 1) {
-	rb_raise(rb_eArgError, "wrong number of arguments (%d for 0..1)", argc);
+    else {
+	rb_check_arity(argc, 0, 1);
     }
     vmax = argv[0];
     if (NIL_P(vmax)) {
 	v = Qnil;
     }
-    else if (TYPE(vmax) != T_FLOAT && (v = rb_check_to_integer(vmax, "to_int"), !NIL_P(v))) {
+    else if (!RB_TYPE_P(vmax, T_FLOAT) && (v = rb_check_to_integer(vmax, "to_int"), !NIL_P(v))) {
 	v = rand_int(&rnd->mt, v, 1);
     }
     else if (v = rb_check_to_float(vmax), !NIL_P(v)) {
@@ -1180,9 +1247,27 @@ random_rand(int argc, VALUE *argv, VALUE obj)
 
 /*
  * call-seq:
- *     prng1 == prng2 -> true or false
+ *   prng1 == prng2 -> true or false
  *
- * Returns true if the generators' states equal.
+ * Returns true if the two generators have the same internal state, otherwise
+ * false.  Equivalent generators will return the same sequence of
+ * pseudo-random numbers.  Two generators will generally have the same state
+ * only if they were initialized with the same seed
+ *
+ *   Random.new == Random.new             # => false
+ *   Random.new(1234) == Random.new(1234) # => true
+ *
+ * and have the same invocation history.
+ *
+ *   prng1 = Random.new(1234)
+ *   prng2 = Random.new(1234)
+ *   prng1 == prng2 # => true
+ *
+ *   prng1.rand     # => 0.1915194503788923
+ *   prng1 == prng2 # => false
+ *
+ *   prng2.rand     # => 0.1915194503788923
+ *   prng1 == prng2 # => true
  */
 static VALUE
 random_equal(VALUE self, VALUE other)
@@ -1199,31 +1284,34 @@ random_equal(VALUE self, VALUE other)
 }
 
 /*
- *  call-seq:
- *     rand(max=0)    -> number
+ * call-seq:
+ *   rand(max=0)    -> number
  *
+ * If called without an argument, or if <tt>max.to_i.abs == 0</tt>, rand
+ * returns a pseudo-random floating point number between 0.0 and 1.0,
+ * including 0.0 and excluding 1.0.
  *
- *  If <i>max</i> is +Range+, returns a pseudorandom number where
- *  range.member(number) == true.
+ *   rand        #=> 0.2725926052826416
  *
- *  Or else converts _max_ to an integer using max1 =
- *  max<code>.to_i.abs</code>.
+ * When +max.abs+ is greater than or equal to 1, +rand+ returns a pseudo-random
+ * integer greater than or equal to 0 and less than +max.to_i.abs+.
  *
- *  Then if _max_ is +nil+ the result is zero, returns a pseudorandom floating
- *  point number greater than or equal to 0.0 and less than 1.0.
+ *   rand(100)   #=> 12
  *
- *  Otherwise, returns a pseudorandom integer greater than or equal to zero and
- *  less than max1.
+ * When +max+ is a Range, +rand+ returns a random number where
+ * range.member?(number) == true.
  *
- *  <code>Kernel::srand</code> may be used to ensure repeatable sequences of
- *  random numbers between different runs of the program. Ruby currently uses
- *  a modified Mersenne Twister with a period of 2**19937-1.
+ * Negative or floating point values for +max+ are allowed, but may give
+ * surprising results.
  *
- *     srand 1234                 #=> 0
- *     [ rand,  rand ]            #=> [0.191519450163469, 0.49766366626136]
- *     [ rand(10), rand(1000) ]   #=> [6, 817]
- *     srand 1234                 #=> 1234
- *     [ rand,  rand ]            #=> [0.191519450163469, 0.49766366626136]
+ *   rand(-100) # => 87
+ *   rand(-0.5) # => 0.8130921818028143
+ *   rand(1.9)  # equivalent to rand(1), which is always 0
+ *
+ * Kernel.srand may be used to ensure that sequences of random numbers are
+ * reproducible between different runs of a program.
+ *
+ * See also Random.rand.
  */
 
 static VALUE
@@ -1247,19 +1335,17 @@ rb_f_rand(int argc, VALUE *argv, VALUE obj)
 }
 
 /*
- *  call-seq:
- *     Random.rand -> float
- *     Random.rand(limit) -> number
- *
- *     Alias of _Random::DEFAULT.rand_.
+ * call-seq:
+ *   Random.rand -> float
+ *   Random.rand(max) -> number
  *
+ * Alias of Random::DEFAULT.rand.
  */
 
 static VALUE
 random_s_rand(int argc, VALUE *argv, VALUE obj)
 {
-    rand_start(&default_rand);
-    return random_rand(argc, argv, rb_Random_DEFAULT);
+    return rand_random(argc, argv, rand_start(&default_rand));
 }
 
 #define SIP_HASH_STREAMING 0
@@ -1361,6 +1447,30 @@ rb_reset_random_seed(void)
     r->seed = INT2FIX(0);
 }
 
+/*
+ * Document-class: Random
+ *
+ * Random provides an interface to Ruby's pseudo-random number generator, or
+ * PRNG.  The PRNG produces a deterministic sequence of bits which approximate
+ * true randomness. The sequence may be represented by integers, floats, or
+ * binary strings.
+ *
+ * The generator may be initialized with either a system-generated or
+ * user-supplied seed value by using Random.srand.
+ *
+ * The class method Random.rand provides the base functionality of Kernel.rand
+ * along with better handling of floating point values. These are both
+ * interfaces to Random::DEFAULT, the Ruby system PRNG.
+ *
+ * Random.new will create a new PRNG with a state independent of
+ * Random::DEFAULT, allowing multiple generators with different seed values or
+ * sequence positions to exist simultaneously. Random objects can be
+ * marshaled, allowing sequences to be saved and resumed.
+ *
+ * PRNGs are currently implemented as a modified Mersenne Twister with a period
+ * of 2**19937-1.
+ */
+
 void
 Init_Random(void)
 {
@@ -1375,15 +1485,17 @@ Init_Random(void)
     rb_define_method(rb_cRandom, "bytes", random_bytes, 1);
     rb_define_method(rb_cRandom, "seed", random_get_seed, 0);
     rb_define_method(rb_cRandom, "initialize_copy", random_copy, 1);
-    rb_define_method(rb_cRandom, "marshal_dump", random_dump, 0);
-    rb_define_method(rb_cRandom, "marshal_load", random_load, 1);
+    rb_define_private_method(rb_cRandom, "marshal_dump", random_dump, 0);
+    rb_define_private_method(rb_cRandom, "marshal_load", random_load, 1);
     rb_define_private_method(rb_cRandom, "state", random_state, 0);
     rb_define_private_method(rb_cRandom, "left", random_left, 0);
     rb_define_method(rb_cRandom, "==", random_equal, 1);
 
-    rb_Random_DEFAULT = TypedData_Wrap_Struct(rb_cRandom, &random_data_type, &default_rand);
-    rb_global_variable(&rb_Random_DEFAULT);
-    rb_define_const(rb_cRandom, "DEFAULT", rb_Random_DEFAULT);
+    {
+	VALUE rand_default = TypedData_Wrap_Struct(rb_cRandom, &random_data_type, &default_rand);
+	rb_gc_register_mark_object(rand_default);
+	rb_define_const(rb_cRandom, "DEFAULT", rand_default);
+    }
 
     rb_define_singleton_method(rb_cRandom, "srand", rb_f_srand, -1);
     rb_define_singleton_method(rb_cRandom, "rand", random_s_rand, -1);
