@@ -2,7 +2,7 @@
 
   ruby.c -
 
-  $Author: usa $
+  $Author: kazu $
   created at: Tue Aug 10 12:47:31 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -52,6 +52,14 @@
 char *getenv();
 #endif
 
+#define numberof(array) (int)(sizeof(array) / sizeof((array)[0]))
+
+#if defined DISABLE_RUBYGEMS && DISABLE_RUBYGEMS
+#define DEFAULT_RUBYGEMS_ENABLED "disabled"
+#else
+#define DEFAULT_RUBYGEMS_ENABLED "enabled"
+#endif
+
 #define DISABLE_BIT(bit) (1U << disable_##bit)
 enum disable_flag_bits {
     disable_gems,
@@ -62,8 +70,10 @@ enum disable_flag_bits {
 #define DUMP_BIT(bit) (1U << dump_##bit)
 enum dump_flag_bits {
     dump_version,
+    dump_version_v,
     dump_copyright,
     dump_usage,
+    dump_help,
     dump_yydebug,
     dump_syntax,
     dump_parsetree,
@@ -112,7 +122,7 @@ cmdline_options_init(struct cmdline_options *opt)
     return opt;
 }
 
-static NODE *load_file(VALUE, const char *, int, struct cmdline_options *);
+static NODE *load_file(VALUE, VALUE, int, struct cmdline_options *);
 static void forbid_setid(const char *, struct cmdline_options *);
 #define forbid_setid(s) forbid_setid((s), opt)
 
@@ -122,42 +132,78 @@ static struct {
 } origarg;
 
 static void
-usage(const char *name)
+usage(const char *name, int help)
 {
     /* This message really ought to be max 23 lines.
      * Removed -h because the user already knows that option. Others? */
 
-    static const char *const usage_msg[] = {
-	"-0[octal]       specify record separator (\\0, if no argument)",
-	"-a              autosplit mode with -n or -p (splits $_ into $F)",
-	"-c              check syntax only",
-	"-Cdirectory     cd to directory, before executing your script",
-	"-d              set debugging flags (set $DEBUG to true)",
-	"-e 'command'    one line of script. Several -e's allowed. Omit [programfile]",
-	"-Eex[:in]       specify the default external and internal character encodings",
-	"-Fpattern       split() pattern for autosplit (-a)",
-	"-i[extension]   edit ARGV files in place (make backup if extension supplied)",
-	"-Idirectory     specify $LOAD_PATH directory (may be used more than once)",
-	"-l              enable line ending processing",
-	"-n              assume 'while gets(); ... end' loop around your script",
-	"-p              assume loop like -n but print line also like sed",
-	"-rlibrary       require the library, before executing your script",
-	"-s              enable some switch parsing for switches after script name",
-	"-S              look for the script using PATH environment variable",
-	"-T[level=1]     turn on tainting checks",
-	"-v              print version number, then turn on verbose mode",
-	"-w              turn warnings on for your script",
-	"-W[level=2]     set warning level; 0=silence, 1=medium, 2=verbose",
-	"-x[directory]   strip off text before #!ruby line and perhaps cd to directory",
-	"--copyright     print the copyright",
-	"--version       print the version",
-	NULL
+    struct message {
+	const char *str;
+	unsigned short namelen, secondlen;
     };
-    const char *const *p = usage_msg;
+#define M(shortopt, longopt, desc) { \
+    shortopt " " longopt " " desc, \
+    (unsigned short)sizeof(shortopt), \
+    (unsigned short)sizeof(longopt), \
+}
+    static const struct message usage_msg[] = {
+	M("-0[octal]",	   "",			   "specify record separator (\\0, if no argument)"),
+	M("-a",		   "",			   "autosplit mode with -n or -p (splits $_ into $F)"),
+	M("-c",		   "",			   "check syntax only"),
+	M("-Cdirectory",   "",			   "cd to directory before executing your script"),
+	M("-d",		   ", --debug",		   "set debugging flags (set $DEBUG to true)"),
+	M("-e 'command'",  "",			   "one line of script. Several -e's allowed. Omit [programfile]"),
+	M("-Eex[:in]",     ", --encoding=ex[:in]", "specify the default external and internal character encodings"),
+	M("-Fpattern",	   "",			   "split() pattern for autosplit (-a)"),
+	M("-i[extension]", "",			   "edit ARGV files in place (make backup if extension supplied)"),
+	M("-Idirectory",   "",			   "specify $LOAD_PATH directory (may be used more than once)"),
+	M("-l",		   "",			   "enable line ending processing"),
+	M("-n",		   "",			   "assume 'while gets(); ... end' loop around your script"),
+	M("-p",		   "",			   "assume loop like -n but print line also like sed"),
+	M("-rlibrary",	   "",			   "require the library before executing your script"),
+	M("-s",		   "",			   "enable some switch parsing for switches after script name"),
+	M("-S",		   "",			   "look for the script using PATH environment variable"),
+	M("-T[level=1]",   "",			   "turn on tainting checks"),
+	M("-v",		   ", --verbose",	   "print version number, then turn on verbose mode"),
+	M("-w",		   "",			   "turn warnings on for your script"),
+	M("-W[level=2]",   "",			   "set warning level; 0=silence, 1=medium, 2=verbose"),
+	M("-x[directory]", "",			   "strip off text before #!ruby line and perhaps cd to directory"),
+	M("-h",		   "",			   "show this message, --help for more info"),
+    };
+    static const struct message help_msg[] = {
+	M("--copyright",                   "", "print the copyright"),
+	M("--enable=feature[,...]",	   ", --disable=feature[,...]",
+	  "enable or disable features"),
+	M("--external-encoding=encoding",  ", --internal-encoding=encoding",
+	  "specify the default external or internal character encoding"),
+	M("--version",                     "", "print the version"),
+	M("--help",			   "", "show this message, -h for short message"),
+    };
+    static const struct message features[] = {
+	M("gems",    "",        "rubygems (default: "DEFAULT_RUBYGEMS_ENABLED")"),
+	M("rubyopt", "",        "RUBYOPT environment variable (default: enabled)"),
+    };
+    int i, w = 16, num = numberof(usage_msg) - (help ? 1 : 0);
+#define SHOW(m) do { \
+	int wrap = help && (m).namelen + (m).secondlen - 2 > w; \
+	printf("  %.*s%-*.*s%-*s%s\n", (m).namelen-1, (m).str, \
+	       (wrap ? 0 : w - (m).namelen + 1), \
+	       (help ? (m).secondlen-1 : 0), (m).str + (m).namelen, \
+	       (wrap ? w + 3 : 0), (wrap ? "\n" : ""), \
+	       (m).str + (m).namelen + (m).secondlen); \
+    } while (0)
 
     printf("Usage: %s [switches] [--] [programfile] [arguments]\n", name);
-    while (*p)
-	printf("  %s\n", *p++);
+    for (i = 0; i < num; ++i)
+	SHOW(usage_msg[i]);
+
+    if (!help) return;
+
+    for (i = 0; i < numberof(help_msg); ++i)
+	SHOW(help_msg[i]);
+    puts("Features:");
+    for (i = 0; i < numberof(features); ++i)
+	SHOW(features[i]);
 }
 
 #ifdef MANGLED_PATH
@@ -423,7 +469,7 @@ ruby_init_loadpath_safe(int safe_level)
 	    p -= bindir_len;
 	    *p = 0;
 	}
-	else if (p - libpath >= libdir_len && !STRCASECMP(p - libdir_len, libdir)) {
+	else if (p - libpath >= libdir_len && !strcmp(p - libdir_len, libdir)) {
 	    p -= libdir_len;
 	    *p = 0;
 	}
@@ -476,12 +522,15 @@ static void
 add_modules(VALUE *req_list, const char *mod)
 {
     VALUE list = *req_list;
+    VALUE feature;
 
     if (!list) {
 	*req_list = list = rb_ary_new();
 	RBASIC(list)->klass = 0;
     }
-    rb_ary_push(list, rb_obj_freeze(rb_str_new2(mod)));
+    feature = rb_str_new2(mod);
+    RBASIC(feature)->klass = 0;
+    rb_ary_push(list, feature);
 }
 
 static void
@@ -491,21 +540,33 @@ require_libraries(VALUE *req_list)
     VALUE self = rb_vm_top_self();
     ID require;
     rb_thread_t *th = GET_THREAD();
-    rb_block_t *prev_base_block = th->base_block;
+    rb_encoding *extenc = rb_default_external_encoding();
     int prev_parse_in_eval = th->parse_in_eval;
-    th->base_block = 0;
     th->parse_in_eval = 0;
 
     Init_ext();		/* should be called here for some reason :-( */
     CONST_ID(require, "require");
     while (list && RARRAY_LEN(list) > 0) {
 	VALUE feature = rb_ary_shift(list);
+	rb_enc_associate(feature, extenc);
+	RBASIC(feature)->klass = rb_cString;
+	OBJ_FREEZE(feature);
 	rb_funcall2(self, require, 1, &feature);
     }
     *req_list = 0;
 
     th->parse_in_eval = prev_parse_in_eval;
-    th->base_block = prev_base_block;
+}
+
+static rb_env_t*
+toplevel_context(VALUE toplevel_binding)
+{
+    rb_env_t *env;
+    rb_binding_t *bind;
+
+    GetBindingPtr(toplevel_binding, bind);
+    GetEnvPtr(bind->env, env);
+    return env;
 }
 
 static void
@@ -669,6 +730,7 @@ dump_option(const char *str, int len, void *arg)
     SET_WHEN_DUMP(version);
     SET_WHEN_DUMP(copyright);
     SET_WHEN_DUMP(usage);
+    SET_WHEN_DUMP(help);
     SET_WHEN_DUMP(yydebug);
     SET_WHEN_DUMP(syntax);
     SET_WHEN_DUMP(parsetree);
@@ -712,7 +774,7 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
 
     for (argc--, argv++; argc > 0; argc--, argv++) {
 	const char *const arg = argv[0];
-	if (arg[0] != '-' || !arg[1])
+	if (!arg || arg[0] != '-' || !arg[1])
 	    break;
 
 	s = arg + 1;
@@ -751,7 +813,7 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
 		s++;
 		goto reswitch;
 	    }
-	    ruby_show_version();
+	    opt->dump |= DUMP_BIT(version_v);
 	    opt->verbose = 1;
 	  case 'w':
 	    ruby_verbose = Qtrue;
@@ -1053,7 +1115,7 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
 	    }
 	    else if (strcmp("help", s) == 0) {
 		if (envopt) goto noenvopt_long;
-		opt->dump |= DUMP_BIT(usage);
+		opt->dump |= DUMP_BIT(help);
 		goto switch_end;
 	    }
 	    else {
@@ -1149,7 +1211,7 @@ uscore_get(void)
     VALUE line;
 
     line = rb_lastline_get();
-    if (TYPE(line) != T_STRING) {
+    if (!RB_TYPE_P(line, T_STRING)) {
 	rb_raise(rb_eTypeError, "$_ value need to be String (%s given)",
 		 NIL_P(line) ? "nil" : rb_obj_classname(line));
     }
@@ -1167,11 +1229,9 @@ uscore_get(void)
  */
 
 static VALUE
-rb_f_sub(argc, argv)
-    int argc;
-    VALUE *argv;
+rb_f_sub(int argc, VALUE *argv)
 {
-    VALUE str = rb_funcall3(uscore_get(), rb_intern("sub"), argc, argv);
+    VALUE str = rb_funcall_passing_block(uscore_get(), rb_intern("sub"), argc, argv);
     rb_lastline_set(str);
     return str;
 }
@@ -1188,11 +1248,9 @@ rb_f_sub(argc, argv)
  */
 
 static VALUE
-rb_f_gsub(argc, argv)
-    int argc;
-    VALUE *argv;
+rb_f_gsub(int argc, VALUE *argv)
 {
-    VALUE str = rb_funcall3(uscore_get(), rb_intern("gsub"), argc, argv);
+    VALUE str = rb_funcall_passing_block(uscore_get(), rb_intern("gsub"), argc, argv);
     rb_lastline_set(str);
     return str;
 }
@@ -1210,7 +1268,7 @@ rb_f_gsub(argc, argv)
 static VALUE
 rb_f_chop(void)
 {
-    VALUE str = rb_funcall3(uscore_get(), rb_intern("chop"), 0, 0);
+    VALUE str = rb_funcall_passing_block(uscore_get(), rb_intern("chop"), 0, 0);
     rb_lastline_set(str);
     return str;
 }
@@ -1228,15 +1286,16 @@ rb_f_chop(void)
  */
 
 static VALUE
-rb_f_chomp(argc, argv)
-    int argc;
-    VALUE *argv;
+rb_f_chomp(int argc, VALUE *argv)
 {
-    VALUE str = rb_funcall3(uscore_get(), rb_intern("chomp"), argc, argv);
+    VALUE str = rb_funcall_passing_block(uscore_get(), rb_intern("chomp"), argc, argv);
     rb_lastline_set(str);
     return str;
 }
 
+/* blank function in dmyext.c or generated by enc/make_encmake.rb */
+extern void Init_enc(void);
+
 static VALUE
 process_options(int argc, char **argv, struct cmdline_options *opt)
 {
@@ -1248,13 +1307,13 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
     char fbuf[MAXPATHLEN];
     int i = (int)proc_options(argc, argv, opt, 0);
     rb_thread_t *th = GET_THREAD();
-    rb_env_t *env = 0;
+    VALUE toplevel_binding = Qundef;
 
     argc -= i;
     argv += i;
 
-    if (opt->dump & DUMP_BIT(usage)) {
-	usage(origarg.argv[0]);
+    if (opt->dump & (DUMP_BIT(usage)|DUMP_BIT(help))) {
+	usage(origarg.argv[0], (opt->dump & DUMP_BIT(help)));
 	return Qtrue;
     }
 
@@ -1274,9 +1333,12 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	    opt->intern.enc.name = int_enc_name;
     }
 
-    if (opt->dump & DUMP_BIT(version)) {
+    if (opt->src.enc.name)
+	rb_warning("-K is specified; it is for 1.8 compatibility and may cause odd behavior");
+
+    if (opt->dump & (DUMP_BIT(version) | DUMP_BIT(version_v))) {
 	ruby_show_version();
-	return Qtrue;
+	if (opt->dump & DUMP_BIT(version)) return Qtrue;
     }
     if (opt->dump & DUMP_BIT(copyright)) {
 	ruby_show_copyright();
@@ -1295,7 +1357,7 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	}
 	else {
 	    opt->script = argv[0];
-	    if (opt->script[0] == '\0') {
+	    if (!opt->script || opt->script[0] == '\0') {
 		opt->script = "-";
 	    }
 	    else if (opt->do_search) {
@@ -1323,6 +1385,7 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 #endif
 
     ruby_init_loadpath_safe(opt->safe_level);
+    Init_enc();
     rb_enc_find_index("encdb");
     lenc = rb_locale_encoding();
     rb_enc_associate(rb_progname, lenc);
@@ -1359,7 +1422,8 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	long i;
 	VALUE load_path = GET_VM()->load_path;
 	for (i = 0; i < RARRAY_LEN(load_path); ++i) {
-	    rb_enc_associate(RARRAY_PTR(load_path)[i], lenc);
+	    RARRAY_PTR(load_path)[i] =
+		rb_enc_associate(rb_str_dup(RARRAY_PTR(load_path)[i]), lenc);
 	}
     }
     if (!(opt->disable & DISABLE_BIT(gems))) {
@@ -1373,16 +1437,10 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
     ruby_set_argv(argc, argv);
     process_sflag(&opt->sflag);
 
-    {
-	/* set eval context */
-	VALUE toplevel_binding = rb_const_get(rb_cObject, rb_intern("TOPLEVEL_BINDING"));
-	rb_binding_t *bind;
-
-	GetBindingPtr(toplevel_binding, bind);
-	GetEnvPtr(bind->env, env);
-    }
+    toplevel_binding = rb_const_get(rb_cObject, rb_intern("TOPLEVEL_BINDING"));
 
 #define PREPARE_PARSE_MAIN(expr) do { \
+    rb_env_t *env = toplevel_context(toplevel_binding); \
     th->parse_in_eval--; \
     th->base_block = &env->block; \
     expr; \
@@ -1400,9 +1458,9 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	    eenc = lenc;
 	}
 	rb_enc_associate(opt->e_script, eenc);
-	rb_vm_set_progname(rb_progname = opt->script_name);
+        ruby_set_script_name(opt->script_name);
 	require_libraries(&opt->req_list);
-	rb_vm_set_progname(rb_progname = progname);
+        ruby_set_script_name(progname);
 
 	PREPARE_PARSE_MAIN({
 	    tree = rb_parser_compile_string(parser, opt->script, opt->e_script, 1);
@@ -1414,11 +1472,10 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	}
 
 	PREPARE_PARSE_MAIN({
-	    tree = load_file(parser, opt->script, 1, opt);
+	    tree = load_file(parser, opt->script_name, 1, opt);
 	});
     }
-    rb_progname = opt->script_name;
-    rb_vm_set_progname(rb_progname);
+    ruby_set_script_name(opt->script_name);
     if (opt->dump & DUMP_BIT(yydebug)) return Qtrue;
 
     if (opt->ext.enc.index >= 0) {
@@ -1476,8 +1533,9 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 
     PREPARE_PARSE_MAIN({
 	VALUE path = Qnil;
-	if (!opt->e_script && strcmp(opt->script, "-"))
+	if (!opt->e_script && strcmp(opt->script, "-")) {
 	    path = rb_realpath_internal(Qnil, opt->script_name, 1);
+	}
 	iseq = rb_iseq_new_main(tree, opt->script_name, path);
     });
 
@@ -1499,7 +1557,7 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 
 struct load_file_arg {
     VALUE parser;
-    const char *fname;
+    VALUE fname;
     int script;
     struct cmdline_options *opt;
 };
@@ -1510,7 +1568,9 @@ load_file_internal(VALUE arg)
     extern VALUE rb_stdin;
     struct load_file_arg *argp = (struct load_file_arg *)arg;
     VALUE parser = argp->parser;
-    const char *fname = argp->fname;
+    VALUE fname_v = rb_str_encode_ospath(argp->fname);
+    const char *fname = StringValueCStr(fname_v);
+    const char *orig_fname = StringValueCStr(argp->fname);
     int script = argp->script;
     struct cmdline_options *opt = argp->opt;
     VALUE f;
@@ -1520,8 +1580,6 @@ load_file_internal(VALUE arg)
     ID set_encoding;
     int xflag = 0;
 
-    if (!fname)
-	rb_load_fail(fname);
     if (strcmp(fname, "-") == 0) {
 	f = rb_stdin;
     }
@@ -1536,11 +1594,21 @@ load_file_internal(VALUE arg)
 	    }
 	}
 #endif
-	if ((fd = open(fname, mode)) < 0) {
-	    rb_load_fail(fname);
+	if ((fd = rb_cloexec_open(fname, mode, 0)) < 0) {
+	    rb_load_fail(fname_v, strerror(errno));
 	}
         rb_update_max_fd(fd);
-
+#if !defined DOSISH && !defined __CYGWIN__
+	{
+	    struct stat st;
+	    if (fstat(fd, &st) != 0)
+		rb_load_fail(fname_v, strerror(errno));
+	    if (S_ISDIR(st.st_mode)) {
+		errno = EISDIR;
+		rb_load_fail(fname_v, strerror(EISDIR));
+	    }
+	}
+#endif
 	f = rb_io_fdopen(fd, mode, fname);
     }
 
@@ -1571,7 +1639,7 @@ load_file_internal(VALUE arg)
 		    }
 		}
 	    }
-	    rb_raise(rb_eLoadError, "no Ruby script found in input");
+	    rb_loaderror("no Ruby script found in input");
 	}
 
 	c = rb_io_getbyte(f);
@@ -1621,7 +1689,7 @@ load_file_internal(VALUE arg)
 	    if (f != rb_stdin) rb_io_close(f);
 	    f = Qnil;
 	}
-	rb_vm_set_progname(rb_progname = opt->script_name);
+        ruby_set_script_name(opt->script_name);
 	require_libraries(&opt->req_list);	/* Why here? unnatural */
     }
     if (opt->src.enc.index >= 0) {
@@ -1631,15 +1699,15 @@ load_file_internal(VALUE arg)
 	enc = rb_locale_encoding();
     }
     else {
-	enc = rb_usascii_encoding();
+	enc = rb_utf8_encoding();
     }
     if (NIL_P(f)) {
 	f = rb_str_new(0, 0);
 	rb_enc_associate(f, enc);
-	return (VALUE)rb_parser_compile_string(parser, fname, f, line_start);
+	return (VALUE)rb_parser_compile_string(parser, orig_fname, f, line_start);
     }
     rb_funcall(f, set_encoding, 2, rb_enc_from_encoding(enc), rb_str_new_cstr("-"));
-    tree = rb_parser_compile_file(parser, fname, f, line_start);
+    tree = rb_parser_compile_file(parser, orig_fname, f, line_start);
     rb_funcall(f, set_encoding, 1, rb_parser_encoding(parser));
     if (script && tree && rb_parser_end_seen_p(parser)) {
 	/*
@@ -1669,7 +1737,7 @@ restore_lineno(VALUE lineno)
 }
 
 static NODE *
-load_file(VALUE parser, const char *fname, int script, struct cmdline_options *opt)
+load_file(VALUE parser, VALUE fname, int script, struct cmdline_options *opt)
 {
     struct load_file_arg arg;
     arg.parser = parser;
@@ -1683,8 +1751,9 @@ void *
 rb_load_file(const char *fname)
 {
     struct cmdline_options opt;
+    VALUE fname_v = rb_str_new_cstr(fname);
 
-    return load_file(rb_parser_new(), fname, 0, cmdline_options_init(&opt));
+    return load_file(rb_parser_new(), fname_v, 0, cmdline_options_init(&opt));
 }
 
 static void
@@ -1704,6 +1773,11 @@ set_arg0(VALUE val, ID id)
     rb_progname = rb_obj_freeze(rb_external_str_new(s, i));
 }
 
+/*! Sets the current script name to this value.
+ *
+ * This is similiar to <code>$0 = name</code> in Ruby level but also affects
+ * <code>Method#location</code> and others.
+ */
 void
 ruby_script(const char *name)
 {
@@ -1713,6 +1787,17 @@ ruby_script(const char *name)
     }
 }
 
+/*! Sets the current script name to this value.
+ *
+ * Same as ruby_script() but accepts a VALUE.
+ */
+void
+ruby_set_script_name(VALUE name)
+{
+    rb_progname = rb_str_dup(name);
+    rb_vm_set_progname(rb_progname);
+}
+
 static void
 init_ids(struct cmdline_options *opt)
 {
@@ -1758,10 +1843,12 @@ opt_W_getter(ID id, void *data)
 	return INT2FIX(1);
       case Qtrue:
 	return INT2FIX(2);
+      default:
+	return Qnil;
     }
-    return Qnil;		/* not reached */
 }
 
+/*! Defines built-in variables */
 void
 ruby_prog_init(void)
 {
@@ -1811,8 +1898,9 @@ ruby_process_options(int argc, char **argv)
 {
     struct cmdline_options opt;
     VALUE iseq;
+    const char *script_name = (argc > 0 && argv[0]) ? argv[0] : "ruby";
 
-    ruby_script(argv[0]);  /* for the time being */
+    ruby_script(script_name);  /* for the time being */
     rb_argv0 = rb_str_new4(rb_progname);
     rb_gc_register_mark_object(rb_argv0);
     iseq = process_options(argc, argv, cmdline_options_init(&opt));
@@ -1857,6 +1945,12 @@ fill_standard_fds(void)
     }
 }
 
+/*! Initializes the process for ruby(1).
+ *
+ * This function assumes this process is ruby(1) and it has just started.
+ * Usually programs that embeds CRuby interpreter should not call this function,
+ * and should do their own initialization.
+ */
 void
 ruby_sysinit(int *argc, char ***argv)
 {
