@@ -17,8 +17,13 @@
 #include <stdio.h>
 #include <errno.h>
 #include "ruby_atomic.h"
+#include "eval_intern.h"
 
-#if !defined(_WIN32) && !defined(HAVE_GCC_ATOMIC_BUILTINS)
+#if defined(__native_client__) && defined(NACL_NEWLIB)
+# include "nacl/signal.h"
+#endif
+
+#ifdef NEED_RUBY_ATOMIC_OPS
 rb_atomic_t
 ruby_atomic_exchange(rb_atomic_t *ptr, rb_atomic_t val)
 {
@@ -26,18 +31,23 @@ ruby_atomic_exchange(rb_atomic_t *ptr, rb_atomic_t val)
     *ptr = val;
     return old;
 }
+
+rb_atomic_t
+ruby_atomic_compare_and_swap(rb_atomic_t *ptr, rb_atomic_t cmp,
+			     rb_atomic_t newval)
+{
+    rb_atomic_t old = *ptr;
+    if (old == cmp) {
+      *ptr = newval;
+    }
+    return old;
+}
 #endif
 
 #if defined(__BEOS__) || defined(__HAIKU__)
 #undef SIGBUS
 #endif
 
-#ifdef HAVE_PTHREAD_SIGMASK
-#define USE_TRAP_MASK 1
-#else
-#define USE_TRAP_MASK 0
-#endif
-
 #ifndef NSIG
 # define NSIG (_SIGMAX + 1)      /* For QNX */
 #endif
@@ -203,6 +213,26 @@ signo2signm(int no)
     return 0;
 }
 
+/*
+ * call-seq:
+ *     Signal.signame(signo)  ->  string
+ *
+ *  convert signal number to signal name
+ *
+ *     Signal.trap("INT") { |signo| puts Signal.signame(signo) }
+ *     Process.kill("INT", 0)
+ *
+ *  <em>produces:</em>
+ *
+ *     INT
+ */
+static VALUE
+sig_signame(VALUE recv, VALUE signo)
+{
+    const char *signame = signo2signm(NUM2INT(signo));
+    return rb_str_new_cstr(signame);
+}
+
 const char *
 ruby_signal_name(int no)
 {
@@ -231,10 +261,7 @@ esignal_init(int argc, VALUE *argv, VALUE self)
 	if (!NIL_P(sig)) argnum = 2;
 	else sig = argv[0];
     }
-    if (argc < 1 || argnum < argc) {
-	rb_raise(rb_eArgError, "wrong number of arguments (%d for %d)",
-		 argc, argnum);
-    }
+    rb_check_arity(argc, 1, argnum);
     if (argnum == 2) {
 	signo = NUM2INT(sig);
 	if (signo < 0 || signo > NSIG) {
@@ -303,11 +330,11 @@ ruby_default_signal(int sig)
  *  call-seq:
  *     Process.kill(signal, pid, ...)    -> fixnum
  *
- *  Sends the given signal to the specified process id(s), or to the
- *  current process if _pid_ is zero. _signal_ may be an
- *  integer signal number or a POSIX signal name (either with or without
- *  a +SIG+ prefix). If _signal_ is negative (or starts
- *  with a minus sign), kills process groups instead of
+ *  Sends the given signal to the specified process id(s) if _pid_ is positive.
+ *  If _pid_ is zero _signal_ is sent to all processes whose group ID is equal
+ *  to the group ID of the process. _signal_ may be an integer signal number or
+ *  a POSIX signal name (either with or without a +SIG+ prefix). If _signal_ is
+ *  negative (or starts with a minus sign), kills process groups instead of
  *  processes. Not all signals are available on all platforms.
  *
  *     pid = fork do
@@ -336,7 +363,7 @@ ruby_default_signal(int sig)
 VALUE
 rb_f_kill(int argc, VALUE *argv)
 {
-#ifndef HAS_KILLPG
+#ifndef HAVE_KILLPG
 #define killpg(pg, sig) kill(-(pg), (sig))
 #endif
     int negative = 0;
@@ -346,8 +373,8 @@ rb_f_kill(int argc, VALUE *argv)
     const char *s;
 
     rb_secure(2);
-    if (argc < 2)
-	rb_raise(rb_eArgError, "wrong number of arguments (%d for at least 2)", argc);
+    rb_check_arity(argc, 2, UNLIMITED_ARGUMENTS);
+
     switch (TYPE(argv[0])) {
       case T_FIXNUM:
 	sig = FIX2INT(argv[0]);
@@ -367,7 +394,7 @@ rb_f_kill(int argc, VALUE *argv)
 	}
 	if (strncmp("SIG", s, 3) == 0)
 	    s += 3;
-	if((sig = signm2signo(s)) == 0)
+	if ((sig = signm2signo(s)) == 0)
 	    rb_raise(rb_eArgError, "unsupported name `SIG%s'", s);
 
 	if (negative)
@@ -398,7 +425,6 @@ rb_f_kill(int argc, VALUE *argv)
 		rb_sys_fail(0);
 	}
     }
-    rb_thread_polling();
     return INT2FIX(i-1);
 }
 
@@ -420,9 +446,28 @@ typedef RETSIGTYPE ruby_sigaction_t(int);
 #define SIGINFO_ARG
 #endif
 
-#ifdef POSIX_SIGNAL
-
 #ifdef USE_SIGALTSTACK
+int rb_sigaltstack_size(void)
+{
+    /* XXX: BSD_vfprintf() uses >1500KiB stack and x86-64 need >5KiB stack. */
+    int size = 8192;
+
+#ifdef MINSIGSTKSZ
+    if (size < MINSIGSTKSZ)
+	size = MINSIGSTKSZ;
+#endif
+#if defined(HAVE_SYSCONF) && defined(_SC_PAGE_SIZE)
+    {
+	int pagesize;
+	pagesize = (int)sysconf(_SC_PAGE_SIZE);
+	if (size < pagesize)
+	    size = pagesize;
+    }
+#endif
+
+    return size;
+}
+
 /* alternate stack for SIGSEGV */
 void
 rb_register_sigaltstack(rb_thread_t *th)
@@ -433,13 +478,14 @@ rb_register_sigaltstack(rb_thread_t *th)
 	rb_bug("rb_register_sigaltstack: th->altstack not initialized\n");
 
     newSS.ss_sp = th->altstack;
-    newSS.ss_size = ALT_STACK_SIZE;
+    newSS.ss_size = rb_sigaltstack_size();
     newSS.ss_flags = 0;
 
     sigaltstack(&newSS, &oldSS); /* ignore error. */
 }
 #endif /* USE_SIGALTSTACK */
 
+#ifdef POSIX_SIGNAL
 static sighandler_t
 ruby_signal(int signum, sighandler_t handler)
 {
@@ -463,7 +509,11 @@ ruby_signal(int signum, sighandler_t handler)
 	sigact.sa_flags |= SA_NOCLDWAIT;
 #endif
 #if defined(SA_ONSTACK) && defined(USE_SIGALTSTACK)
-    if (signum == SIGSEGV || signum == SIGBUS)
+    if (signum == SIGSEGV
+#ifdef SIGBUS
+	|| signum == SIGBUS
+#endif
+       )
 	sigact.sa_flags |= SA_ONSTACK;
 #endif
     if (sigaction(signum, &sigact, &old) < 0) {
@@ -471,7 +521,10 @@ ruby_signal(int signum, sighandler_t handler)
 	    rb_bug_errno("sigaction", errno);
 	}
     }
-    return old.sa_handler;
+    if (old.sa_flags & SA_SIGINFO)
+	return (sighandler_t)old.sa_sigaction;
+    else
+	return old.sa_handler;
 }
 
 sighandler_t
@@ -512,30 +565,24 @@ rb_signal_buff_size(void)
     return signal_buff.size;
 }
 
-#if USE_TRAP_MASK
-static sigset_t trap_last_mask;
-#endif
-
 #if HAVE_PTHREAD_H
 #include <pthread.h>
 #endif
 
-void
+static void
 rb_disable_interrupt(void)
 {
-#if USE_TRAP_MASK
+#ifdef HAVE_PTHREAD_SIGMASK
     sigset_t mask;
     sigfillset(&mask);
-    sigdelset(&mask, SIGVTALRM);
-    sigdelset(&mask, SIGSEGV);
     pthread_sigmask(SIG_SETMASK, &mask, NULL);
 #endif
 }
 
-void
+static void
 rb_enable_interrupt(void)
 {
-#if USE_TRAP_MASK
+#ifdef HAVE_PTHREAD_SIGMASK
     sigset_t mask;
     sigemptyset(&mask);
     pthread_sigmask(SIG_SETMASK, &mask, NULL);
@@ -550,12 +597,8 @@ rb_get_next_signal(void)
     if (signal_buff.size != 0) {
 	for (i=1; i<RUBY_NSIG; i++) {
 	    if (signal_buff.cnt[i] > 0) {
-		rb_disable_interrupt();
-		{
-		    ATOMIC_DEC(signal_buff.cnt[i]);
-		    ATOMIC_DEC(signal_buff.size);
-		}
-		rb_enable_interrupt();
+		ATOMIC_DEC(signal_buff.cnt[i]);
+		ATOMIC_DEC(signal_buff.size);
 		sig = i;
 		break;
 	    }
@@ -565,18 +608,22 @@ rb_get_next_signal(void)
 }
 
 
-#ifdef USE_SIGALTSTACK
+#if defined(USE_SIGALTSTACK) || defined(_WIN32)
 static void
 check_stack_overflow(const void *addr)
 {
     int ruby_stack_overflowed_p(const rb_thread_t *, const void *);
     NORETURN(void ruby_thread_stack_overflow(rb_thread_t *th));
-    rb_thread_t *th = GET_THREAD();
+    rb_thread_t *th = ruby_current_thread;
     if (ruby_stack_overflowed_p(th, addr)) {
 	ruby_thread_stack_overflow(th);
     }
 }
+#ifdef _WIN32
+#define CHECK_STACK_OVERFLOW() check_stack_overflow(0)
+#else
 #define CHECK_STACK_OVERFLOW() check_stack_overflow(info->si_addr)
+#endif
 #else
 #define CHECK_STACK_OVERFLOW() (void)0
 #endif
@@ -618,8 +665,10 @@ static RETSIGTYPE
 sigsegv(int sig SIGINFO_ARG)
 {
     if (segv_received) {
+	ssize_t RB_UNUSED_VAR(err);
 	char msg[] = "SEGV received in SEGV handler\n";
-	write(2, msg, sizeof(msg));
+
+	err = write(2, msg, sizeof(msg));
 	ruby_abort();
     }
 
@@ -634,8 +683,33 @@ sigsegv(int sig SIGINFO_ARG)
 static void
 signal_exec(VALUE cmd, int safe, int sig)
 {
-    VALUE signum = INT2NUM(sig);
-    rb_eval_cmd(cmd, rb_ary_new3(1, signum), safe);
+    rb_thread_t *cur_th = GET_THREAD();
+    volatile unsigned long old_interrupt_mask = cur_th->interrupt_mask;
+    int state;
+
+    /*
+     * workaround the following race:
+     * 1. signal_enque queues signal for execution
+     * 2. user calls trap(sig, "IGNORE"), setting SIG_IGN
+     * 3. rb_signal_exec runs on queued signal
+     */
+    if (IMMEDIATE_P(cmd))
+	return;
+
+    cur_th->interrupt_mask |= TRAP_INTERRUPT_MASK;
+    TH_PUSH_TAG(cur_th);
+    if ((state = EXEC_TAG()) == 0) {
+	VALUE signum = INT2NUM(sig);
+	rb_eval_cmd(cmd, rb_ary_new3(1, signum), safe);
+    }
+    TH_POP_TAG();
+    cur_th = GET_THREAD();
+    cur_th->interrupt_mask = old_interrupt_mask;
+
+    if (state) {
+	/* XXX: should be replaced with rb_threadptr_pending_interrupt_enque() */
+	JUMP_TAG(state);
+    }
 }
 
 void
@@ -692,15 +766,6 @@ rb_signal_exec(rb_thread_t *th, int sig)
     }
 }
 
-struct trap_arg {
-#if USE_TRAP_MASK
-    sigset_t mask;
-#endif
-    int sig;
-    sighandler_t func;
-    VALUE cmd;
-};
-
 static sighandler_t
 default_handler(int sig)
 {
@@ -735,9 +800,6 @@ default_handler(int sig)
 #ifdef SIGSEGV
       case SIGSEGV:
         func = (sighandler_t)sigsegv;
-# ifdef USE_SIGALTSTACK
-        rb_register_sigaltstack(GET_THREAD());
-# endif
         break;
 #endif
 #ifdef SIGPIPE
@@ -785,7 +847,7 @@ trap_handler(VALUE *cmd, int sig)
 		if (strncmp(RSTRING_PTR(command), "SIG_IGN", 7) == 0) {
 sig_ign:
                     func = SIG_IGN;
-                    *cmd = 0;
+                    *cmd = Qtrue;
 		}
 		else if (strncmp(RSTRING_PTR(command), "SIG_DFL", 7) == 0) {
 sig_dfl:
@@ -811,6 +873,7 @@ sig_dfl:
 	else {
 	    rb_proc_t *proc;
 	    GetProcPtr(*cmd, proc);
+	    (void)proc;
 	}
     }
 
@@ -850,21 +913,28 @@ trap_signm(VALUE vsig)
 }
 
 static VALUE
-trap(struct trap_arg *arg)
+trap(int sig, sighandler_t func, VALUE command)
 {
-    sighandler_t oldfunc, func = arg->func;
-    VALUE oldcmd, command = arg->cmd;
-    int sig = arg->sig;
+    sighandler_t oldfunc;
+    VALUE oldcmd;
     rb_vm_t *vm = GET_VM();
 
+    /*
+     * Be careful. ruby_signal() and trap_list[sig].cmd must be changed
+     * atomically. In current implementation, we only need to don't call
+     * RUBY_VM_CHECK_INTS().
+     */
     oldfunc = ruby_signal(sig, func);
     oldcmd = vm->trap_list[sig].cmd;
     switch (oldcmd) {
       case 0:
+      case Qtrue:
 	if (oldfunc == SIG_IGN) oldcmd = rb_str_new2("IGNORE");
 	else if (oldfunc == sighandler) oldcmd = rb_str_new2("DEFAULT");
 	else oldcmd = Qnil;
 	break;
+      case Qnil:
+	break;
       case Qundef:
 	oldcmd = rb_str_new2("EXIT");
 	break;
@@ -872,30 +942,38 @@ trap(struct trap_arg *arg)
 
     vm->trap_list[sig].cmd = command;
     vm->trap_list[sig].safe = rb_safe_level();
-    /* enable at least specified signal. */
-#if USE_TRAP_MASK
-    sigdelset(&arg->mask, sig);
-#endif
+
     return oldcmd;
 }
 
-#if USE_TRAP_MASK
-static VALUE
-trap_ensure(struct trap_arg *arg)
+static int
+reserved_signal_p(int signo)
 {
-    /* enable interrupt */
-    pthread_sigmask(SIG_SETMASK, &arg->mask, NULL);
-    trap_last_mask = arg->mask;
-    return 0;
-}
+/* Synchronous signal can't deliver to main thread */
+#ifdef SIGSEGV
+    if (signo == SIGSEGV)
+	return 1;
+#endif
+#ifdef SIGBUS
+    if (signo == SIGBUS)
+	return 1;
+#endif
+#ifdef SIGILL
+    if (signo == SIGILL)
+	return 1;
+#endif
+#ifdef SIGFPE
+    if (signo == SIGFPE)
+	return 1;
 #endif
 
-void
-rb_trap_restore_mask(void)
-{
-#if USE_TRAP_MASK
-    pthread_sigmask(SIG_SETMASK, &trap_last_mask, NULL);
+/* used ubf internal see thread_pthread.c. */
+#ifdef SIGVTALRM
+    if (signo == SIGVTALRM)
+	return 1;
 #endif
+
+    return 0;
 }
 
 /*
@@ -932,39 +1010,36 @@ rb_trap_restore_mask(void)
 static VALUE
 sig_trap(int argc, VALUE *argv)
 {
-    struct trap_arg arg;
+    int sig;
+    sighandler_t func;
+    VALUE cmd;
 
     rb_secure(2);
-    if (argc < 1 || argc > 2) {
-	rb_raise(rb_eArgError, "wrong number of arguments (%d for 1..2)", argc);
+    rb_check_arity(argc, 1, 2);
+
+    sig = trap_signm(argv[0]);
+    if (reserved_signal_p(sig)) {
+        const char *name = signo2signm(sig);
+        if (name)
+            rb_raise(rb_eArgError, "can't trap reserved signal: SIG%s", name);
+        else
+            rb_raise(rb_eArgError, "can't trap reserved signal: %d", sig);
     }
 
-    arg.sig = trap_signm(argv[0]);
     if (argc == 1) {
-	arg.cmd = rb_block_proc();
-	arg.func = sighandler;
+	cmd = rb_block_proc();
+	func = sighandler;
     }
     else {
-	arg.cmd = argv[1];
-	arg.func = trap_handler(&arg.cmd, arg.sig);
+	cmd = argv[1];
+	func = trap_handler(&cmd, sig);
     }
 
-    if (OBJ_TAINTED(arg.cmd)) {
+    if (OBJ_TAINTED(cmd)) {
 	rb_raise(rb_eSecurityError, "Insecure: tainted signal trap");
     }
-#if USE_TRAP_MASK
-    {
-      sigset_t fullmask;
 
-      /* disable interrupt */
-      sigfillset(&fullmask);
-      pthread_sigmask(SIG_BLOCK, &fullmask, &arg.mask);
-
-      return rb_ensure(trap, (VALUE)&arg, trap_ensure, (VALUE)&arg);
-    }
-#else
-    return trap(&arg);
-#endif
+    return trap(sig, func, cmd);
 }
 
 /*
@@ -993,10 +1068,14 @@ install_sighandler(int signum, sighandler_t handler)
 {
     sighandler_t old;
 
+    /* At this time, there is no subthread. Then sigmask guarantee atomics. */
+    rb_disable_interrupt();
     old = ruby_signal(signum, handler);
+    /* signal handler should be inherited during exec. */
     if (old != SIG_DFL) {
 	ruby_signal(signum, old);
     }
+    rb_enable_interrupt();
 }
 
 #if defined(SIGCLD) || defined(SIGCHLD)
@@ -1004,27 +1083,15 @@ static void
 init_sigchld(int sig)
 {
     sighandler_t oldfunc;
-#if USE_TRAP_MASK
-    sigset_t mask;
-    sigset_t fullmask;
-
-    /* disable interrupt */
-    sigfillset(&fullmask);
-    pthread_sigmask(SIG_BLOCK, &fullmask, &mask);
-#endif
 
+    rb_disable_interrupt();
     oldfunc = ruby_signal(sig, SIG_DFL);
     if (oldfunc != SIG_DFL && oldfunc != SIG_IGN) {
 	ruby_signal(sig, oldfunc);
     } else {
 	GET_VM()->trap_list[sig].cmd = 0;
     }
-
-#if USE_TRAP_MASK
-    sigdelset(&mask, sig);
-    pthread_sigmask(SIG_SETMASK, &mask, NULL);
-    trap_last_mask = mask;
-#endif
+    rb_enable_interrupt();
 }
 #endif
 
@@ -1090,6 +1157,7 @@ Init_signal(void)
     rb_define_global_function("trap", sig_trap, -1);
     rb_define_module_function(mSignal, "trap", sig_trap, -1);
     rb_define_module_function(mSignal, "list", sig_list, 0);
+    rb_define_module_function(mSignal, "signame", sig_signame, 1);
 
     rb_define_method(rb_eSignal, "initialize", esignal_init, -1);
     rb_define_method(rb_eSignal, "signo", esignal_signo, 0);
