@@ -2,7 +2,7 @@
 
   sprintf.c -
 
-  $Author: yugui $
+  $Author: nagachika $
   created at: Fri Oct 15 10:39:26 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -30,9 +30,7 @@ static void fmt_setup(char*,size_t,int,int,int,int);
 static char*
 remove_sign_bits(char *str, int base)
 {
-    char *s, *t;
-
-    s = t = str;
+    char *t = str;
 
     if (base == 16) {
 	while (*t == 'f') {
@@ -94,13 +92,13 @@ sign_bits(int base, const char *p)
 
 #define PUSH(s, l) do { \
     CHECK(l);\
-    memcpy(&buf[blen], s, l);\
+    memcpy(&buf[blen], (s), (l));\
     blen += (l);\
 } while (0)
 
 #define FILL(c, l) do { \
     CHECK(l);\
-    memset(&buf[blen], c, l);\
+    memset(&buf[blen], (c), (l));\
     blen += (l);\
 } while (0)
 
@@ -112,29 +110,29 @@ sign_bits(int base, const char *p)
     (posarg = nextarg++, GETNTHARG(posarg)))
 
 #define GETPOSARG(n) (posarg > 0 ? \
-    (rb_raise(rb_eArgError, "numbered(%d) after unnumbered(%d)", n, posarg), 0) : \
+    (rb_raise(rb_eArgError, "numbered(%d) after unnumbered(%d)", (n), posarg), 0) : \
     posarg == -2 ? \
-    (rb_raise(rb_eArgError, "numbered(%d) after named", n), 0) : \
-    ((n < 1) ? (rb_raise(rb_eArgError, "invalid index - %d$", n), 0) : \
+    (rb_raise(rb_eArgError, "numbered(%d) after named", (n)), 0) : \
+    (((n) < 1) ? (rb_raise(rb_eArgError, "invalid index - %d$", (n)), 0) : \
 	       (posarg = -1, GETNTHARG(n))))
 
 #define GETNTHARG(nth) \
-    ((nth >= argc) ? (rb_raise(rb_eArgError, "too few arguments"), 0) : argv[nth])
+    (((nth) >= argc) ? (rb_raise(rb_eArgError, "too few arguments"), 0) : argv[(nth)])
 
-#define GETNAMEARG(id, name, len) ( \
+#define GETNAMEARG(id, name, len, enc) ( \
     posarg > 0 ? \
-    (rb_raise(rb_eArgError, "named%.*s after unnumbered(%d)", (len), (name), posarg), 0) : \
+    (rb_enc_raise((enc), rb_eArgError, "named%.*s after unnumbered(%d)", (len), (name), posarg), 0) : \
     posarg == -1 ? \
-    (rb_raise(rb_eArgError, "named%.*s after numbered", (len), (name)), 0) :	\
-    (posarg = -2, rb_hash_lookup2(get_hash(&hash, argc, argv), id, Qundef)))
+    (rb_enc_raise((enc), rb_eArgError, "named%.*s after numbered", (len), (name)), 0) :	\
+    (posarg = -2, rb_hash_lookup2(get_hash(&hash, argc, argv), (id), Qundef)))
 
 #define GETNUM(n, val) \
     for (; p < end && rb_enc_isdigit(*p, enc); p++) {	\
-	int next_n = 10 * n + (*p - '0'); \
-        if (next_n / 10 != n) {\
+	int next_n = 10 * (n) + (*p - '0'); \
+        if (next_n / 10 != (n)) {\
 	    rb_raise(rb_eArgError, #val " too big"); \
 	} \
-	n = next_n; \
+	(n) = next_n; \
     } \
     if (p >= end) { \
 	rb_raise(rb_eArgError, "malformed format string - %%*[0-9]"); \
@@ -143,7 +141,7 @@ sign_bits(int base, const char *p)
 #define GETASTER(val) do { \
     t = p++; \
     n = 0; \
-    GETNUM(n, val); \
+    GETNUM(n, (val)); \
     if (*p == '$') { \
 	tmp = GETPOSARG(n); \
     } \
@@ -151,7 +149,7 @@ sign_bits(int base, const char *p)
 	tmp = GETARG(); \
 	p = t; \
     } \
-    val = NUM2INT(tmp); \
+    (val) = NUM2INT(tmp); \
 } while (0)
 
 static VALUE
@@ -163,7 +161,7 @@ get_hash(volatile VALUE *hash, int argc, const VALUE *argv)
     if (argc != 2) {
 	rb_raise(rb_eArgError, "one hash required");
     }
-    tmp = rb_check_convert_type(argv[1], T_HASH, "Hash", "to_hash");
+    tmp = rb_check_hash_type(argv[1]);
     if (NIL_P(tmp)) {
 	rb_raise(rb_eArgError, "one hash required");
     }
@@ -424,7 +422,7 @@ get_hash(volatile VALUE *hash, int argc, const VALUE *argv)
  *  For more complex formatting, Ruby supports a reference by name.
  *  %<name>s style uses format style, but %{name} style doesn't.
  *
- *  Exapmles:
+ *  Examples:
  *    sprintf("%<foo>d : %<bar>f", { :foo => 1, :bar => 2 })
  *      #=> 1 : 2.000000
  *    sprintf("%{foo}f", { :foo => 1 })
@@ -568,6 +566,7 @@ rb_str_format(int argc, const VALUE *argv, VALUE fmt)
 	    {
 		const char *start = p;
 		char term = (*p == '<') ? '>' : '}';
+		int len;
 
 		for (; p < end && *p != term; ) {
 		    p += rb_enc_mbclen(p, end, enc);
@@ -575,14 +574,27 @@ rb_str_format(int argc, const VALUE *argv, VALUE fmt)
 		if (p >= end) {
 		    rb_raise(rb_eArgError, "malformed name - unmatched parenthesis");
 		}
-		if (id) {
-		    rb_raise(rb_eArgError, "name%.*s after <%s>",
-			     (int)(p - start + 1), start, rb_id2name(id));
+#if SIZEOF_INT < SIZEOF_SIZE_T
+		if ((size_t)(p - start) >= INT_MAX) {
+		    const int message_limit = 20;
+		    len = (int)(rb_enc_right_char_head(start, start + message_limit, p, enc) - start);
+		    rb_enc_raise(enc, rb_eArgError,
+				 "too long name (%"PRIdSIZE" bytes) - %.*s...%c",
+				 (size_t)(p - start - 2), len, start, term);
 		}
-		id = rb_intern3(start + 1, p - start - 1, enc);
-		nextvalue = GETNAMEARG(ID2SYM(id), start, (int)(p - start + 1));
+#endif
+		len = (int)(p - start + 1); /* including parenthesis */
+		if (id) {
+		    rb_enc_raise(enc, rb_eArgError, "named%.*s after <%s>",
+				 len, start, rb_id2name(id));
+		}
+		nextvalue = GETNAMEARG((id = rb_check_id_cstr(start + 1,
+							      len - 2 /* without parenthesis */,
+							      enc),
+					ID2SYM(id)),
+				       start, len, enc);
 		if (nextvalue == Qundef) {
-		    rb_raise(rb_eKeyError, "key%.*s not found", (int)(p - start + 1), start);
+		    rb_enc_raise(enc, rb_eKeyError, "key%.*s not found", len, start);
 		}
 		if (term == '}') goto format_s;
 		p++;
@@ -643,6 +655,7 @@ rb_str_format(int argc, const VALUE *argv, VALUE fmt)
 			rb_raise(rb_eArgError, "%%c requires a character");
 		    }
 		    c = rb_enc_codepoint_len(RSTRING_PTR(tmp), RSTRING_END(tmp), &n, enc);
+		    RB_GC_GUARD(tmp);
 		}
 		else {
 		    c = NUM2INT(val);
@@ -713,6 +726,7 @@ rb_str_format(int argc, const VALUE *argv, VALUE fmt)
 			}
 			CHECK(len);
 			memcpy(&buf[blen], RSTRING_PTR(str), len);
+			RB_GC_GUARD(str);
 			blen += len;
 			if (flags&FMINUS) {
 			    CHECK(width);
@@ -725,6 +739,7 @@ rb_str_format(int argc, const VALUE *argv, VALUE fmt)
 		    }
 		}
 		PUSH(RSTRING_PTR(str), len);
+		RB_GC_GUARD(str);
 		rb_enc_associate(result, enc);
 	    }
 	    break;
@@ -738,7 +753,6 @@ rb_str_format(int argc, const VALUE *argv, VALUE fmt)
 	  case 'B':
 	  case 'u':
 	    {
-		volatile VALUE tmp1;
 		volatile VALUE val = GETARG();
 		char fbuf[32], nbuf[64], *s;
 		const char *prefix = 0;
@@ -746,7 +760,7 @@ rb_str_format(int argc, const VALUE *argv, VALUE fmt)
 		char sc = 0;
 		long v = 0;
 		int base, bignum = 0;
-		int len, pos;
+		int len;
 
 		switch (*p) {
 		  case 'd':
@@ -889,7 +903,7 @@ rb_str_format(int argc, const VALUE *argv, VALUE fmt)
 			    val = rb_big_clone(val);
 			    rb_big_2comp(val);
 			}
-			tmp1 = tmp = rb_big2str0(val, base, RBIGNUM_SIGN(val));
+			tmp = rb_big2str0(val, base, RBIGNUM_SIGN(val));
 			s = RSTRING_PTR(tmp);
 			if (*s == '-') {
 			    dots = 1;
@@ -910,7 +924,6 @@ rb_str_format(int argc, const VALUE *argv, VALUE fmt)
 		    len = rb_long2int(RSTRING_END(tmp) - s);
 		}
 
-		pos = -1;
 		if (dots) {
 		    prec -= 2;
 		    width -= 2;
@@ -984,6 +997,7 @@ rb_str_format(int argc, const VALUE *argv, VALUE fmt)
 		    }
 		}
 		PUSH(s, len);
+		RB_GC_GUARD(tmp);
 		CHECK(width);
 		while (width-- > 0) {
 		    buf[blen++] = ' ';
@@ -1068,6 +1082,7 @@ rb_str_format(int argc, const VALUE *argv, VALUE fmt)
     }
 
   sprint_exit:
+    RB_GC_GUARD(fmt);
     /* XXX - We cannot validate the number of arguments if (digit)$ style used.
      */
     if (posarg >= 0 && nextarg < argc) {
@@ -1121,11 +1136,21 @@ fmt_setup(char *buf, size_t size, int c, int flags, int width, int prec)
 #  define quad_t LONG_LONG
 #  define u_quad_t unsigned LONG_LONG
 # endif
+#elif SIZEOF_LONG != SIZEOF_LONG_LONG && SIZEOF_LONG_LONG == 8
+# define _HAVE_SANE_QUAD_
+# define quad_t LONG_LONG
+# define u_quad_t unsigned LONG_LONG
 #endif
 #define FLOATING_POINT 1
 #define BSD__dtoa ruby_dtoa
+#define BSD__hdtoa ruby_hdtoa
 #include "vsnprintf.c"
 
+typedef struct {
+    rb_printf_buffer base;
+    volatile VALUE value;
+} rb_printf_buffer_extra;
+
 static int
 ruby__sfvwrite(register rb_printf_buffer *fp, register struct __suio *uio)
 {
@@ -1149,27 +1174,75 @@ ruby__sfvwrite(register rb_printf_buffer *fp, register struct __suio *uio)
 	len -= n;
     }
     fp->_p = (unsigned char *)buf;
+    rb_str_set_len(result, buf - RSTRING_PTR(result));
     return 0;
 }
 
+static char *
+ruby__sfvextra(rb_printf_buffer *fp, size_t valsize, void *valp, long *sz, int sign)
+{
+    VALUE value, result = (VALUE)fp->_bf._base;
+    rb_encoding *enc;
+    char *cp;
+
+    if (valsize != sizeof(VALUE)) return 0;
+    value = *(VALUE *)valp;
+    if (RBASIC(result)->klass) {
+	rb_raise(rb_eRuntimeError, "rb_vsprintf reentered");
+    }
+    if (sign == '+') {
+	value = rb_inspect(value);
+    }
+    else {
+	value = rb_obj_as_string(value);
+    }
+    enc = rb_enc_compatible(result, value);
+    if (enc) {
+	rb_enc_associate(result, enc);
+    }
+    else {
+	enc = rb_enc_get(result);
+	value = rb_str_conv_enc_opts(value, rb_enc_get(value), enc,
+				     ECONV_UNDEF_REPLACE|ECONV_INVALID_REPLACE,
+				     Qnil);
+	*(volatile VALUE *)valp = value;
+    }
+    StringValueCStr(value);
+    RSTRING_GETMEM(value, cp, *sz);
+    ((rb_printf_buffer_extra *)fp)->value = value;
+    OBJ_INFECT(result, value);
+    return cp;
+}
+
 VALUE
 rb_enc_vsprintf(rb_encoding *enc, const char *fmt, va_list ap)
 {
-    rb_printf_buffer f;
+    rb_printf_buffer_extra buffer;
+#define f buffer.base
     VALUE result;
 
     f._flags = __SWR | __SSTR;
     f._bf._size = 0;
     f._w = 120;
     result = rb_str_buf_new(f._w);
-    if (enc) rb_enc_associate(result, enc);
+    if (enc) {
+	if (rb_enc_mbminlen(enc) > 1) {
+	    /* the implementation deeply depends on plain char */
+	    rb_raise(rb_eArgError, "cannot construct wchar_t based encoding string: %s",
+		     rb_enc_name(enc));
+	}
+	rb_enc_associate(result, enc);
+    }
     f._bf._base = (unsigned char *)result;
     f._p = (unsigned char *)RSTRING_PTR(result);
     RBASIC(result)->klass = 0;
     f.vwrite = ruby__sfvwrite;
+    f.vextra = ruby__sfvextra;
+    buffer.value = 0;
     BSD_vfprintf(&f, fmt, ap);
     RBASIC(result)->klass = rb_cString;
     rb_str_resize(result, (char *)f._p - RSTRING_PTR(result));
+#undef f
 
     return result;
 }
@@ -1209,7 +1282,8 @@ rb_sprintf(const char *format, ...)
 VALUE
 rb_str_vcatf(VALUE str, const char *fmt, va_list ap)
 {
-    rb_printf_buffer f;
+    rb_printf_buffer_extra buffer;
+#define f buffer.base
     VALUE klass;
 
     StringValue(str);
@@ -1222,9 +1296,12 @@ rb_str_vcatf(VALUE str, const char *fmt, va_list ap)
     klass = RBASIC(str)->klass;
     RBASIC(str)->klass = 0;
     f.vwrite = ruby__sfvwrite;
+    f.vextra = ruby__sfvextra;
+    buffer.value = 0;
     BSD_vfprintf(&f, fmt, ap);
     RBASIC(str)->klass = klass;
     rb_str_resize(str, (char *)f._p - RSTRING_PTR(str));
+#undef f
 
     return str;
 }
