@@ -44,10 +44,26 @@ class TestBigDecimal < Test::Unit::TestCase
                  "Expected #{x.inspect} to be negative zero")
   end
 
+  def test_not_equal
+    assert_not_equal BigDecimal("1"), BigDecimal.allocate
+  end
+
   def test_global_new
     assert_equal(1, BigDecimal("1"))
     assert_equal(1, BigDecimal("1", 1))
     assert_raise(ArgumentError) { BigDecimal("1", -1) }
+    assert_raise(ArgumentError) { BigDecimal(4.2) }
+    begin
+      BigDecimal(4.2)
+    rescue ArgumentError => error
+      assert_match(/Float/, error.message)
+    end
+    assert_raise(ArgumentError) { BigDecimal(42.quo(7)) }
+    begin
+      BigDecimal(42.quo(7))
+    rescue ArgumentError => error
+      assert_match(/Rational/, error.message)
+    end
   end
 
   def test_global_new_with_integer
@@ -84,6 +100,13 @@ class TestBigDecimal < Test::Unit::TestCase
     end
   end
 
+  def test_global_new_with_tainted_string
+    Thread.new {
+      $SAFE = 1
+      BigDecimal('1'.taint)
+    }.join
+  end
+
   def test_new
     assert_equal(1, BigDecimal.new("1"))
     assert_equal(1, BigDecimal.new("1", 1))
@@ -134,6 +157,13 @@ class TestBigDecimal < Test::Unit::TestCase
     end
   end
 
+  def test_new_with_tainted_string
+    Thread.new {
+      $SAFE = 1
+      BigDecimal.new('1'.taint)
+    }.join
+  end
+
   def _test_mode(type)
     BigDecimal.mode(type, true)
     assert_raise(FloatDomainError) { yield }
@@ -536,15 +566,40 @@ class TestBigDecimal < Test::Unit::TestCase
     assert_kind_of(Float,   x .to_f)
     assert_kind_of(Float, (-x).to_f)
 
+    bug6944 = '[ruby-core:47342]'
+
     BigDecimal.mode(BigDecimal::EXCEPTION_UNDERFLOW, true)
-    assert_raise(FloatDomainError) {
-      BigDecimal("1e#{Float::MIN_10_EXP - 2*Float::DIG}").to_f }
-    assert_raise(FloatDomainError) {
-      BigDecimal("-1e#{Float::MIN_10_EXP - 2*Float::DIG}").to_f }
+    x = "1e#{Float::MIN_10_EXP - 2*Float::DIG}"
+    assert_raise(FloatDomainError, x) {BigDecimal(x).to_f}
+    x = "-#{x}"
+    assert_raise(FloatDomainError, x) {BigDecimal(x).to_f}
+    x = "1e#{Float::MIN_10_EXP - Float::DIG}"
+    assert_nothing_raised(FloatDomainError, x) {
+      assert_in_delta(0.0, BigDecimal(x).to_f, 10**Float::MIN_10_EXP, bug6944)
+    }
+    x = "-#{x}"
+    assert_nothing_raised(FloatDomainError, x) {
+      assert_in_delta(0.0, BigDecimal(x).to_f, 10**Float::MIN_10_EXP, bug6944)
+    }
 
     BigDecimal.mode(BigDecimal::EXCEPTION_UNDERFLOW, false)
-    assert_equal( 0.0, BigDecimal("1e#{Float::MIN_10_EXP - 2*Float::DIG}").to_f)
-    assert_equal(-0.0, BigDecimal("-1e#{Float::MIN_10_EXP - 2*Float::DIG}").to_f)
+    x = "1e#{Float::MIN_10_EXP - 2*Float::DIG}"
+    assert_equal( 0.0, BigDecimal(x).to_f, x)
+    x = "-#{x}"
+    assert_equal(-0.0, BigDecimal(x).to_f, x)
+    x = "1e#{Float::MIN_10_EXP - Float::DIG}"
+    assert_nothing_raised(FloatDomainError, x) {
+      assert_in_delta(0.0, BigDecimal(x).to_f, 10**Float::MIN_10_EXP, bug6944)
+    }
+    x = "-#{x}"
+    assert_nothing_raised(FloatDomainError, x) {
+      assert_in_delta(0.0, BigDecimal(x).to_f, 10**Float::MIN_10_EXP, bug6944)
+    }
+
+    assert_equal( 0.0, BigDecimal(  '9e-325').to_f)
+    assert_equal( 0.0, BigDecimal( '10e-325').to_f)
+    assert_equal(-0.0, BigDecimal( '-9e-325').to_f)
+    assert_equal(-0.0, BigDecimal('-10e-325').to_f)
   end
 
   def test_coerce
@@ -558,6 +613,10 @@ class TestBigDecimal < Test::Unit::TestCase
 
     a, b = BigDecimal("0.11111").coerce(1.quo(3))
     assert_equal(BigDecimal("0." + "3"*a.precs[0]), a)
+
+    assert_nothing_raised(TypeError, '#7176') do
+      BigDecimal.new('1') + Rational(1)
+    end
   end
 
   def test_uplus
@@ -593,6 +652,14 @@ class TestBigDecimal < Test::Unit::TestCase
     assert_equal(BigDecimal.new((2**100-1).to_s), x - 1)
   end
 
+  def test_sub_with_float
+    assert_kind_of(BigDecimal, BigDecimal.new("3") - 1.0)
+  end
+
+  def test_sub_with_rational
+    assert_kind_of(BigDecimal, BigDecimal.new("3") - 1.quo(3))
+  end
+
   def test_mult
     x = BigDecimal.new((2**100).to_s)
     assert_equal(BigDecimal.new((2**100 * 3).to_s), (x * 3).to_i)
@@ -601,6 +668,14 @@ class TestBigDecimal < Test::Unit::TestCase
     assert_equal(BigDecimal.new((2**200).to_s), (x * x).to_i)
   end
 
+  def test_mult_with_float
+    assert_kind_of(BigDecimal, BigDecimal.new("3") * 1.5)
+  end
+
+  def test_mult_with_rational
+    assert_kind_of(BigDecimal, BigDecimal.new("3") * 1.quo(3))
+  end
+
   def test_div
     x = BigDecimal.new((2**100).to_s)
     assert_equal(BigDecimal.new((2**100 / 3).to_s), (x / 3).to_i)
@@ -610,6 +685,14 @@ class TestBigDecimal < Test::Unit::TestCase
     assert_equal(-2, BigDecimal.new("2") / -1)
   end
 
+  def test_div_with_float
+    assert_kind_of(BigDecimal, BigDecimal.new("3") / 1.5)
+  end
+
+  def test_div_with_rational
+    assert_kind_of(BigDecimal, BigDecimal.new("3") / 1.quo(3))
+  end
+
   def test_mod
     x = BigDecimal.new((2**100).to_s)
     assert_equal(1, x % 3)
@@ -618,6 +701,14 @@ class TestBigDecimal < Test::Unit::TestCase
     assert_equal(-1, (-x) % -3)
   end
 
+  def test_mod_with_float
+    assert_kind_of(BigDecimal, BigDecimal.new("3") % 1.5)
+  end
+
+  def test_mod_with_rational
+    assert_kind_of(BigDecimal, BigDecimal.new("3") % 1.quo(3))
+  end
+
   def test_remainder
     x = BigDecimal.new((2**100).to_s)
     assert_equal(1, x.remainder(3))
@@ -626,6 +717,14 @@ class TestBigDecimal < Test::Unit::TestCase
     assert_equal(-1, (-x).remainder(-3))
   end
 
+  def test_remainder_with_float
+    assert_kind_of(BigDecimal, BigDecimal.new("3").remainder(1.5))
+  end
+
+  def test_remainder_with_rational
+    assert_kind_of(BigDecimal, BigDecimal.new("3").remainder(1.quo(3)))
+  end
+
   def test_divmod
     x = BigDecimal.new((2**100).to_s)
     assert_equal([(x / 3).floor, 1], x.divmod(3))
@@ -664,6 +763,10 @@ class TestBigDecimal < Test::Unit::TestCase
     assert_equal(400000000000000000000000000000, x.div(3, 1))
     assert_equal(420000000000000000000000000000, x.div(3, 2))
     assert_equal(423000000000000000000000000000, x.div(3, 3))
+    BigDecimal.save_exception_mode do
+      BigDecimal.mode(BigDecimal::EXCEPTION_INFINITY, false)
+      assert_equal(0, BigDecimal("0").div(BigDecimal("Infinity")))
+    end
   end
 
   def test_abs_bigdecimal
@@ -811,6 +914,12 @@ class TestBigDecimal < Test::Unit::TestCase
     assert_match(/^#<BigDecimal:[0-9a-f]+,'0.12345678E4',#{prec}\(#{maxprec}\)>$/, x.inspect)
   end
 
+  def test_power
+    assert_nothing_raised(TypeError, '[ruby-core:47632]') do
+      1000.times { BigDecimal.new('1001.10')**0.75 }
+    end
+  end
+
   def test_power_with_nil
     assert_raise(TypeError) do
       BigDecimal(3) ** nil
@@ -1089,28 +1198,29 @@ class TestBigDecimal < Test::Unit::TestCase
 
   def test_split_under_gc_stress
     bug3258 = '[ruby-dev:41213]'
-    stress, GC.stress = GC.stress, true
+    expect = 10.upto(20).map{|i|[1, "1", 10, i+1].inspect}
+    assert_in_out_err(%w[-rbigdecimal --disable-gems], <<-EOS, expect, [], bug3258)
+    GC.stress = true
     10.upto(20) do |i|
-      b = BigDecimal.new("1"+"0"*i)
-      assert_equal([1, "1", 10, i+1], b.split, bug3258)
+      p BigDecimal.new("1"+"0"*i).split
     end
-  ensure
-    GC.stress = stress
+    EOS
   end
 
   def test_coerce_under_gc_stress
-    expect = ":too_long_to_embed_as_string can't be coerced into BigDecimal"
-    under_gc_stress do
+    assert_in_out_err(%w[-rbigdecimal --disable-gems], <<-EOS, [], [])
+      expect = ":too_long_to_embed_as_string can't be coerced into BigDecimal"
       b = BigDecimal.new("1")
+      GC.stress = true
       10.times do
         begin
           b.coerce(:too_long_to_embed_as_string)
         rescue => e
-          assert_instance_of TypeError, e
-          assert_equal expect, e.message
+          raise unless e.is_a?(TypeError)
+          raise "'\#{expect}' is expected, but '\#{e.message}'" unless e.message == expect
         end
       end
-    end
+    EOS
   end
 
   def test_INFINITY
@@ -1174,17 +1284,17 @@ class TestBigDecimal < Test::Unit::TestCase
   end
 
   def test_BigMath_exp_under_gc_stress
-    expect = ":too_long_to_embed_as_string can't be coerced into BigDecimal"
-    under_gc_stress do
+    assert_in_out_err(%w[-rbigdecimal --disable-gems], <<-EOS, [], [])
+      expect = ":too_long_to_embed_as_string can't be coerced into BigDecimal"
       10.times do
         begin
           BigMath.exp(:too_long_to_embed_as_string, 6)
         rescue => e
-          assert_instance_of ArgumentError, e
-          assert_equal expect, e.message
+          raise unless e.is_a?(ArgumentError)
+          raise "'\#{expect}' is expected, but '\#{e.message}'" unless e.message == expect
         end
       end
-    end
+    EOS
   end
 
   def test_BigMath_log_with_nil
@@ -1273,23 +1383,32 @@ class TestBigDecimal < Test::Unit::TestCase
   end
 
   def test_BigMath_log_under_gc_stress
-    expect = ":too_long_to_embed_as_string can't be coerced into BigDecimal"
-    under_gc_stress do
+    assert_in_out_err(%w[-rbigdecimal --disable-gems], <<-EOS, [], [])
+      expect = ":too_long_to_embed_as_string can't be coerced into BigDecimal"
       10.times do
         begin
           BigMath.log(:too_long_to_embed_as_string, 6)
         rescue => e
-          assert_instance_of ArgumentError, e
-          assert_equal expect, e.message
+          raise unless e.is_a?(ArgumentError)
+          raise "'\#{expect}' is expected, but '\#{e.message}'" unless e.message == expect
         end
       end
+    EOS
+  end
+
+  def test_dup
+    [1, -1, 2**100, -2**100].each do |i|
+      x = BigDecimal(i)
+      assert_equal(x, x.dup)
     end
   end
 
-  def test_to_d
-    bug6093 = '[ruby-core:42969]'
-    code = "exit(BigDecimal.new('10.0') == 10.0.to_d)"
-    assert_ruby_status(%w[-rbigdecimal -rbigdecimal/util -rmathn -], code, bug6093)
+  def test_dup_subclass
+    c = Class.new(BigDecimal)
+    x = c.new(1)
+    y = x.dup
+    assert_equal(1, y)
+    assert_kind_of(c, y)
   end
 
   def test_to_d
@@ -1297,4 +1416,10 @@ class TestBigDecimal < Test::Unit::TestCase
     code = "exit(BigDecimal.new('10.0') == 10.0.to_d)"
     assert_ruby_status(%w[-rbigdecimal -rbigdecimal/util -rmathn -], code, bug6093)
   end
+
+  def test_bug6406
+    assert_in_out_err(%w[-rbigdecimal --disable-gems], <<-EOS, [], [])
+    Thread.current.keys.to_s
+    EOS
+  end
 end
