@@ -1,3 +1,4 @@
+# -*- coding: us-ascii -*-
 require 'test/unit'
 require 'erb'
 
@@ -37,6 +38,27 @@ class TestERB < Test::Unit::TestCase
     }
     assert_match(/\Atest filename:1\b/, e.backtrace[0])
   end
+
+  def test_html_escape
+    assert_equal(" !&quot;\#$%&amp;&#39;()*+,-./0123456789:;&lt;=&gt;?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~",
+                 ERB::Util.html_escape(" !\"\#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~"))
+
+    assert_equal("", ERB::Util.html_escape(""))
+    assert_equal("abc", ERB::Util.html_escape("abc"))
+    assert_equal("&lt;&lt;", ERB::Util.html_escape("<\<"))
+
+    assert_equal("", ERB::Util.html_escape(nil))
+    assert_equal("123", ERB::Util.html_escape(123))
+  end
+
+  def test_concurrent_default_binding
+    template1 = 'one <%= ERB.new(template2).result %>'
+
+    eval 'template2 = "two"', TOPLEVEL_BINDING
+
+    bug7046 = '[ruby-core:47638]'
+    assert_equal("one two", ERB.new(template1).result, bug7046)
+  end
 end
 
 class TestERBCore < Test::Unit::TestCase
@@ -44,26 +66,32 @@ class TestERBCore < Test::Unit::TestCase
     @erb = ERB
   end
 
-  def test_01
-    _test_01(nil)
-    _test_01(0)
-    _test_01(1)
-    _test_01(2)
-    _test_01(3)
+  def test_core
+    _test_core(nil)
+    _test_core(0)
+    _test_core(1)
+    _test_core(2)
+    orig = $VERBOSE
+    begin
+      $VERBOSE = false
+      _test_core(3)
+    ensure
+      $VERBOSE = orig
+    end
   end
 
-  def _test_01(safe)
+  def _test_core(safe)
     erb = @erb.new("hello")
-    assert_equal(erb.result, "hello")
+    assert_equal("hello", erb.result)
 
     erb = @erb.new("hello", safe, 0)
-    assert_equal(erb.result, "hello")
+    assert_equal("hello", erb.result)
 
     erb = @erb.new("hello", safe, 1)
-    assert_equal(erb.result, "hello")
+    assert_equal("hello", erb.result)
 
     erb = @erb.new("hello", safe, 2)
-    assert_equal(erb.result, "hello")
+    assert_equal("hello", erb.result)
 
     src = <<EOS
 %% hi
@@ -73,7 +101,7 @@ class TestERBCore < Test::Unit::TestCase
 * <%= n %>
 <% end %>
 EOS
-      
+
     ans = <<EOS
 %% hi
 = hello
@@ -118,7 +146,7 @@ EOS
 % n=0
 * 2
 EOS
-      
+
     erb = @erb.new(src, safe, 2)
     assert_equal(ans, erb.result)
     erb = @erb.new(src, safe, '<>')
@@ -157,14 +185,14 @@ EOS
     assert_equal(ans, erb.result)
   end
 
-  def test_02_safe_04
+  def test_safe_04
     erb = @erb.new('<%=$SAFE%>', 4)
-    assert_equal(erb.result(TOPLEVEL_BINDING.taint), '4')
+    assert_equal('4', erb.result(TOPLEVEL_BINDING.taint))
   end
 
   class Foo; end
 
-  def test_03_def_class
+  def test_def_class
     erb = @erb.new('hello')
     cls = erb.def_class
     assert_equal(Object, cls.superclass)
@@ -177,31 +205,31 @@ EOS
     assert(cls.new.respond_to?('erb'))
   end
 
-  def test_04_percent
+  def test_percent
     src = <<EOS
 %n = 1
 <%= n%>
 EOS
-    assert_equal("1\n", ERB.new(src, nil, '%').result)
+    assert_equal("1\n", ERB.new(src, nil, '%').result(binding))
 
     src = <<EOS
 <%
 %>
 EOS
     ans = "\n"
-    assert_equal(ans, ERB.new(src, nil, '%').result)
+    assert_equal(ans, ERB.new(src, nil, '%').result(binding))
 
     src = "<%\n%>"
     # ans = "\n"
     ans = ""
-    assert_equal(ans, ERB.new(src, nil, '%').result)
+    assert_equal(ans, ERB.new(src, nil, '%').result(binding))
 
     src = <<EOS
 <%
 n = 1
 %><%= n%>
 EOS
-    assert_equal("1\n", ERB.new(src, nil, '%').result)
+    assert_equal("1\n", ERB.new(src, nil, '%').result(binding))
 
     src = <<EOS
 %n = 1
@@ -209,35 +237,59 @@ EOS
 n.times do |i|%>
 %% %%><%%<%= i%><%
 end%>
+%%%
 EOS
     ans = <<EOS
-% 
+%\s
 % %%><%0
 % %%><%1
+%%
 EOS
-    assert_equal(ans, ERB.new(src, nil, '%').result)
+    assert_equal(ans, ERB.new(src, nil, '%').result(binding))
   end
 
-  class Bar; end
-
-  def test_05_def_method
-    assert(! Bar.new.respond_to?('hello'))
-    Bar.module_eval do
+  def test_def_erb_method
+    klass = Class.new
+    klass.module_eval do
       extend ERB::DefMethod
       fname = File.join(File.dirname(File.expand_path(__FILE__)), 'hello.erb')
       def_erb_method('hello', fname)
     end
-    assert(Bar.new.respond_to?('hello'))
+    assert(klass.new.respond_to?('hello'))
 
-    assert(! Bar.new.respond_to?('hello_world'))
+    assert(! klass.new.respond_to?('hello_world'))
     erb = @erb.new('hello, world')
-    Bar.module_eval do
+    klass.module_eval do
       def_erb_method('hello_world', erb)
     end
-    assert(Bar.new.respond_to?('hello_world'))    
+    assert(klass.new.respond_to?('hello_world'))
   end
 
-  def test_06_escape
+  def test_def_method_without_filename
+    klass = Class.new
+    erb = ERB.new("<% raise ::TestERB::MyError %>")
+    erb.filename = "test filename"
+    assert(! klass.new.respond_to?('my_error'))
+    erb.def_method(klass, 'my_error')
+    e = assert_raise(::TestERB::MyError) {
+       klass.new.my_error
+    }
+    assert_match(/\A\(ERB\):1\b/, e.backtrace[0])
+  end
+
+  def test_def_method_with_fname
+    klass = Class.new
+    erb = ERB.new("<% raise ::TestERB::MyError %>")
+    erb.filename = "test filename"
+    assert(! klass.new.respond_to?('my_error'))
+    erb.def_method(klass, 'my_error', 'test fname')
+    e = assert_raise(::TestERB::MyError) {
+       klass.new.my_error
+    }
+    assert_match(/\Atest fname:1\b/, e.backtrace[0])
+  end
+
+  def test_escape
     src = <<EOS
 1.<%% : <%="<%%"%>
 2.%%> : <%="%%>"%>
@@ -274,9 +326,9 @@ EOS
     assert_equal(ans, ERB.new(src, nil, '%').result)
   end
 
-  def test_07_keep_lineno
+  def test_keep_lineno
     src = <<EOS
-Hello, 
+Hello,\s
 % x = "World"
 <%= x%>
 % raise("lineno")
@@ -292,21 +344,21 @@ EOS
 
     src = <<EOS
 %>
-Hello, 
+Hello,\s
 <% x = "World%%>
 "%>
 <%= x%>
 EOS
 
     ans = <<EOS
-%>Hello, 
+%>Hello,\s
 World%>
 EOS
     assert_equal(ans, ERB.new(src, nil, '>').result)
 
     ans = <<EOS
 %>
-Hello, 
+Hello,\s
 
 World%>
 EOS
@@ -314,7 +366,7 @@ EOS
 
     ans = <<EOS
 %>
-Hello, 
+Hello,\s
 
 World%>
 
@@ -322,7 +374,7 @@ EOS
     assert_equal(ans, ERB.new(src).result)
 
     src = <<EOS
-Hello, 
+Hello,\s
 <% x = "World%%>
 "%>
 <%= x%>
@@ -357,7 +409,7 @@ EOS
 % y = 'Hello'
 <%- x = "World%%>
 "-%>
-<%= x %><%- x = nil -%> 
+<%= x %><%- x = nil -%>\s
 <% raise("lineno") %>
 EOS
 
@@ -378,7 +430,7 @@ EOS
     end
   end
 
-  def test_08_explicit
+  def test_explicit
     src = <<EOS
 <% x = %w(hello world) -%>
 NotSkip <%- y = x -%> NotSkip
@@ -393,21 +445,46 @@ NotSkip <%- y = x -%> NotSkip
    <%- up = w.upcase -%>
    * <%= up %>
  <%- end -%>
-KeepNewLine <%- z = nil -%> 
+KeepNewLine <%- z = nil -%>\s
 EOS
 
    ans = <<EOS
 NotSkip  NotSkip
   * HELLO
   * WORLD
- NotSkip 
+ NotSkip\s
    * hello
    * HELLO
    * world
    * WORLD
-KeepNewLine  
+KeepNewLine \s
 EOS
    assert_equal(ans, ERB.new(src, nil, '-').result)
    assert_equal(ans, ERB.new(src, nil, '-%').result)
   end
+
+  def test_url_encode
+    assert_equal("Programming%20Ruby%3A%20%20The%20Pragmatic%20Programmer%27s%20Guide",
+                 ERB::Util.url_encode("Programming Ruby:  The Pragmatic Programmer's Guide"))
+
+    assert_equal("%A5%B5%A5%F3%A5%D7%A5%EB",
+                 ERB::Util.url_encode("\xA5\xB5\xA5\xF3\xA5\xD7\xA5\xEB".force_encoding("EUC-JP")))
+  end
+
+  def test_percent_after_etag
+    assert_equal("1%", @erb.new("<%= 1 %>%", nil, "%").result)
+  end
+end
+
+class TestERBCoreWOStrScan < TestERBCore
+  def setup
+    @save_map = ERB::Compiler::Scanner.instance_variable_get('@scanner_map')
+    map = {[nil, false]=>ERB::Compiler::SimpleScanner}
+    ERB::Compiler::Scanner.instance_variable_set('@scanner_map', map)
+    super
+  end
+
+  def teardown
+    ERB::Compiler::Scanner.instance_variable_set('@scanner_map', @save_map)
+  end
 end
