@@ -1,13 +1,14 @@
-# $Id: test_fileutils.rb 13973 2007-11-19 15:57:52Z akr $
+# $Id: test_fileutils.rb 41946 2013-07-13 14:32:56Z nagachika $
 
 require 'fileutils'
-require 'fileasserts'
+require_relative 'fileasserts'
 require 'pathname'
 require 'tmpdir'
 require 'test/unit'
 
 class TestFileUtils < Test::Unit::TestCase
   TMPROOT = "#{Dir.tmpdir}/fileutils.rb.#{$$}"
+  include Test::Unit::FileAssertions
 end
 
 prevdir = Dir.pwd
@@ -16,11 +17,11 @@ Dir.mkdir tmproot unless File.directory?(tmproot)
 Dir.chdir tmproot
 
 def have_drive_letter?
-  /djgpp|mswin(?!ce)|mingw|bcc|emx/ =~ RUBY_PLATFORM
+  /mswin(?!ce)|mingw|bcc|emx/ =~ RUBY_PLATFORM
 end
 
 def have_file_perm?
-  /djgpp|mswin|mingw|bcc|wince|emx/ !~ RUBY_PLATFORM
+  /mswin|mingw|bcc|emx/ !~ RUBY_PLATFORM
 end
 
 $fileutils_rb_have_symlink = nil
@@ -77,7 +78,7 @@ class TestFileUtils
   include FileUtils
 
   def check_singleton(name)
-    assert_equal true, ::FileUtils.public_methods.include?(name.to_sym)
+    assert_respond_to ::FileUtils, name
   end
 
   def my_rm_rf(path)
@@ -104,9 +105,8 @@ class TestFileUtils
   end
 
   def teardown
-    tmproot = Dir.pwd
     Dir.chdir @prevdir
-    my_rm_rf tmproot
+    my_rm_rf TMPROOT
   end
 
 
@@ -173,17 +173,11 @@ class TestFileUtils
     assert_equal Dir.pwd, pwd()
 
     cwd = Dir.pwd
-if have_drive_letter?
-    cd('C:/') {
-      assert_equal 'C:/', pwd()
+    root = have_drive_letter? ? 'C:/' : '/'
+    cd(root) {
+      assert_equal root, pwd()
     }
     assert_equal cwd, pwd()
-else
-    cd('/') {
-      assert_equal '/', pwd()
-    }
-    assert_equal cwd, pwd()
-end
   end
 
   def test_cmp
@@ -192,7 +186,7 @@ end
     TARGETS.each do |fname|
       assert cmp(fname, fname), 'not same?'
     end
-    assert_raises(ArgumentError) {
+    assert_raise(ArgumentError) {
       cmp TARGETS[0], TARGETS[0], :undefinedoption => true
     }
 
@@ -223,28 +217,64 @@ end
       assert_same_entry srcpath, destpath
     end
 
+    assert_raise(Errno::ENOENT) {
+      cp 'tmp/cptmp', 'tmp/cptmp_new'
+    }
+    assert_file_not_exist('tmp/cptmp_new')
+
     # src==dest (1) same path
     touch 'tmp/cptmp'
-    assert_raises(ArgumentError) {
+    assert_raise(ArgumentError) {
       cp 'tmp/cptmp', 'tmp/cptmp'
     }
-if have_symlink?
+  end
+
+  def test_cp_preserve_permissions
+    bug4507 = '[ruby-core:35518]'
+    touch 'tmp/cptmp'
+    chmod 0755, 'tmp/cptmp'
+    cp 'tmp/cptmp', 'tmp/cptmp2'
+    assert_equal(File.stat('tmp/cptmp').mode,
+                 File.stat('tmp/cptmp2').mode,
+                 bug4507)
+  end
+
+  def test_cp_preserve_permissions_dir
+    bug7246 = '[ruby-core:48603]'
+    mkdir 'tmp/cptmp'
+    mkdir 'tmp/cptmp/d1'
+    chmod 0745, 'tmp/cptmp/d1'
+    mkdir 'tmp/cptmp/d2'
+    chmod 0700, 'tmp/cptmp/d2'
+    cp_r 'tmp/cptmp', 'tmp/cptmp2', :preserve => true
+    assert_equal(File.stat('tmp/cptmp/d1').mode,
+                 File.stat('tmp/cptmp2/d1').mode,
+                 bug7246)
+    assert_equal(File.stat('tmp/cptmp/d2').mode,
+                 File.stat('tmp/cptmp2/d2').mode,
+                 bug7246)
+  end
+
+  def test_cp_symlink
+    touch 'tmp/cptmp'
     # src==dest (2) symlink and its target
     File.symlink 'cptmp', 'tmp/cptmp_symlink'
-    assert_raises(ArgumentError) {
+    assert_raise(ArgumentError) {
       cp 'tmp/cptmp', 'tmp/cptmp_symlink'
     }
-    assert_raises(ArgumentError) {
+    assert_raise(ArgumentError) {
       cp 'tmp/cptmp_symlink', 'tmp/cptmp'
     }
     # src==dest (3) looped symlink
     File.symlink 'symlink', 'tmp/symlink'
-    assert_raises(Errno::ELOOP) {
+    assert_raise(Errno::ELOOP) {
       cp 'tmp/symlink', 'tmp/symlink'
     }
-end
+  end if have_symlink?
 
+  def test_cp_pathname
     # pathname
+    touch 'tmp/cptmp'
     assert_nothing_raised {
       cp 'tmp/cptmp', Pathname.new('tmp/tmpdest')
       cp Pathname.new('tmp/cptmp'), 'tmp/tmpdest'
@@ -283,7 +313,17 @@ end
     my_rm_rf 'tmp/cpr_src'
     my_rm_rf 'tmp/cpr_dest'
 
-if have_symlink?
+    bug3588 = '[ruby-core:31360]'
+    assert_nothing_raised(ArgumentError, bug3588) do
+      cp_r 'tmp', 'tmp2'
+    end
+    assert_directory 'tmp2/tmp'
+    assert_raise(ArgumentError, bug3588) do
+      cp_r 'tmp2', 'tmp2/new_tmp2'
+    end
+  end
+
+  def test_cp_r_symlink
     # symlink in a directory
     mkdir 'tmp/cpr_src'
     ln_s 'SLdest', 'tmp/cpr_src/symlink'
@@ -295,12 +335,25 @@ if have_symlink?
     ln_s 'cpr_src', 'tmp/cpr_src2'
     cp_r 'tmp/cpr_src2', 'tmp/cpr_dest2'
     assert_directory 'tmp/cpr_dest2'
-    #assert_not_symlink 'tmp/cpr_dest2'
-    assert_symlink 'tmp/cpr_dest2'   # 2005-05-26: feature change
+    assert_not_symlink 'tmp/cpr_dest2'
     assert_symlink 'tmp/cpr_dest2/symlink'
     assert_equal 'SLdest', File.readlink('tmp/cpr_dest2/symlink')
-end
+  end if have_symlink?
 
+  def test_cp_r_symlink_preserve
+    mkdir 'tmp/cross'
+    mkdir 'tmp/cross/a'
+    mkdir 'tmp/cross/b'
+    touch 'tmp/cross/a/f'
+    touch 'tmp/cross/b/f'
+    ln_s '../a/f', 'tmp/cross/b/l'
+    ln_s '../b/f', 'tmp/cross/a/l'
+    assert_nothing_raised {
+      cp_r 'tmp/cross', 'tmp/cross2', :preserve => true
+    }
+  end if have_symlink?
+
+  def test_cp_r_pathname
     # pathname
     touch 'tmp/cprtmp'
     assert_nothing_raised {
@@ -328,35 +381,39 @@ end
 
     mkdir 'tmp/tmpdir'
     mkdir_p 'tmp/dest2/tmpdir'
-    assert_raises(Errno::EEXIST) {
+    assert_raise(Errno::EEXIST) {
       mv 'tmp/tmpdir', 'tmp/dest2'
     }
     mkdir 'tmp/dest2/tmpdir/junk'
-    assert_raises(Errno::EEXIST, "[ruby-talk:124368]") {
+    assert_raise(Errno::EEXIST, "[ruby-talk:124368]") {
       mv 'tmp/tmpdir', 'tmp/dest2'
     }
 
     # src==dest (1) same path
     touch 'tmp/cptmp'
-    assert_raises(ArgumentError) {
+    assert_raise(ArgumentError) {
       mv 'tmp/cptmp', 'tmp/cptmp'
     }
-if have_symlink?
+  end
+
+  def test_mv_symlink
+    touch 'tmp/cptmp'
     # src==dest (2) symlink and its target
     File.symlink 'cptmp', 'tmp/cptmp_symlink'
-    assert_raises(ArgumentError) {
+    assert_raise(ArgumentError) {
       mv 'tmp/cptmp', 'tmp/cptmp_symlink'
     }
-    assert_raises(ArgumentError) {
+    assert_raise(ArgumentError) {
       mv 'tmp/cptmp_symlink', 'tmp/cptmp'
     }
     # src==dest (3) looped symlink
     File.symlink 'symlink', 'tmp/symlink'
-    assert_raises(Errno::ELOOP) {
+    assert_raise(Errno::ELOOP) {
       mv 'tmp/symlink', 'tmp/symlink'
     }
-end
+  end if have_symlink?
 
+  def test_mv_pathname
     # pathname
     assert_nothing_raised {
       touch 'tmp/mvtmpsrc'
@@ -398,14 +455,14 @@ end
       rm_f 'tmp/rmsrc'
       assert_file_not_exist 'tmp/rmsrc'
     end
+  end
 
-if have_symlink?
+  def test_rm_symlink
     File.open('tmp/lnf_symlink_src', 'w') {|f| f.puts 'dummy' }
     File.symlink 'tmp/lnf_symlink_src', 'tmp/lnf_symlink_dest'
     rm_f 'tmp/lnf_symlink_dest'
     assert_file_not_exist 'tmp/lnf_symlink_dest'
     assert_file_exist     'tmp/lnf_symlink_src'
-end
 
     rm_f 'notexistdatafile'
     rm_f 'tmp/notexistdatafile'
@@ -413,7 +470,9 @@ end
     Dir.mkdir 'tmpdatadir'
     # rm_f 'tmpdatadir'
     Dir.rmdir 'tmpdatadir'
+  end if have_symlink?
 
+  def test_rm_f_2
     Dir.mkdir 'tmp/tmpdir'
     File.open('tmp/tmpdir/a', 'w') {|f| f.puts 'dummy' }
     File.open('tmp/tmpdir/c', 'w') {|f| f.puts 'dummy' }
@@ -421,7 +480,9 @@ end
     assert_file_not_exist 'tmp/tmpdir/a'
     assert_file_not_exist 'tmp/tmpdir/c'
     Dir.rmdir 'tmp/tmpdir'
+  end
 
+  def test_rm_pathname
     # pathname
     touch 'tmp/rmtmp1'
     touch 'tmp/rmtmp2'
@@ -435,6 +496,11 @@ end
     assert_file_not_exist 'tmp/rmtmp2'
     assert_file_not_exist 'tmp/rmtmp3'
     assert_file_exist 'tmp/rmtmp4'
+
+    # [ruby-dev:39345]
+    touch 'tmp/[rmtmp]'
+    FileUtils.rm_f 'tmp/[rmtmp]'
+    assert_file_not_exist 'tmp/[rmtmp]'
   end
 
   def test_rm_r
@@ -475,16 +541,18 @@ end
     assert_file_not_exist 'tmp/tmpdir/a'
     assert_file_not_exist 'tmp/tmpdir/c'
     Dir.rmdir 'tmp/tmpdir'
+  end
 
-if have_symlink?
+  def test_rm_r_symlink
     # [ruby-talk:94635] a symlink to the directory
     Dir.mkdir 'tmp/tmpdir'
     File.symlink '..', 'tmp/tmpdir/symlink_to_dir'
     rm_r 'tmp/tmpdir'
     assert_file_not_exist 'tmp/tmpdir'
     assert_file_exist     'tmp'
-end
+  end if have_symlink?
 
+  def test_rm_r_pathname
     # pathname
     Dir.mkdir 'tmp/tmpdir1'; touch 'tmp/tmpdir1/tmp'
     Dir.mkdir 'tmp/tmpdir2'; touch 'tmp/tmpdir2/tmp'
@@ -538,16 +606,18 @@ end
     assert_file_not_exist 'tmp/tmpdir/a'
     assert_file_not_exist 'tmp/tmpdir/c'
     Dir.rmdir 'tmp/tmpdir'
+  end
 
-if have_symlink?
+  def test_remove_entry_secure_symlink
     # [ruby-talk:94635] a symlink to the directory
     Dir.mkdir 'tmp/tmpdir'
     File.symlink '..', 'tmp/tmpdir/symlink_to_dir'
     remove_entry_secure 'tmp/tmpdir'
     assert_file_not_exist 'tmp/tmpdir'
     assert_file_exist     'tmp'
-end
+  end if have_symlink?
 
+  def test_remove_entry_secure_pathname
     # pathname
     Dir.mkdir 'tmp/tmpdir1'; touch 'tmp/tmpdir1/tmp'
     assert_nothing_raised {
@@ -571,7 +641,6 @@ end
     assert_file_not_exist 'tmp/mvdest'
   end
 
-if have_hardlink?
   def test_ln
     TARGETS.each do |fname|
       ln fname, 'tmp/lndest'
@@ -589,16 +658,19 @@ if have_hardlink?
 
     # src==dest (1) same path
     touch 'tmp/cptmp'
-    assert_raises(Errno::EEXIST) {
+    assert_raise(Errno::EEXIST) {
       ln 'tmp/cptmp', 'tmp/cptmp'
     }
-if have_symlink?
+  end if have_hardlink?
+
+  def test_ln_symlink
+    touch 'tmp/cptmp'
     # src==dest (2) symlink and its target
     File.symlink 'cptmp', 'tmp/symlink'
-    assert_raises(Errno::EEXIST) {
+    assert_raise(Errno::EEXIST) {
       ln 'tmp/cptmp', 'tmp/symlink'   # normal file -> symlink
     }
-    assert_raises(Errno::EEXIST) {
+    assert_raise(Errno::EEXIST) {
       ln 'tmp/symlink', 'tmp/cptmp'   # symlink -> normal file
     }
     # src==dest (3) looped symlink
@@ -608,8 +680,9 @@ if have_symlink?
     rescue => err
       assert_kind_of SystemCallError, err
     end
-end
+  end if have_symlink?
 
+  def test_ln_pathname
     # pathname
     touch 'tmp/lntmp'
     assert_nothing_raised {
@@ -617,10 +690,8 @@ end
       ln 'tmp/lntmp', Pathname.new('tmp/lndesttmp2')
       ln Pathname.new('tmp/lntmp'), Pathname.new('tmp/lndesttmp3')
     }
-  end
-end
+  end if have_hardlink?
 
-if have_symlink?
   def test_ln_s
     check_singleton :ln_s
 
@@ -642,10 +713,8 @@ if have_symlink?
       ln_s 'lnsdest', Pathname.new('tmp/symlink_tmp2')
       ln_s Pathname.new('lnsdest'), Pathname.new('tmp/symlink_tmp3')
     }
-  end
-end
+  end if have_symlink?
 
-if have_symlink?
   def test_ln_sf
     check_singleton :ln_sf
 
@@ -667,8 +736,7 @@ if have_symlink?
       ln_sf 'lns_dest', Pathname.new('tmp/symlink_tmp2')
       ln_sf Pathname.new('lns_dest'), Pathname.new('tmp/symlink_tmp3')
     }
-  end
-end
+  end if have_symlink?
 
   def test_mkdir
     check_singleton :mkdir
@@ -690,20 +758,22 @@ end
     assert_directory 'tmp/tmp'
     assert_equal 0700, (File.stat('tmp/tmp').mode & 0777) if have_file_perm?
     Dir.rmdir 'tmp/tmp'
+  end
 
-if have_file_perm?
+  def test_mkdir_file_perm
     mkdir 'tmp/tmp', :mode => 07777
     assert_directory 'tmp/tmp'
     assert_equal 07777, (File.stat('tmp/tmp').mode & 07777)
     Dir.rmdir 'tmp/tmp'
-end
+  end if have_file_perm?
 
-if lf_in_path_allowed?
+  def test_mkdir_lf_in_path
     mkdir "tmp-first-line\ntmp-second-line"
     assert_directory "tmp-first-line\ntmp-second-line"
     Dir.rmdir "tmp-first-line\ntmp-second-line"
-end
+  end if lf_in_path_allowed?
 
+  def test_mkdir_pathname
     # pathname
     assert_nothing_raised {
       mkdir Pathname.new('tmp/tmpdirtmp')
@@ -761,15 +831,17 @@ end
     # (rm(1) try to chdir to parent directory, it fails to remove directory.)
     Dir.rmdir 'tmp/tmp'
     Dir.rmdir 'tmp'
+  end
 
-if have_file_perm?
+  def test_mkdir_p_file_perm
     mkdir_p 'tmp/tmp/tmp', :mode => 07777
     assert_directory 'tmp/tmp/tmp'
     assert_equal 07777, (File.stat('tmp/tmp/tmp').mode & 07777)
     Dir.rmdir 'tmp/tmp/tmp'
     Dir.rmdir 'tmp/tmp'
-end
+  end if have_file_perm?
 
+  def test_mkdir_p_pathname
     # pathname
     assert_nothing_raised {
       mkdir_p Pathname.new('tmp/tmp/tmp')
@@ -796,26 +868,30 @@ end
 
     # src==dest (1) same path
     touch 'tmp/cptmp'
-    assert_raises(ArgumentError) {
+    assert_raise(ArgumentError) {
       install 'tmp/cptmp', 'tmp/cptmp'
     }
-if have_symlink?
+  end
+
+  def test_install_symlink
+    touch 'tmp/cptmp'
     # src==dest (2) symlink and its target
     File.symlink 'cptmp', 'tmp/cptmp_symlink'
-    assert_raises(ArgumentError) {
+    assert_raise(ArgumentError) {
       install 'tmp/cptmp', 'tmp/cptmp_symlink'
     }
-    assert_raises(ArgumentError) {
+    assert_raise(ArgumentError) {
       install 'tmp/cptmp_symlink', 'tmp/cptmp'
     }
     # src==dest (3) looped symlink
     File.symlink 'symlink', 'tmp/symlink'
-    assert_raises(Errno::ELOOP) {
+    assert_raise(Errno::ELOOP) {
       # File#install invokes open(2), always ELOOP must be raised
       install 'tmp/symlink', 'tmp/symlink'
     }
-end
+  end if have_symlink?
 
+  def test_install_pathname
     # pathname
     assert_nothing_raised {
       rm_f 'tmp/a'; touch 'tmp/a'
@@ -835,7 +911,6 @@ end
     }
   end
 
-if have_file_perm?
   def test_chmod
     check_singleton :chmod
 
@@ -844,7 +919,38 @@ if have_file_perm?
     assert_equal 0700, File.stat('tmp/a').mode & 0777
     chmod 0500, 'tmp/a'
     assert_equal 0500, File.stat('tmp/a').mode & 0777
-  end
+  end if have_file_perm?
+
+  def test_chmod_symbol_mode
+    check_singleton :chmod
+
+    touch 'tmp/a'
+    chmod "u=wrx,g=,o=", 'tmp/a'
+    assert_equal 0700, File.stat('tmp/a').mode & 0777
+    chmod "u=rx,go=", 'tmp/a'
+    assert_equal 0500, File.stat('tmp/a').mode & 0777
+    chmod "+wrx", 'tmp/a'
+    assert_equal 0777, File.stat('tmp/a').mode & 0777
+    chmod "u+s,o=s", 'tmp/a'
+    assert_equal 04770, File.stat('tmp/a').mode & 07777
+    chmod "u-w,go-wrx", 'tmp/a'
+    assert_equal 04500, File.stat('tmp/a').mode & 07777
+    chmod "+s", 'tmp/a'
+    assert_equal 06500, File.stat('tmp/a').mode & 07777
+
+    # FreeBSD ufs and tmpfs don't allow to change sticky bit against
+    # regular file. It's slightly strange. Anyway it's no effect bit.
+    # see /usr/src/sys/ufs/ufs/ufs_chmod()
+    # NetBSD, OpenBSD and Solaris also denies it.
+    if /freebsd|netbsd|openbsd|solaris/ !~ RUBY_PLATFORM
+      chmod "u+t,o+t", 'tmp/a'
+      assert_equal 07500, File.stat('tmp/a').mode & 07777
+      chmod "a-t,a-s", 'tmp/a'
+      assert_equal 0500, File.stat('tmp/a').mode & 07777
+    end
+
+  end if have_file_perm?
+
 
   def test_chmod_R
     check_singleton :chmod_R
@@ -862,18 +968,72 @@ if have_file_perm?
     assert_equal 0500, File.stat('tmp/dir/dir').mode & 0777
     assert_equal 0500, File.stat('tmp/dir/dir/file').mode & 0777
     chmod_R 0700, 'tmp/dir'   # to remove
-  end
+  end if have_file_perm?
+
+  def test_chmod_symbol_mode_R
+    check_singleton :chmod_R
+
+    mkdir_p 'tmp/dir/dir'
+    touch %w( tmp/dir/file tmp/dir/dir/file )
+    chmod_R "u=wrx,g=,o=", 'tmp/dir'
+    assert_equal 0700, File.stat('tmp/dir').mode & 0777
+    assert_equal 0700, File.stat('tmp/dir/file').mode & 0777
+    assert_equal 0700, File.stat('tmp/dir/dir').mode & 0777
+    assert_equal 0700, File.stat('tmp/dir/dir/file').mode & 0777
+    chmod_R "u=xr,g+X,o=", 'tmp/dir'
+    assert_equal 0510, File.stat('tmp/dir').mode & 0777
+    assert_equal 0500, File.stat('tmp/dir/file').mode & 0777
+    assert_equal 0510, File.stat('tmp/dir/dir').mode & 0777
+    assert_equal 0500, File.stat('tmp/dir/dir/file').mode & 0777
+    chmod_R 0700, 'tmp/dir'   # to remove
+  end if have_file_perm?
+
+  def test_chmod_verbose
+    check_singleton :chmod
+
+    stderr_back = $stderr
+    read, $stderr = IO.pipe
+    th = Thread.new { read.read }
+
+    touch 'tmp/a'
+    chmod 0700, 'tmp/a', verbose: true
+    assert_equal 0700, File.stat('tmp/a').mode & 0777
+    chmod 0500, 'tmp/a', verbose: true
+    assert_equal 0500, File.stat('tmp/a').mode & 0777
+
+    $stderr.close
+    lines = th.value.lines.map {|l| l.chomp }
+    assert_equal(["chmod 700 tmp/a", "chmod 500 tmp/a"], lines)
+  ensure
+    $stderr = stderr_back if stderr_back
+  end if have_file_perm?
+
+  def test_s_chmod_verbose
+    output_back = FileUtils.instance_variable_get(:@fileutils_output)
+    read, write = IO.pipe
+    FileUtils.instance_variable_set(:@fileutils_output, write)
+    th = Thread.new { read.read }
+
+    touch 'tmp/a'
+    FileUtils.chmod 0700, 'tmp/a', verbose: true
+    assert_equal 0700, File.stat('tmp/a').mode & 0777
+
+    write.close
+    lines = th.value.lines.map {|l| l.chomp }
+    assert_equal(["chmod 700 tmp/a"], lines)
+  ensure
+    FileUtils.instance_variable_set(:@fileutils_output, output_back) if output_back
+  end if have_file_perm?
 
   # FIXME: How can I test this method?
   def test_chown
     check_singleton :chown
-  end
+  end if have_file_perm?
 
   # FIXME: How can I test this method?
   def test_chown_R
     check_singleton :chown_R
-  end
-end
+  end if have_file_perm?
 
   def test_copy_entry
     check_singleton :copy_entry
@@ -883,7 +1043,9 @@ end
       assert_same_file srcpath, destpath
       assert_equal File.stat(srcpath).ftype, File.stat(destpath).ftype
     end
-if have_symlink?
+  end
+
+  def test_copy_entry_symlink
     # root is a symlink
     File.symlink 'somewhere', 'tmp/symsrc'
     copy_entry 'tmp/symsrc', 'tmp/symdest'
@@ -898,8 +1060,7 @@ if have_symlink?
     assert_not_symlink 'tmp/dirdest'
     assert_symlink 'tmp/dirdest/sym'
     assert_equal 'somewhere', File.readlink('tmp/dirdest/sym')
-end
-  end
+  end if have_symlink?
 
   def test_copy_file
     check_singleton :copy_file
@@ -914,20 +1075,21 @@ end
     check_singleton :copy_stream
     # IO
     each_srcdest do |srcpath, destpath|
-      File.open(srcpath) {|src|
-        File.open(destpath, 'w') {|dest|
+      File.open(srcpath, 'rb') {|src|
+        File.open(destpath, 'wb') {|dest|
           copy_stream src, dest
         }
       }
       assert_same_file srcpath, destpath
     end
+  end
 
+  def test_copy_stream_duck
+    check_singleton :copy_stream
     # duck typing test  [ruby-dev:25369]
-    my_rm_rf 'tmp'
-    Dir.mkdir 'tmp'
     each_srcdest do |srcpath, destpath|
-      File.open(srcpath) {|src|
-        File.open(destpath, 'w') {|dest|
+      File.open(srcpath, 'rb') {|src|
+        File.open(destpath, 'wb') {|dest|
           copy_stream Stream.new(src), Stream.new(dest)
         }
       }
@@ -940,13 +1102,14 @@ end
     File.open('data/tmp', 'w') {|f| f.puts 'dummy' }
     remove_file 'data/tmp'
     assert_file_not_exist 'data/tmp'
-if have_file_perm?
+  end
+
+  def test_remove_file_file_perm
     File.open('data/tmp', 'w') {|f| f.puts 'dummy' }
     File.chmod 0, 'data/tmp'
     remove_file 'data/tmp'
     assert_file_not_exist 'data/tmp'
-end
-  end
+  end if have_file_perm?
 
   def test_remove_dir
     check_singleton :remove_dir
@@ -954,13 +1117,14 @@ end
     File.open('data/tmpdir/a', 'w') {|f| f.puts 'dummy' }
     remove_dir 'data/tmpdir'
     assert_file_not_exist 'data/tmpdir'
-if have_file_perm?
+  end
+
+  def test_remove_dir_file_perm
     Dir.mkdir 'data/tmpdir'
     File.chmod 0555, 'data/tmpdir'
     remove_dir 'data/tmpdir'
     assert_file_not_exist 'data/tmpdir'
-end
-  end
+  end if have_file_perm?
 
   def test_compare_file
     check_singleton :compare_file
@@ -977,8 +1141,8 @@ end
       @f = f
     end
 
-    def read(n)
-      @f.read(n)
+    def read(*args)
+      @f.read(*args)
     end
 
     def write(str)
@@ -1005,8 +1169,30 @@ end
       uptodate? 'tmp/a', ['tmp/b', Pathname.new('tmp/c')]
       uptodate? Pathname.new('tmp/a'), [Pathname.new('tmp/b'), Pathname.new('tmp/c')]
     }
+    # [Bug #6708] [ruby-core:46256]
+    assert_raises_with_message(ArgumentError, "wrong number of arguments (3 for 2)") {
+      uptodate?('new',['old', 'oldest'], {})
+    }
   end
 
+  def assert_raises_with_message(klass, message)
+    begin
+      yield
+      flunk("Expected Exception #{klass} didn't raise")
+    rescue klass => ex
+      if message.kind_of? String
+        flag = !!(ex.message == message)
+        assert(flag, "Expected Exception(#{klass}) was raised, but the message doesn't match")
+      elsif message.kind_of? Regexp
+        flag = !!(ex.message =~ message)
+        assert(flag, "Expected Exception(#{klass}) was raised, but the message doesn't match")
+      else
+        raise
+      end
+    end
+  end
+  private :assert_raises_with_message
+
   def test_cd
     check_singleton :cd
   end
