@@ -3,40 +3,44 @@ require 'tempfile'
 require_relative 'envutil'
 
 class TestException < Test::Unit::TestCase
-  def test_exception
+  def test_exception_rescued
     begin
       raise "this must be handled"
       assert(false)
     rescue
       assert(true)
     end
+  end
 
-    $bad = true
+  def test_exception_retry
+    bad = true
     begin
       raise "this must be handled no.2"
     rescue
-      if $bad
-        $bad = false
+      if bad
+        bad = false
         retry
         assert(false)
       end
     end
     assert(true)
+  end
 
-    # exception in rescue clause
-    $string = "this must be handled no.3"
+  def test_exception_in_rescue
+    string = "this must be handled no.3"
     e = assert_raise(RuntimeError) do
       begin
         raise "exception in rescue clause"
       rescue
-        raise $string
+        raise string
       end
       assert(false)
     end
-    assert_equal($string, e.message)
+    assert_equal(string, e.message)
+  end
 
-    # exception in ensure clause
-    $string = "exception in ensure clause"
+  def test_exception_in_ensure
+    string = "exception in ensure clause"
     e = assert_raise(RuntimeError) do
       begin
         raise "this must be handled no.4"
@@ -47,55 +51,72 @@ class TestException < Test::Unit::TestCase
       end
       assert(false)
     end
-    assert_equal($string, e.message)
+    assert_equal(string, e.message)
+  end
 
-    $bad = true
+  def test_exception_ensure
+    bad = true
     begin
       begin
         raise "this must be handled no.5"
       ensure
-        $bad = false
+        bad = false
       end
     rescue
     end
-    assert(!$bad)
+    assert(!bad)
+  end
 
-    $bad = true
+  def test_exception_ensure_2   # just duplication?
+    bad = true
     begin
       begin
         raise "this must be handled no.6"
       ensure
-        $bad = false
+        bad = false
       end
     rescue
     end
-    assert(!$bad)
+    assert(!bad)
+  end
 
-    $bad = true
+  def test_break_ensure
+    bad = true
     while true
       begin
         break
       ensure
-        $bad = false
+        bad = false
       end
     end
-    assert(!$bad)
-
-    assert(catch(:foo) {
-         loop do
-           loop do
-    	 throw :foo, true
-    	 break
-           end
-           break
-           assert(false)			# should no reach here
-         end
-         false
-       })
-
+    assert(!bad)
   end
 
-  def test_else
+  def test_catch_throw
+    assert(catch(:foo) {
+             loop do
+               loop do
+                 throw :foo, true
+                 break
+               end
+               break
+               assert(false)			# should no reach here
+             end
+             false
+           })
+  end
+
+  def test_catch_throw_in_require
+    bug7185 = '[ruby-dev:46234]'
+    t = Tempfile.open(["dep", ".rb"])
+    t.puts("throw :extdep, 42")
+    t.close
+    assert_equal(42, catch(:extdep) {require t.path}, bug7185)
+  ensure
+    t.close! if t
+  end
+
+  def test_else_no_exception
     begin
       assert(true)
     rescue
@@ -103,7 +124,9 @@ class TestException < Test::Unit::TestCase
     else
       assert(true)
     end
+  end
 
+  def test_else_raised
     begin
       assert(true)
       raise
@@ -113,7 +136,9 @@ class TestException < Test::Unit::TestCase
     else
       assert(false)
     end
+  end
 
+  def test_else_nested_no_exception
     begin
       assert(true)
       begin
@@ -129,7 +154,9 @@ class TestException < Test::Unit::TestCase
     else
       assert(true)
     end
+  end
 
+  def test_else_nested_rescued
     begin
       assert(true)
       begin
@@ -147,7 +174,9 @@ class TestException < Test::Unit::TestCase
     else
       assert(true)
     end
+  end
 
+  def test_else_nested_unrescued
     begin
       assert(true)
       begin
@@ -165,7 +194,9 @@ class TestException < Test::Unit::TestCase
     else
       assert(false)
     end
+  end
 
+  def test_else_nested_rescued_reraise
     begin
       assert(true)
       begin
@@ -193,6 +224,25 @@ class TestException < Test::Unit::TestCase
     assert_raise(ArgumentError) { raise 1, 1, 1, 1 }
   end
 
+  def test_type_error_message_encoding
+    c = eval("Module.new do break class C\u{4032}; self; end; end")
+    o = c.new
+    e = assert_raise(TypeError) do
+      ""[o]
+    end
+    assert_match(/C\u{4032}/, e.message)
+    c.class_eval {def to_int; self; end}
+    e = assert_raise(TypeError) do
+      ""[o]
+    end
+    assert_match(/C\u{4032}/, e.message)
+    c.class_eval {def to_a; self; end}
+    assert_raise(TypeError) do
+      [*o]
+    end
+    assert_match(/C\u{4032}/, e.message)
+  end
+
   def test_errat
     assert_in_out_err([], "p $@", %w(nil), [])
 
@@ -245,7 +295,7 @@ class TestException < Test::Unit::TestCase
   end
 
   def test_thread_signal_location
-    stdout, stderr, status = EnvUtil.invoke_ruby("-d", <<-RUBY, false, true)
+    _, stderr, _ = EnvUtil.invoke_ruby("-d", <<-RUBY, false, true)
 Thread.start do
   begin
     Process.kill(:INT, $$)
@@ -297,13 +347,25 @@ end.join
       exit
     rescue SystemExit => e
     end
-    assert(e.success?)
+    assert_send([e, :success?], "success by default")
+
+    begin
+      exit(true)
+    rescue SystemExit => e
+    end
+    assert_send([e, :success?], "true means success")
+
+    begin
+      exit(false)
+    rescue SystemExit => e
+    end
+    assert_not_send([e, :success?], "false means failure")
 
     begin
       abort
     rescue SystemExit => e
     end
-    assert(!e.success?)
+    assert_not_send([e, :success?], "abort means failure")
   end
 
   def test_nomethoderror
@@ -318,20 +380,104 @@ end.join
     assert_equal(Encoding.find("locale"), Errno::EINVAL.new.message.encoding)
   end
 
+  def test_too_many_args_in_eval
+    bug5720 = '[ruby-core:41520]'
+    arg_string = (0...140000).to_a.join(", ")
+    assert_raise(SystemStackError, bug5720) {eval "raise(#{arg_string})"}
+  end
+
+  def test_systemexit_new
+    e0 = SystemExit.new
+    assert_equal(0, e0.status)
+    assert_equal("SystemExit", e0.message)
+    ei = SystemExit.new(3)
+    assert_equal(3, ei.status)
+    assert_equal("SystemExit", ei.message)
+    es = SystemExit.new("msg")
+    assert_equal(0, es.status)
+    assert_equal("msg", es.message)
+    eis = SystemExit.new(7, "msg")
+    assert_equal(7, eis.status)
+    assert_equal("msg", eis.message)
+
+    bug5728 = '[ruby-dev:44951]'
+    et = SystemExit.new(true)
+    assert_equal(true, et.success?, bug5728)
+    assert_equal("SystemExit", et.message, bug5728)
+    ef = SystemExit.new(false)
+    assert_equal(false, ef.success?, bug5728)
+    assert_equal("SystemExit", ef.message, bug5728)
+    ets = SystemExit.new(true, "msg")
+    assert_equal(true, ets.success?, bug5728)
+    assert_equal("msg", ets.message, bug5728)
+    efs = SystemExit.new(false, "msg")
+    assert_equal(false, efs.success?, bug5728)
+    assert_equal("msg", efs.message, bug5728)
+  end
+
   def test_exception_in_name_error_to_str
     bug5575 = '[ruby-core:41612]'
-    t = Tempfile.new(["test_exception_in_name_error_to_str", ".rb"])
-    t.puts <<-EOC
+    t = nil
+    Tempfile.open(["test_exception_in_name_error_to_str", ".rb"]) do |f|
+      t = f
+      t.puts <<-EOC
       begin
         BasicObject.new.inspect
       rescue
         $!.inspect
       end
     EOC
-    t.close
+    end
     assert_nothing_raised(NameError, bug5575) do
       load(t.path)
     end
+  ensure
+    t.close(true) if t
+  end
+
+  def test_equal
+    bug5865 = '[ruby-core:41979]'
+    assert_equal(RuntimeError.new("a"), RuntimeError.new("a"), bug5865)
+    assert_not_equal(RuntimeError.new("a"), StandardError.new("a"), bug5865)
+  end
+
+  def test_exception_in_exception_equal
+    bug5865 = '[ruby-core:41979]'
+    t = nil
+    Tempfile.open(["test_exception_in_exception_equal", ".rb"]) do |f|
+      t = f
+      t.puts <<-EOC
+      o = Object.new
+      def o.exception(arg)
+      end
+      _ = RuntimeError.new("a") == o
+    EOC
+    end
+    assert_nothing_raised(ArgumentError, bug5865) do
+      load(t.path)
+    end
+  ensure
+    t.close(true) if t
+  end
+
+  Bug4438 = '[ruby-core:35364]'
+
+  def test_rescue_single_argument
+    assert_raise(TypeError, Bug4438) do
+      begin
+        raise
+      rescue 1
+      end
+    end
+  end
+
+  def test_rescue_splat_argument
+    assert_raise(TypeError, Bug4438) do
+      begin
+        raise
+      rescue *Array(1)
+      end
+    end
   end
 
   def test_to_s_taintness_propagation
@@ -345,7 +491,7 @@ end.join
       assert_equal(false, s.tainted?,
                    "#{exc}#to_s should not propagate taintness")
     end
-    
+
     o = Object.new
     def o.to_str
       "foo"
