@@ -64,25 +64,6 @@ class TestModule < Test::Unit::TestCase
 
   # Support stuff
 
-  def remove_pp_mixins(list)
-    list.reject {|c| c == PP::ObjectMixin }
-  end
-
-  def remove_json_mixins(list)
-    list.reject {|c| c.to_s.start_with?("JSON") }
-  end
-
-  def remove_rake_mixins(list)
-    list.reject {|c|
-      name = c.name
-      name.start_with?("Rake") or name.start_with?("FileUtils")
-    }
-  end
-
-  def remove_minitest_mixins(list)
-    list.reject {|c| c.to_s.start_with?("MiniTest") }
-  end
-
   module Mixin
     MIXIN = 1
     def mixin
@@ -220,10 +201,11 @@ class TestModule < Test::Unit::TestCase
     assert_equal([User, Mixin],      User.ancestors)
     assert_equal([Mixin],            Mixin.ancestors)
 
-    assert_equal([Object, Kernel, BasicObject],
-                 remove_minitest_mixins(remove_rake_mixins(remove_json_mixins(remove_pp_mixins(Object.ancestors)))))
-    assert_equal([String, Comparable, Object, Kernel, BasicObject],
-                 remove_minitest_mixins(remove_rake_mixins(remove_json_mixins(remove_pp_mixins(String.ancestors)))))
+    ancestors = Object.ancestors
+    mixins = ancestors - [Object, Kernel, BasicObject]
+    mixins << JSON::Ext::Generator::GeneratorMethods::String if defined?(JSON::Ext::Generator::GeneratorMethods::String)
+    assert_equal([Object, Kernel, BasicObject], ancestors - mixins)
+    assert_equal([String, Comparable, Object, Kernel, BasicObject], String.ancestors - mixins)
   end
 
   CLASS_EVAL = 2
@@ -258,11 +240,61 @@ class TestModule < Test::Unit::TestCase
     assert(!Math.const_defined?("IP"))
   end
 
+  def test_bad_constants
+    [
+      "#<Class:0x7b8b718b>",
+      ":Object",
+      "",
+      ":",
+      ["String::", "[Bug #7573]"],
+    ].each do |name, msg|
+      e = assert_raises(NameError, "#{msg}#{': ' if msg}wrong constant name #{name.dump}") {
+        Object.const_get name
+      }
+      assert_equal("wrong constant name %s" % name, e.message)
+    end
+  end
+
+  def test_leading_colons
+    assert_equal Object, AClass.const_get('::Object')
+  end
+
   def test_const_get
     assert_equal(Math::PI, Math.const_get("PI"))
     assert_equal(Math::PI, Math.const_get(:PI))
   end
 
+  def test_nested_get
+    assert_equal Other, Object.const_get([self.class, Other].join('::'))
+    assert_equal User::USER, self.class.const_get([User, 'USER'].join('::'))
+  end
+
+  def test_nested_get_symbol
+    const = [self.class, Other].join('::').to_sym
+    assert_raise(NameError) {Object.const_get(const)}
+
+    const = [User, 'USER'].join('::').to_sym
+    assert_raise(NameError) {self.class.const_get(const)}
+  end
+
+  def test_nested_get_const_missing
+    classes = []
+    klass = Class.new {
+      define_singleton_method(:const_missing) { |name|
+	classes << name
+	klass
+      }
+    }
+    klass.const_get("Foo::Bar::Baz")
+    assert_equal [:Foo, :Bar, :Baz], classes
+  end
+
+  def test_nested_bad_class
+    assert_raises(TypeError) do
+      self.class.const_get([User, 'USER', 'Foo'].join('::'))
+    end
+  end
+
   def test_const_set
     assert(!Other.const_defined?(:KOALA))
     Other.const_set(:KOALA, 99)
@@ -277,24 +309,39 @@ class TestModule < Test::Unit::TestCase
     assert_equal([:MIXIN, :USER], User.constants.sort)
   end
 
-  def test_self_initialize_copy
-    bug9535 = '[ruby-dev:47989] [Bug #9535]'
-    m = Module.new do
-      def foo
-        :ok
-      end
-      initialize_copy(self)
-    end
-    assert_equal(:ok, Object.new.extend(m).foo, bug9535)
+  def test_dup
+    bug6454 = '[ruby-core:45132]'
+
+    a = Module.new
+    Other.const_set :BUG6454, a
+
+    original = Other::BUG6454.inspect
+
+    b = a.dup
+    Other.const_set :BUG6454_dup, b
+
+    assert_equal "TestModule::Other::BUG6454_dup", b.inspect, bug6454
+  end
+
+  def test_dup_anonymous
+    bug6454 = '[ruby-core:45132]'
+
+    a = Module.new
+    original = a.inspect
+
+    b = a.dup
+
+    refute_equal original, b.inspect, bug6454
   end
 
   def test_included_modules
     assert_equal([], Mixin.included_modules)
     assert_equal([Mixin], User.included_modules)
-    assert_equal([Kernel],
-                 remove_minitest_mixins(remove_rake_mixins(remove_json_mixins(remove_pp_mixins(Object.included_modules)))))
-    assert_equal([Comparable, Kernel],
-                 remove_minitest_mixins(remove_rake_mixins(remove_json_mixins(remove_pp_mixins(String.included_modules)))))
+
+    mixins = Object.included_modules - [Kernel]
+    mixins << JSON::Ext::Generator::GeneratorMethods::String if defined?(JSON::Ext::Generator::GeneratorMethods::String)
+    assert_equal([Kernel], Object.included_modules - mixins)
+    assert_equal([Comparable, Kernel], String.included_modules - mixins)
   end
 
   def test_instance_methods
@@ -366,6 +413,27 @@ class TestModule < Test::Unit::TestCase
     assert_equal("TestModule::User",   User.name)
   end
 
+  def test_classpath
+    m = Module.new
+    n = Module.new
+    m.const_set(:N, n)
+    assert_nil(m.name)
+    assert_nil(n.name)
+    assert_equal([:N], m.constants)
+    m.module_eval("module O end")
+    assert_equal([:N, :O], m.constants)
+    m.module_eval("class C; end")
+    assert_equal([:N, :O, :C], m.constants)
+    assert_nil(m::N.name)
+    assert_match(/\A#<Module:.*>::O\z/, m::O.name)
+    assert_match(/\A#<Module:.*>::C\z/, m::C.name)
+    self.class.const_set(:M, m)
+    prefix = self.class.name + "::M::"
+    assert_equal(prefix+"N", m.const_get(:N).name)
+    assert_equal(prefix+"O", m.const_get(:O).name)
+    assert_equal(prefix+"C", m.const_get(:C).name)
+  end
+
   def test_private_class_method
     assert_raise(ExpectedException) { AClass.cm1 }
     assert_raise(ExpectedException) { AClass.cm3 }
@@ -487,6 +555,9 @@ class TestModule < Test::Unit::TestCase
     assert_raise(NameError) { c2::Bar }
     assert_raise(NameError) { c2.const_get(:Bar) }
     assert_raise(NameError) { c2.const_get(:Bar, false) }
+    assert_raise(NameError) { c2.const_get("Bar", false) }
+    assert_raise(NameError) { c2.const_get("BaR11", false) }
+    assert_raise(NameError) { Object.const_get("BaR11", false) }
 
     c1.instance_eval do
       def const_missing(x)
@@ -498,6 +569,11 @@ class TestModule < Test::Unit::TestCase
     assert_equal(:Bar, c2::Bar)
     assert_equal(:Bar, c2.const_get(:Bar))
     assert_equal(:Bar, c2.const_get(:Bar, false))
+    assert_equal(:Bar, c2.const_get("Bar"))
+    assert_equal(:Bar, c2.const_get("Bar", false))
+
+    v = c2.const_get("Bar11", false)
+    assert_equal("Bar11".to_sym, v)
 
     assert_raise(NameError) { c1.const_get(:foo) }
   end
@@ -508,8 +584,23 @@ class TestModule < Test::Unit::TestCase
   end
 
   def test_const_get_invalid_name
+    c1 = Class.new
+    assert_raise(NameError) { c1.const_get(:foo) }
+    bug5084 = '[ruby-dev:44200]'
+    assert_raise(TypeError, bug5084) { c1.const_get(1) }
+    bug7574 = '[ruby-dev:46749]'
+    e = assert_raise(NameError) { Object.const_get("String\0") }
+    assert_equal("wrong constant name \"String\\u0000\"", e.message, bug7574)
+  end
+
+  def test_const_defined_invalid_name
     c1 = Class.new
     assert_raise(NameError) { c1.const_defined?(:foo) }
+    bug5084 = '[ruby-dev:44200]'
+    assert_raise(TypeError, bug5084) { c1.const_defined?(1) }
+    bug7574 = '[ruby-dev:46749]'
+    e = assert_raise(NameError) { Object.const_defined?("String\0") }
+    assert_equal("wrong constant name \"String\\u0000\"", e.message, bug7574)
   end
 
   def test_const_get_no_inherited
@@ -647,7 +738,7 @@ class TestModule < Test::Unit::TestCase
     end
 
     %w(object_id __send__ initialize).each do |n|
-      assert_in_out_err([], <<-INPUT, [], /warning: undefining `#{n}' may cause serious problems$/)
+      assert_in_out_err([], <<-INPUT, [], %r"warning: undefining `#{n}' may cause serious problems$")
         $VERBOSE = false
         Class.new.instance_eval { undef_method(:#{n}) }
       INPUT
@@ -680,19 +771,6 @@ class TestModule < Test::Unit::TestCase
     m.instance_eval { remove_const(:Foo) }
   end
 
-  class Bug9413
-    class << self
-      Foo = :foo
-    end
-  end
-
-  def test_singleton_constants
-    bug9413 = '[ruby-core:59763] [Bug #9413]'
-    c = Bug9413.singleton_class
-    assert_include(c.constants(true), :Foo, bug9413)
-    assert_include(c.constants(false), :Foo, bug9413)
-  end
-
   def test_frozen_class
     m = Module.new
     m.freeze
@@ -1021,7 +1099,7 @@ class TestModule < Test::Unit::TestCase
       Module.new do
         define_method(:foo) do end
         alias bar foo
-        alias barf oo
+        alias bar foo
       end
     end
     assert_equal("", stderr)
@@ -1087,6 +1165,14 @@ class TestModule < Test::Unit::TestCase
     assert_equal(1, c.x, bug3406)
   end
 
+  def test_attr_writer_with_no_arguments
+    bug8540 = "[ruby-core:55543]"
+    c = Class.new do
+      attr_writer :foo
+    end
+    assert_raise(ArgumentError) { c.new.send :foo= }
+  end
+
   def test_private_constant
     c = Class.new
     c.const_set(:FOO, "foo")
@@ -1114,6 +1200,15 @@ class TestModule < Test::Unit::TestCase
     assert_equal("bar", c.class_eval("BAR"))
   end
 
+  def test_private_constant_with_no_args
+    assert_in_out_err([], <<-RUBY, [], ["-:3: warning: private_constant with no argument is just ignored"])
+      $-w = true
+      class X
+        private_constant
+      end
+    RUBY
+  end
+
   class PrivateClass
   end
   private_constant :PrivateClass
@@ -1245,6 +1340,402 @@ class TestModule < Test::Unit::TestCase
     assert_in_out_err([], src, ["NameError"], [])
   end
 
+  module M0
+    def m1; [:M0] end
+  end
+  module M1
+    def m1; [:M1, *super] end
+  end
+  module M2
+    def m1; [:M2, *super] end
+  end
+  M3 = Module.new do
+    def m1; [:M3, *super] end
+  end
+  module M4
+    def m1; [:M4, *super] end
+  end
+  class C
+    def m1; end
+  end
+  class C0 < C
+    include M0
+    prepend M1
+    def m1; [:C0, *super] end
+  end
+  class C1 < C0
+    prepend M2, M3
+    include M4
+    def m1; [:C1, *super] end
+  end
+
+  def test_prepend
+    obj = C0.new
+    expected = [:M1,:C0,:M0]
+    assert_equal(expected, obj.m1)
+    obj = C1.new
+    expected = [:M2,:M3,:C1,:M4,:M1,:C0,:M0]
+    assert_equal(expected, obj.m1)
+  end
+
+  def test_prepend_inheritance
+    bug6654 = '[ruby-core:45914]'
+    a = labeled_module("a")
+    b = labeled_module("b") {include a}
+    c = labeled_class("c") {prepend b}
+    assert_operator(c, :<, b, bug6654)
+    assert_operator(c, :<, a, bug6654)
+    bug8357 = '[ruby-core:54736] [Bug #8357]'
+    b = labeled_module("b") {prepend a}
+    c = labeled_class("c") {include b}
+    assert_operator(c, :<, b, bug8357)
+    assert_operator(c, :<, a, bug8357)
+    bug8357 = '[ruby-core:54742] [Bug #8357]'
+    assert_kind_of(b, c.new, bug8357)
+  end
+
+  def test_prepend_instance_methods
+    bug6655 = '[ruby-core:45915]'
+    assert_equal(Object.instance_methods, Class.new {prepend Module.new}.instance_methods, bug6655)
+  end
+
+  def test_prepend_singleton_methods
+    o = Object.new
+    o.singleton_class.class_eval {prepend Module.new}
+    assert_equal([], o.singleton_methods)
+  end
+
+  def test_prepend_remove_method
+    c = Class.new do
+      prepend Module.new {def foo; end}
+    end
+    assert_raise(NameError) do
+      c.class_eval do
+        remove_method(:foo)
+      end
+    end
+    c.class_eval do
+      def foo; end
+    end
+    removed = nil
+    c.singleton_class.class_eval do
+      define_method(:method_removed) {|id| removed = id}
+    end
+    assert_nothing_raised(NoMethodError, NameError, '[Bug #7843]') do
+      c.class_eval do
+        remove_method(:foo)
+      end
+    end
+    assert_equal(:foo, removed)
+  end
+
+  def test_prepend_class_ancestors
+    bug6658 = '[ruby-core:45919]'
+    m = labeled_module("m")
+    c = labeled_class("c") {prepend m}
+    assert_equal([m, c], c.ancestors[0, 2], bug6658)
+
+    bug6662 = '[ruby-dev:45868]'
+    c2 = labeled_class("c2", c)
+    anc = c2.ancestors
+    assert_equal([c2, m, c, Object], anc[0..anc.index(Object)], bug6662)
+  end
+
+  def test_prepend_module_ancestors
+    bug6659 = '[ruby-dev:45861]'
+    m0 = labeled_module("m0") {def x; [:m0, *super] end}
+    m1 = labeled_module("m1") {def x; [:m1, *super] end; prepend m0}
+    m2 = labeled_module("m2") {def x; [:m2, *super] end; prepend m1}
+    c0 = labeled_class("c0") {def x; [:c0] end}
+    c1 = labeled_class("c1") {def x; [:c1] end; prepend m2}
+    c2 = labeled_class("c2", c0) {def x; [:c2, *super] end; include m2}
+
+    assert_equal([m0, m1], m1.ancestors, bug6659)
+
+    bug6662 = '[ruby-dev:45868]'
+    assert_equal([m0, m1, m2], m2.ancestors, bug6662)
+    assert_equal([m0, m1, m2, c1], c1.ancestors[0, 4], bug6662)
+    assert_equal([:m0, :m1, :m2, :c1], c1.new.x)
+    assert_equal([c2, m0, m1, m2, c0], c2.ancestors[0, 5], bug6662)
+    assert_equal([:c2, :m0, :m1, :m2, :c0], c2.new.x)
+
+    m3 = labeled_module("m3") {include m1; prepend m1}
+    assert_equal([m3, m0, m1], m3.ancestors)
+    m3 = labeled_module("m3") {prepend m1; include m1}
+    assert_equal([m0, m1, m3], m3.ancestors)
+    m3 = labeled_module("m3") {prepend m1; prepend m1}
+    assert_equal([m0, m1, m3], m3.ancestors)
+    m3 = labeled_module("m3") {include m1; include m1}
+    assert_equal([m3, m0, m1], m3.ancestors)
+  end
+
+  def labeled_module(name, &block)
+    Module.new do
+      singleton_class.class_eval {define_method(:to_s) {name}; alias inspect to_s}
+      class_eval(&block) if block
+    end
+  end
+
+  def labeled_class(name, superclass = Object, &block)
+    Class.new(superclass) do
+      singleton_class.class_eval {define_method(:to_s) {name}; alias inspect to_s}
+      class_eval(&block) if block
+    end
+  end
+
+  def test_prepend_instance_methods_false
+    bug6660 = '[ruby-dev:45863]'
+    assert_equal([:m1], Class.new{ prepend Module.new; def m1; end }.instance_methods(false), bug6660)
+    assert_equal([:m1], Class.new(Class.new{def m2;end}){ prepend Module.new; def m1; end }.instance_methods(false), bug6660)
+  end
+
+  def test_cyclic_prepend
+    bug7841 = '[ruby-core:52205] [Bug #7841]'
+    m1 = Module.new
+    m2 = Module.new
+    m1.instance_eval { prepend(m2) }
+    assert_raise(ArgumentError, bug7841) do
+      m2.instance_eval { prepend(m1) }
+    end
+  end
+
+  def test_prepend_optmethod
+    bug7983 = '[ruby-dev:47124] [Bug #7983]'
+    assert_separately [], %{
+      module M
+        def /(other)
+          to_f / other
+        end
+      end
+      Fixnum.send(:prepend, M)
+      assert_equal(0.5, 1 / 2, "#{bug7983}")
+    }
+    assert_equal(0, 1 / 2)
+  end
+
+  def test_prepend_visibility
+    bug8005 = '[ruby-core:53106] [Bug #8005]'
+    c = Class.new do
+      prepend Module.new {}
+      def foo() end
+      protected :foo
+    end
+    a = c.new
+    assert_respond_to a, [:foo, true]
+    assert_nothing_raised(NoMethodError) {a.send :foo}
+  end
+
+  def test_prepend_visibility_inherited
+    bug8238 = '[ruby-core:54105] [Bug #8238]'
+    assert_separately [], <<-"end;", timeout: 3
+      class A
+        def foo() A; end
+        private :foo
+      end
+      class B < A
+        public :foo
+        prepend Module.new
+      end
+      assert_equal(A, B.new.foo, "#{bug8238}")
+    end;
+  end
+
+  def test_prepend_included_modules
+    bug8025 = '[ruby-core:53158] [Bug #8025]'
+    mixin = labeled_module("mixin")
+    c = labeled_module("c") {prepend mixin}
+    im = c.included_modules
+    assert_not_include(im, c, bug8025)
+    assert_include(im, mixin, bug8025)
+    c1 = labeled_class("c1") {prepend mixin}
+    c2 = labeled_class("c2", c1)
+    im = c2.included_modules
+    assert_not_include(im, c1, bug8025)
+    assert_not_include(im, c2, bug8025)
+    assert_include(im, mixin, bug8025)
+  end
+
+  def test_class_variables
+    m = Module.new
+    m.class_variable_set(:@@foo, 1)
+    m2 = Module.new
+    m2.send(:include, m)
+    m2.class_variable_set(:@@bar, 2)
+    assert_equal([:@@foo], m.class_variables)
+    assert_equal([:@@bar, :@@foo], m2.class_variables)
+    assert_equal([:@@bar, :@@foo], m2.class_variables(true))
+    assert_equal([:@@bar], m2.class_variables(false))
+  end
+
+  Bug6891 = '[ruby-core:47241]'
+
+  def test_extend_module_with_protected_method
+    list = []
+
+    x = Class.new {
+      @list = list
+
+      extend Module.new {
+        protected
+
+        def inherited(klass)
+          @list << "protected"
+          super(klass)
+        end
+      }
+
+      extend Module.new {
+        def inherited(klass)
+          @list << "public"
+          super(klass)
+        end
+      }
+    }
+
+    assert_nothing_raised(NoMethodError, Bug6891) {Class.new(x)}
+    assert_equal(['public', 'protected'], list)
+  end
+
+  def test_extend_module_with_protected_bmethod
+    list = []
+
+    x = Class.new {
+      extend Module.new {
+        protected
+
+        define_method(:inherited) do |klass|
+          list << "protected"
+          super(klass)
+        end
+      }
+
+      extend Module.new {
+        define_method(:inherited) do |klass|
+          list << "public"
+          super(klass)
+        end
+      }
+    }
+
+    assert_nothing_raised(NoMethodError, Bug6891) {Class.new(x)}
+    assert_equal(['public', 'protected'], list)
+  end
+
+  def test_invalid_attr
+    %w[
+      foo?
+      @foo
+      @@foo
+      $foo
+    ].each do |name|
+      assert_raises(NameError) do
+        Module.new { attr_accessor name.to_sym }
+      end
+    end
+  end
+
+  class AttrTest
+    class << self
+      attr_accessor :cattr
+    end
+    attr_accessor :iattr
+    def ivar
+      @ivar
+    end
+  end
+
+  def test_uninitialized_instance_variable
+    a = AttrTest.new
+    stderr = EnvUtil.verbose_warning do
+      assert_nil(a.ivar)
+    end
+    assert_match(/instance variable @ivar not initialized/, stderr)
+    a.instance_variable_set(:@ivar, 42)
+    stderr = EnvUtil.verbose_warning do
+      assert_equal(42, a.ivar)
+    end
+    assert_equal("", stderr)
+  end
+
+  def test_uninitialized_attr
+    a = AttrTest.new
+    stderr = EnvUtil.verbose_warning do
+      assert_nil(a.iattr)
+    end
+    assert_equal("", stderr)
+    a.iattr = 42
+    stderr = EnvUtil.verbose_warning do
+      assert_equal(42, a.iattr)
+    end
+    assert_equal("", stderr)
+  end
+
+  def test_uninitialized_attr_class
+    stderr = EnvUtil.verbose_warning do
+      assert_nil(AttrTest.cattr)
+    end
+    assert_equal("", stderr)
+    AttrTest.cattr = 42
+    stderr = EnvUtil.verbose_warning do
+      assert_equal(42, AttrTest.cattr)
+    end
+    assert_equal("", stderr)
+  end
+
+  def test_uninitialized_attr_non_object
+    a = Class.new(Array) do
+      attr_accessor :iattr
+    end.new
+    stderr = EnvUtil.verbose_warning do
+      assert_nil(a.iattr)
+    end
+    assert_equal("", stderr)
+    a.iattr = 42
+    stderr = EnvUtil.verbose_warning do
+      assert_equal(42, a.iattr)
+    end
+    assert_equal("", stderr)
+  end
+
+  def test_remove_const
+    m = Module.new
+    assert_raise(NameError){ m.instance_eval { remove_const(:__FOO__) } }
+  end
+
+  def test_private_top_methods
+    assert_top_method_is_private(:include)
+    assert_top_method_is_private(:public)
+    assert_top_method_is_private(:private)
+    assert_top_method_is_private(:define_method)
+  end
+
+  module PrivateConstantReopen
+    PRIVATE_CONSTANT = true
+    private_constant :PRIVATE_CONSTANT
+  end
+
+  def test_private_constant_reopen
+    assert_raise(NameError) do
+      eval <<-EOS, TOPLEVEL_BINDING
+        module TestModule::PrivateConstantReopen::PRIVATE_CONSTANT
+        end
+      EOS
+    end
+    assert_raise(NameError) do
+      eval <<-EOS, TOPLEVEL_BINDING
+        class TestModule::PrivateConstantReopen::PRIVATE_CONSTANT
+        end
+      EOS
+    end
+  end
+
+  def test_anonymous_module_public_class_method
+    bug8284 = '[ruby-core:54404] [Bug #8284]'
+    assert_raise(NoMethodError) {Object.define_method}
+    Module.new.public_class_method(:define_method)
+    assert_raise(NoMethodError, bug8284) {Object.define_method}
+  end
+
   def test_include_module_with_constants_invalidates_method_cache
     assert_in_out_err([], <<-RUBY, %w(123 456), [])
       A = 123
@@ -1264,4 +1755,16 @@ class TestModule < Test::Unit::TestCase
       puts Foo.a
     RUBY
   end
+
+  private
+
+  def assert_top_method_is_private(method)
+    top = eval("self", TOPLEVEL_BINDING)
+    methods = top.singleton_class.private_instance_methods(false)
+    assert(methods.include?(method), "#{method} should be private")
+
+    assert_in_out_err([], <<-INPUT, [], /private method `#{method}' called for main:Object \(NoMethodError\)/)
+      self.#{method}
+    INPUT
+  end
 end
