@@ -1,3 +1,4 @@
+# coding: US-ASCII
 require 'test/unit'
 require 'stringio'
 
@@ -12,7 +13,7 @@ class TestParse < Test::Unit::TestCase
   end
 
   def test_else_without_rescue
-    x = eval <<-END
+    x = eval <<-END, nil, __FILE__, __LINE__+1
       begin
       else
         42
@@ -20,10 +21,10 @@ class TestParse < Test::Unit::TestCase
     END
     assert_equal(42, x)
   end
-  
+
   def test_alias_backref
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         alias $foo $1
       END
     end
@@ -36,7 +37,7 @@ class TestParse < Test::Unit::TestCase
     a = false
     b = c = d = true
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         a &&= t.foo 42
         b &&= t.foo 42
         c &&= t.foo nil
@@ -51,7 +52,7 @@ class TestParse < Test::Unit::TestCase
 
     a = [nil, nil, true, true]
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         a[0] ||= t.foo 42
         a[1] &&= t.foo 42
         a[2] ||= t.foo 42
@@ -67,7 +68,7 @@ class TestParse < Test::Unit::TestCase
     o.foo = o.Foo = o::baz = nil
     o.bar = o.Bar = o::qux = 1
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         o.foo ||= t.foo 42
         o.bar &&= t.foo 42
         o.Foo ||= t.foo 42
@@ -81,7 +82,7 @@ class TestParse < Test::Unit::TestCase
     assert_equal([42, 42], [o::baz, o::qux])
 
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         $1 ||= t.foo 42
       END
     end
@@ -90,7 +91,7 @@ class TestParse < Test::Unit::TestCase
 
     a = b = nil
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         a = t.bar "foo" do
           "bar"
         end.gsub "ob", "OB"
@@ -104,16 +105,16 @@ class TestParse < Test::Unit::TestCase
 
     a = nil
     assert_nothing_raised do
-      t.instance_eval <<-END
-        a = bar "foo" { "bar" }
+      t.instance_eval <<-END, __FILE__, __LINE__+1
+        a = bar "foo" do "bar" end
       END
     end
     assert_equal("foobar", a)
 
     a = nil
     assert_nothing_raised do
-      eval <<-END
-        a = t::bar "foo" { "bar" }
+      eval <<-END, nil, __FILE__, __LINE__+1
+        a = t::bar "foo" do "bar" end
       END
     end
     assert_equal("foobar", a)
@@ -136,7 +137,7 @@ class TestParse < Test::Unit::TestCase
     end
 
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         c::foo, c::bar = 1, 2
         c.Foo, c.Bar = 1, 2
         c::FOO, c::BAR = 1, 2
@@ -149,7 +150,7 @@ class TestParse < Test::Unit::TestCase
 
   def test_dynamic_constant_assignment
     assert_raise(SyntaxError) do
-      Object.new.instance_eval <<-END
+      Object.new.instance_eval <<-END, __FILE__, __LINE__+1
         def foo
           self::FOO, self::BAR = 1, 2
           ::FOO, ::BAR = 1, 2
@@ -158,13 +159,13 @@ class TestParse < Test::Unit::TestCase
     end
 
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         $1, $2 = 1, 2
       END
     end
 
     assert_raise(SyntaxError) do
-      Object.new.instance_eval <<-END
+      Object.new.instance_eval <<-END, __FILE__, __LINE__+1
         def foo
           ::FOO = 1
         end
@@ -172,16 +173,18 @@ class TestParse < Test::Unit::TestCase
     end
 
     c = Class.new
-    assert_raise(SyntaxError) do
-      eval <<-END
+    assert_nothing_raised(SyntaxError) do
+      eval <<-END, nil, __FILE__, __LINE__+1
+      if false
         c::FOO &= 1
         ::FOO &= 1
+      end
       END
     end
 
     c = Class.new
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         $1 &= 1
       END
     end
@@ -189,13 +192,13 @@ class TestParse < Test::Unit::TestCase
 
   def test_class_module
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         class foo; end
       END
     end
-    
+
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         def foo
           class Foo; end
           module Bar; end
@@ -204,7 +207,7 @@ class TestParse < Test::Unit::TestCase
     end
 
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         class Foo Bar; end
       END
     end
@@ -217,7 +220,7 @@ class TestParse < Test::Unit::TestCase
 
     a = nil
     assert_nothing_raised do
-      o.instance_eval <<-END
+      o.instance_eval <<-END, __FILE__, __LINE__+1
         undef >, /
       END
     end
@@ -231,7 +234,7 @@ class TestParse < Test::Unit::TestCase
     o.foo = o.Foo = o::baz = nil
     o.bar = o.Bar = o::qux = 1
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         o.foo ||= 42
         o.bar &&= 42
         o.Foo ||= 42
@@ -246,7 +249,7 @@ class TestParse < Test::Unit::TestCase
 
     a = nil
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         a = -2.0 ** 2
       END
     end
@@ -259,8 +262,8 @@ class TestParse < Test::Unit::TestCase
 
     a = nil
     assert_nothing_raised do
-      eval <<-END
-        o.foo 1 {|; a| a = 42 }
+      eval <<-END, nil, __FILE__, __LINE__+1
+        o.foo 1 do|; a| a = 42 end
       END
     end
     assert_nil(a)
@@ -268,25 +271,25 @@ class TestParse < Test::Unit::TestCase
 
   def test_bad_arg
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         def foo(FOO); end
       END
     end
 
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         def foo(@foo); end
       END
     end
 
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         def foo($foo); end
       END
     end
 
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         def foo(@@foo); end
       END
     end
@@ -295,7 +298,7 @@ class TestParse < Test::Unit::TestCase
     def o.foo(*r); yield(*r); end
 
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         o.foo 1 {|; @a| @a = 42 }
       END
     end
@@ -304,7 +307,7 @@ class TestParse < Test::Unit::TestCase
   def test_do_lambda
     a = b = nil
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         a = -> do
           b = 42
         end
@@ -320,7 +323,7 @@ class TestParse < Test::Unit::TestCase
 
     a = b = nil
     assert_nothing_raised do
-      o.instance_eval <<-END
+      o.instance_eval <<-END, __FILE__, __LINE__+1
         a = foo 1 do 42 end.to_s
         b = foo 1 do 42 end::to_s
       END
@@ -332,7 +335,7 @@ class TestParse < Test::Unit::TestCase
   def test_call_method
     a = b = nil
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         a = proc {|x| x + "bar" }.("foo")
         b = proc {|x| x + "bar" }::("foo")
       END
@@ -350,7 +353,7 @@ class TestParse < Test::Unit::TestCase
   def test_words
     assert_equal([], %W( ))
   end
-  
+
   def test_dstr
     @@foo = 1
     assert_equal("foo 1 bar", "foo #@@foo bar")
@@ -358,6 +361,17 @@ class TestParse < Test::Unit::TestCase
     assert_equal("foo 1 bar", "foo #$1 bar")
   end
 
+  def test_dstr_disallowd_variable
+    bug8375 = '[ruby-core:54885] [Bug #8375]'
+    %w[@ @1 @@. @@ @@1 @@. $ $%].each do |src|
+      src = '#'+src+' '
+      str = assert_nothing_raised(SyntaxError, "#{bug8375} #{src.dump}") do
+        break eval('"'+src+'"')
+      end
+      assert_equal(src, str, bug8375)
+    end
+  end
+
   def test_dsym
     assert_nothing_raised { eval(':""') }
   end
@@ -366,7 +380,7 @@ class TestParse < Test::Unit::TestCase
     o = Object.new
 
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         def o.foo(a=42,*r,z,&b); b.call(r.inject(a*1000+z*100, :+)); end
       END
     end
@@ -377,7 +391,7 @@ class TestParse < Test::Unit::TestCase
     assert_raise(ArgumentError) { o.foo() }
 
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         def o.foo(a=42,z,&b); b.call(a*1000+z*100); end
       END
     end
@@ -386,7 +400,7 @@ class TestParse < Test::Unit::TestCase
     assert_raise(ArgumentError) { o.foo() }
 
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         def o.foo(*r,z,&b); b.call(r.inject(z*100, :+)); end
       END
     end
@@ -398,19 +412,19 @@ class TestParse < Test::Unit::TestCase
 
   def test_duplicate_argument
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         1.times {|&b?| }
       END
     end
 
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         1.times {|a, a|}
       END
     end
 
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         def foo(a, a); end
       END
     end
@@ -418,7 +432,7 @@ class TestParse < Test::Unit::TestCase
 
   def test_define_singleton_error
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         def ("foo").foo; end
       END
     end
@@ -428,17 +442,17 @@ class TestParse < Test::Unit::TestCase
     t = Object.new
 
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         def t.`(x); "foo" + x + "bar"; end
       END
     end
     a = b = nil
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         a = t.` "zzz"
         1.times {|;z| t.` ("zzz") }
       END
-      t.instance_eval <<-END
+      t.instance_eval <<-END, __FILE__, __LINE__+1
         b = `zzz`
       END
     end
@@ -491,14 +505,16 @@ class TestParse < Test::Unit::TestCase
   end
 
   def test_symbol
-    assert_raise(SyntaxError) do
-      eval ":'foo\0bar'"
+    bug = '[ruby-dev:41447]'
+    sym = "foo\0bar".to_sym
+    assert_nothing_raised(SyntaxError, bug) do
+      assert_equal(sym, eval(":'foo\0bar'"))
     end
-    assert_raise(SyntaxError) do
-      eval ':"foo\u0000bar"'
+    assert_nothing_raised(SyntaxError, bug) do
+      assert_equal(sym, eval(':"foo\u0000bar"'))
     end
-    assert_raise(SyntaxError) do
-      eval ':"foo\u{0}bar"'
+    assert_nothing_raised(SyntaxError, bug) do
+      assert_equal(sym, eval(':"foo\u{0}bar"'))
     end
     assert_raise(SyntaxError) do
       eval ':"foo\u{}bar"'
@@ -507,7 +523,7 @@ class TestParse < Test::Unit::TestCase
 
   def test_parse_string
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
 /
       END
     end
@@ -517,22 +533,23 @@ class TestParse < Test::Unit::TestCase
     x = nil
 
     assert_raise(SyntaxError) do
-      eval %q(
-<<FOO
+      eval %Q(
+<\<FOO
       )
     end
 
-    assert_raise(SyntaxError) do
-      eval %q(
+    assert_nothing_raised(SyntaxError) do
+      x = eval %q(
 <<FOO
 #$
 FOO
       )
     end
+    assert_equal "\#$\n", x
 
     assert_raise(SyntaxError) do
-      eval %q(
-<<"
+      eval %Q(
+<\<\"
       )
     end
 
@@ -548,17 +565,18 @@ FOO
       )
     end
 
-    assert_raise(SyntaxError) do
-      eval %q(
+    assert_nothing_raised(SyntaxError) do
+      x = eval %q(
 <<FOO
 #$
 foo
 FOO
       )
     end
+    assert_equal "\#$\nfoo\n", x
 
     assert_nothing_raised do
-      eval "x = <<FOO\r\n1\r\nFOO"
+      eval "x = <<""FOO\r\n1\r\nFOO"
     end
     assert_equal("1\n", x)
   end
@@ -566,7 +584,7 @@ FOO
   def test_magic_comment
     x = nil
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
 # coding = utf-8
 x = __ENCODING__
       END
@@ -574,7 +592,7 @@ x = __ENCODING__
     assert_equal(Encoding.find("UTF-8"), x)
 
     assert_raise(ArgumentError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
 # coding = foobarbazquxquux_dummy_enconding
 x = __ENCODING__
       END
@@ -593,7 +611,7 @@ x = __ENCODING__
   def test_dot_in_next_line
     x = nil
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         x = 1
         .to_s
       END
@@ -609,7 +627,7 @@ x = __ENCODING__
 
   def test_embedded_rd
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
 =begin
       END
     end
@@ -640,8 +658,11 @@ x = __ENCODING__
   end
 
   def test_invalid_char
+    bug10117 = '[ruby-core:64243] [Bug #10117]'
+    invalid_char = /Invalid char `\\x01'/
     x = 1
-    assert_equal(1, eval("\x01x"))
+    assert_in_out_err(%W"-e \x01x", "", [], invalid_char, bug10117)
+    assert_syntax_error("\x01x", invalid_char, bug10117)
     assert_equal(nil, eval("\x04x"))
   end
 
@@ -673,7 +694,7 @@ x = __ENCODING__
       eval %q(__ENCODING__ = 1)
     end
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         def foo
           FOO = 1
         end
@@ -683,7 +704,7 @@ x = __ENCODING__
 
   def test_block_dup
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         foo(&proc{}) {}
       END
     end
@@ -691,7 +712,7 @@ x = __ENCODING__
 
   def test_set_backref
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         $& = 1
       END
     end
@@ -704,7 +725,7 @@ x = __ENCODING__
     end
     r = nil
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         o[&proc{|x| r = x }] = 1
       END
     end
@@ -736,23 +757,23 @@ x = __ENCODING__
 
     o = Object.new
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         x = def o.foo; end
       END
     end
-    assert_equal($stderr.string.lines.to_a.size, 14)
+    assert_equal(14, $stderr.string.lines.to_a.size)
     $stderr = stderr
   end
 
   def test_assign_in_conditional
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         (x, y = 1, 2) ? 1 : 2
       END
     end
 
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         if @x = true
           1
         else
@@ -764,33 +785,33 @@ x = __ENCODING__
 
   def test_literal_in_conditional
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         "foo" ? 1 : 2
       END
     end
 
     assert_nothing_raised do
       x = "bar"
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         /foo#{x}baz/ ? 1 : 2
       END
     end
 
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         (true..false) ? 1 : 2
       END
     end
 
     assert_nothing_raised do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         ("foo".."bar") ? 1 : 2
       END
     end
 
     assert_nothing_raised do
       x = "bar"
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         :"foo#{"x"}baz" ? 1 : 2
       END
     end
@@ -798,7 +819,7 @@ x = __ENCODING__
 
   def test_no_blockarg
     assert_raise(SyntaxError) do
-      eval <<-END
+      eval <<-END, nil, __FILE__, __LINE__+1
         yield(&:+)
       END
     end
@@ -823,4 +844,18 @@ x = __ENCODING__
       c.instance_eval { remove_class_variable(:@var) }
     end
   end
+
+  def test_method_block_location
+    bug5614 = '[ruby-core:40936]'
+    expected = nil
+    e = assert_raise(NoMethodError) do
+      1.times do
+        expected = __LINE__+1
+      end.print do
+        #
+      end
+    end
+    actual = e.backtrace.first[/\A#{Regexp.quote(__FILE__)}:(\d+):/o, 1].to_i
+    assert_equal(expected, actual, bug5614)
+  end
 end
