@@ -1,5 +1,5 @@
 require 'test/unit'
-require 'tmpdir'
+require 'tempfile'
 require 'pathname'
 require 'timeout'
 require_relative 'envutil'
@@ -303,6 +303,78 @@ class TestProcess < Test::Unit::TestCase
     end
   end
 
+  def _test_execopts_env_popen(cmd)
+    message = cmd.inspect
+    IO.popen({"FOO"=>"BAR"}, cmd) {|io|
+      assert_equal('FOO=BAR', io.read[/^FOO=.*/], message)
+    }
+
+    old = ENV["hmm"]
+    begin
+      ENV["hmm"] = "fufu"
+      IO.popen(cmd) {|io| assert_match(/^hmm=fufu$/, io.read, message)}
+      IO.popen({"hmm"=>""}, cmd) {|io| assert_match(/^hmm=$/, io.read, message)}
+      IO.popen({"hmm"=>nil}, cmd) {|io| assert_not_match(/^hmm=/, io.read, message)}
+      ENV["hmm"] = ""
+      IO.popen(cmd) {|io| assert_match(/^hmm=$/, io.read, message)}
+      IO.popen({"hmm"=>""}, cmd) {|io| assert_match(/^hmm=$/, io.read, message)}
+      IO.popen({"hmm"=>nil}, cmd) {|io| assert_not_match(/^hmm=/, io.read, message)}
+      ENV["hmm"] = nil
+      IO.popen(cmd) {|io| assert_not_match(/^hmm=/, io.read, message)}
+      IO.popen({"hmm"=>""}, cmd) {|io| assert_match(/^hmm=$/, io.read, message)}
+      IO.popen({"hmm"=>nil}, cmd) {|io| assert_not_match(/^hmm=/, io.read, message)}
+    ensure
+      ENV["hmm"] = old
+    end
+  end
+
+  def test_execopts_env_popen_vector
+    _test_execopts_env_popen(ENVCOMMAND)
+  end
+
+  def test_execopts_env_popen_string
+    with_tmpchdir do |d|
+      open('test-script', 'w') do |f|
+        ENVCOMMAND.each_with_index do |cmd, i|
+          next if i.zero? or cmd == "-e"
+          f.puts cmd
+        end
+      end
+      _test_execopts_env_popen("#{RUBY} test-script")
+    end
+  end
+
+  def test_execopts_preserve_env_on_exec_failure
+    with_tmpchdir {|d|
+      write_file 's', <<-"End"
+        ENV["mgg"] = nil
+        prog = "./nonexistent"
+        begin
+          Process.exec({"mgg" => "mggoo"}, [prog, prog])
+        rescue Errno::ENOENT
+        end
+        open('out', 'w') {|f|
+          f.print ENV["mgg"].inspect
+        }
+      End
+      system(RUBY, 's')
+      assert_equal(nil.inspect, File.read('out'),
+        "[ruby-core:44093] [ruby-trunk - Bug #6249]")
+    }
+  end
+
+  def test_execopts_env_single_word
+    with_tmpchdir {|d|
+      open("test_execopts_env_single_word.rb", "w") {|f|
+        f.puts "print ENV['hgga']"
+      }
+      system({"hgga"=>"ugu"}, RUBY,
+             :in => 'test_execopts_env_single_word.rb',
+             :out => 'test_execopts_env_single_word.out')
+      assert_equal('ugu', File.read('test_execopts_env_single_word.out'))
+    }
+  end
+
   def test_execopts_unsetenv_others
     h = {}
     MANDATORY_ENVS.each {|k| e = ENV[k] and h[k] = e}
@@ -327,6 +399,16 @@ class TestProcess < Test::Unit::TestCase
     }
   end
 
+  def test_execopts_open_chdir
+    with_tmpchdir {|d|
+      Dir.mkdir "foo"
+      system(*PWD, :chdir => "foo", :out => "open_chdir_test")
+      assert_file.exist?("open_chdir_test")
+      assert_file.not_exist?("foo/open_chdir_test")
+      assert_equal("#{d}/foo", File.read("open_chdir_test").chomp)
+    }
+  end
+
   UMASK = [RUBY, '-e', 'printf "%04o\n", File.umask']
 
   def test_execopts_umask
@@ -465,8 +547,8 @@ class TestProcess < Test::Unit::TestCase
           h = {}
           ios.length.times {|i| h[ios[i]] = ios[(i-1)%ios.length] }
           h2 = h.invert
-          rios = pipes.map {|r, w| r }
-          wios = pipes.map {|r, w| w }
+          _rios = pipes.map {|r, w| r }
+          wios  = pipes.map {|r, w| w }
           child_wfds = wios.map {|w| h2[w].fileno }
           pid = spawn(RUBY, "-e",
                   "[#{child_wfds.join(',')}].each {|fd| IO.new(fd, 'w').puts fd }", h)
@@ -481,8 +563,8 @@ class TestProcess < Test::Unit::TestCase
           h = {}
           ios.length.times {|i| h[ios[i]] = ios[(i+1)%ios.length] }
           h2 = h.invert
-          rios = pipes.map {|r, w| r }
-          wios = pipes.map {|r, w| w }
+          _rios = pipes.map {|r, w| r }
+          wios  = pipes.map {|r, w| w }
           child_wfds = wios.map {|w| h2[w].fileno }
           pid = spawn(RUBY, "-e",
                   "[#{child_wfds.join(',')}].each {|fd| IO.new(fd, 'w').puts fd }", h)
@@ -590,7 +672,6 @@ class TestProcess < Test::Unit::TestCase
   end
 
   def test_popen_fork
-    return if /freebsd/ =~ RUBY_PLATFORM # this test freeze in FreeBSD
     IO.popen("-") {|io|
       if !io
         puts "fooo"
@@ -604,7 +685,7 @@ class TestProcess < Test::Unit::TestCase
   def test_fd_inheritance
     skip "inheritance of fd other than stdin,stdout and stderr is not supported" if windows?
     with_pipe {|r, w|
-      system(RUBY, '-e', 'IO.new(ARGV[0].to_i, "w").puts(:ba)', w.fileno.to_s)
+      system(RUBY, '-e', 'IO.new(ARGV[0].to_i, "w").puts(:ba)', w.fileno.to_s, w=>w)
       w.close
       assert_equal("ba\n", r.read)
     }
@@ -620,8 +701,9 @@ class TestProcess < Test::Unit::TestCase
 	write_file("s", <<-"End")
 	  exec(#{RUBY.dump}, '-e',
 	       'IO.new(ARGV[0].to_i, "w").puts("bu") rescue nil',
-	       #{w.fileno.to_s.dump})
+	       #{w.fileno.to_s.dump}, :close_others=>false)
 	End
+        w.close_on_exec = false
 	Process.wait spawn(RUBY, "s", :close_others=>false)
 	w.close
 	assert_equal("bu\n", r.read)
@@ -661,6 +743,7 @@ class TestProcess < Test::Unit::TestCase
         File.unlink("err")
       }
       with_pipe {|r, w|
+        w.close_on_exec = false
         Process.wait spawn(RUBY, '-e', 'IO.new(ARGV[0].to_i, "w").puts("bi")', w.fileno.to_s, :close_others=>false)
         w.close
         assert_equal("bi\n", r.read)
@@ -687,6 +770,7 @@ class TestProcess < Test::Unit::TestCase
         Process.wait
       }
       with_pipe {|r, w|
+        w.close_on_exec = false
         io = IO.popen([RUBY, "-e", "STDERR.reopen(STDOUT); IO.new(#{w.fileno}, 'w').puts('mo')", :close_others=>false])
         w.close
         errmsg = io.read
@@ -695,6 +779,7 @@ class TestProcess < Test::Unit::TestCase
         Process.wait
       }
       with_pipe {|r, w|
+        w.close_on_exec = false
         io = IO.popen([RUBY, "-e", "STDERR.reopen(STDOUT); IO.new(#{w.fileno}, 'w').puts('mo')", :close_others=>nil])
         w.close
         errmsg = io.read
@@ -721,6 +806,19 @@ class TestProcess < Test::Unit::TestCase
     end
   end
 
+  def test_execopts_redirect_tempfile
+    bug6269 = '[ruby-core:44181]'
+    Tempfile.open("execopts") do |tmp|
+      pid = assert_nothing_raised(ArgumentError, bug6269) do
+        break spawn(RUBY, "-e", "print $$", out: tmp)
+      end
+      Process.wait(pid)
+      tmp.rewind
+      assert_equal(pid.to_s, tmp.read)
+      tmp.close(true)
+    end
+  end
+
   def test_execopts_duplex_io
     IO.popen("#{RUBY} -e ''", "r+") {|duplex|
       assert_raise(ArgumentError) { system("#{RUBY} -e ''", duplex=>STDOUT) }
@@ -826,6 +924,25 @@ class TestProcess < Test::Unit::TestCase
     }
   end
 
+  def test_popen_wordsplit_beginning_and_trailing_spaces
+    with_tmpchdir {|d|
+      write_file("script", <<-'End')
+        print "fufumm pid=#{$$} ppid=#{Process.ppid}"
+        exit 7
+      End
+      str = " #{RUBY} script "
+      io = IO.popen(str)
+      pid = io.pid
+      result = io.read
+      io.close
+      status = $?
+      assert_equal(pid, status.pid)
+      assert(status.exited?)
+      assert_equal(7, status.exitstatus)
+      assert_equal("fufumm pid=#{status.pid} ppid=#{$$}", result)
+    }
+  end
+
   def test_exec_wordsplit
     with_tmpchdir {|d|
       write_file("script", <<-'End')
@@ -1159,9 +1276,29 @@ class TestProcess < Test::Unit::TestCase
   end
 
   def test_geteuid
+    assert_kind_of(Integer, Process.euid)
+  end
+
+  def test_seteuid
+    assert_nothing_raised(TypeError) {Process.euid += 0}
+  rescue NotImplementedError
+  end
+
+  def test_seteuid_name
+    user = ENV["USER"] or return
+    assert_nothing_raised(TypeError) {Process.euid = user}
+  rescue NotImplementedError
+  end
+
+  def test_getegid
     assert_kind_of(Integer, Process.egid)
   end
 
+  def test_setegid
+    assert_nothing_raised(TypeError) {Process.egid += 0}
+  rescue NotImplementedError
+  end
+
   def test_uid_re_exchangeable_p
     r = Process::UID.re_exchangeable?
     assert(true == r || false == r)
@@ -1233,22 +1370,48 @@ class TestProcess < Test::Unit::TestCase
       open("tmp_script.#{$$}", "w") {|f| f.puts "echo $#: $@"; f.chmod(0755)}
       result = IO.popen(["./tmp_script.#{$$}", "a b", "c"]) {|f| f.read}
       assert_equal("2: a b c\n", result, feature)
+
+      open("tmp_script.#{$$}", "w") {|f| f.puts "echo $hghg"; f.chmod(0755)}
+      result = IO.popen([{"hghg" => "mogomogo"}, "./tmp_script.#{$$}", "a b", "c"]) {|f| f.read}
+      assert_equal("mogomogo\n", result, feature)
+
     end
   end if File.executable?("/bin/sh")
 
-  def test_too_long_path
+  def test_spawn_too_long_path
     bug4314 = '[ruby-core:34842]'
-    exs = [Errno::ENOENT]
-    exs << Errno::E2BIG if defined?(Errno::E2BIG)
-    assert_raise(*exs, bug4314) {Process.spawn("a" * 10_000_000)}
+    assert_fail_too_long_path(%w"echo", bug4314)
   end
 
-  def test_too_long_path2
-    skip
+  def test_aspawn_too_long_path
     bug4315 = '[ruby-core:34833]'
+    assert_fail_too_long_path(%w"echo |", bug4315)
+  end
+
+  def assert_fail_too_long_path((cmd, sep), mesg)
+    sep ||= ""
+    min = 1_000 / (cmd.size + sep.size)
+    cmds = Array.new(min, cmd)
     exs = [Errno::ENOENT]
     exs << Errno::E2BIG if defined?(Errno::E2BIG)
-    assert_raise(*exs, bug4315) {Process.spawn('"a"|'*10_000_000)}
+    EnvUtil.suppress_warning do
+      assert_raise(*exs, mesg) do
+        begin
+          loop do
+            Process.spawn(cmds.join(sep), [STDOUT, STDERR]=>:close)
+            min = [cmds.size, min].max
+            cmds *= 100
+          end
+        rescue NoMemoryError
+          size = cmds.size
+          raise if min >= size - 1
+          min = [min, size /= 2].max
+          cmds[size..-1] = []
+          raise if size < 250
+          retry
+        end
+      end
+    end
   end
 
   def test_system_sigpipe
@@ -1308,7 +1471,7 @@ class TestProcess < Test::Unit::TestCase
       assert_equal("ok?\n", data)
     end
 
-    if File.directory?("/proc/self/task")
+    if File.directory?("/proc/self/task") && /netbsd[a-z]*[1-6]/ !~ RUBY_PLATFORM
       def test_daemon_no_threads
         pid, data = IO.popen("-", "r+") do |f|
           break f.pid, f.readlines if f
@@ -1319,21 +1482,16 @@ class TestProcess < Test::Unit::TestCase
         assert_equal(2, data.size, bug4920)
         assert_not_include(data.map(&:to_i), pid)
       end
-    else # darwin
-      def test_daemon_no_threads
-        data = Timeout.timeout(3) do
-          IO.popen("-") do |f|
-            break f.readlines.map(&:chomp) if f
-            th = Thread.start {sleep 3}
-            Process.daemon(true, true)
-            puts Thread.list.size, th.status.inspect
-          end
-        end
-        assert_equal(["1", "false"], data)
-      end
     end
   end
 
+  def test_popen_cloexec
+    return unless defined? Fcntl::FD_CLOEXEC
+    IO.popen([RUBY, "-e", ""]) {|io|
+      assert(io.close_on_exec?)
+    }
+  end
+
   def test_execopts_new_pgroup
     return unless windows?
 
@@ -1342,4 +1500,116 @@ class TestProcess < Test::Unit::TestCase
     assert_nothing_raised { spawn(*TRUECOMMAND, :new_pgroup=>true) }
     assert_nothing_raised { IO.popen([*TRUECOMMAND, :new_pgroup=>true]) {} }
   end
+
+  def test_execopts_uid
+    feature6975 = '[ruby-core:47414]'
+
+    [30000, [Process.uid, ENV["USER"]]].each do |uid, user|
+      if user
+        assert_nothing_raised(feature6975) do
+          begin
+            system(*TRUECOMMAND, uid: user)
+          rescue Errno::EPERM, NotImplementedError
+          end
+        end
+      end
+
+      assert_nothing_raised(feature6975) do
+        begin
+          system(*TRUECOMMAND, uid: uid)
+        rescue Errno::EPERM, NotImplementedError
+        end
+      end
+
+      assert_nothing_raised(feature6975) do
+        begin
+          u = IO.popen([RUBY, "-e", "print Process.uid", uid: user||uid], &:read)
+          assert_equal(uid.to_s, u, feature6975)
+        rescue Errno::EPERM, NotImplementedError
+        end
+      end
+    end
+  end
+
+  def test_execopts_gid
+    skip "Process.groups not implemented on Windows platform" if windows?
+    feature6975 = '[ruby-core:47414]'
+
+    [30000, *Process.groups.map {|g| g = Etc.getgrgid(g); [g.name, g.gid]}].each do |group, gid|
+      assert_nothing_raised(feature6975) do
+        begin
+          system(*TRUECOMMAND, gid: group)
+        rescue Errno::EPERM, NotImplementedError
+        end
+      end
+
+      gid = "#{gid || group}"
+      assert_nothing_raised(feature6975) do
+        begin
+          g = IO.popen([RUBY, "-e", "print Process.gid", gid: group], &:read)
+          assert_equal(gid, g, feature6975)
+        rescue Errno::EPERM, NotImplementedError
+        end
+      end
+    end
+  end
+
+  def test_sigpipe
+    system(RUBY, "-e", "")
+    with_pipe {|r, w|
+      r.close
+      assert_raise(Errno::EPIPE) { w.print "a" }
+    }
+  end
+
+  def test_sh_comment
+    IO.popen("echo a # fofoof") {|f|
+      assert_equal("a\n", f.read)
+    }
+  end if File.executable?("/bin/sh")
+
+  def test_sh_env
+    IO.popen("foofoo=barbar env") {|f|
+      lines = f.readlines
+      assert_operator(lines, :include?, "foofoo=barbar\n")
+    }
+  end if File.executable?("/bin/sh")
+
+  def test_sh_exec
+    IO.popen("exec echo exexexec") {|f|
+      assert_equal("exexexec\n", f.read)
+    }
+  end if File.executable?("/bin/sh")
+
+  def test_setsid
+    return unless Process.respond_to?(:setsid)
+    return unless Process.respond_to?(:getsid)
+    # OpenBSD doesn't allow Process::getsid(pid) when pid is in
+    # different session.
+    return if /openbsd/ =~ RUBY_PLATFORM
+
+    IO.popen([RUBY, "-e", <<EOS]) do|io|
+	Marshal.dump(Process.getsid, STDOUT)
+	newsid = Process.setsid
+	Marshal.dump(newsid, STDOUT)
+	STDOUT.flush
+	# getsid() on MacOS X return ESRCH when target process is zombie
+	# even if it is valid process id.
+	sleep
+EOS
+      begin
+        # test Process.getsid() w/o arg
+        assert_equal(Marshal.load(io), Process.getsid)
+
+        # test Process.setsid return value and Process::getsid(pid)
+        assert_equal(Marshal.load(io), Process.getsid(io.pid))
+      ensure
+        Process.kill(:KILL, io.pid) rescue nil
+        Process.wait(io.pid)
+      end
+    end
+  end
+
+
+
 end
