@@ -1,4 +1,6 @@
+# coding: US-ASCII
 require 'test/unit'
+require 'envutil'
 
 class TestRegexp < Test::Unit::TestCase
   def setup
@@ -10,6 +12,21 @@ class TestRegexp < Test::Unit::TestCase
     $VERBOSE = @verbose
   end
 
+  def test_has_NOENCODING
+    assert Regexp::NOENCODING
+    re = //n
+    assert_equal Regexp::NOENCODING, re.options
+  end
+
+  def test_ruby_dev_999
+    assert_match(/(?<=a).*b/, "aab")
+    assert_match(/(?<=\u3042).*b/, "\u3042ab")
+  end
+
+  def test_ruby_core_27247
+    assert_match(/(a){2}z/, "aaz")
+  end
+
   def test_ruby_dev_24643
     assert_nothing_raised("[ruby-dev:24643]") {
       /(?:(?:[a]*[a])?b)*a*$/ =~ "aabaaca"
@@ -61,6 +78,11 @@ class TestRegexp < Test::Unit::TestCase
     end
   end
 
+  def test_word_boundary
+    assert_match(/\u3042\b /, "\u3042 ")
+    assert_not_match(/\u3042\ba/, "\u3042a")
+  end
+
   def test_named_capture
     m = /&(?<foo>.*?);/.match("aaa &amp; yyy")
     assert_equal("amp", m["foo"])
@@ -104,6 +126,12 @@ class TestRegexp < Test::Unit::TestCase
     assert_equal({}, /(.)(.)/.named_captures)
 
     assert_equal("a[b]c", "abc".sub(/(?<x>[bc])/, "[\\k<x>]"))
+
+    assert_equal("o", "foo"[/(?<bar>o)/, "bar"])
+
+    s = "foo"
+    s[/(?<bar>o)/, "bar"] = "baz"
+    assert_equal("fbazo", s)
   end
 
   def test_assign_named_capture
@@ -128,19 +156,62 @@ class TestRegexp < Test::Unit::TestCase
   end
 
   def test_source
+    bug5484 = '[ruby-core:40364]'
     assert_equal('', //.source)
+    assert_equal('\:', /\:/.source, bug5484)
+    assert_equal(':', %r:\::.source, bug5484)
+  end
+
+  def test_source_escaped
+    expected, result = "$*+.?^|".each_char.map {|c|
+      [
+        ["\\#{c}", "\\#{c}", 1],
+        begin
+          re = eval("%r#{c}\\#{c}#{c}", nil, __FILE__, __LINE__)
+          t = eval("/\\#{c}/", nil, __FILE__, __LINE__).source
+        rescue SyntaxError => e
+          [e, t, nil]
+        else
+          [re.source, t, re =~ "a#{c}a"]
+        end
+      ]
+    }.transpose
+    assert_equal(expected, result)
+  end
+
+  def test_source_escaped_paren
+    bug7610 = '[ruby-core:51088]'
+    s = '\(a\)'
+    assert_equal(/#{s}/, eval("%r(#{s})"), bug7610)
+  end
+
+  def test_source_unescaped
+    expected, result = "!\"#%&',-/:;=@_`~".each_char.map {|c|
+      [
+        ["#{c}", "\\#{c}", 1],
+        begin
+          re = eval("%r#{c}\\#{c}#{c}", nil, __FILE__, __LINE__)
+          t = eval("%r{\\#{c}}", nil, __FILE__, __LINE__).source
+        rescue SyntaxError => e
+          [e, t, nil]
+        else
+          [re.source, t, re =~ "a#{c}a"]
+        end
+      ]
+    }.transpose
+    assert_equal(expected, result)
   end
 
   def test_inspect
     assert_equal('//', //.inspect)
     assert_equal('//i', //i.inspect)
     assert_equal('/\//i', /\//i.inspect)
-    assert_equal('/\//i', /#{'/'}/i.inspect)
+    assert_equal('/\//i', %r"#{'/'}"i.inspect)
     assert_equal('/\/x/i', /\/x/i.inspect)
     assert_equal('/\x00/i', /#{"\0"}/i.inspect)
     assert_equal("/\n/i", /#{"\n"}/i.inspect)
-    s = [0xff].pack("C")
-    assert_equal('/\/'+s+'/i', /\/#{s}/i.inspect)
+    s = [0xf1, 0xf2, 0xf3].pack("C*")
+    assert_equal('/\/\xF1\xF2\xF3/i', /\/#{s}/i.inspect)
   end
 
   def test_char_to_option
@@ -276,11 +347,16 @@ class TestRegexp < Test::Unit::TestCase
     assert_equal("bar", "foobarbaz"[Regexp.new("b..", nil, "n")])
     assert_equal(//n, Regexp.new("", nil, "n"))
 
+    arg_encoding_none = 32 # ARG_ENCODING_NONE is implementation defined value
+    assert_equal(arg_encoding_none, Regexp.new("", nil, "n").options)
+    assert_equal(arg_encoding_none, Regexp.new("", nil, "N").options)
+
     assert_raise(RegexpError) { Regexp.new(")(") }
   end
 
   def test_unescape
     assert_raise(ArgumentError) { s = '\\'; /#{ s }/ }
+    assert_equal(/\xFF/n, /#{ s="\\xFF" }/n)
     assert_equal(/\177/, (s = '\177'; /#{ s }/))
     assert_raise(ArgumentError) { s = '\u'; /#{ s }/ }
     assert_raise(ArgumentError) { s = '\u{ ffffffff }'; /#{ s }/ }
@@ -298,7 +374,7 @@ class TestRegexp < Test::Unit::TestCase
     assert_equal(/a/, eval(%q(s="\u0061";/#{s}/n)))
     assert_raise(RegexpError) { s = "\u3042"; eval(%q(/#{s}/n)) }
     assert_raise(RegexpError) { s = "\u0061"; eval(%q(/\u3042#{s}/n)) }
-    assert_raise(ArgumentError) { s1=[0xff].pack("C"); s2="\u3042"; eval(%q(/#{s1}#{s2}/)) }
+    assert_raise(RegexpError) { s1=[0xff].pack("C"); s2="\u3042"; eval(%q(/#{s1}#{s2}/)) }
 
     assert_raise(ArgumentError) { s = '\x'; /#{ s }/ }
 
@@ -337,9 +413,11 @@ class TestRegexp < Test::Unit::TestCase
   end
 
   def test_equal
-    assert_equal(true, /abc/ == /abc/)
-    assert_equal(false, /abc/ == /abc/m)
-    assert_equal(false, /abc/ == /abd/)
+    bug5484 = '[ruby-core:40364]'
+    assert_equal(/abc/, /abc/)
+    assert_not_equal(/abc/, /abc/m)
+    assert_not_equal(/abc/, /abd/)
+    assert_equal(/\/foo/, Regexp.new('/foo'), bug5484)
   end
 
   def test_match
@@ -465,23 +543,27 @@ class TestRegexp < Test::Unit::TestCase
       /foo/.match("foo")
     end.value
     assert(m.tainted?)
+    assert_nothing_raised('[ruby-core:26137]') {
+      m = proc {$SAFE = 4; %r"#{ }"o}.call
+    }
+    assert(m.tainted?)
   end
 
-  def check(re, ss, fs = [])
+  def check(re, ss, fs = [], msg = nil)
     re = Regexp.new(re) unless re.is_a?(Regexp)
     ss = [ss] unless ss.is_a?(Array)
     ss.each do |e, s|
       s ||= e
-      assert_match(re, s)
+      assert_match(re, s, msg)
       m = re.match(s)
-      assert_equal(e, m[0])
+      assert_equal(e, m[0], msg)
     end
     fs = [fs] unless fs.is_a?(Array)
-    fs.each {|s| assert_no_match(re, s) }
+    fs.each {|s| assert_no_match(re, s, msg) }
   end
 
   def failcheck(re)
-    assert_raise(RegexpError) { /#{ re }/ }
+    assert_raise(RegexpError) { %r"#{ re }" }
   end
 
   def test_parse
@@ -558,6 +640,9 @@ class TestRegexp < Test::Unit::TestCase
     failcheck('(?<!.*)')
     check(/(?<=A|B.)C/, [%w(C AC), %w(C BXC)], %w(C BC))
     check(/(?<!A|B.)C/, [%w(C C), %w(C BC)], %w(AC BXC))
+
+    assert_not_match(/(?<!aa|b)c/i, "Aac")
+    assert_not_match(/(?<!b|aa)c/i, "Aac")
   end
 
   def test_parse_kg
@@ -579,7 +664,7 @@ class TestRegexp < Test::Unit::TestCase
     failcheck('()\g<-2>')
     check(/\A(?<x>.)(?<x>.)\k<x>\z/, %w(aba abb), %w(abc .. ....))
     check(/\A(?<x>.)(?<x>.)\k<x>\z/i, %w(aba ABa abb ABb), %w(abc .. ....))
-    check(/\k\g/, "kg")
+    check('\k\g', "kg")
     failcheck('(.\g<1>)')
     failcheck('(.\g<2>)')
     failcheck('(?=\g<1>)')
@@ -622,8 +707,8 @@ class TestRegexp < Test::Unit::TestCase
     check(/\u3042\d/, ["\u30421", "\u30422"])
 
     # CClassTable cache test
-    assert(/\u3042\d/.match("\u30421"))
-    assert(/\u3042\d/.match("\u30422"))
+    assert_match(/\u3042\d/, "\u30421")
+    assert_match(/\u3042\d/, "\u30422")
   end
 
   def test_char_class
@@ -654,11 +739,18 @@ class TestRegexp < Test::Unit::TestCase
     check(/\A[a-b-]\z/, %w(a b -), ["", "c"])
     check('\A[a-b-&&\w]\z', %w(a b), ["", "-"])
     check('\A[a-b-&&\W]\z', "-", ["", "a", "b"])
-    check('\A[a-c-e]\z', %w(a b c e), %w(- d)) # is it OK?
+    check('\A[a-c-e]\z', %w(a b c e -), %w(d))
     check(/\A[a-f&&[^b-c]&&[^e]]\z/, %w(a d f), %w(b c e g 0))
     check(/\A[[^b-c]&&[^e]&&a-f]\z/, %w(a d f), %w(b c e g 0))
     check(/\A[\n\r\t]\z/, ["\n", "\r", "\t"])
     failcheck('[9-1]')
+
+    assert_match(/\A\d+\z/, "0123456789")
+    assert_no_match(/\d/, "\uff10\uff11\uff12\uff13\uff14\uff15\uff16\uff17\uff18\uff19")
+    assert_match(/\A\w+\z/, "09azAZ_")
+    assert_no_match(/\w/, "\uff10\uff19\uff41\uff5a\uff21\uff3a")
+    assert_match(/\A\s+\z/, "\r\n\v\f\r\s")
+    assert_no_match(/\s/, "\u0085")
   end
 
   def test_posix_bracket
@@ -671,6 +763,12 @@ class TestRegexp < Test::Unit::TestCase
     failcheck('[[:alpha')
     failcheck('[[:alpha:')
     failcheck('[[:alp:]]')
+
+    assert_match(/\A[[:digit:]]+\z/, "\uff10\uff11\uff12\uff13\uff14\uff15\uff16\uff17\uff18\uff19")
+    assert_match(/\A[[:alnum:]]+\z/, "\uff10\uff19\uff41\uff5a\uff21\uff3a")
+    assert_match(/\A[[:space:]]+\z/, "\r\n\v\f\r\s\u0085")
+    assert_match(/\A[[:ascii:]]+\z/, "\x00\x7F")
+    assert_no_match(/[[:ascii:]]/, "\x80\xFF")
   end
 
   def test_backward
@@ -725,6 +823,7 @@ class TestRegexp < Test::Unit::TestCase
     assert_match(/^\u3042{0}\p{Any}$/, "a")
     assert_match(/^\u3042{0}\p{Any}$/, "\u3041")
     assert_match(/^\u3042{0}\p{Any}$/, "\0")
+    assert_match(/^\p{Lo}{4}$/u, "\u3401\u4E01\u{20001}\u{2A701}")
     assert_no_match(/^\u3042{0}\p{Any}$/, "\0\0")
     assert_no_match(/^\u3042{0}\p{Any}$/, "")
     assert_raise(SyntaxError) { eval('/^\u3042{0}\p{' + "\u3042" + '}$/') }
@@ -755,7 +854,86 @@ class TestRegexp < Test::Unit::TestCase
     #assert_match(/^(\ufb05)\1\1$/i, "\ufb05\ufb06st") # this must be bug...
     assert_match(/^\ufb05{3}$/i, "\ufb05\ufb06st")
     assert_match(/^\u03b9\u0308\u0301$/i, "\u0390")
-    assert_nothing_raised { 0x03ffffff.chr("utf-8").size }
-    assert_nothing_raised { 0x7fffffff.chr("utf-8").size }
+  end
+
+  def test_unicode_age
+    assert_match(/^\p{Age=6.0}$/u, "\u261c")
+    assert_match(/^\p{Age=1.1}$/u, "\u261c")
+    assert_no_match(/^\P{age=6.0}$/u, "\u261c")
+
+    assert_match(/^\p{age=6.0}$/u, "\u31f6")
+    assert_match(/^\p{age=3.2}$/u, "\u31f6")
+    assert_no_match(/^\p{age=3.1}$/u, "\u31f6")
+    assert_no_match(/^\p{age=3.0}$/u, "\u31f6")
+    assert_no_match(/^\p{age=1.1}$/u, "\u31f6")
+
+    assert_match(/^\p{age=6.0}$/u, "\u2754")
+    assert_no_match(/^\p{age=5.0}$/u, "\u2754")
+    assert_no_match(/^\p{age=4.0}$/u, "\u2754")
+    assert_no_match(/^\p{age=3.0}$/u, "\u2754")
+    assert_no_match(/^\p{age=2.0}$/u, "\u2754")
+    assert_no_match(/^\p{age=1.1}$/u, "\u2754")
+  end
+
+  def test_matchdata
+    a = "haystack".match(/hay/)
+    b = "haystack".match(/hay/)
+    assert_equal(a, b, '[ruby-core:24748]')
+    h = {a => 42}
+    assert_equal(42, h[b], '[ruby-core:24748]')
+  end
+
+  def test_regexp_poped
+    assert_nothing_raised { eval("a = 1; /\#{ a }/; a") }
+    assert_nothing_raised { eval("a = 1; /\#{ a }/o; a") }
+  end
+
+  def test_invalid_fragment
+    bug2547 = '[ruby-core:27374]'
+    assert_raise(SyntaxError, bug2547) {eval('/#{"\\\\"}y/')}
+  end
+
+  def test_dup_warn
+    assert_warning(/duplicated/) { Regexp.new('[\u3042\u3043\u3042]') }
+    assert_warning(/duplicated/) { Regexp.new('[\u3042\u3043\u3043]') }
+    assert_warning(/\A\z/) { Regexp.new('[\u3042\u3044\u3043]') }
+    assert_warning(/\A\z/) { Regexp.new('[\u3042\u3045\u3043]') }
+    assert_warning(/\A\z/) { Regexp.new('[\u3042\u3045\u3044]') }
+    assert_warning(/\A\z/) { Regexp.new('[\u3042\u3045\u3043-\u3044]') }
+    assert_warning(/duplicated/) { Regexp.new('[\u3042\u3045\u3042-\u3043]') }
+    assert_warning(/duplicated/) { Regexp.new('[\u3042\u3045\u3044-\u3045]') }
+    assert_warning(/\A\z/) { Regexp.new('[\u3042\u3046\u3044]') }
+    assert_warning(/duplicated/) { Regexp.new('[\u1000-\u2000\u3042-\u3046\u3044]') }
+    assert_warning(/duplicated/) { Regexp.new('[\u3044\u3041-\u3047]') }
+    assert_warning(/duplicated/) { Regexp.new('[\u3042\u3044\u3046\u3041-\u3047]') }
+
+    bug7471 = '[ruby-core:50344]'
+    EnvUtil.verbose_warning do
+      assert_warning(/\A\z/, bug7471) { Regexp.new('[\D]') =~ "\u3042" }
+    end
+  end
+
+  def test_property_warn
+    assert_in_out_err('-w', 'x=/\p%s/', [], %r"warning: invalid Unicode Property \\p: /\\p%s/")
+  end
+
+  def test_invalid_escape_error
+    bug3539 = '[ruby-core:31048]'
+    error = assert_raise(SyntaxError) {eval('/\x/', nil, bug3539)}
+    assert_match(/invalid hex escape/, error.message)
+    assert_equal(1, error.message.scan(/.*invalid .*escape.*/i).size, bug3539)
+  end
+
+  def test_raw_hyphen_and_tk_char_type_after_range
+    bug6853 = '[ruby-core:47115]'
+    # use Regexp.new instead of literal to ignore a parser warning.
+    check(Regexp.new('[0-1-\\s]'), [' ', '-'], ['2', 'a'], bug6853)
+  end
+
+  def test_error_message_on_failed_conversion
+    bug7539 = '[ruby-core:50733]'
+    assert_equal false, /x/=== 42
+    err = assert_raise(TypeError){ Regexp.quote(42) }
+    assert_equal 'no implicit conversion of Fixnum into String', err.message, bug7539
   end
 end
