@@ -1,3 +1,4 @@
+# -*- coding: us-ascii -*-
 require 'test/unit'
 
 require 'tmpdir'
@@ -29,11 +30,16 @@ class TestRubyOptions < Test::Unit::TestCase
   def test_usage
     assert_in_out_err(%w(-h)) do |r, e|
       assert_operator(r.size, :<=, 24)
+      longer = r[1..-1].select {|x| x.size > 80}
+      assert_equal([], longer)
       assert_equal([], e)
     end
+  end
 
+  def test_usage_long
     assert_in_out_err(%w(--help)) do |r, e|
-      assert_operator(r.size, :<=, 24)
+      longer = r[1..-1].select {|x| x.size > 80}
+      assert_equal([], longer)
       assert_equal([], e)
     end
   end
@@ -121,9 +127,17 @@ class TestRubyOptions < Test::Unit::TestCase
     assert_in_out_err(%w(-KU), "p '\u3042'") do |r, e|
       assert_equal("\"\u3042\"", r.join.force_encoding(Encoding::UTF_8))
     end
-    assert_in_out_err(%w(-KE -e) + [""], "", [], [])
-    assert_in_out_err(%w(-KS -e) + [""], "", [], [])
-    assert_in_out_err(%w(-KN -e) + [""], "", [], [])
+    line = '-eputs"\xc2\xa1".encoding'
+    env = {'RUBYOPT' => nil}
+    assert_in_out_err([env, '-Ke', line], "", ["EUC-JP"], [])
+    assert_in_out_err([env, '-KE', line], "", ["EUC-JP"], [])
+    assert_in_out_err([env, '-Ks', line], "", ["Windows-31J"], [])
+    assert_in_out_err([env, '-KS', line], "", ["Windows-31J"], [])
+    assert_in_out_err([env, '-Ku', line], "", ["UTF-8"], [])
+    assert_in_out_err([env, '-KU', line], "", ["UTF-8"], [])
+    assert_in_out_err([env, '-Kn', line], "", ["ASCII-8BIT"], [])
+    assert_in_out_err([env, '-KN', line], "", ["ASCII-8BIT"], [])
+    assert_in_out_err([env, '-wKe', line], "", ["EUC-JP"], /-K/)
   end
 
   def test_version
@@ -191,14 +205,21 @@ class TestRubyOptions < Test::Unit::TestCase
   end
 
   def test_encoding
-    assert_in_out_err(%w(-Eutf-8), "p '\u3042'", [], /invalid multibyte char/)
-
     assert_in_out_err(%w(--encoding), "", [], /missing argument for --encoding/)
 
     assert_in_out_err(%w(--encoding test_ruby_test_rubyoptions_foobarbazqux), "", [],
                       /unknown encoding name - test_ruby_test_rubyoptions_foobarbazqux \(RuntimeError\)/)
 
-    assert_in_out_err(%w(--encoding utf-8), "p '\u3042'", [], /invalid multibyte char/)
+    if /mswin|mingw/ =~ RUBY_PLATFORM &&
+      (str = "\u3042".force_encoding(Encoding.find("locale"))).valid_encoding?
+      # This result depends on locale because LANG=C doesn't affect locale
+      # on Windows.
+      out, err = [str], []
+    else
+      out, err = [], /invalid multibyte char/
+    end
+    assert_in_out_err(%w(-Eutf-8), "puts '\u3042'", out, err)
+    assert_in_out_err(%w(--encoding utf-8), "puts '\u3042'", out, err)
   end
 
   def test_syntax_check
@@ -287,7 +308,7 @@ class TestRubyOptions < Test::Unit::TestCase
     assert_in_out_err([], "#! /test_r_u_b_y_test_r_u_b_y_options_foobarbazqux -foo -bar\r\np 1\r\n",
                       [], /: no Ruby script found in input/)
 
-    assert_in_out_err([], "#!ruby -KU -Eutf-8\r\np \"\u3042\"\r\n") do |r, e|
+    assert_in_out_err([{'RUBYOPT' => nil}], "#!ruby -KU -Eutf-8\r\np \"\u3042\"\r\n") do |r, e|
       assert_equal("\"\u3042\"", r.join.force_encoding(Encoding::UTF_8))
       assert_equal([], e)
     end
@@ -321,12 +342,33 @@ class TestRubyOptions < Test::Unit::TestCase
     t.puts "  end"
     t.puts "end"
     t.close
-    err = ["#{t.path}:1: warning: found = in conditional, should be ==",
-           "#{t.path}:4: warning: found = in conditional, should be =="]
-    err = /\A(#{Regexp.quote(t.path)}):1(: warning: found = in conditional, should be ==)\n\1:4\2\Z/
+    warning = ' warning: found = in conditional, should be =='
+    err = ["#{t.path}:1:#{warning}",
+           "#{t.path}:4:#{warning}",
+          ]
     bug2136 = '[ruby-dev:39363]'
     assert_in_out_err(["-w", t.path], "", [], err, bug2136)
     assert_in_out_err(["-wr", t.path, "-e", ""], "", [], err, bug2136)
+
+    t.open
+    t.truncate(0)
+    t.puts "if a = ''; end"
+    t.puts "if a = []; end"
+    t.puts "if a = [1]; end"
+    t.puts "if a = [a]; end"
+    t.puts "if a = {}; end"
+    t.puts "if a = {1=>2}; end"
+    t.puts "if a = {3=>a}; end"
+    t.close
+    err = ["#{t.path}:1:#{warning}",
+           "#{t.path}:2:#{warning}",
+           "#{t.path}:3:#{warning}",
+           "#{t.path}:5:#{warning}",
+           "#{t.path}:6:#{warning}",
+          ]
+    feature4299 = '[ruby-dev:43083]'
+    assert_in_out_err(["-w", t.path], "", [], err, feature4299)
+    assert_in_out_err(["-wr", t.path, "-e", ""], "", [], err, feature4299)
   ensure
     t.close(true) if t
   end
@@ -373,7 +415,7 @@ class TestRubyOptions < Test::Unit::TestCase
     rubybin = Regexp.quote(EnvUtil.rubybin)
     pat = Regexp.quote(notexist)
     bug1573 = '[ruby-core:23717]'
-    assert_equal(false, File.exist?(notexist))
+    assert_file.not_exist?(notexist)
     assert_in_out_err(["-r", notexist, "-ep"], "", [], /.* -- #{pat} \(LoadError\)/, bug1573)
     assert_in_out_err([notexist], "", [], /#{rubybin}:.* -- #{pat} \(LoadError\)/, bug1573)
   end
@@ -439,10 +481,11 @@ class TestRubyOptions < Test::Unit::TestCase
       opts[:rlimit_core] = 0
       additional = ""
     end
-    assert_in_out_err(["-e", "Process.kill :SEGV, $$"], "", [],
+    expected_stderr =
       %r(\A
       -e:(?:1:)?\s\[BUG\]\sSegmentation\sfault\n
       #{ Regexp.quote(RUBY_DESCRIPTION) }\n\n
+      (?:--\s(?:.+\n)*\n)?
       --\sControl\sframe\sinformation\s-+\n
       (?:c:.*\n)*
       (?:
@@ -459,13 +502,30 @@ class TestRubyOptions < Test::Unit::TestCase
       \[NOTE\]\n
       You\smay\shave\sencountered\sa\sbug\sin\sthe\sRuby\sinterpreter\sor\sextension\slibraries.\n
       Bug\sreports\sare\swelcome.\n
-      For\sdetails:\shttp:\/\/www.ruby-lang.org/bugreport.html\n
+      (?:.*\n)?
+      For\sdetails:\shttp:\/\/.*\.ruby-lang\.org/.*\n
       \n
       (?:#{additional})
       \z
-      )x,
-      nil,
-      opts)
+      )x
+    assert_in_out_err(["-e", "Process.kill :SEGV, $$"], "", [], expected_stderr, nil, opts)
+
+    bug7402 = '[ruby-core:49573]'
+    status = assert_in_out_err(['-e', 'class Bogus; def to_str; exit true; end; end',
+                                '-e', '$".unshift Bogus.new',
+                                '-e', 'Process.kill :SEGV, $$'],
+                               "", //, /#<Bogus:/,
+                               nil,
+                               opts)
+    assert_not_predicate(status, :success?, "segv but success #{bug7402}")
+
+    bug7597 = '[ruby-dev:46786]'
+    t = Tempfile.new(["test_ruby_test_bug7597", ".rb"])
+    t.write "f" * 100
+    t.flush
+    assert_in_out_err(["-e", "$0=ARGV[0]; Process.kill :SEGV, $$", t.path],
+                      "", [], expected_stderr, bug7597, opts)
+    t.close(true)
   end
 
   def test_DATA
@@ -489,6 +549,13 @@ class TestRubyOptions < Test::Unit::TestCase
     assert_in_out_err(["-we", "1.times do\n  a=1\nend"], "", [], [], feature3446)
     assert_in_out_err(["-we", "def foo\n  1.times do\n    a=1\n  end\nend"], "", [], ["-e:3: warning: assigned but unused variable - a"], feature3446)
     assert_in_out_err(["-we", "def foo\n""  1.times do |a| end\n""end"], "", [], [])
+    feature6693 = '[ruby-core:46160]'
+    assert_in_out_err(["-we", "def foo\n  _a=1\nend"], "", [], [], feature6693)
+    bug7408 = '[ruby-core:49659]'
+    assert_in_out_err(["-we", "def foo\n  a=1\n :a\nend"], "", [], ["-e:2: warning: assigned but unused variable - a"], bug7408)
+    feature7730 = '[ruby-core:51580]'
+    assert_in_out_err(["-w", "-"], "a=1", [], ["-:1: warning: assigned but unused variable - a"], feature7730)
+    assert_in_out_err(["-w", "-"], "eval('a=1')", [], [], feature7730)
   end
 
   def test_shadowing_variable
@@ -500,6 +567,9 @@ class TestRubyOptions < Test::Unit::TestCase
                       ["-e:3: warning: shadowing outer local variable - a",
                        "-e:2: warning: assigned but unused variable - a",
                       ], bug4130)
+    feature6693 = '[ruby-core:46160]'
+    assert_in_out_err(["-we", "def foo\n""  _a=1\n""  1.times do |_a| end\n""end"],
+                      "", [], [], feature6693)
   end
 
   def test_script_from_stdin
@@ -554,6 +624,11 @@ class TestRubyOptions < Test::Unit::TestCase
     end
   end
 
+  def test_script_is_directory
+    feature2408 = '[ruby-core:26925]'
+    assert_in_out_err(%w[.], "", [], /Is a directory -- \./, feature2408)
+  end
+
   def test_pflag_gsub
     bug7157 = '[ruby-core:47967]'
     assert_in_out_err(['-p', '-e', 'gsub(/t.*/){"TEST"}'], %[test], %w[TEST], [], bug7157)
