@@ -1,3 +1,4 @@
+# -*- coding: us-ascii -*-
 require 'test/unit'
 require 'thread'
 require_relative 'envutil'
@@ -7,7 +8,6 @@ class TestThread < Test::Unit::TestCase
     Threads = []
     def self.new(*)
       th = super
-      th.abort_on_exception = true
       Threads << th
       th
     end
@@ -27,6 +27,79 @@ class TestThread < Test::Unit::TestCase
     end
   end
 
+  def test_main_thread_variable_in_enumerator
+    assert_equal Thread.main, Thread.current
+
+    Thread.current.thread_variable_set :foo, "bar"
+
+    thread, value = Fiber.new {
+      Fiber.yield [Thread.current, Thread.current.thread_variable_get(:foo)]
+    }.resume
+
+    assert_equal Thread.current, thread
+    assert_equal Thread.current.thread_variable_get(:foo), value
+  end
+
+  def test_thread_variable_in_enumerator
+    Thread.new {
+      Thread.current.thread_variable_set :foo, "bar"
+
+      thread, value = Fiber.new {
+        Fiber.yield [Thread.current, Thread.current.thread_variable_get(:foo)]
+      }.resume
+
+      assert_equal Thread.current, thread
+      assert_equal Thread.current.thread_variable_get(:foo), value
+    }.join
+  end
+
+  def test_thread_variables
+    assert_equal [], Thread.new { Thread.current.thread_variables }.join.value
+
+    t = Thread.new {
+      Thread.current.thread_variable_set(:foo, "bar")
+      Thread.current.thread_variables
+    }
+    assert_equal [:foo], t.join.value
+  end
+
+  def test_thread_variable?
+    refute Thread.new { Thread.current.thread_variable?("foo") }.join.value
+    t = Thread.new {
+      Thread.current.thread_variable_set("foo", "bar")
+    }.join
+
+    assert t.thread_variable?("foo")
+    assert t.thread_variable?(:foo)
+    refute t.thread_variable?(:bar)
+  end
+
+  def test_thread_variable_strings_and_symbols_are_the_same_key
+    t = Thread.new {}.join
+    t.thread_variable_set("foo", "bar")
+    assert_equal "bar", t.thread_variable_get(:foo)
+  end
+
+  def test_thread_variable_frozen
+    t = Thread.new { }.join
+    t.freeze
+    assert_raises(RuntimeError) do
+      t.thread_variable_set(:foo, "bar")
+    end
+  end
+
+  def test_thread_variable_security
+    t = Thread.new { sleep }
+
+    assert_raises(SecurityError) do
+      Thread.new { $SAFE = 4; t.thread_variable_get(:foo) }.join
+    end
+
+    assert_raises(SecurityError) do
+      Thread.new { $SAFE = 4; t.thread_variable_set(:foo, :baz) }.join
+    end
+  end
+
   def test_mutex_synchronize
     m = Mutex.new
     r = 0
@@ -47,172 +120,9 @@ class TestThread < Test::Unit::TestCase
     assert_equal(max * max * max, r)
   end
 
-  def test_condvar
-    mutex = Mutex.new
-    condvar = ConditionVariable.new
-    result = []
-    mutex.synchronize do
-      t = Thread.new do
-        mutex.synchronize do
-          result << 1
-          condvar.signal
-        end
-      end
-
-      result << 0
-      condvar.wait(mutex)
-      result << 2
-      t.join
-    end
-    assert_equal([0, 1, 2], result)
-  end
-
-  def test_condvar_wait_not_owner
-    mutex = Mutex.new
-    condvar = ConditionVariable.new
-
-    assert_raise(ThreadError) { condvar.wait(mutex) }
-  end
-
-  def test_condvar_wait_exception_handling
-    # Calling wait in the only thread running should raise a ThreadError of
-    # 'stopping only thread'
-    mutex = Mutex.new
-    condvar = ConditionVariable.new
-
-    locked = false
-    thread = Thread.new do
-      Thread.current.abort_on_exception = false
-      mutex.synchronize do
-        begin
-          condvar.wait(mutex)
-        rescue Exception
-          locked = mutex.locked?
-          raise
-        end
-      end
-    end
-
-    until thread.stop?
-      sleep(0.1)
-    end
-
-    thread.raise Interrupt, "interrupt a dead condition variable"
-    assert_raise(Interrupt) { thread.value }
-    assert(locked)
-  end
-
-  def test_condvar_wait_and_broadcast
-    nr_threads = 3
-    threads = Array.new
-    mutex = Mutex.new
-    condvar = ConditionVariable.new
-    result = []
-
-    nr_threads.times do |i|
-      threads[i] = Thread.new do
-        mutex.synchronize do
-          result << "C1"
-          condvar.wait mutex
-          result << "C2"
-        end
-      end
-    end
-    sleep 0.1
-    mutex.synchronize do
-      result << "P1"
-      condvar.broadcast
-      result << "P2"
-    end
-    nr_threads.times do |i|
-      threads[i].join
-    end
-
-    assert_equal ["C1", "C1", "C1", "P1", "P2", "C2", "C2", "C2"], result
-  end
-
-#  Hmm.. don't we have a way of catch fatal exception?
-#
-#  def test_cv_wait_deadlock
-#    mutex = Mutex.new
-#    cv = ConditionVariable.new
-#
-#    assert_raises(fatal) {
-#      mutex.lock
-#      cv.wait mutex
-#      mutex.unlock
-#    }
-#  end
-
-  def test_condvar_wait_deadlock_2
-    nr_threads = 3
-    threads = Array.new
-    mutex = Mutex.new
-    condvar = ConditionVariable.new
-
-    nr_threads.times do |i|
-      if (i != 0)
-        mutex.unlock
-      end
-      threads[i] = Thread.new do
-        mutex.synchronize do
-          condvar.wait mutex
-        end
-      end
-      mutex.lock
-    end
-
-    assert_raise(Timeout::Error) do
-      Timeout.timeout(0.1) { condvar.wait mutex }
-    end
-    mutex.unlock rescue
-    threads[i].each.join
-  end
-
-  def test_condvar_timed_wait
-    mutex = Mutex.new
-    condvar = ConditionVariable.new
-    timeout = 0.3
-    locked = false
-
-    t0 = Time.now
-    mutex.synchronize do
-      begin
-        condvar.wait(mutex, timeout)
-      ensure
-        locked = mutex.locked?
-      end
-    end
-    t1 = Time.now
-    t = t1-t0
-
-    assert_block { timeout*0.9 < t && t < timeout*1.1 }
-    assert(locked)
-  end
-
-  def test_condvar_nolock
-    mutex = Mutex.new
-    condvar = ConditionVariable.new
-
-    assert_raise(ThreadError) { condvar.wait(mutex) }
-  end
-
-  def test_condvar_nolock_2
-    mutex = Mutex.new
-    condvar = ConditionVariable.new
-
-    Thread.new do
-      assert_raise(ThreadError) {condvar.wait(mutex)}
-    end.join
-  end
-
-  def test_condvar_nolock_3
-    mutex = Mutex.new
-    condvar = ConditionVariable.new
-
-    Thread.new do
-      assert_raise(ThreadError) {condvar.wait(mutex, 0.1)}
-    end.join
+  def test_mutex_synchronize_yields_no_block_params
+    bug8097 = '[ruby-core:53424] [Bug #8097]'
+    assert_empty(Mutex.new.synchronize {|*params| break params}, bug8097)
   end
 
   def test_local_barrier
@@ -222,19 +132,18 @@ class TestThread < Test::Unit::TestCase
     require 'envutil'
     $:.shift
     3.times {
-      result = `#{EnvUtil.rubybin} #{lbtest}`
-      assert(!$?.coredump?, '[ruby-dev:30653]')
-      assert_equal("exit.", result[/.*\Z/], '[ruby-dev:30653]')
+      `#{EnvUtil.rubybin} #{lbtest}`
+      assert_not_predicate($?, :coredump?, '[ruby-dev:30653]')
     }
   end
 
   def test_priority
     c1 = c2 = 0
     t1 = Thread.new { loop { c1 += 1 } }
-    t1.priority = -1
+    t1.priority = 3
     t2 = Thread.new { loop { c2 += 1 } }
     t2.priority = -3
-    assert_equal(-1, t1.priority)
+    assert_equal(3, t1.priority)
     assert_equal(-3, t2.priority)
     sleep 0.5
     5.times do
@@ -431,30 +340,27 @@ class TestThread < Test::Unit::TestCase
     a = ::Thread.new { raise("die now") }
     b = Thread.new { Thread.stop }
     c = Thread.new { Thread.exit }
-    d = Thread.new { sleep }
     e = Thread.current
     sleep 0.5
 
     assert_equal(nil, a.status)
-    assert(a.stop?)
+    assert_predicate(a, :stop?)
 
     assert_equal("sleep", b.status)
-    assert(b.stop?)
+    assert_predicate(b, :stop?)
 
     assert_equal(false, c.status)
     assert_match(/^#<TestThread::Thread:.* dead>$/, c.inspect)
-    assert(c.stop?)
+    assert_predicate(c, :stop?)
 
-    d.kill
-    assert_equal(["aborting", false], [d.status, d.stop?])
-
-    assert_equal(["run", false], [e.status, e.stop?])
+    es1 = e.status
+    es2 = e.stop?
+    assert_equal(["run", false], [es1, es2])
 
   ensure
     a.kill if a
     b.kill if b
     c.kill if c
-    d.kill if d
   end
 
   def test_safe_level
@@ -593,7 +499,7 @@ class TestThread < Test::Unit::TestCase
       raise "recursive_outer should short circuit intermediate calls"
     end
     assert_nothing_raised {arr.hash}
-    assert(obj[:visited])
+    assert(obj[:visited], "obj.hash was not called")
   end
 
   def test_thread_instance_variable
@@ -608,66 +514,201 @@ class TestThread < Test::Unit::TestCase
   def test_no_valid_cfp
     skip 'with win32ole, cannot run this testcase because win32ole redefines Thread#intialize' if defined?(WIN32OLE)
     bug5083 = '[ruby-dev:44208]'
-    error = assert_raise(RuntimeError) do
-      Thread.new(&Module.method(:nesting)).join
-    end
-    assert_equal("Can't call on top of Fiber or Thread", error.message, bug5083)
+    assert_equal([], Thread.new(&Module.method(:nesting)).value)
     error = assert_raise(RuntimeError) do
       Thread.new(:to_s, &Module.method(:undef_method)).join
     end
     assert_equal("Can't call on top of Fiber or Thread", error.message, bug5083)
   end
-end
 
-class TestThreadGroup < Test::Unit::TestCase
-  def test_thread_init
-    thgrp = ThreadGroup.new
-    Thread.new{
-      thgrp.add(Thread.current)
-      assert_equal(thgrp, Thread.new{sleep 1}.group)
-    }.join
+  def make_handle_interrupt_test_thread1 flag
+    r = []
+    ready_p = false
+    th = Thread.new{
+      begin
+        Thread.handle_interrupt(RuntimeError => flag){
+          begin
+            ready_p = true
+            sleep 0.5
+          rescue
+            r << :c1
+          end
+        }
+      rescue
+        r << :c2
+      end
+    }
+    Thread.pass until ready_p
+    th.raise
+    begin
+      th.join
+    rescue
+      r << :c3
+    end
+    r
   end
 
-  def test_frozen_thgroup
-    thgrp = ThreadGroup.new
-
-    t = Thread.new{1}
-    Thread.new{
-      thgrp.add(Thread.current)
-      thgrp.freeze
-      assert_raise(ThreadError) do
-        Thread.new{1}.join
-      end
-      assert_raise(ThreadError) do
-        thgrp.add(t)
-      end
-      assert_raise(ThreadError) do
-        ThreadGroup.new.add Thread.current
-      end
-    }.join
-    t.join
+  def test_handle_interrupt
+    [[:never, :c2],
+     [:immediate, :c1],
+     [:on_blocking, :c1]].each{|(flag, c)|
+      assert_equal([flag, c], [flag] + make_handle_interrupt_test_thread1(flag))
+    }
+    # TODO: complex cases are needed.
   end
 
-  def test_enclosed_thgroup
-    thgrp = ThreadGroup.new
-    assert_equal(false, thgrp.enclosed?)
+  def test_handle_interrupt_invalid_argument
+    assert_raise(ArgumentError) {
+      Thread.handle_interrupt(RuntimeError => :immediate) # no block
+    }
+    assert_raise(ArgumentError) {
+      Thread.handle_interrupt(RuntimeError => :xyzzy) {}
+    }
+    assert_raise(TypeError) {
+      Thread.handle_interrupt([]) {} # array
+    }
+  end
 
-    t = Thread.new{1}
-    Thread.new{
-      thgrp.add(Thread.current)
-      thgrp.enclose
-      assert_equal(true, thgrp.enclosed?)
-      assert_nothing_raised do
-        Thread.new{1}.join
+  def for_test_handle_interrupt_with_return
+    Thread.handle_interrupt(Object => :never){
+      Thread.current.raise RuntimeError.new("have to be rescured")
+      return
+    }
+  rescue
+  end
+
+  def test_handle_interrupt_with_return
+    assert_nothing_raised do
+      for_test_handle_interrupt_with_return
+      _dummy_for_check_ints=nil
+    end
+  end
+
+  def test_handle_interrupt_with_break
+    assert_nothing_raised do
+      begin
+        Thread.handle_interrupt(Object => :never){
+          Thread.current.raise RuntimeError.new("have to be rescured")
+          break
+        }
+      rescue
       end
-      assert_raise(ThreadError) do
-        thgrp.add t
+      _dummy_for_check_ints=nil
+    end
+  end
+
+  def test_handle_interrupt_blocking
+    r=:ng
+    e=Class.new(Exception)
+    th_s = Thread.current
+    begin
+      th = Thread.start{
+        Thread.handle_interrupt(Object => :on_blocking){
+          begin
+            Thread.current.raise RuntimeError
+            r=:ok
+            sleep
+          ensure
+            th_s.raise e
+          end
+        }
+      }
+      sleep 1
+      r=:ng
+      th.raise RuntimeError
+      th.join
+    rescue e
+    end
+    assert_equal(:ok,r)
+  end
+
+  def test_handle_interrupt_and_io
+    assert_in_out_err([], <<-INPUT, %w(ok), [])
+      th_waiting = true
+
+      t = Thread.new {
+        Thread.handle_interrupt(RuntimeError => :on_blocking) {
+          nil while th_waiting
+          # async interrupt should be raised _before_ writing puts arguments
+          puts "ng"
+        }
+      }
+
+      sleep 0.1
+      t.raise RuntimeError
+      th_waiting = false
+      t.join rescue nil
+      puts "ok"
+    INPUT
+  end
+
+  def test_handle_interrupt_and_p
+    assert_in_out_err([], <<-INPUT, %w(:ok :ok), [])
+      th_waiting = true
+
+      t = Thread.new {
+        Thread.handle_interrupt(RuntimeError => :on_blocking) {
+          nil while th_waiting
+          # p shouldn't provide interruptible point
+          p :ok
+          p :ok
+        }
+      }
+
+      sleep 0.1
+      t.raise RuntimeError
+      th_waiting = false
+      t.join rescue nil
+    INPUT
+  end
+
+  def test_handle_interrupted?
+    q = Queue.new
+    Thread.handle_interrupt(RuntimeError => :never){
+      th = Thread.new{
+        q.push :e
+        begin
+          begin
+            sleep 0.5
+          rescue => e
+            q.push :ng1
+          end
+          begin
+            Thread.handle_interrupt(Object => :immediate){} if Thread.pending_interrupt?
+          rescue RuntimeError => e
+            q.push :ok
+          end
+        rescue => e
+          q.push :ng2
+        ensure
+          q.push :ng3
+        end
+      }
+      q.pop
+      th.raise
+      th.join
+      assert_equal(:ok, q.pop)
+    }
+  end
+
+  def test_thread_timer_and_ensure
+    assert_normal_exit(<<_eom, 'r36492', timeout: 3)
+    flag = false
+    t = Thread.new do
+      begin
+        sleep
+      ensure
+        1 until flag
       end
-      assert_raise(ThreadError) do
-        ThreadGroup.new.add Thread.current
-      end
-    }.join
+    end
+
+    Thread.pass until t.status == "sleep"
+
+    t.kill
+    t.alive? == true
+    flag = true
     t.join
+_eom
   end
 
   def test_uninitialized
@@ -693,7 +734,7 @@ class TestThreadGroup < Test::Unit::TestCase
     cmd = 'r,=IO.pipe; Thread.start {Thread.pass until Thread.main.stop?; puts; STDOUT.flush}; r.read'
     opt = {}
     opt[:new_pgroup] = true if /mswin|mingw/ =~ RUBY_PLATFORM
-    s, err = EnvUtil.invoke_ruby(['-e', cmd], "", true, true, opt) do |in_p, out_p, err_p, cpid|
+    s, _err = EnvUtil.invoke_ruby(['-e', cmd], "", true, true, opt) do |in_p, out_p, err_p, cpid|
       out_p.gets
       pid = cpid
       Process.kill(:SIGINT, pid)
@@ -701,14 +742,200 @@ class TestThreadGroup < Test::Unit::TestCase
       [$?, err_p.read]
     end
     t1 = Time.now.to_f
-    assert_equal(pid, s.pid)
+    assert_equal(pid, s.pid, bug5757)
     unless /mswin|mingw/ =~ RUBY_PLATFORM
       # status of signal is not supported on Windows
       assert_equal([false, true, false, Signal.list["INT"]],
                    [s.exited?, s.signaled?, s.stopped?, s.termsig],
                    "[s.exited?, s.signaled?, s.stopped?, s.termsig]")
     end
-    assert_in_delta(t1 - t0, 1, 1)
+    assert_in_delta(t1 - t0, 1, 1, bug5757)
+  end
+
+  def test_thread_join_in_trap
+    assert_nothing_raised{
+      t = Thread.new{ sleep 0.2; Process.kill(:INT, $$) }
+
+      Signal.trap :INT do
+        t.join
+      end
+
+      t.join
+    }
+
+    assert_equal(:normal_end,
+                 begin
+                   t = Thread.new{ sleep 0.2; Process.kill(:INT, $$); :normal_end }
+
+                   Signal.trap :INT do
+                     t.value
+                   end
+                   t.value
+                 end
+                 )
+  end
+
+  def test_thread_join_current
+    assert_raises(ThreadError) do
+      Thread.current.join
+    end
+  end
+
+  def test_thread_join_main_thread
+    assert_raises(ThreadError) do
+      Thread.new(Thread.current) {|t|
+        t.join
+      }.join
+    end
+  end
+
+  def test_main_thread_status_at_exit
+    assert_in_out_err([], <<-INPUT, %w(false), [])
+Thread.new(Thread.current) {|mth|
+  begin
+    sleep 0.1
+  ensure
+    p mth.alive?
+  end
+}
+    INPUT
+  end
+
+  def test_thread_status_in_trap
+    # when running trap handler, Thread#status must show "run"
+    # Even though interrupted from sleeping function
+    assert_in_out_err([], <<-INPUT, %w(sleep run), [])
+      Signal.trap(:INT) {
+        puts Thread.current.status
+      }
+
+      Thread.new(Thread.current) {|mth|
+        sleep 0.01
+        puts mth.status
+        Process.kill(:INT, $$)
+      }
+      sleep 0.1
+    INPUT
+  end
+
+  # Bug #7450
+  def test_thread_status_raise_after_kill
+    ary = []
+
+    t = Thread.new {
+      begin
+        ary << Thread.current.status
+        sleep #1
+      ensure
+        begin
+          ary << Thread.current.status
+          sleep #2
+        ensure
+          ary << Thread.current.status
+        end
+      end
+    }
+
+    begin
+      sleep 0.01
+      t.kill  # wake up sleep #1
+      sleep 0.01
+      t.raise "wakeup" # wake up sleep #2
+      sleep 0.01
+      assert_equal(ary, ["run", "aborting", "aborting"])
+    ensure
+      t.join rescue nil
+    end
+  end
+
+  def test_mutex_owned
+    mutex = Mutex.new
+
+    assert_equal(mutex.owned?, false)
+    mutex.synchronize {
+      # Now, I have the mutex
+      assert_equal(mutex.owned?, true)
+    }
+    assert_equal(mutex.owned?, false)
+  end
+
+  def test_mutex_owned2
+    begin
+      mutex = Mutex.new
+      th = Thread.new {
+        # lock forever
+        mutex.lock
+        sleep
+      }
+
+      sleep 0.01 until th.status == "sleep"
+      # acquired another thread.
+      assert_equal(mutex.locked?, true)
+      assert_equal(mutex.owned?, false)
+    ensure
+      th.kill if th
+    end
+  end
+
+  def test_mutex_unlock_on_trap
+    assert_in_out_err([], <<-INPUT, %w(locked unlocked false), [])
+      m = Mutex.new
+
+      Signal.trap("INT") { |signo|
+        m.unlock
+        puts "unlocked"
+      }
+
+      m.lock
+      puts "locked"
+      Process.kill("INT", $$)
+      sleep 0.2
+      puts m.locked?
+    INPUT
+  end
+
+  def invoke_rec script, vm_stack_size, machine_stack_size, use_length = true
+    env = {}
+    env['RUBY_THREAD_VM_STACK_SIZE'] = vm_stack_size.to_s if vm_stack_size
+    env['RUBY_THREAD_MACHINE_STACK_SIZE'] = machine_stack_size.to_s if machine_stack_size
+    out, = EnvUtil.invoke_ruby([env, '-e', script], '', true, true, :timeout => 50)
+    use_length ? out.length : out
+  end
+
+  def test_stack_size
+    h_default = eval(invoke_rec('p RubyVM::DEFAULT_PARAMS', nil, nil, false))
+    h_0 = eval(invoke_rec('p RubyVM::DEFAULT_PARAMS', 0, 0, false))
+    h_large = eval(invoke_rec('p RubyVM::DEFAULT_PARAMS', 1024 * 1024 * 10, 1024 * 1024 * 10, false))
+
+    assert_operator(h_default[:thread_vm_stack_size], :>, h_0[:thread_vm_stack_size],
+                    "0 thread_vm_stack_size")
+    assert_operator(h_default[:thread_vm_stack_size], :<, h_large[:thread_vm_stack_size],
+                    "large thread_vm_stack_size")
+    assert_operator(h_default[:thread_machine_stack_size], :>=, h_0[:thread_machine_stack_size],
+                    "0 thread_machine_stack_size")
+    assert_operator(h_default[:thread_machine_stack_size], :<=, h_large[:thread_machine_stack_size],
+                    "large thread_machine_stack_size")
+
+    # check VM machine stack size
+    script = 'def rec; print "."; STDOUT.flush; rec; end; rec'
+    size_default = invoke_rec script, nil, nil
+    assert_operator(size_default, :>, 0, "default size")
+    size_0 = invoke_rec script, 0, nil
+    assert_operator(size_default, :>, size_0, "0 size")
+    size_large = invoke_rec script, 1024 * 1024 * 10, nil
+    assert_operator(size_default, :<, size_large, "large size")
+
+    return if /mswin|mingw/ =~ RUBY_PLATFORM
+
+    # check machine stack size
+    # Note that machine stack size may not change size (depend on OSs)
+    script = 'def rec; print "."; STDOUT.flush; 1.times{1.times{1.times{rec}}}; end; Thread.new{rec}.join'
+    vm_stack_size = 1024 * 1024
+    size_default = invoke_rec script, vm_stack_size, nil
+    size_0 = invoke_rec script, vm_stack_size, 0
+    assert_operator(size_default, :>=, size_0, "0 size")
+    size_large = invoke_rec script, vm_stack_size, 1024 * 1024 * 10
+    assert_operator(size_default, :<=, size_large, "large size")
   end
 
   def test_blocking_mutex_unlocked_on_fork
