@@ -1,3 +1,4 @@
+# -*- coding: utf-8 -*-
 #
 # test/strscan/test_stringscanner.rb
 #
@@ -31,8 +32,8 @@ class TestStringScanner < Test::Unit::TestCase
   def test_s_allocate
     s = StringScanner.allocate
     assert_equal '#<StringScanner (uninitialized)>', s.inspect.sub(/StringScanner_C/, 'StringScanner')
-    assert_raises(UNINIT_ERROR) { s.eos? }
-    assert_raises(UNINIT_ERROR) { s.scan(/a/) }
+    assert_raise(UNINIT_ERROR) { s.eos? }
+    assert_raise(UNINIT_ERROR) { s.scan(/a/) }
     s.string = 'test'
     assert_equal '#<StringScanner 0/4 @ "test">', s.inspect.sub(/StringScanner_C/, 'StringScanner')
     assert_nothing_raised(UNINIT_ERROR) { s.eos? }
@@ -168,6 +169,26 @@ class TestStringScanner < Test::Unit::TestCase
     assert_equal 0, s.pos
   end
 
+  def test_string_set_is_equal
+    name = 'tenderlove'
+
+    s = StringScanner.new(name)
+    assert_equal name.object_id, s.string.object_id
+
+    s.string = name
+    assert_equal name.object_id, s.string.object_id
+  end
+
+  def test_string_append
+    s = StringScanner.new('tender')
+    s << 'love'
+    assert_equal 'tenderlove', s.string
+
+    s.string = 'tender'
+    s << 'love'
+    assert_equal 'tenderlove', s.string
+  end
+
   def test_pos
     s = StringScanner.new('test string')
     assert_equal 0, s.pos
@@ -179,6 +200,17 @@ class TestStringScanner < Test::Unit::TestCase
     assert_equal 11, s.pos
   end
 
+  def test_pos_unicode
+    s = StringScanner.new("abcädeföghi")
+    assert_equal 0, s.charpos
+    assert_equal "abcä", s.scan_until(/ä/)
+    assert_equal 4, s.charpos
+    assert_equal "defö", s.scan_until(/ö/)
+    assert_equal 8, s.charpos
+    s.terminate
+    assert_equal 11, s.charpos
+  end
+
   def test_concat
     s = StringScanner.new('a')
     s.scan(/a/)
@@ -528,9 +560,8 @@ class TestStringScanner < Test::Unit::TestCase
     s.terminate
     assert_nil s.matched_size
 
-    # obsolete
     s = StringScanner.new('test string')
-    assert_nil s.matchedsize
+    assert_nil s.matched_size
     s.scan(/test/)
     assert_equal 4, s.matched_size
     s.terminate
@@ -545,7 +576,130 @@ class TestStringScanner < Test::Unit::TestCase
   def test_generic_regexp
     ss = StringScanner.new("\xA1\xA2".force_encoding("euc-jp"))
     t = ss.scan(/./)
-    assert_equal("\xa1\xa1".force_encoding("euc-jp"), t)
+    assert_equal("\xa1\xa2".force_encoding("euc-jp"), t)
   end
 
+  def test_set_pos
+    s = StringScanner.new("test string")
+    s.pos = 7
+    assert_equal("ring", s.rest)
+  end
+
+  def test_match_p
+    s = StringScanner.new("test string")
+    assert_equal(4, s.match?(/\w+/))
+    assert_equal(4, s.match?(/\w+/))
+    assert_equal(nil, s.match?(/\s+/))
+  end
+
+  def test_check
+    s = StringScanner.new("Foo Bar Baz")
+    assert_equal("Foo", s.check(/Foo/))
+    assert_equal(0, s.pos)
+    assert_equal("Foo", s.matched)
+    assert_equal(nil, s.check(/Bar/))
+    assert_equal(nil, s.matched)
+  end
+
+  def test_scan_full
+    s = StringScanner.new("Foo Bar Baz")
+    assert_equal(4, s.scan_full(/Foo /, false, false))
+    assert_equal(0, s.pos)
+    assert_equal(nil, s.scan_full(/Baz/, false, false))
+    assert_equal("Foo ", s.scan_full(/Foo /, false, true))
+    assert_equal(0, s.pos)
+    assert_equal(nil, s.scan_full(/Baz/, false, false))
+    assert_equal(4, s.scan_full(/Foo /, true, false))
+    assert_equal(4, s.pos)
+    assert_equal(nil, s.scan_full(/Baz /, false, false))
+    assert_equal("Bar ", s.scan_full(/Bar /, true, true))
+    assert_equal(8, s.pos)
+    assert_equal(nil, s.scan_full(/az/, false, false))
+  end
+
+  def test_exist_p
+    s = StringScanner.new("test string")
+    assert_equal(3, s.exist?(/s/))
+    assert_equal(0, s.pos)
+    s.scan(/test/)
+    assert_equal(2, s.exist?(/s/))
+    assert_equal(4, s.pos)
+    assert_equal(nil, s.exist?(/e/))
+  end
+
+  def test_skip_until
+    s = StringScanner.new("Foo Bar Baz")
+    assert_equal(3, s.skip_until(/Foo/))
+    assert_equal(3, s.pos)
+    assert_equal(4, s.skip_until(/Bar/))
+    assert_equal(7, s.pos)
+    assert_equal(nil, s.skip_until(/Qux/))
+  end
+
+  def test_check_until
+    s = StringScanner.new("Foo Bar Baz")
+    assert_equal("Foo", s.check_until(/Foo/))
+    assert_equal(0, s.pos)
+    assert_equal("Foo Bar", s.check_until(/Bar/))
+    assert_equal(0, s.pos)
+    assert_equal(nil, s.check_until(/Qux/))
+  end
+
+  def test_search_full
+    s = StringScanner.new("Foo Bar Baz")
+    assert_equal(8, s.search_full(/Bar /, false, false))
+    assert_equal(0, s.pos)
+    assert_equal("Foo Bar ", s.search_full(/Bar /, false, true))
+    assert_equal(0, s.pos)
+    assert_equal(8, s.search_full(/Bar /, true, false))
+    assert_equal(8, s.pos)
+    assert_equal("Baz", s.search_full(/az/, true, true))
+    assert_equal(11, s.pos)
+  end
+
+  def test_peek
+    s = StringScanner.new("test string")
+    assert_equal("test st", s.peek(7))
+    assert_equal("test st", s.peek(7))
+    s.scan(/test/)
+    assert_equal(" stri", s.peek(5))
+    assert_equal(" string", s.peek(10))
+    s.scan(/ string/)
+    assert_equal("", s.peek(10))
+  end
+
+  def test_unscan
+    s = StringScanner.new('test string')
+    assert_equal("test", s.scan(/\w+/))
+    s.unscan
+    assert_equal("te", s.scan(/../))
+    assert_equal(nil, s.scan(/\d/))
+    assert_raise(ScanError) { s.unscan }
+  end
+
+  def test_rest
+    s = StringScanner.new('test string')
+    assert_equal("test string", s.rest)
+    s.scan(/test/)
+    assert_equal(" string", s.rest)
+    s.scan(/ string/)
+    assert_equal("", s.rest)
+    s.scan(/ string/)
+  end
+
+  def test_rest_size
+    s = StringScanner.new('test string')
+    assert_equal(11, s.rest_size)
+    s.scan(/test/)
+    assert_equal(7, s.rest_size)
+    s.scan(/ string/)
+    assert_equal(0, s.rest_size)
+    s.scan(/ string/)
+  end
+
+  def test_inspect2
+    s = StringScanner.new('test string test')
+    s.scan(/test strin/)
+    assert_equal('#<StringScanner 10/16 "...strin" @ "g tes...">', s.inspect)
+  end
 end
