@@ -3,7 +3,7 @@
 
   thread_pthread.c -
 
-  $Author: naruse $
+  $Author: nagachika $
 
   Copyright (C) 2004-2007 Koichi Sasada
 
@@ -24,6 +24,18 @@
 #elif HAVE_SYS_FCNTL_H
 #include <sys/fcntl.h>
 #endif
+#ifdef HAVE_SYS_PRCTL_H
+#include <sys/prctl.h>
+#endif
+#if defined(__native_client__) && defined(NACL_NEWLIB)
+# include "nacl/select.h"
+#endif
+#if HAVE_POLL
+#include <poll.h>
+#endif
+#if defined(HAVE_SYS_TIME_H)
+#include <sys/time.h>
+#endif
 
 static void native_mutex_lock(pthread_mutex_t *lock);
 static void native_mutex_unlock(pthread_mutex_t *lock);
@@ -35,17 +47,30 @@ static void native_cond_broadcast(rb_thread_cond_t *cond);
 static void native_cond_wait(rb_thread_cond_t *cond, pthread_mutex_t *mutex);
 static void native_cond_initialize(rb_thread_cond_t *cond, int flags);
 static void native_cond_destroy(rb_thread_cond_t *cond);
+static void rb_thread_wakeup_timer_thread_low(void);
 static pthread_t timer_thread_id;
 
 #define RB_CONDATTR_CLOCK_MONOTONIC 1
 
 #if defined(HAVE_PTHREAD_CONDATTR_SETCLOCK) && defined(HAVE_CLOCKID_T) && \
-    defined(CLOCK_REALTIME) && defined(CLOCK_MONOTONIC) && defined(HAVE_CLOCK_GETTIME)
+    defined(CLOCK_REALTIME) && defined(CLOCK_MONOTONIC) && \
+    defined(HAVE_CLOCK_GETTIME) && defined(HAVE_PTHREAD_CONDATTR_INIT)
 #define USE_MONOTONIC_COND 1
 #else
 #define USE_MONOTONIC_COND 0
 #endif
 
+#if defined(HAVE_POLL) && defined(HAVE_FCNTL) && defined(F_GETFL) && defined(F_SETFL) && defined(O_NONBLOCK) && !defined(__native_client__)
+/* The timer thread sleeps while only one Ruby thread is running. */
+# define USE_SLEEPY_TIMER_THREAD 1
+#else
+# define USE_SLEEPY_TIMER_THREAD 0
+#endif
+
+#ifndef ARRAY_SIZE
+#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
+#endif
+
 static void
 gvl_acquire_common(rb_vm_t *vm)
 {
@@ -53,8 +78,12 @@ gvl_acquire_common(rb_vm_t *vm)
 
 	vm->gvl.waiting++;
 	if (vm->gvl.waiting == 1) {
-	    /* transit to polling mode */
-	    rb_thread_wakeup_timer_thread();
+	    /*
+	     * Wake up timer thread iff timer thread is slept.
+	     * When timer thread is polling mode, we don't want to
+	     * make confusing timer thread interval time.
+	     */
+	    rb_thread_wakeup_timer_thread_low();
 	}
 
 	while (vm->gvl.acquired) {
@@ -140,6 +169,7 @@ gvl_init(rb_vm_t *vm)
     vm->gvl.acquired = 0;
     vm->gvl.waiting = 0;
     vm->gvl.need_yield = 0;
+    vm->gvl.wait_yield = 0;
 }
 
 static void
@@ -232,12 +262,13 @@ native_mutex_destroy(pthread_mutex_t *lock)
 static void
 native_cond_initialize(rb_thread_cond_t *cond, int flags)
 {
+#ifdef HAVE_PTHREAD_COND_INIT
     int r;
+# if USE_MONOTONIC_COND
     pthread_condattr_t attr;
 
     pthread_condattr_init(&attr);
 
-#if USE_MONOTONIC_COND
     cond->clockid = CLOCK_REALTIME;
     if (flags & RB_CONDATTR_CLOCK_MONOTONIC) {
 	r = pthread_condattr_setclock(&attr, CLOCK_MONOTONIC);
@@ -245,23 +276,29 @@ native_cond_initialize(rb_thread_cond_t *cond, int flags)
 	    cond->clockid = CLOCK_MONOTONIC;
 	}
     }
-#endif
 
     r = pthread_cond_init(&cond->cond, &attr);
+    pthread_condattr_destroy(&attr);
+# else
+    r = pthread_cond_init(&cond->cond, NULL);
+# endif
     if (r != 0) {
 	rb_bug_errno("pthread_cond_init", r);
     }
 
     return;
- }
+#endif
+}
 
 static void
 native_cond_destroy(rb_thread_cond_t *cond)
 {
+#ifdef HAVE_PTHREAD_COND_INIT
     int r = pthread_cond_destroy(&cond->cond);
     if (r != 0) {
 	rb_bug_errno("pthread_cond_destroy", r);
     }
+#endif
 }
 
 /*
@@ -358,7 +395,7 @@ native_cond_timeout(rb_thread_cond_t *cond, struct timespec timeout_rel)
     }
 
     if (cond->clockid != CLOCK_REALTIME)
-	rb_bug("unsupported clockid %d", cond->clockid);
+	rb_bug("unsupported clockid %"PRIdVALUE, (SIGNED_VALUE)cond->clockid);
 #endif
 
     ret = gettimeofday(&tv, 0);
@@ -436,7 +473,9 @@ Init_native_thread(void)
 #ifdef USE_SIGNAL_THREAD_LIST
     native_mutex_initialize(&signal_thread_list_lock);
 #endif
+#ifndef __native_client__
     posix_signal(SIGVTALRM, null_func);
+#endif
 }
 
 static void
@@ -452,7 +491,9 @@ native_thread_destroy(rb_thread_t *th)
     native_cond_destroy(&th->native_thread_data.sleep_cond);
 }
 
+#ifndef USE_THREAD_CACHE
 #define USE_THREAD_CACHE 0
+#endif
 
 #if USE_THREAD_CACHE
 static rb_thread_t *register_cached_thread_and_wait(void);
@@ -462,12 +503,24 @@ static rb_thread_t *register_cached_thread_and_wait(void);
 #define STACKADDR_AVAILABLE 1
 #elif defined HAVE_PTHREAD_GET_STACKADDR_NP && defined HAVE_PTHREAD_GET_STACKSIZE_NP
 #define STACKADDR_AVAILABLE 1
+#undef MAINSTACKADDR_AVAILABLE
+#define MAINSTACKADDR_AVAILABLE 0
+void *pthread_get_stackaddr_np(pthread_t);
+size_t pthread_get_stacksize_np(pthread_t);
 #elif defined HAVE_THR_STKSEGMENT || defined HAVE_PTHREAD_STACKSEG_NP
 #define STACKADDR_AVAILABLE 1
 #elif defined HAVE_PTHREAD_GETTHRDS_NP
 #define STACKADDR_AVAILABLE 1
 #endif
 
+#ifndef MAINSTACKADDR_AVAILABLE
+# ifdef STACKADDR_AVAILABLE
+#   define MAINSTACKADDR_AVAILABLE 1
+# else
+#   define MAINSTACKADDR_AVAILABLE 0
+# endif
+#endif
+
 #ifdef STACKADDR_AVAILABLE
 /*
  * Get the initial address and size of current thread's stack
@@ -477,41 +530,38 @@ get_stack(void **addr, size_t *size)
 {
 #define CHECK_ERR(expr)				\
     {int err = (expr); if (err) return err;}
-#if defined HAVE_PTHREAD_GETATTR_NP || defined HAVE_PTHREAD_ATTR_GET_NP || \
-    (defined HAVE_PTHREAD_GET_STACKADDR_NP && defined HAVE_PTHREAD_GET_STACKSIZE_NP)
+#ifdef HAVE_PTHREAD_GETATTR_NP /* Linux */
     pthread_attr_t attr;
     size_t guard = 0;
-
-# ifdef HAVE_PTHREAD_GETATTR_NP /* Linux */
     STACK_GROW_DIR_DETECTION;
     CHECK_ERR(pthread_getattr_np(pthread_self(), &attr));
-#   ifdef HAVE_PTHREAD_ATTR_GETSTACK
+# ifdef HAVE_PTHREAD_ATTR_GETSTACK
     CHECK_ERR(pthread_attr_getstack(&attr, addr, size));
     STACK_DIR_UPPER((void)0, (void)(*addr = (char *)*addr + *size));
-#   else
+# else
     CHECK_ERR(pthread_attr_getstackaddr(&attr, addr));
     CHECK_ERR(pthread_attr_getstacksize(&attr, size));
-#   endif
-# elif defined HAVE_PTHREAD_ATTR_GET_NP /* FreeBSD, DragonFly BSD, NetBSD */
-    CHECK_ERR(pthread_attr_init(&attr));
-    CHECK_ERR(pthread_attr_get_np(pthread_self(), &attr));
-#   ifdef HAVE_PTHREAD_ATTR_GETSTACK
-    CHECK_ERR(pthread_attr_getstack(&attr, addr, size));
-    STACK_DIR_UPPER((void)0, (void)(*addr = (char *)*addr + *size));
-#   else
-    CHECK_ERR(pthread_attr_getstackaddr(&attr, addr));
-    CHECK_ERR(pthread_attr_getstacksize(&attr, size));
-    STACK_DIR_UPPER((void)0, (void)(*addr = (char *)*addr + *size));
-#   endif
-# else /* MacOS X */
-    pthread_t th = pthread_self();
-    *addr = pthread_get_stackaddr_np(th);
-    *size = pthread_get_stacksize_np(th);
-    CHECK_ERR(pthread_attr_init(&attr));
 # endif
     CHECK_ERR(pthread_attr_getguardsize(&attr, &guard));
     *size -= guard;
     pthread_attr_destroy(&attr);
+#elif defined HAVE_PTHREAD_ATTR_GET_NP /* FreeBSD, DragonFly BSD, NetBSD */
+    pthread_attr_t attr;
+    CHECK_ERR(pthread_attr_init(&attr));
+    CHECK_ERR(pthread_attr_get_np(pthread_self(), &attr));
+# ifdef HAVE_PTHREAD_ATTR_GETSTACK
+    CHECK_ERR(pthread_attr_getstack(&attr, addr, size));
+    STACK_DIR_UPPER((void)0, (void)(*addr = (char *)*addr + *size));
+# else
+    CHECK_ERR(pthread_attr_getstackaddr(&attr, addr));
+    CHECK_ERR(pthread_attr_getstacksize(&attr, size));
+    STACK_DIR_UPPER((void)0, (void)(*addr = (char *)*addr + *size));
+# endif
+    pthread_attr_destroy(&attr);
+#elif (defined HAVE_PTHREAD_GET_STACKADDR_NP && defined HAVE_PTHREAD_GET_STACKSIZE_NP) /* MacOS X */
+    pthread_t th = pthread_self();
+    *addr = pthread_get_stackaddr_np(th);
+    *size = pthread_get_stacksize_np(th);
 #elif defined HAVE_THR_STKSEGMENT || defined HAVE_PTHREAD_STACKSEG_NP
     stack_t stk;
 # if defined HAVE_THR_STKSEGMENT /* Solaris */
@@ -553,7 +603,28 @@ static struct {
 extern void *STACK_END_ADDRESS;
 #endif
 
+enum {
+    RUBY_STACK_SPACE_LIMIT = 1024 * 1024, /* 1024KB */
+    RUBY_STACK_SPACE_RATIO = 5
+};
+
+static size_t
+space_size(size_t stack_size)
+{
+    size_t space_size = stack_size / RUBY_STACK_SPACE_RATIO;
+    if (space_size > RUBY_STACK_SPACE_LIMIT) {
+	return RUBY_STACK_SPACE_LIMIT;
+    }
+    else {
+	return space_size;
+    }
+}
+
 #undef ruby_init_stack
+/* Set stack bottom of Ruby implementation.
+ *
+ * You must call this function before any heap allocation by Ruby implementation.
+ * Or GC will break living objects */
 void
 ruby_init_stack(volatile VALUE *addr
 #ifdef __ia64
@@ -579,19 +650,60 @@ ruby_init_stack(volatile VALUE *addr
     }
 #endif
     {
-	size_t size = 0;
-	size_t space = 0;
-#if defined(HAVE_PTHREAD_ATTR_GET_NP)
-	void* addr;
-	get_stack(&addr, &size);
+#if defined(PTHREAD_STACK_DEFAULT)
+# if PTHREAD_STACK_DEFAULT < RUBY_STACK_SPACE*5
+#  error "PTHREAD_STACK_DEFAULT is too small"
+# endif
+	size_t size = PTHREAD_STACK_DEFAULT;
+#else
+	size_t size = RUBY_VM_THREAD_VM_STACK_SIZE;
+#endif
+	size_t space = space_size(size);
+#if MAINSTACKADDR_AVAILABLE
+	void* stackaddr;
+	STACK_GROW_DIR_DETECTION;
+	if (get_stack(&stackaddr, &size) == 0) {
+            space = STACK_DIR_UPPER((char *)addr - (char *)stackaddr, (char *)stackaddr - (char *)addr);
+        }
+	native_main_thread.stack_maxsize = size - space;
 #elif defined(HAVE_GETRLIMIT)
+	int pagesize = getpagesize();
 	struct rlimit rlim;
+        STACK_GROW_DIR_DETECTION;
 	if (getrlimit(RLIMIT_STACK, &rlim) == 0) {
 	    size = (size_t)rlim.rlim_cur;
 	}
+	addr = native_main_thread.stack_start;
+	if (IS_STACK_DIR_UPPER()) {
+	    space = ((size_t)((char *)addr + size) / pagesize) * pagesize - (size_t)addr;
+	}
+	else {
+	    space = (size_t)addr - ((size_t)((char *)addr - size) / pagesize + 1) * pagesize;
+	}
+	native_main_thread.stack_maxsize = space;
 #endif
-	space = size > 5 * 1024 * 1024 ? 1024 * 1024 : size / 5;
-	native_main_thread.stack_maxsize = size - space;
+    }
+
+    /* If addr is out of range of main-thread stack range estimation,  */
+    /* it should be on co-routine (alternative stack). [Feature #2294] */
+    {
+	void *start, *end;
+	STACK_GROW_DIR_DETECTION;
+
+	if (IS_STACK_DIR_UPPER()) {
+	    start = native_main_thread.stack_start;
+	    end = (char *)native_main_thread.stack_start + native_main_thread.stack_maxsize;
+	}
+	else {
+	    start = (char *)native_main_thread.stack_start - native_main_thread.stack_maxsize;
+	    end = native_main_thread.stack_start;
+	}
+
+	if ((void *)addr < start || (void *)addr > end) {
+	    /* out of range */
+	    native_main_thread.stack_start = (VALUE *)addr;
+	    native_main_thread.stack_maxsize = 0; /* unknown */
+	}
     }
 }
 
@@ -685,11 +797,15 @@ register_cached_thread_and_wait(void)
 {
     rb_thread_cond_t cond = { PTHREAD_COND_INITIALIZER, };
     volatile rb_thread_t *th_area = 0;
+    struct timeval tv;
+    struct timespec ts;
     struct cached_thread_entry *entry =
       (struct cached_thread_entry *)malloc(sizeof(struct cached_thread_entry));
 
-    struct timeval tv;
-    struct timespec ts;
+    if (entry == 0) {
+	return 0; /* failed -> terminate thread immediately */
+    }
+
     gettimeofday(&tv, 0);
     ts.tv_sec = tv.tv_sec + 60;
     ts.tv_nsec = tv.tv_usec * 1000;
@@ -757,24 +873,6 @@ use_cached_thread(rb_thread_t *th)
     return result;
 }
 
-enum {
-#ifdef __SYMBIAN32__
-    RUBY_STACK_MIN_LIMIT = 64 * 1024,  /* 64KB: Let's be slightly more frugal on mobile platform */
-#else
-    RUBY_STACK_MIN_LIMIT = 512 * 1024, /* 512KB */
-#endif
-    RUBY_STACK_SPACE_LIMIT = 1024 * 1024
-};
-
-#ifdef PTHREAD_STACK_MIN
-#define RUBY_STACK_MIN ((RUBY_STACK_MIN_LIMIT < PTHREAD_STACK_MIN) ? \
-			PTHREAD_STACK_MIN * 2 : RUBY_STACK_MIN_LIMIT)
-#else
-#define RUBY_STACK_MIN (RUBY_STACK_MIN_LIMIT)
-#endif
-#define RUBY_STACK_SPACE (RUBY_STACK_MIN/5 > RUBY_STACK_SPACE_LIMIT ? \
-			  RUBY_STACK_SPACE_LIMIT : RUBY_STACK_MIN/5)
-
 static int
 native_thread_create(rb_thread_t *th)
 {
@@ -785,8 +883,8 @@ native_thread_create(rb_thread_t *th)
     }
     else {
 	pthread_attr_t attr;
-	const size_t stack_size = RUBY_STACK_MIN;
-	const size_t space = RUBY_STACK_SPACE;
+	const size_t stack_size = th->vm->default_params.thread_machine_stack_size;
+	const size_t space = space_size(stack_size);
 
         th->machine_stack_maxsize = stack_size - space;
 #ifdef __ia64
@@ -794,21 +892,27 @@ native_thread_create(rb_thread_t *th)
         th->machine_register_stack_maxsize = th->machine_stack_maxsize;
 #endif
 
+#ifdef HAVE_PTHREAD_ATTR_INIT
 	CHECK_ERR(pthread_attr_init(&attr));
 
-#ifdef PTHREAD_STACK_MIN
+# ifdef PTHREAD_STACK_MIN
 	thread_debug("create - stack size: %lu\n", (unsigned long)stack_size);
 	CHECK_ERR(pthread_attr_setstacksize(&attr, stack_size));
-#endif
+# endif
 
-#ifdef HAVE_PTHREAD_ATTR_SETINHERITSCHED
+# ifdef HAVE_PTHREAD_ATTR_SETINHERITSCHED
 	CHECK_ERR(pthread_attr_setinheritsched(&attr, PTHREAD_INHERIT_SCHED));
-#endif
+# endif
 	CHECK_ERR(pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED));
 
 	err = pthread_create(&th->thread_id, &attr, thread_start_func_1, th);
+#else
+	err = pthread_create(&th->thread_id, NULL, thread_start_func_1, th);
+#endif
 	thread_debug("create: %p (%d)\n", (void *)th, err);
+#ifdef HAVE_PTHREAD_ATTR_INIT
 	CHECK_ERR(pthread_attr_destroy(&attr));
+#endif
     }
     return err;
 }
@@ -853,6 +957,12 @@ native_thread_apply_priority(rb_thread_t *th)
 
 #endif /* USE_NATIVE_THREAD_PRIORITY */
 
+static int
+native_fd_select(int n, rb_fdset_t *readfds, rb_fdset_t *writefds, rb_fdset_t *exceptfds, struct timeval *timeout, rb_thread_t *th)
+{
+    return rb_fd_select(n, readfds, writefds, exceptfds, timeout);
+}
+
 static void
 ubf_pthread_cond_signal(void *ptr)
 {
@@ -942,7 +1052,7 @@ print_signal_list(char *str)
     struct signal_thread_list *list =
       signal_thread_list_anchor.next;
     thread_debug("list (%s)> ", str);
-    while(list){
+    while (list) {
 	thread_debug("%p (%p), ", list->th, list->th->thread_id);
 	list = list->next;
     }
@@ -1016,7 +1126,8 @@ ubf_select(void *ptr)
 }
 
 static void
-ping_signal_thread_list(void) {
+ping_signal_thread_list(void)
+{
     if (signal_thread_list_anchor.next) {
 	FGLOCK(&signal_thread_list_lock, {
 	    struct signal_thread_list *list;
@@ -1039,23 +1150,29 @@ check_signal_thread_list(void)
 	return 0;
 }
 #else /* USE_SIGNAL_THREAD_LIST */
-static void add_signal_thread_list(rb_thread_t *th) { }
-static void remove_signal_thread_list(rb_thread_t *th) { }
+#define add_signal_thread_list(th) (void)(th)
+#define remove_signal_thread_list(th) (void)(th)
 #define ubf_select 0
 static void ping_signal_thread_list(void) { return; }
 static int check_signal_thread_list(void) { return 0; }
 #endif /* USE_SIGNAL_THREAD_LIST */
 
-static int timer_thread_pipe[2] = {-1, -1};
-static int timer_thread_pipe_owner_process;
-
 #define TT_DEBUG 0
-
 #define WRITE_CONST(fd, str) (void)(write((fd),(str),sizeof(str)-1)<0)
 
+/* 100ms.  10ms is too small for user level thread scheduling
+ * on recent Linux (tested on 2.6.35)
+ */
+#define TIME_QUANTUM_USEC (100 * 1000)
+
+#if USE_SLEEPY_TIMER_THREAD
+static int timer_thread_pipe[2] = {-1, -1};
+static int timer_thread_pipe_low[2] = {-1, -1}; /* low priority */
+static int timer_thread_pipe_owner_process;
+
 /* only use signal-safe system calls here */
-void
-rb_thread_wakeup_timer_thread(void)
+static void
+rb_thread_wakeup_timer_thread_fd(int fd)
 {
     ssize_t result;
 
@@ -1063,7 +1180,7 @@ rb_thread_wakeup_timer_thread(void)
     if (timer_thread_pipe_owner_process == getpid()) {
 	const char *buff = "!";
       retry:
-	if ((result = write(timer_thread_pipe[1], buff, 1)) <= 0) {
+	if ((result = write(fd, buff, 1)) <= 0) {
 	    switch (errno) {
 	      case EINTR: goto retry;
 	      case EAGAIN:
@@ -1082,97 +1199,208 @@ rb_thread_wakeup_timer_thread(void)
     }
 }
 
+void
+rb_thread_wakeup_timer_thread(void)
+{
+    rb_thread_wakeup_timer_thread_fd(timer_thread_pipe[1]);
+}
+
+static void
+rb_thread_wakeup_timer_thread_low(void)
+{
+    rb_thread_wakeup_timer_thread_fd(timer_thread_pipe_low[1]);
+}
+
 /* VM-dependent API is not available for this function */
 static void
-consume_communication_pipe(void)
+consume_communication_pipe(int fd)
 {
 #define CCP_READ_BUFF_SIZE 1024
     /* buffer can be shared because no one refers to them. */
     static char buff[CCP_READ_BUFF_SIZE];
     ssize_t result;
 
-  retry:
-    result = read(timer_thread_pipe[0], buff, CCP_READ_BUFF_SIZE);
-    if (result < 0) {
-	switch (errno) {
-	  case EINTR: goto retry;
-	  default:
-	    rb_async_bug_errno("consume_communication_pipe: read\n", errno);
+    while (1) {
+	result = read(fd, buff, sizeof(buff));
+	if (result == 0) {
+	    return;
+	}
+	else if (result < 0) {
+	    switch (errno) {
+	    case EINTR:
+		continue; /* retry */
+	    case EAGAIN:
+		return;
+	    default:
+		rb_async_bug_errno("consume_communication_pipe: read\n", errno);
+	    }
 	}
     }
 }
 
 static void
-close_communication_pipe(void)
+close_communication_pipe(int pipes[2])
 {
-    if (close(timer_thread_pipe[0]) < 0) {
+    if (close(pipes[0]) < 0) {
 	rb_bug_errno("native_stop_timer_thread - close(ttp[0])", errno);
     }
-    if (close(timer_thread_pipe[1]) < 0) {
+    if (close(pipes[1]) < 0) {
 	rb_bug_errno("native_stop_timer_thread - close(ttp[1])", errno);
     }
-    timer_thread_pipe[0] = timer_thread_pipe[1] = -1;
+    pipes[0] = pipes[1] = -1;
 }
 
-/* 100ms.  10ms is too small for user level thread scheduling
- * on recent Linux (tested on 2.6.35)
+static void
+set_nonblock(int fd)
+{
+    int oflags;
+    int err;
+
+    oflags = fcntl(fd, F_GETFL);
+    if (oflags == -1)
+	rb_sys_fail(0);
+    oflags |= O_NONBLOCK;
+    err = fcntl(fd, F_SETFL, oflags);
+    if (err == -1)
+	rb_sys_fail(0);
+}
+
+static void
+setup_communication_pipe_internal(int pipes[2])
+{
+    int err;
+
+    if (pipes[0] != -1) {
+	/* close pipe of parent process */
+	close_communication_pipe(pipes);
+    }
+
+    err = rb_cloexec_pipe(pipes);
+    if (err != 0) {
+	rb_bug_errno("setup_communication_pipe: Failed to create communication pipe for timer thread", errno);
+    }
+    rb_update_max_fd(pipes[0]);
+    rb_update_max_fd(pipes[1]);
+    set_nonblock(pipes[0]);
+    set_nonblock(pipes[1]);
+}
+
+/* communication pipe with timer thread and signal handler */
+static void
+setup_communication_pipe(void)
+{
+    if (timer_thread_pipe_owner_process == getpid()) {
+	/* already set up. */
+	return;
+    }
+    setup_communication_pipe_internal(timer_thread_pipe);
+    setup_communication_pipe_internal(timer_thread_pipe_low);
+
+    /* validate pipe on this process */
+    timer_thread_pipe_owner_process = getpid();
+}
+
+/**
+ * Let the timer thread sleep a while.
+ *
+ * The timer thread sleeps until woken up by rb_thread_wakeup_timer_thread() if only one Ruby thread is running.
+ * @pre the calling context is in the timer thread.
  */
-#define TIME_QUANTUM_USEC (100 * 1000)
+static inline void
+timer_thread_sleep(rb_global_vm_lock_t* gvl)
+{
+    int result;
+    int need_polling;
+    struct pollfd pollfds[2];
+
+    pollfds[0].fd = timer_thread_pipe[0];
+    pollfds[0].events = POLLIN;
+    pollfds[1].fd = timer_thread_pipe_low[0];
+    pollfds[1].events = POLLIN;
+
+    need_polling = check_signal_thread_list();
+
+    if (gvl->waiting > 0 || need_polling) {
+	/* polling (TIME_QUANTUM_USEC usec) */
+	result = poll(pollfds, 1, TIME_QUANTUM_USEC/1000);
+    }
+    else {
+	/* wait (infinite) */
+	result = poll(pollfds, ARRAY_SIZE(pollfds), -1);
+    }
+
+    if (result == 0) {
+	/* maybe timeout */
+    }
+    else if (result > 0) {
+	consume_communication_pipe(timer_thread_pipe[0]);
+	consume_communication_pipe(timer_thread_pipe_low[0]);
+    }
+    else { /* result < 0 */
+	switch (errno) {
+	case EBADF:
+	case EINVAL:
+	case ENOMEM: /* from Linux man */
+	case EFAULT: /* from FreeBSD man */
+	    rb_async_bug_errno("thread_timer: select", errno);
+	default:
+	    /* ignore */;
+	}
+    }
+}
+
+#else /* USE_SLEEPY_TIMER_THREAD */
+# define PER_NANO 1000000000
+void rb_thread_wakeup_timer_thread(void) {}
+static void rb_thread_wakeup_timer_thread_low(void) {}
+
+static pthread_mutex_t timer_thread_lock;
+static rb_thread_cond_t timer_thread_cond;
+
+static inline void
+timer_thread_sleep(rb_global_vm_lock_t* unused)
+{
+    struct timespec ts;
+    ts.tv_sec = 0;
+    ts.tv_nsec = TIME_QUANTUM_USEC * 1000;
+    ts = native_cond_timeout(&timer_thread_cond, ts);
+
+    native_cond_timedwait(&timer_thread_cond, &timer_thread_lock, &ts);
+}
+#endif /* USE_SLEEPY_TIMER_THREAD */
 
 static void *
 thread_timer(void *p)
 {
     rb_global_vm_lock_t *gvl = (rb_global_vm_lock_t *)p;
-    int result;
-    struct timeval timeout;
 
     if (TT_DEBUG) WRITE_CONST(2, "start timer thread\n");
 
+#if defined(__linux__) && defined(PR_SET_NAME)
+    prctl(PR_SET_NAME, "ruby-timer-thr");
+#endif
+
+#if !USE_SLEEPY_TIMER_THREAD
+    native_mutex_initialize(&timer_thread_lock);
+    native_cond_initialize(&timer_thread_cond, RB_CONDATTR_CLOCK_MONOTONIC);
+    native_mutex_lock(&timer_thread_lock);
+#endif
     while (system_working > 0) {
-	fd_set rfds;
-	int need_polling;
 
 	/* timer function */
 	ping_signal_thread_list();
 	timer_thread_function(0);
-	need_polling = check_signal_thread_list();
 
 	if (TT_DEBUG) WRITE_CONST(2, "tick\n");
 
-	/* wait */
-	FD_ZERO(&rfds);
-	FD_SET(timer_thread_pipe[0], &rfds);
-
-	if (gvl->waiting > 0 || need_polling) {
-	    timeout.tv_sec = 0;
-	    timeout.tv_usec = TIME_QUANTUM_USEC;
-
-	    /* polling (TIME_QUANTUM_USEC usec) */
-	    result = select(timer_thread_pipe[0] + 1, &rfds, 0, 0, &timeout);
-	}
-	else {
-	    /* wait (infinite) */
-	    result = select(timer_thread_pipe[0] + 1, &rfds, 0, 0, 0);
-	}
-
-	if (result == 0) {
-	    /* maybe timeout */
-	}
-	else if (result > 0) {
-	    consume_communication_pipe();
-	}
-	else { /* result < 0 */
-	  switch (errno) {
-	    case EBADF:
-	    case EINVAL:
-	    case ENOMEM: /* from Linux man */
-	    case EFAULT: /* from FreeBSD man */
-	      rb_async_bug_errno("thread_timer: select", errno);
-	    default:
-	      /* ignore */;
-	  }
-	}
+        /* wait */
+	timer_thread_sleep(gvl);
     }
+#if !USE_SLEEPY_TIMER_THREAD
+    native_mutex_unlock(&timer_thread_lock);
+    native_cond_destroy(&timer_thread_cond);
+    native_mutex_destroy(&timer_thread_lock);
+#endif
 
     if (TT_DEBUG) WRITE_CONST(2, "finish timer thread\n");
     return NULL;
@@ -1181,14 +1409,17 @@ thread_timer(void *p)
 static void
 rb_thread_create_timer_thread(void)
 {
-    rb_enable_interrupt();
-
     if (!timer_thread_id) {
-	pthread_attr_t attr;
 	int err;
+#ifdef HAVE_PTHREAD_ATTR_INIT
+	pthread_attr_t attr;
 
-	pthread_attr_init(&attr);
-#ifdef PTHREAD_STACK_MIN
+	err = pthread_attr_init(&attr);
+	if (err != 0) {
+	    fprintf(stderr, "[FATAL] Failed to initialize pthread attr(errno: %d)\n", err);
+	    exit(EXIT_FAILURE);
+        }
+# ifdef PTHREAD_STACK_MIN
 	if (PTHREAD_STACK_MIN < 4096 * 3) {
 	    /* Allocate the machine stack for the timer thread
              * at least 12KB (3 pages).  FreeBSD 8.2 AMD64 causes
@@ -1201,55 +1432,30 @@ rb_thread_create_timer_thread(void)
 	    pthread_attr_setstacksize(&attr,
 				      PTHREAD_STACK_MIN + (THREAD_DEBUG ? BUFSIZ : 0));
 	}
+# endif
 #endif
 
-	/* communication pipe with timer thread and signal handler */
-	if (timer_thread_pipe_owner_process != getpid()) {
-	    if (timer_thread_pipe[0] != -1) {
-		/* close pipe of parent process */
-		close_communication_pipe();
-	    }
-
-	    err = pipe(timer_thread_pipe);
-	    if (err != 0) {
-		rb_bug_errno("thread_timer: Failed to create communication pipe for timer thread", errno);
-	    }
-            rb_update_max_fd(timer_thread_pipe[0]);
-            rb_update_max_fd(timer_thread_pipe[1]);
-#if defined(HAVE_FCNTL) && defined(F_GETFL) && defined(F_SETFL)
-	    {
-		int oflags;
-#if defined(O_NONBLOCK)
-		oflags = fcntl(timer_thread_pipe[1], F_GETFL);
-		oflags |= O_NONBLOCK;
-		fcntl(timer_thread_pipe[1], F_SETFL, oflags);
-#endif /* defined(O_NONBLOCK) */
-#if defined(FD_CLOEXEC)
-		oflags = fcntl(timer_thread_pipe[0], F_GETFD);
-		fcntl(timer_thread_pipe[0], F_SETFD, oflags | FD_CLOEXEC);
-		oflags = fcntl(timer_thread_pipe[1], F_GETFD);
-		fcntl(timer_thread_pipe[1], F_SETFD, oflags | FD_CLOEXEC);
-#endif /* defined(FD_CLOEXEC) */
-	    }
-#endif /* defined(HAVE_FCNTL) && defined(F_GETFL) && defined(F_SETFL) */
-
-	    /* validate pipe on this process */
-	    timer_thread_pipe_owner_process = getpid();
-	}
+#if USE_SLEEPY_TIMER_THREAD
+	setup_communication_pipe();
+#endif /* USE_SLEEPY_TIMER_THREAD */
 
 	/* create timer thread */
 	if (timer_thread_id) {
 	    rb_bug("rb_thread_create_timer_thread: Timer thread was already created\n");
 	}
+#ifdef HAVE_PTHREAD_ATTR_INIT
 	err = pthread_create(&timer_thread_id, &attr, thread_timer, &GET_VM()->gvl);
+#else
+	err = pthread_create(&timer_thread_id, NULL, thread_timer, &GET_VM()->gvl);
+#endif
 	if (err != 0) {
 	    fprintf(stderr, "[FATAL] Failed to create timer thread (errno: %d)\n", err);
 	    exit(EXIT_FAILURE);
 	}
+#ifdef HAVE_PTHREAD_ATTR_INIT
 	pthread_attr_destroy(&attr);
+#endif
     }
-
-    rb_disable_interrupt(); /* only timer thread recieve signal */
 }
 
 static int
@@ -1308,9 +1514,9 @@ ruby_stack_overflowed_p(const rb_thread_t *th, const void *addr)
     else {
 	return 0;
     }
-    size /= 5;
+    size /= RUBY_STACK_SPACE_RATIO;
     if (size > water_mark) size = water_mark;
-    if (STACK_DIR_UPPER(1, 0)) {
+    if (IS_STACK_DIR_UPPER()) {
 	if (size > ~(size_t)base+1) size = ~(size_t)base+1;
 	if (addr > base && addr <= (void *)((char *)base + size)) return 1;
     }
@@ -1325,13 +1531,19 @@ ruby_stack_overflowed_p(const rb_thread_t *th, const void *addr)
 int
 rb_reserved_fd_p(int fd)
 {
-    if (fd == timer_thread_pipe[0] ||
-	fd == timer_thread_pipe[1]) {
+#if USE_SLEEPY_TIMER_THREAD
+    if (fd == timer_thread_pipe[0]     ||
+	fd == timer_thread_pipe[1]     ||
+	fd == timer_thread_pipe_low[0] ||
+	fd == timer_thread_pipe_low[1]) {
 	return 1;
     }
     else {
 	return 0;
     }
+#else
+    return 0;
+#endif
 }
 
 #endif /* THREAD_SYSTEM_DEPENDENT_IMPLEMENTATION */
